SystemOrganization addCategory: #'Croquet-Objects'!

!PositionableStream methodsFor: '*Croquet-Objects' stamp: 'ar 1/13/2006 13:27'!
upToAnyOf: aCharacterSet 
	"Answer string until aCharacterSet found. The answer INCLUDES the 
	character "
	"'test' readStream upToAnyOf: {$e} asCharacterSet"
	| pos buffer count |
	pos := self position.
	buffer := self next: 100.
	(count := String
					findFirstInString: buffer
					inSet: aCharacterSet byteArrayMap
					startingAt: 1) > 0
		ifTrue: [self position: pos + count.
			^ buffer copyFrom: 1 to: count].
	self atEnd
		ifTrue: ["Never found it, and hit end of file"
			^ buffer , Character end asString].
	"Never found it, but there's more..."
	^ buffer
		, (self upToAnyOf: aCharacterSet)! !

Object subclass: #TBox
	instanceVariableNames: 'minCorner maxCorner'
	classVariableNames: 'Epsilon'
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TBox commentStamp: 'das 10/10/2004 10:22' prior: 0!
TBox has a number of uses. It is an axis aligned bounding box (AABB) defined by two 3D corners. It is used by the TBoundSphere to help define the boundSphere tree. It is used by the TParticle as a parameter bounds for it's instance variables. It is also used by the TQuadTree for it's hierarchical representations.

DAS!

!TBox class methodsFor: 'class initialization' stamp: 'ar 6/19/2006 12:20'!
initialize
	"TBox initialize"

	"Bullet-proof initialization against missing Float>>epsilon"
	Epsilon := [Float epsilon] on: Error do:[:ex | ex return: 0.000000000001].! !

!TBox class methodsFor: 'instance creation' stamp: 'das 4/23/2002 00:29'!
min: mn max: mx

	^ self new min:mn max: mx.! !

!TBox class methodsFor: 'instance creation' stamp: 'ar 9/30/2004 19:20'!
new
	^self basicNew initialize! !

!TBox class methodsFor: 'instance creation' stamp: 'ar 10/8/2002 10:43'!
origin: origin corner: corner

	^ self new min: origin max: corner.! !

!TBox methodsFor: 'math' stamp: 'das 2/10/2003 07:53'!
* aNumber

	
	^ TBox min: (minCorner * aNumber) max: (maxCorner * aNumber).
! !

!TBox methodsFor: 'math' stamp: 'das 2/10/2003 08:18'!
addBox: aBox

	^ TBox min: minCorner + aBox min max: maxCorner + aBox max.! !

!TBox methodsFor: 'accessing' stamp: 'das 5/16/2005 08:52'!
atRandom: segments

	| dm |
	dm := (maxCorner - minCorner)/segments.
	^ Vector3 x:(dm x * (segments atRandom) + minCorner x) 
				y:(dm y * (segments atRandom) + minCorner y)
				z:(dm z * (segments atRandom) + minCorner z).
! !

!TBox methodsFor: 'comparing' stamp: 'das 4/24/2002 10:30'!
axisTestXv1: v1 v2: v2 a: a b: b fa: fa fb: fb

	| p0 p1 min max rad |

	p0 := (a * v1 y) - (b * v1 z).
	p1 := (a * v2 y) - (b * v2 z).
	p0 < p1 ifTrue:[min := p0. max := p1.] ifFalse:[min := p1. max := p0.].
	rad := ((fa * self extentY) + (fb * self extentZ))/2.0.
	(min > rad or:[max < (rad negated)]) ifTrue:[^ false.].
	^ true.! !

!TBox methodsFor: 'comparing' stamp: 'das 4/27/2002 13:08'!
axisTestYv1: v1 v2: v2 a: a b: b fa: fa fb: fb

	| p0 p1 min max rad |

	p0 := (b * v1 z) - (a * v1 x).
	p1 := (b * v2 z) - (a * v2 x).
	p0 < p1 ifTrue:[min := p0. max := p1.] ifFalse:[min := p1. max := p0.].
	rad := ((fa * self extentX) + (fb * self extentZ))/2.0.
	(min > rad or:[max < (rad negated)]) ifTrue:[^ false.].
	^ true.! !

!TBox methodsFor: 'comparing' stamp: 'das 4/24/2002 11:05'!
axisTestZv1: v1 v2: v2 a: a b: b fa: fa fb: fb

	| p0 p1 min max rad |

	p0 := (a * v1 x) - (b * v1 y).
	p1 := (a * v2 x) - (b * v2 y).
	p0 < p1 ifTrue:[min := p0. max := p1.] ifFalse:[min := p1. max := p0.].
	rad := ((fa * self extentX) + (fb * self extentY))/2.0.
	(min > rad or:[max < (rad negated)]) ifTrue:[^ false.].
	^ true.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 00:27'!
center

	^ (minCorner + maxCorner)/2.0.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 00:36'!
centerX

	^ (minCorner x + maxCorner x)/2.0.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 00:36'!
centerY

	^ (minCorner y + maxCorner y)/2.0.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 00:36'!
centerZ

	^ (minCorner z + maxCorner z)/2.0.! !

!TBox methodsFor: 'private' stamp: 'das 5/21/2006 13:01'!
clipPoly: clipPoly plane: pl d: d

	| poly v1 d1 d2 plane |

"negative d1, d2 means the vector is 'outside' the plane"

	poly := OrderedCollection new.
	plane := pl asFloatArray.
	v1 := clipPoly last.
	d1 := (plane dot: v1) + d.
	clipPoly do:[ :v2 | 
		d2 := (plane dot:v2)+d.
		d2>= 0 ifTrue:[           
			d1>=0 ifTrue:[poly add: v2] "**** d1<=0 - d2<=0"
			ifFalse:[poly add:(self splitv1:v1 v2:v2 d1:d1 negated d2:d2).
					poly add:v2.] "****d1>=0 - d2>=0"
		]ifFalse:[
			d1>=0 ifTrue:[ poly add:(self splitv1: v1 v2: v2 d1:d1 d2:d2 negated).]. "****d1<0 - d2>=0"
		].
		v1 := v2.
		d1 := d2.
	].
	^ poly.! !

!TBox methodsFor: 'comparing' stamp: 'bgf 8/2/2006 19:53'!
confinePoint: aVector3

	"If the point is in the box, return it.
	 Otherwise, return a new near the point, but within the box."

	(self pointInBox: aVector3) ifTrue: [
		^ aVector3
	].

	^ Vector3	x: (aVector3 x min: maxCorner x max: minCorner x)
				y: (aVector3 y min: maxCorner y max: minCorner y)
				z: (aVector3 z min: maxCorner z max: minCorner z).
	! !

!TBox methodsFor: 'accessing' stamp: 'ar 10/8/2002 10:42'!
corner
	^maxCorner! !

!TBox methodsFor: 'accessing' stamp: 'ar 10/8/2002 10:49'!
cornerVertices
	"Answer all the vertices defining the corners of this block"
	| out |
	out := (WriteStream on: (Array new: 8)).
	self cornerVerticesDo:[:vtx| out nextPut: vtx].
	^out contents! !

!TBox methodsFor: 'accessing' stamp: 'ar 10/8/2002 10:48'!
cornerVerticesDo: aBlock
	"Evaluate aBlock with all the vertices defining the corners of this block"
	| min max |
	min := minCorner.
	max := maxCorner.
	aBlock value: min x @ min y @ min z.
	aBlock value: min x @ min y @ max z.
	aBlock value: min x @ max y @ min z.
	aBlock value: min x @ max y @ max z.
	aBlock value: max x @ min y @ min z.
	aBlock value: max x @ min y @ max z.
	aBlock value: max x @ max y @ min z.
	aBlock value: max x @ max y @ max z.
! !

!TBox methodsFor: 'accessing' stamp: 'das 4/25/2002 16:25'!
diagonal

	^ (maxCorner - minCorner)length.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 00:26'!
extent

	^ maxCorner - minCorner.! !

!TBox methodsFor: 'accessing' stamp: 'das 10/14/2003 01:24'!
extentMax


	^ (maxCorner - minCorner) max.
! !

!TBox methodsFor: 'accessing' stamp: 'das 10/14/2003 01:24'!
extentMin

	^ (maxCorner - minCorner) min.

! !

!TBox methodsFor: 'accessing' stamp: 'das 4/24/2002 10:28'!
extentX

	^ maxCorner x - minCorner x.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/24/2002 10:28'!
extentY

	^ maxCorner y - minCorner y.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/24/2002 10:28'!
extentZ

	^ maxCorner z - minCorner z.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 00:19'!
growVertex: v

	minCorner := minCorner min: v.
	maxCorner := maxCorner max: v.! !

!TBox methodsFor: 'initialize' stamp: 'das 5/16/2005 08:52'!
initialize
	| iP iN |

" Defines an infinitely empty box - min is positive infinity, max is negative infinity. We then add vertices to the box via the #growVertex: method to 'grow' it."

	iP := Float infinity.
	iN := Float infinity negated.
	minCorner := Vector3 x: iP y: iP z: iP.
	maxCorner := Vector3 x: iN y: iN z: iN.
! !

!TBox methodsFor: 'comparing' stamp: 'das 4/25/2002 20:52'!
intersectBox: bx
	| mn mx |
	mn := bx min max: self min.
	mx := bx max min: self max.
	^ TBox min: mn max: mx.! !

!TBox methodsFor: 'comparing' stamp: 'das 5/21/2006 13:55'!
intersectT1: t1 T2: t2 T3: t3

	| v1 v2 v3 e1 e2 e3 a1 a2 a3 center tNormal d vmin vmax minCornerVal maxCornerVal t1val t2val t3val extent |
self error:'This code appears to be broken. Use #intersectV1:V2:V3 instead.'.
" This method is based upon:
	Fast 3D Triangle-Box Overlap Testing 
	Tomas Akenine-Moller
	March 2001, updated June 2001.

The order of this code is somewhat different from that presented in the example C source provided with the paper.

Though this is considered to be the fastest test in the graphics literature, I think that there is a better approach by attempting to construct a seperating plane between the two objects. Since I am not currently using it for real-time work, I won't worry about it."

" First, the naive test - is the triangle all on one side of a half-plane?"
	minCornerVal := minCorner x.
	t1val := t1 x. 
	t2val := t2 x. 
	t3val := t3 x. 
	(t1val < minCornerVal and:[ t2val < minCornerVal and:[ t3val < minCornerVal]]) 
		ifTrue:[ ^ false. ].
	maxCornerVal := maxCorner x.
	(t1val > maxCornerVal and:[. t2val > maxCornerVal and:[ t3val > maxCornerVal]]) 
		ifTrue:[ ^ false. ].
	minCornerVal := minCorner y.
	t1val := t1 y. 
	t2val := t2 y. 
	t3val := t3 y. 
	(t1val < minCornerVal and:[ t2val < minCornerVal and:[ t3val < minCornerVal]]) 
		ifTrue:[^ false. ].
	maxCornerVal := maxCorner y.
	(t1val > maxCornerVal and:[ t2val > maxCornerVal and:[ t3val > maxCornerVal]]) 
		ifTrue:[ ^ false. ].
	minCornerVal := minCorner z.
	t1val := t1 z. 
	t2val := t2 z. 
	t3val := t3 z. 
	(t1val < minCornerVal and:[ t2val < minCornerVal and:[ t3val < minCornerVal]]) 
		ifTrue:[ ^ false. ].
	maxCornerVal := maxCorner z.
	(t1val > maxCornerVal and:[ t2val > maxCornerVal and:[ t3val > maxCornerVal]]) 
		ifTrue:[ ^ false. ].
" Second, the cross product (edge from tri, (x y z) direction. "
	center := self center.
	v1 := t1 - center.
	v2 := t2 - center.
	v3 := t3 - center.

	e1 := t2 - t1.
	a1 := e1 abs.
	(self axisTestXv1:v1 v2:v3 a: e1 z  b: e1 y fa: a1 z fb: a1 y) ifFalse:[^ false.].
	(self axisTestYv1:v1 v2:v3 a: e1 z b: e1 x fa: a1 z fb: a1 x) ifFalse:[^ false.].
	(self axisTestZv1:v2 v2:v3 a: e1 y b: e1 x fa: a1 y fb: a1 x) ifFalse:[^ false.].

	e2 := t3 - t2.
	a2 := e2 abs.
	(self axisTestXv1:v1 v2:v3 a: e2 z  b: e2 y fa: a2 z fb: a2 y) ifFalse:[^ false.].
	(self axisTestYv1:v1 v2:v3 a: e2 z b: e2 x fa: a2 z fb: a2 x) ifFalse:[^ false.].
	(self axisTestZv1: v1 v2:v2 a:e2 y b: e2 x fa: a2 y fb: a2 x) ifFalse:[^ false.].

	e3 := t1 - t3.
	a3 := e3 abs.
	(self axisTestXv1:v1 v2:v2 a: e3 z  b: e3 y fa: a3 z fb: a3 y) ifFalse:[^ false.].
	(self axisTestYv1:v1 v2:v2 a: e3 z b: e3 x fa: a3 z fb: a3 x) ifFalse:[^ false.].
	(self axisTestZv1:v2 v2:v3 a: e3 y b: e3 x fa: a3 y fb: a3 x) ifFalse:[^ false.].

" Third, test if the box intersects the plane of the triangle."

	tNormal := e1 cross: e2.
	d := (tNormal dot: v1)negated.
	vmin := Vector3 new.	
	vmax := Vector3 new.
	
	extent := self extentX.
	tNormal x>0.0 ifTrue:[
		vmin x: extent negated.
		vmax x: extent.] 
		ifFalse:[
		vmin x: extent.
		vmax x: extent negated.].
	extent := self extentY.
	tNormal y>0.0 ifTrue:[
		vmin y: extent negated.
		vmax y: extent.] 
		ifFalse:[
		vmin y: extent.
		vmax y: extent negated.].
	extent := self extentZ.
	tNormal z>0.0 ifTrue:[
		vmin z: extent negated.
		vmax z: extent.] 
		ifFalse:[
		vmin z: extent.
		vmax z: extent negated.].
	(tNormal dot: vmin)+d >0.0 ifTrue:[^ false. ].
	(tNormal dot: vmax)+d >= 0.0 ifTrue:[^ true.].
	^ true.
	! !

!TBox methodsFor: 'comparing' stamp: 'das 5/21/2006 13:47'!
intersectV1: v1 V2: v2 V3: v3

	| poly |
	(self less: minCorner x t1: v1 x t2: v2 x t3: v3 x) ifTrue:[^ false].
	(self more: maxCorner x t1: v1 x t2: v2 x t3: v3 x) ifTrue:[^ false].
	(self less: minCorner y t1: v1 y t2: v2 y t3: v3 y) ifTrue:[^ false].
	(self more: maxCorner y t1: v1 y t2: v2 y t3: v3 y) ifTrue:[^ false].
	(self less: minCorner z t1: v1 z t2: v2 z t3: v3 z) ifTrue:[^ false].
	(self more: maxCorner z t1: v1 z t2: v2 z t3: v3 z) ifTrue:[^ false].
	poly := Array with:v1 with:v2 with: v3.

	((poly := self clipPoly: poly plane: #(1 0 0) d: minCorner x negated) size < 3)ifTrue:[^false.].
	((poly := self clipPoly: poly plane: #(-1 0 0) d: maxCorner x) size < 3)ifTrue:[^false.].
	((poly := self clipPoly: poly plane: #(0 1 0) d: minCorner y negated) size < 3)ifTrue:[^false.].
	((poly := self clipPoly: poly plane: #(0 -1 0) d: maxCorner y) size < 3)ifTrue:[^false.].
	((poly := self clipPoly: poly plane: #(0 0 1) d: minCorner z negated) size < 3)ifTrue:[^false.].
	((poly := self clipPoly: poly plane: #(0 0 -1) d: maxCorner z) size < 3)ifTrue:[^false.].

	^ true.! !

!TBox methodsFor: 'private' stamp: 'das 5/18/2006 17:14'!
less: val t1: t1 t2: t2 t3:t3

	^(t1<val and:[t2<val and:[t3<val]])
! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 11:12'!
max

	^ maxCorner.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 11:12'!
max: mx

	maxCorner := mx.! !

!TBox methodsFor: 'accessing' stamp: 'das 5/17/2003 01:25'!
merge: aBox
	"Answer a Box that contains both the receiver and aBox."

	^TBox 
		origin: (self origin min: aBox origin)
		corner: (self corner max: aBox corner)! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 11:11'!
min

	^ minCorner.! !

!TBox methodsFor: 'accessing' stamp: 'das 4/23/2002 11:12'!
min: mn

	minCorner := mn.! !

!TBox methodsFor: 'initialize' stamp: 'das 4/23/2002 00:28'!
min: mn max: mx

	minCorner := mn.
	maxCorner := mx.! !

!TBox methodsFor: 'private' stamp: 'das 5/18/2006 17:14'!
more: val t1: t1 t2: t2 t3:t3

	^(t1>val and:[t2>val and:[t3>val]])! !

!TBox methodsFor: 'accessing' stamp: 'ar 10/8/2002 10:42'!
origin
	^minCorner! !

!TBox methodsFor: 'comparing' stamp: 'das 4/24/2002 11:46'!
pointInBox: v

	v x < minCorner x ifTrue:[^ false.].
	v x > maxCorner x ifTrue:[^ false.].
	v y < minCorner y ifTrue:[^ false.].
	v y > maxCorner y ifTrue:[^ false.].
	v z < minCorner z ifTrue:[^ false.].
	v z > maxCorner z ifTrue:[^ false. ].
	^ true.! !

!TBox methodsFor: 'comparing' stamp: 'das 6/29/2002 16:54'!
pointOverBox: v

	v x < minCorner x ifTrue:[^ false.].
	v x > maxCorner x ifTrue:[^ false.].
"	v y < minCorner y ifTrue:[^ false.].
	v y > maxCorner y ifTrue:[^ false.]."
	v z < minCorner z ifTrue:[^ false.].
	v z > maxCorner z ifTrue:[^ false. ].
	^ true.! !

!TBox methodsFor: 'accessing' stamp: 'das 10/14/2003 01:25'!
radius

	^ self diagonal/2.0.! !

!TBox methodsFor: 'accessing' stamp: 'das 9/14/2002 00:13'!
scale: scale

	minCorner := minCorner * scale.
	maxCorner := maxCorner * scale.! !

!TBox methodsFor: 'accessing' stamp: 'ar 6/19/2006 12:18'!
splitXMax

	| corner |
	corner := self min copy.
	corner x: self centerX - Epsilon.
	^ TBox min: corner max: self max copy.! !

!TBox methodsFor: 'accessing' stamp: 'ar 6/19/2006 12:18'!
splitXMin

	| corner |
	corner := self max copy.
	corner x: self centerX + Epsilon.
	^ TBox min: self min copy max: corner.! !

!TBox methodsFor: 'accessing' stamp: 'ar 6/19/2006 12:18'!
splitYMax

	| corner |
	corner := self min copy.
	corner y: self centerY - Epsilon.
	^ TBox min: corner max: self max copy.! !

!TBox methodsFor: 'accessing' stamp: 'ar 6/19/2006 12:18'!
splitYMin

	| corner |
	corner := self max copy.
	corner y: self centerY + Epsilon.
	^ TBox min: self min copy max: corner.! !

!TBox methodsFor: 'accessing' stamp: 'ar 6/19/2006 12:18'!
splitZMax

	| corner |
	corner := self min copy.
	corner z: self centerZ - Epsilon.
	^ TBox min: corner max: self max copy.! !

!TBox methodsFor: 'accessing' stamp: 'ar 6/19/2006 12:18'!
splitZMin

	| corner |
	corner := self max copy.
	corner z: self centerZ + Epsilon.
	^ TBox min: self min copy max: corner.! !

!TBox methodsFor: 'private' stamp: 'das 5/21/2006 12:11'!
splitv1:v1 v2:v2 d1:d1 d2:d2

	| vx dx |
	vx:= v2-v1.
	dx := d1+d2.
	^ v1+(vx*d1/dx)! !

!TBox methodsFor: 'accessing' stamp: 'ar 10/8/2002 10:47'!
transformedBy: aTransform
	"Answer a box transformed by the given transformation"
	| aVertex min max |
	min := max := nil.
	self cornerVerticesDo:[:vtx| 
		aVertex := aTransform localPointToGlobal: vtx.
		min ifNil:[
			min := max := aVertex.
		] ifNotNil:[
			min := min min: aVertex.
			max := max max: aVertex.
		].
	].
	^TBox origin: min corner: max! !

!TBox methodsFor: 'accessing' stamp: 'ar 10/8/2002 11:01'!
translatedBy: delta
	"Answer a box transformed by the given transformation"
	^TBox origin: minCorner + delta corner: maxCorner + delta! !

!TBox methodsFor: 'comparing' stamp: 'das 6/18/2002 16:51'!
unionBox: bx
	| mn mx |
	mn := bx min min: self min.
	mx := bx max max: self max.
	^ TBox min: mn max: mx.! !

!TBox methodsFor: 'comparing' stamp: 'das 6/29/2002 16:37'!
unionSphere: position radius: radius
	| mn mx |
	mn := (position - radius) min: self min.
	mx := (position + radius) max: self max.
	^ TBox min: mn max: mx.! !

Object subclass: #TFormFind
	instanceVariableNames: 'sha url distance routerAddress routerID bMipmap bShrinkFit extension controllerID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TFormFind commentStamp: '<historical>' prior: 0!
TFormFind is used to find an image file either locally in cache, or remotely in server, or from another user via the router. It is only used by the TFormManager and is created inside of TFormManager>>#resolve:distance:

sha - sha name used for caching files locally.
url - web based access of object
routerAddress, routerID - used to find the router to ask other users for file.
distance - distance from the camera, used to prioritize search.
bMipmap - is the result mipmapped?
bShrinkFit - is the result shrinkfitted?
extension - any extensions, such as color key?!

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 13:19'!
bMipmap

	^ bMipmap.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 13:19'!
bShrinkFit

	^ bShrinkFit.! !

!TFormFind methodsFor: 'accessing' stamp: 'ar 1/30/2006 13:02'!
controllerID
	^controllerID! !

!TFormFind methodsFor: 'accessing' stamp: 'ar 1/30/2006 13:02'!
controllerID: cid
	controllerID := cid! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:17'!
distance

	^distance.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:15'!
distance: d

	distance := d.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/4/2006 22:24'!
extension

	^ extension.! !

!TFormFind methodsFor: 'initialize' stamp: 'das 1/3/2006 13:19'!
initialize:thumb distance: dist

	sha := thumb sha.
	url := thumb url.
	bMipmap := thumb bMipmap.
	bShrinkFit := thumb bShrinkFit.
	extension := thumb extension.
	distance := dist.
	" routerAddress := ? "
	" routerID := ? "
	
! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:16'!
routerAddress

	^routerAddress.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:15'!
routerAddress: ra

	routerAddress := ra.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:16'!
routerID

	^routerID.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:16'!
routerID: rid

	routerID := rid.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:14'!
sha

	^sha.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:14'!
sha: s

	sha := s.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:16'!
url

	^url.! !

!TFormFind methodsFor: 'accessing' stamp: 'das 1/3/2006 11:14'!
url: u

	url := u.! !

Object subclass: #TFormManager
	instanceVariableNames: 'textureDictionary findList sortedFindList mutex updateList cacheMgr loaderProcess loaderSema loaderDelay'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TFormManager commentStamp: 'das 1/5/2006 00:10' prior: 0!
The TFormManager is created in the CroquetHarness, but is also used by the OpenGL object. The reason for this is the harness is used to create Islands and their contents, and OpenGL is used to render the contents, hence the TForm objects used to generate textures must be accessible from both.

TFormManager does a number of things:

- When rendering, it will look up the real TForm in the textureDictionary when provided with a TForm thumbnail.
- If a TForm is requested that is not in the textureDictionary, TFormManager will attempt to load a new copy in the following order:
	> Check the cache directory for a local copy with the name to the sha
	> Check the URL provided as part of the thumbnail object
	> Request a copy of the texture from the originator of the island.

- When it loads a new TForm it does the following:
	> Computes an SHA hash from the file
	> Generates a new TForm object that is placed into a Dictionary that is accessed by OpenGL
	> Loads a bitmap file into a Form that is part of a TForm
	> Generates a TForm thumb object and places the SHA hash into it. This is what is actually returned from the constructor.
	> Saves a copy of the file into the local cache with the name <sha hash value>
	> Saves a copy of the file into the remote super cache with the same name.
!

!TFormManager class methodsFor: 'accessing' stamp: 'ar 11/15/2006 12:14'!
clearDefault
	self default: nil.! !

!TFormManager class methodsFor: 'accessing' stamp: 'ar 11/15/2006 12:14'!
default
	^Default ifNil: [
		Default := self new. 
		TForm defaultTForm. 
		Default]! !

!TFormManager class methodsFor: 'accessing' stamp: 'ar 1/30/2006 22:24'!
default: aFormManager
	Default := aFormManager! !

!TFormManager class methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:51'!
generateSHA: byteArray

	^(SecureHashAlgorithm new hashStream: (ReadStream on: byteArray)) radix:36.! !

!TFormManager methodsFor: 'loading' stamp: 'jcg 8/3/2006 00:08'!
addFile: file mipmap: bMM shrinkFit: bSF extension: ext
	"This is the guts of #addFileNamed:mipmap:shrinkFit:extention:.
	In my opinion, this is preferred interface, because it ought to be safe for any
	API to do operations with an already opened file. Only 'safe' operations (e.g.,
	those explicitly given permission by the user, or those that nag the user, should be allowed
	to open a file on my machine. Besides, the other method could be rewritten to use this one."
	| form tform sha contents |
	contents := file binary contentsOfEntireFile.
	"compute the sha hash"
	sha := TFormManager generateSHA: contents.
	"load the file into a Form"
	tform := textureDictionary at: sha ifAbsentPut:[
		form := Form fromBinaryStream: (ReadStream on:contents).
		form ifNotNil:[
			tform := TForm new 
							initializeSHA: sha 
							form: form 
							mipmap: bMM 
							shrinkFit: bSF 
							extension: ext 
							url: nil.
		] ifNil:[
			Transcript show:'Bad texture file - cannot load';cr.
			^ nil.
		].
		tform.
	].
	"copy the original file into the local cache"
	cacheMgr cache: contents to: sha.
	^tform thumb.! !

!TFormManager methodsFor: 'loading' stamp: 'bgf 6/28/2006 19:24'!
addFileNamed: fName mipmap: bMM shrinkFit: bSF extension: ext

	| file form tform fileName sha contents |
	"find the file"
	fileName := self findTextureFileName: fName.
	fileName ifNil:[ 
		Transcript show: 'Texture not found-  check the name.', fName;cr.
		fileName := self findTextureFileName: 'default.bmp'.
	].
	file := (FileStream readOnlyFileNamed: fileName)binary.
	contents := file contentsOfEntireFile.
	"compute the sha hash"
	sha := TFormManager generateSHA: contents.
	"load the file into a Form"
	tform := textureDictionary at: sha ifAbsentPut:[
		[
		  form := Form fromBinaryStream: (ReadStream on:contents).
		  form ifNotNil:[
			tform := TForm new initialize: sha form: form 
						mipmap: bMM shrinkFit: bSF extension: ext.
		  ] ifNil:[
			Transcript cr; show:'Bad texture file - cannot load'.
			^ nil.
		  ].
		] ifError: [ :msg | 
			Transcript cr; show:'Ignoring texture file: ', msg.
			^ nil.
		].
		tform.
	].
	"copy the original file into the local cache"
	cacheMgr cache: contents to: sha.
	^tform thumb.! !

!TFormManager methodsFor: 'accessing' stamp: 'das 2/10/2006 10:20'!
at: shaName ifAbsentPut: block

	^ textureDictionary at: shaName ifAbsentPut: block.! !

!TFormManager methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:36'!
cacheManager
	^cacheMgr! !

!TFormManager methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:36'!
cacheManager: aCacheMgr
	cacheMgr := aCacheMgr! !

!TFormManager methodsFor: 'initialize' stamp: 'ar 1/30/2006 22:34'!
destroy
	self stopLoaderProcess.! !

!TFormManager methodsFor: 'accessing' stamp: 'ar 2/3/2006 00:10'!
findCount
	^findList size! !

!TFormManager methodsFor: 'loading' stamp: 'das 12/31/2005 15:22'!
findTextureFileName: fName
	| path dir file pn paths files |
	"the following attempts to handle case issues in the various texture file names on Linux for now."
	" Generate a list in paths of ways that people have spelled file and directory names.
	Pending reorganization of the Contents directory!!"
	pn := FileDirectory default.
	paths := OrderedCollection new.
	path := FileDirectory default fullNameFor: fName.
	dir := FileDirectory on: (FileDirectory dirPathFor: path).
	file := FileDirectory localNameFor: path.
	"three forms of the file name: FooBar, foobar and Foobar"
	files := OrderedCollection new.
	files add: file.
	files add: (file := file asLowercase).
	files add: file capitalized.
	"look in file's directory first"
	paths := files collect: [ :fn |
			dir pathFromURI: fn].
	"look in textures subdirectory second"
	#('Textures/' 'textures/') do: [ :casedName |
		paths := paths, (files collect: [ :fn |
			dir pathFromURI: casedName, fn]).
		].
	"if fName has no prefix, look in default Textures directories"
	(FileDirectory dirPathFor: fName) isEmpty ifTrue: [
		#('Textures/' 'textures/') do: [ :casedName |
			paths := paths, (files collect: [ :fn | 
				pn pathFromURI: casedName, fn])].
		#('Content/' 'content/') do: [ :casedDir |
			#('Textures/' 'textures/') do: [ :casedName |
				paths := paths, (files collect: [ :fn |
					pn pathFromURI: casedDir, casedName, fn])]].
		].

    "Also search in top-level Content/Textures directory"
    #('Textures/' 'textures/') do: [ :casedName |
		paths := paths, (files collect: [ :fn |
			pn pathFromURI: 'Content/', casedName, fn]).
		].
   
    "Also search in top-level Textures directory"
    #('Textures/' 'textures/') do: [ :casedName |
		paths := paths, (files collect: [ :fn |
			pn pathFromURI: casedName, fn]).
		].

	"Now search all the paths we constructed"
	paths do: [ :filePath | (FileStream isAFileNamed: filePath) ifTrue:[^filePath] ].
	^ nil.
"	CroquetData logError: 'Substituting default texture for missing texture file: : ', fName.
	^pn pathFromURI: 'Content/Textures/default.bmp'."! !

!TFormManager methodsFor: 'initialize' stamp: 'das 2/11/2006 15:28'!
initialize

	textureDictionary := Dictionary new.
	findList := Dictionary new.
	sortedFindList := SortedCollection sortBlock:[:a :b | a distance <= b distance].
	updateList := OrderedCollection new.
	mutex := TMutex new.
	cacheMgr := TFileCacheManager new.
	loaderSema := Semaphore new.
	loaderDelay := Delay forMilliseconds: 20.
	self startLoaderProcess.! !

!TFormManager methodsFor: 'loader process' stamp: 'ar 1/31/2006 16:09'!
loadNextTFF: tff
	| data controller promise sha |
	"Look in the local file cache"
	self log: tff sha, ': Loading texture...'.
	data := cacheMgr find: tff sha.
	data ifNotNil:[
		"See if cache got corrupted"
		self log: tff sha, ': Found in file cache...'.
		sha := TFormManager generateSHA: data.
		sha = tff sha ifFalse:[
			self log: tff sha, ': Cache invalid (SHA mismatch)...'.
			cacheMgr delete: tff sha.
			data := nil]].
	data ifNotNil:[^self loadNextTFF: tff from: data].
	"Try to get data from a controller"
	controller := TIslandController controllerWithID: tff controllerID.
	controller ifNil:[
		"Try to find another controller later - if this texture is still used we'll get around to it"
		self log: tff sha, ': Controller went away...'.
		mutex critical:[
			findList removeKey: tff sha.
			sortedFindList remove: tff
		].
		^self
	].
	"Ask the controller for the data"
	self log: tff sha, ': Requesting sync...'.
	promise := controller sync: tff sha.
	promise wait.
	data := promise result.
	self log: tff sha, ': Sync complete...'.
	data ifNotNil:["Verify SHA"
		sha := TFormManager generateSHA: data.
		sha = tff sha ifFalse:[
			self log: tff sha, ': Data invalid (SHA mismatch)...'.
			data := nil].
	].
	data ifNil:[^self updateForm: nil tff: tff].
	cacheMgr cache: data to: tff sha.
	self log: tff sha, ': Installing texture...'.
	self loadNextTFF: tff from: data.
! !

!TFormManager methodsFor: 'loader process' stamp: 'ar 1/30/2006 22:57'!
loadNextTFF: tff from: data
	"Load the TForm from the given binary data"
	| form tform |
	form := [Form fromBinaryStream: data readStream] on: Error do:[:ex| ex return: nil].
	tform := TForm new initialize: tff sha form: form 
				mipmap: tff bMipmap shrinkFit: tff bShrinkFit 
				extension: tff extension.
	self updateForm: tform tff: tff.! !

!TFormManager methodsFor: 'loading' stamp: 'ar 1/31/2006 16:08'!
log: aString
	false ifTrue:[
		WorldState addDeferredUIMessage: (MessageSend receiver: Transcript selector: #cr).
		WorldState addDeferredUIMessage: (MessageSend receiver: Transcript selector: #show: argument: aString).
	].
! !

!TFormManager methodsFor: 'accessing' stamp: 'ar 1/31/2006 00:07'!
resolve: thumb distance: dist
	" This returns the real TForm, if it is available. Otherwise, it adds the location info of the TForm to the findList and sortedFindList so that it can be loaded later. There are two search lists, the findList and the sortedFindList. The findList is a simple Dictionary used to determine if we are currently searching for this TForm. The sortedFindList is a SortedCollection sorting on the distance field of the TFormFind. If the thumb sha is not found in the findList then a new TFormFind is created and added to the findList and sortedFindList. If it already exists, then we compare the current distance to the previous TForm distance, and if the current is shorter, we reset this value in the TFormFind and force the sortedFindList to reSort, which will raise the probability that this object gets loaded sooner. If the value of the texture in the textureDictionary is nil, this means that we were unable to find the real texture, so just return the thumb.
DAS"

	| tff tf |
	tf := textureDictionary at: thumb sha ifAbsent:[nil].
	tf ifNotNil:[^tf].
	tff := findList at: thumb sha ifAbsent:[nil].
	tff ifNil:["Create new TFormFind for this texture"
		tff := TFormFind new initialize: thumb distance: dist.
		tff controllerID: Processor activeIsland controllerID.
		mutex critical:[
			findList at: tff sha put: tff.
			sortedFindList add: tff.
			loaderSema signal]].
	tff distance > dist ifTrue:["this texture is not closer to us, increase priority"
		mutex critical:[
			tff distance: dist.
			sortedFindList reSort.
		].
	].
	^thumb! !

!TFormManager methodsFor: 'loader process' stamp: 'ar 1/30/2006 22:48'!
runLoaderProcess
	| tff |
	[true] whileTrue:[
		loaderSema wait.
		mutex critical:[tff := sortedFindList first].
		self loadNextTFF: tff.
	].! !

!TFormManager methodsFor: 'loader process' stamp: 'sn 2/25/2018 13:45'!
startLoaderProcess
	self stopLoaderProcess.
	loaderProcess := self fork: #runLoaderProcess at: Processor userBackgroundPriority.! !

!TFormManager methodsFor: 'loader process' stamp: 'bgf 7/27/2006 16:36'!
step
	"If it needs it, give the loader process a time slice to execute"

	self findCount > 0 ifTrue: [
		loaderDelay wait.
	].! !

!TFormManager methodsFor: 'loader process' stamp: 'ar 1/30/2006 22:34'!
stopLoaderProcess
	loaderProcess ifNotNil:[
		loaderProcess terminate.
		loaderProcess := nil.
	].! !

!TFormManager methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:49'!
substituteForm: aTForm
	"Substitute the given TForm in the form manager"
	mutex critical:[
		textureDictionary at: aTForm sha put: aTForm.
		updateList add: aTForm.
	].! !

!TFormManager methodsFor: 'loader process' stamp: 'ar 2/1/2006 22:49'!
updateForm: aTForm tff: tff
	mutex critical:[
		"The following is written so that TForms that are substituted via #substituteForm: 
		will not be overwritten here if they have not been found anywhere"
		textureDictionary at: tff sha ifAbsentPut:[aTForm].
		findList removeKey: tff sha.
		sortedFindList remove: tff.
		aTForm ifNotNil:[updateList add: aTForm].
	].! !

!TFormManager methodsFor: 'accessing' stamp: 'das 1/5/2006 21:59'!
updateList: ogl
	updateList size= 0 ifTrue:[^self].
	ogl updateTextures: updateList.
	updateList := OrderedCollection new.! !

Object subclass: #TMediaResource
	instanceVariableNames: 'sha'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TMediaResource commentStamp: 'jcg 11/15/2006 09:58' prior: 0!
Base class for resources that can live outside and be shared between islands.!

!TMediaResource methodsFor: 'initialize' stamp: 'jcg 11/14/2006 15:47'!
fakeSHA: fakeSHA
	sha := fakeSHA! !

!TMediaResource methodsFor: 'accessing' stamp: 'jcg 11/14/2006 15:20'!
sha
	^ sha copy! !

Object subclass: #TRenderAlpha
	instanceVariableNames: 'tObject transform distance parent transparency'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TRenderAlpha commentStamp: 'das 10/10/2004 14:05' prior: 0!
This class is used to render the alpha (transparent) objects. As an alpha object is found, a TRenderAlpha object is created that includes:
tObject - the object
transform - the object's OpenGL global transform
distance - the distance from the TCamera used for sorting
parent - its parent frame (if needed - until we remove multiple parents from the system)

DAS!

!TRenderAlpha class methodsFor: 'instance creation' stamp: 'das 6/7/2006 21:54'!
object: obj transform: trans distance: dist parent: par transparency: tr

	^ self new object: obj transform: trans distance: dist parent: par transparency: tr.! !

!TRenderAlpha methodsFor: 'accessing' stamp: ''!
distance

	^ distance.! !

!TRenderAlpha methodsFor: 'accessing' stamp: ''!
object: obj transform: trans distance: dist parent: par

	tObject := obj.
	transform := trans.
	distance := dist.
	parent := par.
	^ self.! !

!TRenderAlpha methodsFor: 'accessing' stamp: 'das 6/7/2006 21:52'!
object: obj transform: trans distance: dist parent: par transparency: transp

	tObject := obj.
	transform := trans.
	distance := dist.
	parent := par.
	transparency := transp.
	^ self.! !

!TRenderAlpha methodsFor: 'accessing' stamp: ''!
parent

	^ parent.! !

!TRenderAlpha methodsFor: 'accessing' stamp: ''!
tObject

	^ tObject.! !

!TRenderAlpha methodsFor: 'accessing' stamp: ''!
transform

	^ transform.! !

!TRenderAlpha methodsFor: 'accessing' stamp: 'das 6/7/2006 21:56'!
transparency

	^ transparency.! !

Object subclass: #TSelection
	instanceVariableNames: 'target frame framePosition portal portalPoint parent distance distanceSquared index normal point globalPoint triangle cameraTransform rayTransform events scale'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TSelection commentStamp: '<historical>' prior: 0!
TSelection is a holder for data generated by the ray testing in Croquet. See TRay.

target - this is the frame object owner
frame - this is the frame object that is selected.
framePosition - this is the ray location relative to the selected object.
parent - this is the parent of the selectedFrame at the time of rendering/selection.
parentCameraTransform - this is the transform of the camera frame at selection time
distance - this is the distance from the ray position.
distanceSquared - this is the squared distance from the ray position.
index - this may be used by the selectedObject. It is usually the surface index.
normal - this is the surface normal at the selectedPoint.
point - this is the selected point in the local selectedFrame reference.
triangle - used by TSelectionTracker
cameraTransform - the camera global transform at the last time of selection.
scale - the current scale at which we picked this object. Used by TPortal3D.

DAS!

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:29'!
cameraTransform

	^ cameraTransform! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
cameraTransform: ct

	cameraTransform := ct! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
distance

	^ distance! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
distance: dist

	distance := dist.
	distanceSquared := distance*distance.! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
distanceSquared

	^ distanceSquared! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
distanceSquared: distSq

	distanceSquared := distSq.
	distance := distanceSquared sqrt.! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/27/2006 14:22'!
eventMask

	^ events! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/27/2006 14:23'!
eventMask: ev

	events := ev.! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
frame
	^frame! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
frame: aFrame
	frame := aFrame! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
framePosition
	^framePosition! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
framePosition: aPoint
	framePosition := aPoint! !

!TSelection methodsFor: 'accessing' stamp: 'das 2/3/2006 22:41'!
globalPoint

	^ globalPoint! !

!TSelection methodsFor: 'accessing' stamp: 'das 2/3/2006 22:41'!
globalPoint: aVector

	globalPoint := aVector! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
index
	^index! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
index: aNumber
	index := aNumber! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
normal
	^normal! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:30'!
normal: aVector
	normal := aVector! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
parent
	^parent! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
parent: aFrame
	parent := aFrame! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
point
	^point! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
point: aVector
	point := aVector! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
portal

	^ portal! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
portal: prtl

	portal := prtl! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
portalPoint

	^ portalPoint! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
portalPoint: prtlPnt

	portalPoint := prtlPnt! !

!TSelection methodsFor: 'accessing' stamp: 'das 1/15/2006 12:43'!
rayTransform

	^ rayTransform.! !

!TSelection methodsFor: 'accessing' stamp: 'das 1/15/2006 12:43'!
rayTransform: rt

	rayTransform := rt.! !

!TSelection methodsFor: 'accessing' stamp: 'das 3/26/2006 16:05'!
scale
	
	^ scale.! !

!TSelection methodsFor: 'accessing' stamp: 'das 3/26/2006 15:47'!
scale: scl
	
	scale := scl.! !

!TSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:17'!
target

	^ target! !

!TSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:16'!
target: aTarget

	target := aTarget! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
triangle
	^triangle! !

!TSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:31'!
triangle: anArray
	triangle := anArray! !

Object subclass: #TTask
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TTask commentStamp: 'das 4/24/2006 17:25' prior: 0!
A TTask is an interface between the users generation of an event and the actual interpretation and utilization of that event. A TTask has a relatively short lifetime - typically defined between a #mouseDown/#pointerDown and a #mouseUp/#pointerUp, or a keyboard event. The default TTaskMain manages events in the normal way.!

!TTask methodsFor: 'events' stamp: 'das 4/23/2006 15:39'!
doKeyDown: event! !

!TTask methodsFor: 'events' stamp: 'das 4/23/2006 15:39'!
doKeyScan: event! !

!TTask methodsFor: 'events' stamp: 'das 4/23/2006 15:39'!
doKeyUp: event! !

!TTask methodsFor: 'events' stamp: 'das 4/21/2006 20:02'!
doPointerDown: event! !

!TTask methodsFor: 'events' stamp: 'das 4/24/2006 18:41'!
doPointerEnter: event! !

!TTask methodsFor: 'events' stamp: 'das 4/24/2006 18:41'!
doPointerLeave: event! !

!TTask methodsFor: 'events' stamp: 'das 4/21/2006 20:02'!
doPointerMove: event! !

!TTask methodsFor: 'events' stamp: 'das 4/24/2006 18:41'!
doPointerOver: event! !

!TTask methodsFor: 'events' stamp: 'das 4/21/2006 20:02'!
doPointerUp: event! !

!TTask methodsFor: 'controls' stamp: 'das 4/23/2006 16:43'!
forcePick
" we only want to select items that are actually selectable"
	^ false.! !

!TTask methodsFor: 'controls' stamp: 'das 4/23/2006 17:20'!
noPickWhileDown

	"perform a pick even when the pointer is down. This allows us to span across objects if we need to drag or draw something."
	^ true.! !

TTask subclass: #TTaskMain
	instanceVariableNames: 'harness'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TTaskMain commentStamp: 'das 4/24/2006 17:26' prior: 0!
TTaskMain is the default TTask for managing events. Basically, all it does is forwards the events to the target object.!

!TTaskMain methodsFor: 'events' stamp: 'das 4/23/2006 16:02'!
doKeyDown: event

	| target |
	(event selection eventMask anyMask: TFrame eventKeyboard) ifTrue:[
		target := event selection target.
		target ifNotNil:[
			event selection target: nil.
			target futureDo: #doKeyDown: at:0 args:{event}.
			event selection target: target.
			^ true.
		].
	].
	^ false.! !

!TTaskMain methodsFor: 'events' stamp: 'das 4/23/2006 16:03'!
doKeyStroke: event

	| target |
	(event selection eventMask anyMask: TFrame eventKeyboard) ifTrue:[
		target := event selection target.
		target ifNotNil:[
			event selection target: nil.
			target futureDo: #doKeyStroke: at:0 args:{event}.
			event selection target: target.
			^ true.
		].
	].
	^ false.! !

!TTaskMain methodsFor: 'events' stamp: 'das 4/23/2006 16:03'!
doKeyUp: event

	| target |
	(event selection eventMask anyMask: TFrame eventKeyboard) ifTrue:[
		target := event selection target.
		target ifNotNil:[
			event selection target: nil.
			target futureDo: #doKeyUp: at:0 args:{event}.
			event selection target: target.
			^ true.
		].
	].
	^ false.! !

!TTaskMain methodsFor: 'events' stamp: 'das 8/10/2006 17:58'!
doPointerDown: event

	| target oTarget |

	oTarget := target := event selection target.
	target ifNil:[
		target := harness avatar replica.
		event selection eventMask: (TFrame eventPointerDown bitOr: TFrame eventBlockRecurse).
	].
 
	(event selection eventMask anyMask: TFrame eventPointerDown) ifTrue:[
		event selection target: nil.
		target futureDo: #doPointerDown: at:0 args:{event}.
	].
	event selection target: oTarget.
! !

!TTaskMain methodsFor: 'events' stamp: 'das 4/23/2006 16:21'!
doPointerEnter: event

	| target |
	(event selection eventMask anyMask: TFrame eventPointerOver) ifTrue:[
		target := event selection target.
		target ifNotNil:[
			event selection target: nil.
			target futureDo: #doPointerEnter: at:0 args:{event}.
			event selection target: target.
			^ true.
		].
	].
	^ false.! !

!TTaskMain methodsFor: 'events' stamp: 'das 4/23/2006 16:19'!
doPointerLeave: event

	| target |
	(event selection eventMask anyMask: TFrame eventPointerOver) ifTrue:[
		target := event selection target.
		target ifNotNil:[
			event selection target: nil.
			target futureDo: #doPointerLeave: at:0 args:{event}.
			event selection target: target.
			^ true.
		].
	].
	^ false.! !

!TTaskMain methodsFor: 'events' stamp: 'das 4/23/2006 16:06'!
doPointerMove: event

	| target |
	(event selection eventMask anyMask: TFrame eventPointerDown) ifTrue:[
		target := event selection target.
		target ifNotNil:[
			event selection target: nil.
			target futureDo: #doPointerMove: at:0 args:{event}.
			event selection target: target.
			^ true.
		].
	].
	^ false.! !

!TTaskMain methodsFor: 'events' stamp: 'das 4/23/2006 16:21'!
doPointerOver: event

	| target |
	(event selection eventMask anyMask: TFrame eventPointerOver) ifTrue:[
		target := event selection target.
		target ifNotNil:[
			event selection target: nil.
			target futureDo: #doPointerOver: at:0 args:{event}.
			event selection target: target.
			^ true.
		].
	].
	^ false.! !

!TTaskMain methodsFor: 'events' stamp: 'das 8/10/2006 17:58'!
doPointerUp: event

	| target oTarget |

	oTarget := target := event selection target.
	target ifNil:[
		target := harness avatar replica.
		event selection eventMask: (TFrame eventPointerDown bitOr: TFrame eventBlockRecurse).
	].

	(event selection eventMask anyMask: TFrame eventPointerDown) ifTrue:[
		event selection target: nil.
		target futureDo: #doPointerUp: at:0 args:{event}.
	].
	event selection target: oTarget.
! !

!TTaskMain methodsFor: 'controls' stamp: 'das 4/23/2006 16:33'!
forcePick
" we only want to select items that are actually selectable"
	^ false.! !

!TTaskMain methodsFor: 'accessing' stamp: 'das 8/10/2006 10:43'!
harness: hrns

	harness := hrns.! !

Object subclass: #TXSelection
	instanceVariableNames: 'target frameID frame framePosition portal portalPoint parent distance distanceSquared index normal point globalPoint triangle cameraTransform rayTransform events scale'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TXSelection commentStamp: 'das 4/3/2006 01:53' prior: 0!
TXSelection acts a selection data holder for TXRay. See description of TSelection.

Same as TSelection except we ensure that we never have access to an object that is inside of an Island. TXSelection (eXternal Selection) either has a copy or a has a reference ID to the original object.!

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
cameraTransform

	^ cameraTransform copy! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
cameraTransform: ct

	cameraTransform := ct copy.! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
distance

	^ distance copy! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
distance: dist

	distance := dist copy.
	distanceSquared := distance*distance.! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
distanceSquared

	^ distanceSquared copy! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
distanceSquared: distSq

	distanceSquared := distSq copy.
	distance := distanceSquared sqrt.! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/27/2006 14:22'!
eventMask

	^ (events ifNil:[0]) copy! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/27/2006 14:23'!
eventMask: ev

	events := ev copy.! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
frame

	^frame! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
frame: aFrame

	frame := aFrame! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
frameID

	^frameID copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
frameID: aFrameID

	frameID := aFrameID copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
framePosition

	^framePosition copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
framePosition: aPoint

	framePosition := aPoint copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 2/3/2006 22:42'!
globalPoint

	^ globalPoint.! !

!TXSelection methodsFor: 'accessing' stamp: 'das 2/3/2006 22:42'!
globalPoint: aVector

	globalPoint := aVector! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
index

	^index copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
index: aNumber

	index := aNumber copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
normal

	^normal copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
normal: aVector

	normal := aVector copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:24'!
parent

	^parent copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:25'!
parent: aFrame

	parent := aFrame copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:25'!
point

	^point copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:25'!
point: aVector

	point := aVector copy! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
portal

	^ portal copy! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
portal: prtl

	portal := prtl copy! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
portalPoint

	^ portalPoint copy! !

!TXSelection methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:28'!
portalPoint: prtlPnt

	portalPoint := prtlPnt copy! !

!TXSelection methodsFor: 'printing' stamp: 'ar 2/2/2006 15:21'!
printOn: aStream
	| ivNames |
	aStream nextPutAll: self class name; nextPut: $(.
	ivNames := self class allInstVarNames.
	1 to: ivNames size do:[:i|
		aStream crtab; nextPutAll: (ivNames at: i); nextPutAll:': '.
		aStream print: (self instVarAt: i).
	].
	aStream cr; nextPut: $).
! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/15/2006 12:42'!
rayTransform

	^rayTransform copy.! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/15/2006 12:41'!
rayTransform: rt

	rayTransform := rt copy.! !

!TXSelection methodsFor: 'accessing' stamp: 'das 3/26/2006 16:05'!
scale

	^ scale.! !

!TXSelection methodsFor: 'accessing' stamp: 'das 3/26/2006 15:47'!
scale: scl

	scale := scl.! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:18'!
target

	^ target copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:18'!
target: aTarget

	target := aTarget copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:25'!
triangle

	^triangle copy! !

!TXSelection methodsFor: 'accessing' stamp: 'das 1/21/2006 15:25'!
triangle: anArray

	triangle := anArray copy! !

Object subclass: #TBoundSphere
	instanceVariableNames: 'localPosition globalPosition radius radiusSquared children vertices frame box normal up side offset'
	classVariableNames: ''
	poolDictionaries: 'OpenGLConstants'
	category: 'Croquet-Objects'!

!TBoundSphere commentStamp: 'das 10/10/2004 10:19' prior: 0!
This is the main bounds test object. It is used for object culling against view planes, ray testing, and for collision detection. It is defined heirarchically for the collision detection and object picking algorithms. At the moment, this is one of the slower objects to calculate with complex meshes. We need to turn much of this code into primitives. Also, there are a number of instance variables that may go away or be used in a slightly different way. 

Currently, the TBoundSphere global position is set at render time. This is because this is a more efficient method and it also presumes that a TFrame may be instanced in a scene (that is, it may appear as the child of more than one object). If we move to a seperate pass for determining ray collisions, then we will no longer be able to instance TFrames, and the global positions of the TBoundSpheres will need to be set by the TFrame when it's own globalPosition is updated.

Other instance variables are:

radius - the radius of the sphere, which should include all of the associated TFrame contents.
radiusSquared - the square radius, here for performance.
children - complex objects often have a need for a heirarchical collection of bound spheres for quick collision tests. These children are also TBoundSpheres.
vertices - if the contained object is a TMesh, then the leaf children contain copies of the vertex information.
box - in computing the bound sphere hierarchy, we utilize TBoxes in an octree subdivision. We keep these around for now, though they can probably be tossed out.
normal - this is to be used by the hierarchical collision detection method. This is the best fit normal to the surface vertices contained in the sphere.
up and side are the two perpendicular vectors to the normal. 
offset - to be used by the hierarchical collision method, this is the offset of the normal from the sphere center representing the closest fit to the contained faces.

DAS
	

!

!TBoundSphere class methodsFor: 'instance creation' stamp: 'das 4/5/2002 10:39'!
calcBoundSphere: verts

	 ^ TBoundSphere new calcBoundSphere: verts.! !

!TBoundSphere class methodsFor: 'instance creation' stamp: 'das 4/24/2002 12:15'!
calcBoundSphere: verts faces: faces

	 ^ TBoundSphere new calcBoundSphere: verts faces: faces.! !

!TBoundSphere class methodsFor: 'instance creation' stamp: 'das 4/25/2002 20:36'!
calcTree: vertices faces: faces box: bx depth: depth

	^ TBoundSphere new calcTree: vertices faces: faces box: bx depth: depth.! !

!TBoundSphere class methodsFor: 'instance creation' stamp: 'das 4/24/2002 12:18'!
calcTree: vertices faces: faces depth: depth

	^ TBoundSphere new calcTree: vertices faces: faces depth: depth.! !

!TBoundSphere class methodsFor: 'instance creation' stamp: ''!
localPosition: pos radius: rad

	^ TBoundSphere new localPosition: pos radius: rad.
! !

!TBoundSphere class methodsFor: 'instance creation' stamp: 'das 6/23/2002 12:26'!
mtfBall: verts

	 ^ self new mtfBall: verts.! !

!TBoundSphere class methodsFor: 'instance creation' stamp: 'das 6/23/2002 12:26'!
mtfBall: verts faces: faces

	 ^ self new mtfBall: verts faces: faces.! !

!TBoundSphere methodsFor: 'access' stamp: 'das 4/24/2002 12:25'!
addChild: child

	children ifNil:[children := OrderedCollection new.].
	children add: child.! !

!TBoundSphere methodsFor: 'access' stamp: 'das 1/24/2005 01:02'!
box

	box ifNil:[
		box := TBox min: (radius@radius@radius) negated max: (radius@radius@radius).].
	^ box.! !

!TBoundSphere methodsFor: 'access' stamp: 'das 4/25/2002 15:55'!
box: bx

	box := bx.! !

!TBoundSphere methodsFor: 'construct' stamp: 'das 5/16/2005 08:52'!
calcBoundSphere: verts

	|  iP iN xmx xmn ymx ymn zmx zmn xspan yspan zspan maxSpan d1 d2 center radSq rad |
	
" This routine is from the Graphics Gems 1, C code on page 723. verts is a Vector3Array. It constructs a bounding sphere from the vertices."

	iP := Float infinity.
	iN := Float infinity negated.

	xmx := Vector3 x: iN y: iN z: iN.
	ymx := Vector3 x: iN y: iN z: iN.
	zmx := Vector3 x: iN y: iN z: iN.
	xmn := Vector3 x: iP y: iP z: iP.
	ymn := Vector3 x: iP y: iP z: iP.
	zmn := Vector3 x: iP y: iP z: iP.

" First pass, find a pair of maximally distant points between each axis."
	verts do:[ :v |
		v x > xmx x ifTrue:[ xmx := v].
		v x < xmn x ifTrue:[ xmn := v].
		v y > ymx y ifTrue:[ ymx := v].
		v y < ymn y ifTrue:[ ymn := v].
		v z > zmx z ifTrue:[ zmx := v].
		v z < zmn z ifTrue:[ zmn := v].].

	xspan := (xmx-xmn) squaredLength.
	yspan := (ymx-ymn) squaredLength.
	zspan := (zmx-zmn) squaredLength.
		
	maxSpan := xspan.
	d1 := xmn.
	d2 := xmx.
	yspan > maxSpan ifTrue:[maxSpan := yspan. d1 := ymn. d2 := ymx.].
	zspan > maxSpan ifTrue:[maxSpan := zspan. d1 := zmn. d2 := zmx.].
	
	center := (d1 + d2)/2.0.
	radSq := (d1 - center) squaredLength.
	rad := radSq sqrt.
	
"Second pass, expand sphere to include outlying points."

	verts do:[ :v |
		d1 := (v-center)squaredLength.
		d1 > radSq ifTrue:[
			d1 := d1 sqrt.
			rad := (d1 + rad)/2.0.
			radSq := rad*rad.
			d2 := d1 - rad.
			center := ((rad * center) + (d2*v))/d1.
			].
		].

	localPosition := center.
	radius := rad.
	radiusSquared := radSq.
"	vertices := verts. " "don't need this and it screws up other things later."
	^  self.! !

!TBoundSphere methodsFor: 'construct' stamp: 'das 5/16/2005 08:52'!
calcBoundSphere: verts faces: faces

	|  v iP iN xmx xmn ymx ymn zmx zmn xspan yspan zspan maxSpan d1 d2 center radSq rad |
	
" This routine is from the Graphics Gems 1, C code on page 723. verts is a Vector3Array. It constructs a bounding sphere based upon the faces and vertices."

	iP := Float infinity.
	iN := Float infinity negated.

	xmx := Vector3 x: iN y: iN z: iN.
	ymx := Vector3 x: iN y: iN z: iN.
	zmx := Vector3 x: iN y: iN z: iN.
	xmn := Vector3 x: iP y: iP z: iP.
	ymn := Vector3 x: iP y: iP z: iP.
	zmn := Vector3 x: iP y: iP z: iP.

" First pass, find a pair of maximally distant points between each axis."
	faces do:[ :f |
		v := verts at:(f+1).
		v x > xmx x ifTrue:[ xmx := v].
		v x < xmn x ifTrue:[ xmn := v].
		v y > ymx y ifTrue:[ ymx := v].
		v y < ymn y ifTrue:[ ymn := v].
		v z > zmx z ifTrue:[ zmx := v].
		v z < zmn z ifTrue:[ zmn := v].].

	xspan := (xmx-xmn) squaredLength.
	yspan := (ymx-ymn) squaredLength.
	zspan := (zmx-zmn) squaredLength.
		
	maxSpan := xspan.
	d1 := xmn.
	d2 := xmx.
	yspan > maxSpan ifTrue:[maxSpan := yspan. d1 := ymn. d2 := ymx.].
	zspan > maxSpan ifTrue:[maxSpan := zspan. d1 := zmn. d2 := zmx.].
	
	center := (d1 + d2)/2.0.
	radSq := (d1 - center) squaredLength.
	rad := radSq sqrt.
	
"Second pass, expand sphere to include outlying points."

	faces do:[ :f |
		v := verts at: (f+1).
		d1 := (v-center)squaredLength.
		d1 > radSq ifTrue:[
			d1 := d1 sqrt.
			rad := (d1 + rad)/2.0.
			radSq := rad*rad.
			d2 := d1 - rad.
			center := ((rad * center) + (d2*v))/d1.
			].
		].

	localPosition := center.
	radius := rad.
	radiusSquared := radSq.
	^  self.! !

!TBoundSphere methodsFor: 'construct' stamp: 'das 5/16/2005 09:10'!
calcSurface: verts faces: faces

"This is for a first approximation - I may need a better solution later."

	| n nabs m nlist ray t |

	normal := Vector3 new.
	nlist := OrderedCollection new.

	1 to: faces size by: 3 do:[ :i |
		
		n := 
		(((verts at:1+(faces at: i)) - (verts at:1+(faces at: i+1))) cross:
		(((verts at:1+(faces at: i)) - (verts at:1+(faces at: i+2))))).
		n squaredLength = 0.0 ifFalse:[
			nlist add: n.
			normal := normal + n. ].].

	normal := normal/( faces size/3).
	normal squaredLength = 0.0 ifTrue:[normal := Vector3 x: 0.0 y: 1.0 z: 0.0].
	normal normalize.
	nabs := normal abs.
	up := Vector3 new.

	nabs x < nabs y ifTrue:[
		nabs x < nabs z ifTrue:[up x:1.0] 
		ifFalse:[up z:1.0].]ifFalse:[
		nabs y < nabs z ifTrue:[up y: 1.0] ifFalse:[up z: 1.0]].
	side := (up cross: normal) normalized.
	up := (normal cross: side) normalized.

	m := Matrix4x4 identity.
	
	m at: 1 at: 1 put: side x.
	m at: 1 at: 2 put: side y.
	m at: 1 at: 3 put: side z.
	m at: 2 at: 1 put: up x.
	m at: 2 at: 2 put: up y.
	m at: 2 at: 3 put: up z.
	m at: 3 at: 1 put: normal x.
	m at: 3 at: 2 put: normal y.
	m at: 3 at: 3 put: normal z.

	offset := 0.
	1 to: faces size by: 3 do:[ :i |
		ray := self ray: m tri: localPosition - (verts at: 1+(faces at:i))
					tri: localPosition - (verts at: 1+(faces at: i+1))
					tri: localPosition - (verts at: 1+(faces at: i+2)).
		
		ray ifNotNil:[ 
			nabs x > nabs y ifTrue:[
				nabs x > nabs z ifTrue:[t := ray x/normal x] 
				ifFalse:[ t := ray z/normal z].]ifFalse:[
				nabs y > nabs z ifTrue:[t := ray y/normal y] ifFalse:[t := ray z/normal z]].


		t abs > offset abs ifTrue:[ offset := t negated.]]].! !

!TBoundSphere methodsFor: 'construct' stamp: 'das 5/21/2006 13:48'!
calcTree: verts faces: faces box: bx depth: depth

	| child faceBox boxes flist |

" See TBoundSphere >>#calcTree:faces:depth for more info."

	vertices := nil.
	(depth = 0 or:[ faces size <= 12])ifFalse:[
	"depth = 0 ifFalse:["
		boxes := self splitBoxes: bx.
" Calculate the list of faces that overlap each box, and recurse."
		boxes do:[:b |
			flist := OrderedCollection new.
			1 to: faces size by:3 do:[:i |
				(b intersectV1: (verts at:1+(faces at: i)) 
					V2:(verts at: 1+(faces at: i+1)) 
					V3:(verts at: 1+(faces at: i+2))) ifTrue:[
						flist add: (faces at: i).
						flist add: (faces at: i+1).
						flist add: (faces at: i+2).].].
			flist size > 1 ifTrue:[
				faceBox := TBox new.
				flist do:[:f | faceBox growVertex: (verts at: 1+f).].
				faceBox := faceBox intersectBox: b.
				child := (TBoundSphere calcTree: verts faces: flist box: faceBox depth: depth-1).
				child children ifNotNil:[self addChild: child].					
				].].]
	ifTrue:[
		vertices := verts. 
		children := faces asIntegerArray. 
		].

" Now calculate the minimal bounding sphere. We do it this way because the polygons actually overlap between boxes. If we were to use the mtfBall approach, the resulting radius could well be larger than this one."
	self box: bx.
	self localPosition: bx center radius: bx diagonal/2.0.
	vertices ifNotNil:[self calcSurface: verts faces: faces.].

				
		

	! !

!TBoundSphere methodsFor: 'construct' stamp: 'das 5/16/2005 08:51'!
calcTree: verts faces: faces depth: depth

	| bx epsilon |

" 
CONSTRUCTION FOR SPHERE HIERARCHY

This method is used to compute a heirarchy of bounding spheres for a complex object. This allows for extremely fast picking and collision detection. The object is broken into octrees if the ratio of sides permits. Long thin objects tend to be split into two parts, flat objects into four.

	1.	This is a top down approach. I use aligned boxes (TBox) as the basis of the
		construction, given the assumption that when the boxes are small enough
		their shape and relative orientation doesn't matter as long as they
		completely cover the object. We will construct spheres from these boxes
		(or cubes) that only contain the contents of the cubes, hence, overlap
		is not much of an issue.
	2.	Once the top level containing box is calculated from the underlying objects
		vertices (TBox>>#max is the maximum value corner and TBox>>#min is the
		minimal value corner), we slice the object into smaller boxes and recurse.
		This slicing is only done along the 'longer' dimensions, hence a long
		thin object may only be sliced once perpendicular to its long axis, 
		a flat object may only be sliced twice each perpendicular to each other.
		We determine whether a dimension is 'sliceable' based upon the ratio of
		its length relative to the other dimensions.
	3.	The recursion first determines which surfaces of the containing box 
		overlap the resulting bounding box. These resulting values are stored as 
		an array for use in the same way by the children of this box.
	4.	We test the depth of the recursion at this stage. If the depth is zero or
		the number of surfaces that we overlap is zero we stop the recursion. If
		this cube does not contain any surfaces its parent will delete it.
	5.	Inside the recursion, we determine which of the set of faces overlap
		this box. Once we know this, we calculate the bounding box of these
		surfaces. We then take the intersection between this box and the 
		previously constructed box from the parent. This can only be the same
		size or smaller, which improves the accuracy of the bounding box 
		representation of the object and further improves its value in 
		collision detection.
	6.	Once we construct the child cubes, we calculate the minimal bounding boxes
		of these in attempt to make the current cube size even smaller.
	7.	The last thing to do here is if this cube covers four faces or less, we
		shouldn't bother testing the children, and go directly to these surfaces.
		We set this cube to be a leaf node and destroy all of the children.
	8.	And now, the things we have been waiting for. We construct the sphere
		tree from the cube tree recursively (of course). The boolean leaf value indicates
		whether or not this is a leaf sphere, hence its children are faces.


We use the TBox class to help us out here. The TBox class defines a min and max corner and has a number of useful methods to simplify the code here."

	bx := TBox new.
	faces do:[:f| bx growVertex: (verts at: 1+f).].
" Grow by epsilon to make sure the box has some depth - otherwise, it will fail to construct good boxes."
	epsilon := 0.000001.
	bx growVertex: bx min - (Vector3 x: epsilon y: epsilon z: epsilon).
	bx growVertex: bx max + (Vector3 x: epsilon y: epsilon z: epsilon).
	self calcTree: verts faces: faces box: bx depth: depth.				
	globalPosition := Vector3 new.

	! !

!TBoundSphere methodsFor: 'access' stamp: 'das 4/24/2002 12:25'!
children

	^ children.! !

!TBoundSphere methodsFor: 'access' stamp: 'das 4/27/2002 15:25'!
children: ch

	children := ch.! !

!TBoundSphere methodsFor: 'collision' stamp: 'das 5/16/2005 08:57'!
collideFloor: floor transform: trans

	| collideLoc rval vrt |
	
"	(trans localPointToGlobal: localPosition) y - floor > radius ifTrue:[ ^ nil. ]."

	collideLoc := Vector4 x: 0.0 y: Float infinity z: 0.0 w:0.0.

	vertices ifNil:[ 
		children do:[ :c | 
			(trans localPointToGlobal: c localPosition) y - floor > c radius ifFalse:[
				rval := c collideFloor: floor transform: trans. 
				rval ifNotNil:[ collideLoc y > rval y ifTrue:[collideLoc := rval.].].].].
			collideLoc y <= floor ifTrue:[^ collideLoc.].
			^ nil.
	] ifNotNil: [
		children ifNil:[
			vertices do:[:v | 
				vrt := trans localPointToGlobal: v. 
				vrt y < floor ifTrue:[ vrt y < collideLoc y ifTrue:[ 
					collideLoc := vrt asVector4. 
					collideLoc w: vrt y - floor].].].] 		
		ifNotNil:[
			children do:[:f | 
				vrt := trans localPointToGlobal: (vertices at: (f+1)).
				vrt y < floor ifTrue:[ vrt y < collideLoc y ifTrue:[ 
					collideLoc := vrt asVector4. 
					collideLoc w: vrt y - floor.].].].].].
	collideLoc y <= floor ifTrue:[ ^ collideLoc.].
	^ nil.
! !

!TBoundSphere methodsFor: 'collision' stamp: 'das 5/16/2005 08:57'!
collidePlane: norm offset: offst 

	| current |
" The plane is defined in the local coordinate frame - no transforms of vertices are necessary. Compare this to #collideFloor:transform: where the plane is the floor, and the test vertices must be transformed into global coordinate space."

	current := Vector4 new.
	current w: Float infinity.

	current := self collidePlane: norm offset: offst current: current.
	current w = Float infinity ifTrue:[^ nil. ].
	^ current.
! !

!TBoundSphere methodsFor: 'collision' stamp: 'das 5/16/2005 08:57'!
collidePlane: norm offset: offst current: current

	| distance rval vrt |
	
" Presumably, we already know that the current sphere intersects this plane - we just want to test its contents."

	rval := current.
	vertices ifNil:[ 
		children do:[ :c | 
			(c localPosition dot: norm)-offst < c radius ifTrue:[
				rval := c collidePlane: norm offset: offst current: rval.].].
			^ rval.
	] ifNotNil: [
		children ifNil:[
			vertices do:[:v | 
				distance := (v dot: norm) - offst. 
				distance < 0.0 ifTrue:[
					distance < rval w ifTrue:[ rval := v asVector4. rval w: distance.].].
				].] 		ifNotNil:[
			children do:[:f | 
				vrt := vertices at: (f+1).
				distance := (vrt dot: norm) - offst.
				distance < 0.0 ifTrue:[
					distance < rval w ifTrue:[ rval :=  vrt asVector4. rval w: distance.].].
				].].].
	^ rval.
! !

!TBoundSphere methodsFor: 'collision' stamp: 'ar 2/11/2003 18:44'!
collideSphere: aSphere
	"Answer whether I collilde into aSphere."
	| myPos itsPos dist pt |
	"Note: For some reason the global position of some spheres is broken. E.g.,
		(frame globalTransform localPointToGlobal: localPosition) distanceTo: globalPosition.
	is NOT zero."
	myPos := self frame globalTransform localPointToGlobal: localPosition.
	itsPos := aSphere frame globalTransform localPointToGlobal: localPosition.
	dist := myPos - itsPos.
	dist := dist dot: dist.
	dist > (self radiusSquared + aSphere radiusSquared) ifTrue:[^nil]. "no intersection"
	"The global spheres intersect. If I have any local spheres, then recurse into them."
	normal ifNil:[ children ifNotNil:[ children size > 0 ifTrue:[
		children do:[:each|
			pt := each collideSphere: aSphere.
			pt ifNotNil:[^pt].
		].
		^nil.
	]]].
	"I am a leaf sphere"
	aSphere isLeafSphere ifTrue:[
		"It's a leaf sphere as well. Answer an approximate collision point (not accurate!!)"
		^(self globalPosition + aSphere globalPosition) * 0.5.
	].
	"Otherwise have the other sphere collide"
	^aSphere collideSphere: self.! !

!TBoundSphere methodsFor: 'access' stamp: 'das 4/27/2002 15:25'!
frame

	^ frame.! !

!TBoundSphere methodsFor: 'construct' stamp: 'das 5/6/2002 16:00'!
frame: frm

	frame := frm.
	children ifNotNil:[ vertices ifNil:[ 
		children do:[:c | c frame: frm.].].].! !

!TBoundSphere methodsFor: 'access' stamp: 'das 4/27/2002 15:25'!
globalPosition

	^ globalPosition.! !

!TBoundSphere methodsFor: 'access' stamp: 'ar 2/11/2003 18:12'!
isLeafSphere
	"Answer whether I am a leaf sphere"
	^normal notNil! !

!TBoundSphere methodsFor: 'access' stamp: ''!
localPosition

	^ localPosition.! !

!TBoundSphere methodsFor: 'access' stamp: ''!
localPosition: pos

	localPosition := pos.! !

!TBoundSphere methodsFor: 'access' stamp: 'das 5/16/2005 08:52'!
localPosition: pos radius: rad

	globalPosition := Vector3 new.
	self localPosition: pos.
	self radius: rad.
! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'das 2/20/2004 11:01'!
mtfBall: aCollection
	"Create a minimum enclosing sphere from aCollection of points.
	See Emo Welzl, 'Smallest enclosing disks (balls and ellipsoids)' (1991),
	H. Maurer (Ed.), New Results and New Trends in Computer Science, LNCS 555
	http://citeseer.nj.nec.com/welzl91smallest.html"
	| border idxList |
	idxList := WordArray new: aCollection size.
	1 to: idxList size do:[:i| idxList at: i put: i].
	self privateMtfRandomize: idxList.
	border := Array new: 4.
	self privateMtfBall0: aCollection with: idxList with: 1 with: border.
	radius := radiusSquared sqrt.

	box := TBox new.
	aCollection do:[ :v | box growVertex:(v).].! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'ar 6/23/2002 15:09'!
mtfBall: aCollection faces: faceCollection
	"Create a minimum enclosing sphere from aCollection of points.
	See Emo Welzl, 'Smallest enclosing disks (balls and ellipsoids)' (1991),
	H. Maurer (Ed.), New Results and New Trends in Computer Science, LNCS 555
	http://citeseer.nj.nec.com/welzl91smallest.html"
	| border idxList |
	idxList := WordArray new: faceCollection size.
	1 to: idxList size do:[:i| idxList at: i put: (faceCollection at: i) + 1].
	self privateMtfRandomize: idxList.
	border := Array new: 4.
	self privateMtfBall0: aCollection with: idxList with: 1 with: border.
	radius := radiusSquared sqrt.! !

!TBoundSphere methodsFor: 'render' stamp: 'tak 10/15/2004 20:57'!
pick: pointer

	| rval |

" There is a problem here. I do not have the norm for these faces, which means that this will
 allow us to pick a triangle on the wrong side of the object. This must be fixed.

Another thing that needs to be looked at is just because we find a triangle to pick does NOT mean that there isn't a closer one. This code exits on the very first triangle - but I am leaving it -as is- for performance reasons. To fix it, look at the next 'clause', where I had the same problem with groups of polys."

	(pointer pickLocalBoundSphere: self) ifFalse:[^false].
	vertices ifNotNil:[^pointer pickTriangles: vertices list: children].
	rval := false.
	children ifNotNil: [children do:[:c | rval := (c pick: pointer) or:[ rval]. ]].
	^rval.		
! !

!TBoundSphere methodsFor: 'render' stamp: 'ar 1/2/2003 02:28'!
pickChildren: pointer

	| rval |

" This method exists to avoid testing the top-most sphere twice. It is just a special case of the #pick: method. Notice the commented out code."

"	(pointer pickLocalBoundSphere: self)ifTrue:["
	vertices ifNotNil:[^pointer pickTriangles: vertices list: children].
	rval := false.
	children do:[:c | rval := (c pick: pointer) or:[ rval]. ].
	^ rval! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'das 5/16/2005 08:52'!
privateMtfBall0: pointList with: indexList with: minIndex with: borderList
	"Note: In this variant the default recursion path is inlined. The recursion exclusively affects changes of the border vertices so its depth is bounded by the max. number of vertices on the border (which is four)."
	| idx p dist2 index |
	localPosition := Vector3 new.
	radiusSquared := -1.0.
	index := indexList size.
	[index >= minIndex] whileTrue:[
		idx := indexList at: index.
		p := pointList at: idx.
		p := p - localPosition.
		dist2 := (p dot: p) * 0.999. "numerical accuracy"
		dist2 <= radiusSquared ifFalse:[
			p := pointList at: idx.
			borderList at: 1 put: p.
			self privateMtfBall1: pointList with: indexList with: index+1 with: borderList.
			"Move p to front"
			indexList replaceFrom: index to: indexList size-1 with: indexList startingAt: index+1.
			indexList at: indexList size put: idx.
		].
		index := index - 1.
	].! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'ar 2/7/2005 09:21'!
privateMtfBall1: pointList with: indexList with: minIndex with: borderList
	"Note: In this variant the default recursion path is inlined. The recursion exclusively affects changes of the border vertices so its depth is bounded by the max. number of vertices on the border (which is four)."
	| idx p dist2 index |
	localPosition := (borderList at: 1).
	radiusSquared := 0.0.
	index := indexList size.
	[index >= minIndex] whileTrue:[
		idx := indexList at: index.
		p := pointList at: idx.
		p := p - localPosition.
		dist2 := (p dot: p) * 0.999. "numerical accuracy"
		dist2 <= radiusSquared ifFalse:[
			p := pointList at: idx.
			borderList at: 2 put: p.
			self privateMtfBall2: pointList with: indexList with: index+1 with: borderList.
			"Move p to front"
			indexList replaceFrom: index to: indexList size-1 with: indexList startingAt: index+1.
			indexList at: indexList size put: idx.
		].
		index := index - 1.
	].! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'ar 2/7/2005 09:21'!
privateMtfBall2: pointList with: indexList with: minIndex with: borderList
	"Note: In this variant the default recursion path is inlined. The recursion exclusively affects changes of the border vertices so its depth is bounded by the max. number of vertices on the border (which is four)."
	| idx p dist2 index |
	self privateMtfSphereFromP1: (borderList at: 1) p2: (borderList at: 2).
	index := indexList size.
	[index >= minIndex] whileTrue:[
		idx := indexList at: index.
		p := pointList at: idx.
		p := p - localPosition.
		dist2 := (p dot: p) * 0.999. "numerical accuracy"
		dist2 <= radiusSquared ifFalse:[
			p := pointList at: idx.
			borderList at: 3 put: p.
			self privateMtfBall3: pointList with: indexList with: index+1 with: borderList.
			"Move p to front"
			indexList replaceFrom: index to: indexList size-1 with: indexList startingAt: index+1.
			indexList at: indexList size put: idx.
		].
		index := index - 1.
	].! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'ar 2/7/2005 09:21'!
privateMtfBall3: pointList with: indexList with: minIndex with: borderList
	"Note: In this variant the default recursion path is inlined. The recursion exclusively affects changes of the border vertices so its depth is bounded by the max. number of vertices on the border (which is four)."
	| idx p dist2 index |
	self privateMtfSphereFromP1: (borderList at: 1) p2: (borderList at: 2) p3: (borderList at: 3).
	index := indexList size.
	[index >= minIndex] whileTrue:[
		idx := indexList at: index.
		p := pointList at: idx.
		p := p - localPosition.
		dist2 := (p dot: p) * 0.999. "numerical accuracy"
		dist2 <= radiusSquared ifFalse:[
			p := pointList at: idx.
			borderList at: 4 put: p.
			self privateMtfSphereFromP1: (borderList at: 1) p2: (borderList at: 2) 
								p3: (borderList at: 3) p4: (borderList at: 4).
			"Move p to front"
			indexList replaceFrom: index to: indexList size-1 with: indexList startingAt: index+1.
			indexList at: indexList size put: idx.
		].
		index := index - 1.
	].! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'ar 6/23/2002 15:09'!
privateMtfRandomize: idxList
	"Randomize input"
	| max rnd j tmp |
	max := idxList size.
	rnd := Random new" seed: 16r2134567".
	1 to: max do:[:i|
		j := (rnd next * i) truncated +1.
		tmp := idxList at: i.
		idxList at: i put: (idxList at: j).
		idxList at: j put: tmp.
	].! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'ar 6/23/2002 15:10'!
privateMtfSphereFromP1: p1 p2: p2
	"Two points"
	| delta center rad2 |
	center := (p1 + p2).
	center *= 0.5.
	delta := center - p1.
	rad2 := delta dot: delta.
	localPosition := center.
	radiusSquared := rad2.! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'das 5/23/2005 10:56'!
privateMtfSphereFromP1: p1 p2: p2 p3: p3
	| d1 d2 d3 v1 v2 v3 m b rad2 center |
	v1 := (p1 + p2). v1 *= 0.5.
	v2 := (p1 + p3). v2 *= 0.5.
	v3 := p1.

	d1 := (p2 - p1).
	d2 := (p3 - p1).
	d3 := d1 cross: d2.
	m := Matrix4x4 new.
	m
		a11: d1 x; a12: d1 y; a13: d1 z;
		a21: d2 x; a22: d2 y; a23: d2 z;
		a31: d3 x; a32: d3 y; a33: d3 z.
	b := Vector3
			x: (v1 dot: d1)
			y: (v2 dot: d2)
			z: (v3 dot: d3).
	center := m solve3x3: b.
	center ifNil:[^self privateMtfSphereFromP1: p1 p2: p2].
	d1 := center - p1.
	rad2 := d1 dot: d1.
	localPosition := center.
	radiusSquared := rad2.
! !

!TBoundSphere methodsFor: 'mini ball' stamp: 'das 5/23/2005 10:56'!
privateMtfSphereFromP1: p1 p2: p2 p3: p3 p4: p4
	| d1 d2 d3 v1 v2 v3 m b rad2 center |
	v1 := (p1 + p2). v1 *= 0.5.
	v2 := (p1 + p3). v2 *= 0.5.
	v3 := (p1 + p4). v3 *= 0.5.

	d1 := (p2 - p1).
	d2 := (p3 - p1).
	d3 := (p4 - p1).
	m := Matrix4x4 new.
	m
		a11: d1 x; a12: d1 y; a13: d1 z;
		a21: d2 x; a22: d2 y; a23: d2 z;
		a31: d3 x; a32: d3 y; a33: d3 z.
	b := Vector3
			x: (v1 dot: d1)
			y: (v2 dot: d2)
			z: (v3 dot: d3).
	center := m solve3x3: b.
	center ifNil:[^self privateMtfSphereFromP1: p1 p2: p2 p3: p3].
	d1 := center - p1.
	rad2 := d1 dot: d1.
	localPosition := center.
	radiusSquared := rad2.
! !

!TBoundSphere methodsFor: 'access' stamp: ''!
radius

	^ radius.! !

!TBoundSphere methodsFor: 'access' stamp: ''!
radius: rad

	radius := rad.
	radiusSquared := rad*rad.! !

!TBoundSphere methodsFor: 'access' stamp: ''!
radiusSquared

	^ radiusSquared.! !

!TBoundSphere methodsFor: 'collision' stamp: 'das 5/13/2002 16:41'!
ray: framePtr tri: p1 tri: p2 tri: p3


" This method works by using the orthogonal picking matrix as orthogonal planes going through the origin at 'origin'. The points of the triangle are tested to the two planes orthoganal to the direction of picking. If they are on either side of both planes, then there is a very high probability that our picking ray interesects the triangle, and we do the actual calculation of the intersection point. Note the comments below. I use a vertical bar to indicate which side of the plane a point is on.  ***** p1 p2 | ***** means that p1 and p2 are on the same side, 
***** p1 | p2 ***** indicates that they are on opposite sides. p1 is always on the left side, because it is the first one I test. This really just means - same side as p1 - no meaning other than that. 

This can be optimized for tri-strips, tri-fans, and even full face based meshes. When I have the time... "

	| d1 d2 d3 pp1 pp2 dd1 dd2 xPlane yPlane  |
		
	xPlane := framePtr row1.
	d1 := xPlane dot: p1.
	d2 := xPlane dot: p2.
	d3 := xPlane dot: p3.
	dd1 := d1*d2.
	dd2 := d1*d3.

" test if all points are on the same side of the x plane."
	dd1 < 0 ifFalse:[ 				"***** p1 p2 |  *****"
		dd2 < 0  ifFalse:[  ^ nil		"***** p1 p2 p3 | *****" ] 
				ifTrue:[ 			"***** p1 p2 | p3 *****"
						pp1 := p1 + ((p3-p1) * (d1/(d1-d3))).
						pp2 := p2 + ((p3-p2) * (d2/(d2-d3))).]]
		ifTrue:[ 				"***** p1 | p2 *****"
			dd2 < 0  ifTrue:[ 		"***** p1 | p2 p3 *****" 
						pp1 := p1 + ((p2-p1) * (d1/(d1-d2))).
						pp2 := p1 + ((p3-p1) * (d1/(d1-d3))).] 
					ifFalse:[ 		"***** p1 p3 | p2 *****"
						pp1 := p1 + ((p2-p1) * (d1/(d1-d2))).
						pp2 := p3 + ((p2-p3) * (d3/(d3-d2))).].].
" At this point, the triangle is intersected by the x plane and we know where. If both points are on the same side of the y plane, we are done. "
	yPlane := framePtr row2.

	d1 := yPlane dot: pp1.
	d2 := yPlane dot: pp2.
" test if both points are on the same side of the y plane."
	d1 * d2 < 0 ifFalse:[ 
	 ^ nil ].
	pp1 := pp1 + ((pp2-pp1) * ( d1/ (d1-d2))).
	^ pp1.
	
	


	! !

!TBoundSphere methodsFor: 'render' stamp: 'das 10/16/2004 15:06'!
render: ogl box: frm

	children ifNotNil:[
		vertices ifNil: [ children do:[ :c | c render: ogl box: frm].].].
	self box ifNotNil:[frm render: ogl box: self box].
! !

!TBoundSphere methodsFor: 'render' stamp: 'das 3/12/2004 12:57'!
render: ogl box: frm depth: depth

	children ifNotNil:[
		vertices ifNil: [ children do:[ :c | c render: ogl box: frm depth: depth-1. ].].].
	depth = 0 ifTrue:[ box ifNotNil:[
		frm render: ogl box: self box.].].! !

!TBoundSphere methodsFor: 'render' stamp: 'das 3/12/2004 12:57'!
render: ogl frame: frm segments: segs

	children ifNotNil:[
		vertices ifNil: [ children do:[ :c | c render: ogl frame: frm segments: segs. ].].].
	frm render: ogl sphere: self segments: segs.! !

!TBoundSphere methodsFor: 'render' stamp: 'das 3/2/2006 16:36'!
renderSurface: ogl

	| mid |
	children ifNotNil:[
		normal ifNil: [ children do:[ :c | c renderSurface: ogl. ].].].
	normal ifNotNil:[
	 		ogl glColor3fv: #(1.0 0 0)asFloatArray.
			ogl glBegin: GLLineStrip.
    
			ogl glVertex3fv:localPosition + (normal*offset) + normal.
			ogl glVertex3fv:localPosition + (normal * offset).
			ogl glVertex3fv:localPosition + (normal*offset) + (up*radius).
		mid := (up+side) normalized.
			ogl glVertex3fv:localPosition + (normal*offset) + (mid*radius).		
			ogl glVertex3fv:localPosition + (normal*offset) + (side*radius).
		mid := (side - up) normalized.
			ogl glVertex3fv:localPosition + (normal*offset) + (mid*radius).		
			ogl glVertex3fv:localPosition + (normal*offset) + (up*radius negated).
		mid :=( side+up)normalized negated.
			ogl glVertex3fv:localPosition + (normal*offset) + (mid*radius).		
			ogl glVertex3fv:localPosition + (normal*offset) + (side*radius negated).
		mid := (up-side)normalized.
			ogl glVertex3fv:localPosition + (normal*offset) + (mid*radius).		
			ogl glVertex3fv:localPosition + (normal*offset) + (up*radius).
		ogl glEnd.].
! !

!TBoundSphere methodsFor: 'render' stamp: 'das 5/18/2006 12:02'!
renderTriangles: ogl


	| index triList v0 v1 v2 |
	vertices ifNil:[ children ifNotNil:[ children do:[:c |c renderTriangles:ogl.]]]
	ifNotNil:[
		index := 0.
		triList := children.
		[index < triList size] whileTrue:[
			v0 := vertices at: (triList at: (index := index+1))+1.
			v1 := vertices at: (triList at: (index := index+1))+1.
			v2 := vertices at: (triList at: (index := index+1))+1.
v0:=v0+(0@1@0).
v1:=v1+(0@1@0).
v2:=v2+(0@1@0).

			ogl
				glBegin: GLTriangles;
					glVertex3fv: v0;
					glVertex3fv: v1;
					glVertex3fv: v2;
				glEnd.
		].
	].
! !

!TBoundSphere methodsFor: 'access' stamp: 'ar 7/22/2006 20:18'!
rigidBody
	"Answer my associated rigid body"
	^frame rigidBody! !

!TBoundSphere methodsFor: 'access' stamp: 'bgf 7/14/2006 23:09'!
scale: scale 
	self radius: radius * scale.
	self localPosition: self localPosition * scale.
	box ifNotNil: [ box scale: scale ].
	children
		ifNotNil: [children
				do: [:each | (each isMemberOf: TBoundSphere)
						ifTrue: [each scale: scale]]]! !

!TBoundSphere methodsFor: 'render' stamp: 'jhk 5/11/2005 15:05'!
setMaxOffset: m verts: verts nabs: nabs face1: face1 face2: face2 face3: face3
   | ray t  |
		ray := self ray: m tri: localPosition - (verts at: 1+face1)
					tri: localPosition - (verts at: 1+face2)
					tri: localPosition - (verts at: 1+face3).
		
		ray ifNotNil:[ 
			nabs x > nabs y 
				ifTrue:[
					nabs x > nabs z 
						ifTrue: [t := ray x/normal x] 
						ifFalse: [ t := ray z/normal z]
					]
				ifFalse:[
					nabs y > nabs z 
						ifTrue:[t := ray y/normal y] 
						ifFalse:[t := ray z/normal z]
					].
			t abs > offset abs ifTrue: [ offset := t negated.]
		]
! !

!TBoundSphere methodsFor: 'construct' stamp: 'das 5/16/2005 08:52'!
splitBoxes: bx

	| epsilon ratio boxes splitBoxes extent |
" Split along the axis where the ratio with the other axes is greater than 0.6. I am sure there is a more efficient number. Just don't know what it is."

"	ratio := 0.737."
	ratio := 0.6.
	extent := bx extent.

" Make sure we don't divide by zero. "
	epsilon := 0.000001.
	extent := extent max: (Vector3 x: epsilon y: epsilon z: epsilon).
	boxes := OrderedCollection new.
	boxes add: bx.

	(((extent x/extent y) > ratio) and:[(extent x/extent z) > ratio]) ifTrue:[ 
		splitBoxes := OrderedCollection new.
		boxes do:[:b | 
			splitBoxes add: b splitXMin.
			splitBoxes add: b splitXMax.
			].
		boxes := splitBoxes.
		].

	(((extent y/extent x) > ratio) and:[(extent y/extent z) > ratio]) ifTrue:[ 
		splitBoxes := OrderedCollection new.
		boxes do:[:b | 
			splitBoxes add: b splitYMin.
			splitBoxes add: b splitYMax.
			].
		boxes := splitBoxes.
		].

	(((extent z/extent x) > ratio) and:[(extent z/extent y) > ratio]) ifTrue:[ 
		splitBoxes := OrderedCollection new.
		boxes do:[:b | 
			splitBoxes add: b splitZMin.
			splitBoxes add: b splitZMax.
			].
		boxes := splitBoxes.
		].

	^ boxes.! !

!TBoundSphere methodsFor: 'access' stamp: 'das 5/6/2002 15:59'!
sumFaces
	| total |

" Returns the number of faces in the leaves."
	children ifNil:[^ 0].
	vertices ifTrue:[^children size/3].
	total := 0.
	children do:[:c | total := total + c sumFaces.].
	^total.! !

!TBoundSphere methodsFor: 'access' stamp: 'das 5/6/2002 16:00'!
sumNodes
	| total |

" Returns the total number of nodes in the tree."
	children ifNil:[^ 1].
	vertices ifNotNil:[^1].
	total := 1.
	children do:[:c | total := total + c sumNodes.].
	^total.! !

!TBoundSphere methodsFor: 'transform' stamp: 'das 4/24/2002 19:51'!
transform: trans

	trans ifNotNil:[
		globalPosition := trans localPointToGlobal: localPosition.] 
	ifNil: [globalPosition := localPosition.].

" As far as I can tell, we never need to transform the children boundSpheres.

	children ifNotNil:[
		leaf ifFalse:[
		children do:[:cs | cs transform: trans.].].]."
! !

!TBoundSphere methodsFor: 'construct' stamp: 'das 9/13/2002 01:42'!
union: sphere

" We calculate the minimal sphere that contains both this sphere and the argument sphere."

" This needs to be performed in global coordinates."

	| vector length norm p1 p2 |
	sphere ifNil:[ ^ TBoundSphere localPosition: (self globalPosition) radius: (self radius).].

	vector := sphere globalPosition - self globalPosition.
	vector length = 0.0 ifTrue:[
		^ TBoundSphere localPosition: (self globalPosition) radius: (self radius max: sphere radius).
		].
	length := vector length.
	norm := vector/length.
	length + sphere radius > self radius ifTrue:[ p2 := norm * (length + sphere radius).] 
		ifFalse:[ ^ TBoundSphere localPosition: (self globalPosition) radius: (self radius).].
	length + self radius > sphere radius ifTrue:[ p1 := norm * (self radius) negated ] 
		ifFalse:[ ^ TBoundSphere localPosition: (sphere globalPosition) radius: (sphere radius).].

	length := ((p1-p2) length)/2.0.
	p1 := ((p1+p2)/2.0) + self globalPosition.
	^ TBoundSphere localPosition: p1 radius: length.

	
! !

!TBoundSphere methodsFor: 'access' stamp: 'ar 6/22/2002 17:51'!
volume
	^4.0 / 3.0 * Float pi * radius * radius * radius! !

TMediaResource subclass: #TForm
	instanceVariableNames: 'url form updateRect extent bMipmap bShrinkFit extension memUsed objectID bThumb static'
	classVariableNames: 'DefaultForm'
	poolDictionaries: 'GLExtConstants OpenGLConstants'
	category: 'Croquet-Objects'!

!TForm commentStamp: '<historical>' prior: 0!
TForm is an imaging object that maps Squeak forms into OpenGL textures. It is also intended to act as a static Croquet object type in that it will typically not be stored as part of an Island but must be loaded either from a local cache, a remote cache, or from another user. It can also generate a thumbnail version of itself that can be copied and stored as part of a TTexture.

sha - the Secure Hash Algorithm value from the original file
form - the Squeak Form object.
updateRect - what part of the OGL texture to update.
extent - the actual original extent of the texture.
bMipmap - boolean flag, determines if mipmapping is on.
bShrinkFit - boolean flag, determines if we shrink or grow the texture to fit the proper power of two extent.
extension - extension bit flags for further bitmap processing
memUsed - the memory required to store the OGL texture.
objectID - a TObjectID
bThumb - indicates whether or not this TForm is a thumb. If it is, then it needs to be resolved to the "real" TForm, which may be either cached or located in a remote server. 

NOTE: Dynamically generated TForms (those not loaded from elsewhere) are the same as themselves. That is, they cannot really generate thumbnails.

DAS!

!TForm class methodsFor: 'instance creation' stamp: 'das 7/4/2004 21:21'!
defaultForm

	^DefaultForm ifNotNil:[DefaultForm] ifNil:[DefaultForm := Form extent: 64@64 depth: 32 bits: (Base64MimeConverter mimeDecode:
   '8AAS/wAAAQf/AQEBFv8AAAEH/wABARL/AAABB/8BAAEK/wAAAQf/AAEBCv8AAAEH/wARAAr/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' as: String) asByteArray].! !

!TForm class methodsFor: 'instance creation' stamp: 'das 2/10/2006 10:22'!
defaultTForm

	| tform |
	tform := TFormManager default at: 'default' ifAbsentPut:[
		TForm fromForm: TForm defaultForm mipmap: true shrinkFit: true extension: nil.].
	^ tform.
! !

!TForm class methodsFor: 'instance creation' stamp: 'jcg 11/15/2006 09:45'!
fromForm: baseForm mipmap: bMM shrinkFit: bSF extension: ext
	^ self new initializeSHA: nil form: baseForm mipmap: bMM shrinkFit: bSF extension: ext.
	! !

!TForm methodsFor: 'accessing' stamp: 'das 6/30/2004 11:29'!
bMipmap

	^ bMipmap.! !

!TForm methodsFor: 'accessing' stamp: 'jcg 11/14/2006 09:47'!
bMipmap: bool
	bMipmap := bool
! !

!TForm methodsFor: 'accessing' stamp: 'das 1/4/2006 22:18'!
bShrinkFit

	^ bShrinkFit.! !

!TForm methodsFor: 'accessing' stamp: 'jcg 11/14/2006 09:47'!
bShrinkFit: bool
	bShrinkFit := bool
! !

!TForm methodsFor: 'accessing' stamp: 'das 1/2/2006 23:06'!
bThumb

	^ bThumb.! !

!TForm methodsFor: 'accessing' stamp: 'das 1/2/2006 23:06'!
bThumb: bool

	bThumb := bool.! !

!TForm methodsFor: 'accessing' stamp: 'das 1/5/2006 16:32'!
boundingBox

	^form boundingBox.! !

!TForm methodsFor: 'initialize' stamp: 'das 7/1/2004 12:50'!
calcMemUsed

	memUsed := form extent area*4.
	bMipmap ifTrue:[memUsed := memUsed + (memUsed//3)].! !

!TForm methodsFor: 'accessing' stamp: 'das 6/30/2004 11:29'!
extension

	^ extension.! !

!TForm methodsFor: 'accessing' stamp: 'das 6/30/2004 11:29'!
extent

	^ extent.! !

!TForm methodsFor: 'accessing' stamp: 'das 1/2/2006 16:45'!
extent: ext

	extent := ext.! !

!TForm methodsFor: 'accessing' stamp: 'das 6/30/2004 11:30'!
form

	^form.! !

!TForm methodsFor: 'initialize' stamp: 'sn 2/24/2018 16:23'!
initialize: shaName form: baseForm mipmap: bMM shrinkFit: bSF extension: ext
	"self deprecatedExplanation: 'method renamed for consistency and understandability'."
	self 
		initializeSHA: shaName 
		form: baseForm 
		mipmap: bMM 
		shrinkFit: bSF 
		extension: ext 
		url: nil
	! !

!TForm methodsFor: 'initialize' stamp: 'sn 2/23/2018 12:47'!
initialize: u sha: shaName form: baseForm mipmap: bMM shrinkFit: bSF extension: ext
	"self deprecatedExplanation: 'method renamed for consistency and understandability'."
	self
		initializeSHA: shaName 
		form: baseForm 
		mipmap: bMM 
		shrinkFit: bSF 
		extension: ext 
		url: u
	! !

!TForm methodsFor: 'initialize' stamp: 'jcg 11/15/2006 09:32'!
initializeForm: baseForm mipmap: bMM shrinkFit: bSF extension: ext

	bMipmap := bMM.
	bShrinkFit := bSF.
	extension := ext.
	objectID := TObjectID for: self.
	self setupForm: baseForm.! !

!TForm methodsFor: 'initialize' stamp: 'jcg 11/15/2006 09:43'!
initializeSHA: shaName form: baseForm mipmap: bMM shrinkFit: bSF
	self 
		initializeSHA: shaName 
		form: baseForm 
		mipmap: bMM 
		shrinkFit: bSF 
		extension: nil 
		url: nil
	! !

!TForm methodsFor: 'initialize' stamp: 'jcg 11/15/2006 09:43'!
initializeSHA: shaName form: baseForm mipmap: bMM shrinkFit: bSF extension: ext
	self 
		initializeSHA: shaName 
		form: baseForm 
		mipmap: bMM 
		shrinkFit: bSF 
		extension: ext 
		url: nil
	! !

!TForm methodsFor: 'initialize' stamp: 'jcg 11/15/2006 09:43'!
initializeSHA: shaName form: baseForm mipmap: bMM shrinkFit: bSF extension: ext url: u
	url := u.
	bMipmap := bMM.
	bShrinkFit := bSF.
	extension := ext.
	objectID := TObjectID new.
	sha := shaName.
	self setupForm: baseForm.! !

!TForm methodsFor: 'accessing' stamp: 'das 1/2/2006 19:13'!
invalidate: ur
	
	updateRect := updateRect ifNil:[ur] ifNotNil:[updateRect quickMerge: ur].! !

!TForm methodsFor: 'accessing' stamp: 'das 1/15/2006 10:02'!
isStatic
	
	^static.! !

!TForm methodsFor: 'accessing' stamp: 'das 1/4/2006 20:21'!
isThumb

	^ bThumb.! !

!TForm methodsFor: 'accessing' stamp: 'das 6/30/2004 11:30'!
magFilter
	^GLLinear! !

!TForm methodsFor: 'accessing' stamp: 'das 6/30/2004 11:30'!
minFilter
	^bMipmap ifTrue:[GLLinearMipmapLinear] ifFalse:[GLLinear]! !

!TForm methodsFor: 'accessing' stamp: 'das 1/5/2006 01:12'!
objectID

	^objectID copy.! !

!TForm methodsFor: 'accessing' stamp: 'das 7/1/2004 09:30'!
resetUpdateRect

	updateRect := nil.! !

!TForm methodsFor: 'accessing' stamp: 'das 6/30/2004 11:30'!
scaleTexture: oldForm to: scaledSize
	| mmForm |
	oldForm extent = scaledSize ifTrue:[^oldForm].
	mmForm := Form extent: scaledSize depth: 32.
	(WarpBlt current toForm: mmForm)
		sourceForm: form destRect: mmForm boundingBox;
		combinationRule: 3;
		cellSize: 2;
		warpBits.
	^mmForm! !

!TForm methodsFor: 'initialize' stamp: 'bgf 5/16/2006 21:55'!
setupForm: baseForm
	| aForm size |
	bThumb := false.
	aForm := baseForm.
	extent := baseForm extent. "before we scale it down"
	aForm depth = 32 ifFalse: [aForm := aForm asFormOfDepth: 32.].
	size := aForm width @ aForm height.
	bShrinkFit
		ifTrue:[size := size asSmallerPowerOfTwo]
		ifFalse: [size := size asLargerPowerOfTwo].
	size = aForm extent ifTrue:[form := aForm]
		ifFalse:[	form := Form extent: size depth: 32.
				aForm displayScaledOn: form].
	extension == #colorKeyZero ifTrue:[form colorKeyZero].
	extension == #colorKey ifTrue:[form colorKeyAlpha].
	extension == nil ifTrue:[form fixAlpha].

	self updateAll.
	self calcMemUsed.
	sha ifNil:[
		sha := TObjectID for: 'TForm setup SHA'. 
		static := false.
	] ifNotNil:[
		static := true
	].
! !

!TForm methodsFor: 'accessing' stamp: 'das 7/1/2004 09:32'!
shrinkFit

^ bShrinkFit.! !

!TForm methodsFor: 'accessing' stamp: 'das 7/1/2004 09:32'!
shrinkFit: bSF

bShrinkFit := bSF.! !

!TForm methodsFor: 'accessing' stamp: 'ar 10/1/2004 15:28'!
target
	^GLTexture2d! !

!TForm methodsFor: 'accessing' stamp: 'ar 10/1/2004 15:17'!
textureForm
	^form! !

!TForm methodsFor: 'initialize' stamp: 'das 1/26/2006 11:44'!
thumb

	| tsize size thumb tf |
"no sha means there is no external reference, hence no reason to have a thumbnail"	
	sha ifNil:[^self]. 
	size := form width @ form height.
	tsize:= (size*32/size maxDimension) asLargerPowerOfTwo max:1@1.
	thumb := Form extent:tsize depth:32.
	form displayScaledOn:thumb.
	tf := TForm new initialize: url sha:sha copy form: thumb mipmap: bMipmap shrinkFit: bShrinkFit extension: extension. 
	tf extent: extent.
	tf bThumb: true.
	^ tf.! !

!TForm methodsFor: 'accessing' stamp: 'das 7/8/2004 18:50'!
trueExtent

	^ form extent.! !

!TForm methodsFor: 'accessing' stamp: 'ar 10/1/2004 15:03'!
updateAll

	self invalidate: form boundingBox.! !

!TForm methodsFor: 'accessing' stamp: 'ar 10/1/2004 15:03'!
updateArea: aRectangle with: aForm
	"the form contains new data to copy into the rectangle area specified"
	| fixBlt |
	fixBlt := BitBlt toForm: aForm.
	fixBlt combinationRule: 40 "fixAlpha:with:".
	fixBlt halftoneForm: (Bitmap with: 16rFF000000).
	fixBlt destRect: fixBlt clipRect; copyBits.

	form copy: aRectangle from: 0@0 in: aForm rule: Form over.
	self invalidate: aRectangle.
	! !

!TForm methodsFor: 'accessing' stamp: 'das 7/1/2004 09:35'!
updateRect

	^ updateRect.! !

!TForm methodsFor: 'accessing' stamp: 'das 1/2/2006 15:15'!
url

	^url.! !

!TForm methodsFor: 'accessing' stamp: 'das 1/2/2006 15:15'!
url: u

	url := u.! !

!Matrix4x4 methodsFor: '*Croquet-Objects' stamp: 'sn 2/26/2018 12:36'!
localBoxToGlobal: aTBox

	^TBox min: (self localPointToGlobal: aTBox min) max: (self localPointToGlobal: aTBox max).! !

TObjectProperties subclass: #TFrameProperties
	instanceVariableNames: 'behaviors player'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TFrameProperties methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:39'!
behaviors
	^behaviors! !

!TFrameProperties methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:39'!
behaviors: anArray
	behaviors := anArray! !

!TFrameProperties methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:39'!
player
	^player! !

!TFrameProperties methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:39'!
player: aPlayer
	player := aPlayer! !

TObject subclass: #TFrame
	instanceVariableNames: 'objectName objectID objectOwner ownedObjects localTransform globalTransform frameRoot frameParent frameChildren frameChanged visible solid oglid currentUser eventMask test hiliteColor'
	classVariableNames: 'EventBlockRecurse EventKeyboard EventMaskBits EventNone EventPointerDown EventPointerOver'
	poolDictionaries: 'GLExtConstants OpenGLConstants'
	category: 'Croquet-Objects'!

TFrame class
	instanceVariableNames: 'protoData'!

!TFrame commentStamp: 'das 1/22/2006 17:07' prior: 0!
The TFrame class is designed as a heirarchical transformation frame. The rendering engine walks through the frame, performing the appropriate transformation associated with it, then it calls the render methods of the tObjects. Two passes are made through the frame heirarchy. The first is for non-alpha objects, the second for alpha. The base frame is owned by a TSpace object. We render from the root up because we want to minimize setting transforms.


" ***** This is how you rotate an object using a quaternion ball. The base position is stored in the selectedPoint and compared with the pointer selectedPoint.
	pointer event2D shiftPressed ifTrue:[
		pointer frame: self pickSphere: Vector3 new radiusSquared: selectedRadiusSquared.
		spin _ self rotFromBallPoints: selectedPoint to: pointer selectedPoint.
		trans _ self translation.
		self translationX: 0.0 y:0.0 z:0.0.
		self localTransform: (self localTransform composeWith: spin).
		self translation: trans.
		]"
" ***** This is how you move an object relative to a specified plane. In this case, it is the plane determined by either the camera or the surface normal when the object is selected. 
	- cameraNorm defines the plane perpendicular to the line of sight of the camera
	- selectedNorm defines the norm of the selected surface of the object.
	- slab frontNorm forces a normal to the front of the slab (for example).
	ifFalse:[
	(pointer frame: self pickPlane: selectedPoint normal: cameraNorm) ifTrue:[
		delta _ selectedPoint - pointer selectedPoint.
		self translation: (self translation - (self orientation localPointToGlobal: delta)).
		^ true.].]."
	^ false.

objectName - a name that can be given to any TFrame by the programmer. This is used by mesh importers to set the corresponding node names.
objectID - a unique value generated by TObjectID. This is used as the default name of the TFrame when we register it with the Island.
objectOwner - this is the TFrame that "owns" this TFrame. In this case, ownership is which frame processes events that are sent to this frame.
ownedObjects - this is a dictionary of all of the objects that this TFrame is referenced by. It utilizes the TFrame objectID as the key.
localTransform - the 4x4 matrix transform of this frame relative to it's parent frame.
globalTransform - the transform relative to the global coordinates (the root).
globalPosition - the position of the frame in global coordinates. This is for faster access.
frameRoot - this is the root frame of the heirarchy, usually a TSpace. The root of a TSpace is self.
frameParent - the parent frame of this TFrame.
frameChildren - an OrderedCollection of children frames.
frameChanged - this flag is set when any change occurs to the transform (among other things). This forces a recalculation of the globalTransform.
visible - boolean flag indicating that this object is (or isn't) visible.
solid - boolean flag indicating that this object is solid, that is, can be picked and walked on.
oglid - a TObjectID used by the renderer to maintain a link between the TFrame and its display lists (if any).
currentUser - probably a bad idea ...
test - used for testing

#start/#stop are called when a TFrame is inserted into a hierarchy. They can be used to start and stop future messages.
DAS!

TFrame class
	instanceVariableNames: 'protoData'!

TFrame subclass: #TAvatarUser
	instanceVariableNames: 'replica replicaClass cameraBehind thirdPerson thirdPersonLag head speedControl controlPitch newPostcard newTransform currentSpace laser lastPosition portalRay modelData costume driving lastRedButton viewAngle viewAspect snapshots newSnapshot lastLength harness lastYellowDownTime potentialYellowClick myFarRef jumpCameraTrans laserColor userID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TAvatarUser commentStamp: '<historical>' prior: 0!
TAvatarUser

I am the user side of the TUser/TAvatar pair. I handle all of the direct control messages generated by the user and interpret them and forward them to the actual TAvatar. I also check for portal collisions so that I can manage the TAvatar transition from space to space. Finally, I manage how the camera is positioned relative to the TAvatar.

!

!TAvatarUser methodsFor: 'portals' stamp: 'sn 2/25/2018 15:48'!
addToNewSpace: aSpace
	"Add this avatar to aSpace"
	| newIsland spaceClass emap |

	"May get called for a postcard jump to somewhere else in the current space;
	 if so, we needn't do anything."
	(aSpace == self currentSpace) ifTrue: [
		newTransform ifNotNil:[
			replica future localTransform: newTransform.
			newTransform := nil.
		].
		^ self.
	].

	"FIXME: We should know that this is a space..."
	spaceClass := aSpace get: #class.
	(spaceClass includesBehavior: TSpace) ifFalse:[^nil].

	"Remember current space"
	currentSpace := aSpace.

	newIsland := aSpace island.
	replica ifNotNil:[
		self stopScript: #onSnapshot:.
		self costume: nil.
		replica future carryingDrop.
		replica future destroy.
		replica := nil.
	].

	"Create new replicated avatar in aSpace"

	replica := newIsland future new: (replicaClass ifNil:[TAvatarReplica]).
	
	replica wait.
	
	aSpace future addChild: replica.
	aSpace future signal: #avatarEnter with: replica.

	replica future userID: userID.
	replica future viewAngle: viewAngle aspect: viewAspect.
	
	"Reposition in proper location/orientation"
	newTransform ifNotNil:[
		replica future localTransform: newTransform.
		newTransform := nil.
	] ifNil: [
		replica future goToSpaceHome.
	].

	emap := self myEventMap.
	emap ifNotNil: [
		emap keysDo:[:eventName| replica on: eventName notify: self]].

	"Copy costume to new island"
	self installModelData.

	"Create new laser"
	laser := newIsland future new: TLaser.
	laser future translation:(Vector3 x: 0.5 y: -0.3 z: -0.2).
	laser future length: 0.5.
	laserColor ifNil:[
		laserColor := aSpace future nextLaserColor.
		laserColor whenResolved:[laserColor := laserColor value].
	].
	laser future color: laserColor.
	replica future laser: laser.

	"If the space defines an initial translation and/or yaw, use it."
	replica future takeInitialPosition.

	"update the current driving state"
	replica future controlPitch: controlPitch.
	replica future resetPitch.
	replica future speedControl: speedControl.
	driving ifTrue:[replica future startDriving.].
	replica future driveStep.
	replica future selfDestruct.
	lastPosition := nil.
	replica wait.

! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 1/29/2006 14:17'!
controlPitch: cp

	controlPitch ~= cp ifTrue:[
		controlPitch := cp.
		replica future controlPitch: controlPitch.
	].
! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 1/28/2006 21:30'!
costume

	^ costume.
! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 6/7/2006 23:22'!
costume: aFrame
	
	costume := aFrame.
	costume ifNotNil:[
		costume future solidTree: false. "don't walk on it!!"
	].
	replica future costume: costume.
! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 1/28/2006 21:18'!
currentSpace

	^ currentSpace.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 4/4/2006 18:01'!
driveStep
	"Always forward the drive step so replica knows that it's still alive"
	replica future driveStep.! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:42'!
globalMatrixOut

	^(replica get: #globalMatrixOut).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:39'!
globalOrientation

	^(replica get: #globalOrientation).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:39'!
globalPitch

	^(replica get: #globalPitch).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:39'!
globalPitchYawRoll

	^(replica get: #globalPitchYawRoll).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/17/2006 20:27'!
globalPosition

	^(replica get: #globalPosition)! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:40'!
globalRoll

	^(replica get: #globalRoll).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/17/2006 20:27'!
globalTransform

	^(replica get: #globalTransform)! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/17/2006 20:24'!
globalTransform: gt
! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:40'!
globalYaw

	^(replica get: #globalYaw).! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 5/8/2006 07:56'!
goToSpaceHome.

	replica future goToSpaceHome.! !

!TAvatarUser methodsFor: 'initialize' stamp: 'bgf 3/12/2006 00:41'!
harness: h
	harness := h.! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 1/28/2006 21:12'!
head
	^ head! !

!TAvatarUser methodsFor: 'accessing' stamp: 'bgf 3/8/2006 10:38'!
head: tframe
	head := tframe.
	replica future head: tframe! !

!TAvatarUser methodsFor: 'initialize' stamp: 'ar 7/8/2006 17:57'!
initialize

	super initialize.
	controlPitch := false.
	cameraBehind := false.
	head := nil.
	thirdPerson := 1.0.	"Slightly above and behind the avatar origin."
	thirdPersonLag := 3.0.	"Moderately high lag, feels like driving."
	speedControl := 0@0.
	currentSpace := nil.
	portalRay := TRay new.
	driving := false.
	lastLength := 0.5.
	self runScript: #onSnapshot: when:{self. #snapshot}.
! !

!TAvatarUser methodsFor: 'initialize' stamp: 'das 6/7/2006 23:22'!
installModelData
	"Install our model in the current space"
	| costumeAndHead p |
	currentSpace ifNil:[^nil].
	"Allow current space to override user's avatar choice"
	(currentSpace get: #defaultAvatar) ifNotNilDo:[:atarData|
		^self costume: (TIslandCopier new import: atarData to: currentSpace island).
	].
	modelData ifNil:[^nil].
	costumeAndHead := TIslandCopier new import: modelData to: currentSpace island.
	self costume: costumeAndHead future first.
	"This is a little awkward since the head can resolve to nil 
	and we need to handle this properly"
	p := costumeAndHead future second.
	p whenResolved:[self head: p value].
! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:40'!
inverseGlobalOrientation

	^(replica get: #inverseGlobalOrientation).! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 5/28/2006 20:21'!
jumpCameraTrans: jct

	jumpCameraTrans := jct.! !

!TAvatarUser methodsFor: 'events' stamp: 'das 1/29/2006 15:11'!
keyDown: event

	replica future keyDown: event.
! !

!TAvatarUser methodsFor: 'events' stamp: 'das 5/23/2006 11:24'!
keyStroke: event

	event keyValue = 1 ifTrue: [self toggleAvatarPerson. ^self. ].			"Ctl-A"
	event keyValue = 8 ifTrue: [replica future goToSpaceHome. ^self ].	"Ctl-H for home"
	event keyValue = 30 ifTrue:[
		event shiftPressed ifFalse:[
			replica future moveUp:1.]
		ifTrue:[self zoomIn.].
		^ self.
	].
	event keyValue = 31 ifTrue:[
		event shiftPressed ifFalse:[
			replica future moveDown:1.]
		ifTrue:[self zoomOut.].
		^ self.
	].
	replica future keyStroke: event.
! !

!TAvatarUser methodsFor: 'events' stamp: 'das 1/29/2006 15:03'!
keyUp: event

	replica future keyUp: event.
! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 3/9/2006 14:42'!
laser
	^laser! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 7/8/2006 17:57'!
laserColor
	^laserColor ifNil:[Color red]! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 7/9/2006 21:12'!
laserColor: aColor
	laserColor := aColor asVectorColor.
	laser future color: laserColor.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 8/10/2006 13:14'!
laserControl: pointer redButton: redButton
	| p0 p1 at up side m st |
	
	redButton ifFalse:[
		lastLength := 0.5.
		laser future length: lastLength.
		laser future orientation: pointer orientation.
		nil. "Don't let above future be in tail position."
	]ifTrue:[
pointer isOverlay ifTrue:[
		laser future length: 0.5.
		laser future orientation: pointer orientation.
		^ self.
	].
pointer selectedTarget ifNotNil:[
		st := pointer selectedTarget.
		pointer selectedPortal ifNotNil:[
			laser future length: lastLength.
			laser future orientation: pointer orientation.
			^ self.
		].
		p0 := laser get:#globalPosition.
		p1 := pointer selectedGlobalPoint.
		p0 := p0-p1.
		at := p0 normalized.
		up := 0@1@0.
		side := (at cross: up) normalized negated.
		up := (side cross: at) normalized negated.
		m := Matrix4x4 identity.
		m a11: side x.
		m a21: side y.
		m a31: side z.
		m a12: up x.
		m a22: up y.
		m a32: up z.
		m a13: at x.
		m a23: at y.
		m a33: at z.
m := self inverseGlobalOrientation composeWith: m.
		laser future orientation: m.
		lastLength := p0 length.
		laser future length: lastLength.
		nil. "Don't let above future be in tail position."
		
	] ifNil:[
		laser future length: 0.5.
		laser future orientation: pointer orientation.
		nil. "Don't let above future be in tail position."
	].
].
		
	head ifNotNil: [
		head future orientation: pointer orientation.
	].! !

!TAvatarUser methodsFor: 'initialize' stamp: 'sn 2/26/2018 02:24'!
loadAvatar: avatarPath
	| box scale dy newModel newHead |
	newModel := TLoadMDL new parseFileNamed: (FileDirectory pathFromURI: avatarPath).
	newModel ifNil:[^nil].
	box := newModel boundingBox.
	scale := 4.708164/box extentY.
	dy := box centerY * scale negated.
	newModel scale: scale.
	box := newModel boundingBox.
	newModel translationX: 0 y: 0.616526961326599+dy-2.3 z: 0.
	newModel solidTree: false.
	newModel frameChildrenDo:[:child |
		(child objectName = 'head' or: [ child objectName = 'neck' ]) ifTrue: [
				newHead := child.
		]
	].
	modelData := TIslandCopier new export: (Array with: newModel with: newHead).
	self installModelData.! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:40'!
localTransform

	^(replica get: #localTransform).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:38'!
localTransform: lt! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 5/23/2006 11:09'!
lockYaw: yw

	replica future lockYaw:yw.! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 3/9/2006 20:00'!
modelData
	^modelData! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 3/9/2006 20:00'!
modelData: data
	modelData := data.! !

!TAvatarUser methodsFor: 'events' stamp: 'bgf 3/14/2006 15:51'!
mouseDown: evt


	"Note when the button went down, so we can identify a button click later in mouseUp:."
	potentialYellowClick := true.
	lastYellowDownTime := Time millisecondClockValue.

	self startDriving.
	self controlPitch: evt shiftPressed.
! !

!TAvatarUser methodsFor: 'events' stamp: 'bgf 3/14/2006 15:52'!
mouseMove: evt

	"Called from the harness when the mouse is dragged with the drive (yellow) button down."
	potentialYellowClick := false.
	self controlPitch: evt shiftPressed.

! !

!TAvatarUser methodsFor: 'events' stamp: 'das 5/23/2006 12:58'!
mouseUp: evt

	| maxDownTimeForClick |
	self stopDriving.
	self controlPitch: evt shiftPressed.

	"UI policy setting... if the button is down longer than 300 ms, don't do a jump for the click."
	maxDownTimeForClick := 360.

	(potentialYellowClick 
	 and:[(Time millisecondClockValue - lastYellowDownTime) abs < maxDownTimeForClick]) 
	ifTrue: [
			self takeMeThere: evt.
	].

	potentialYellowClick := false.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 5/7/2006 22:36'!
moveBack: val
	"move in this direction by val units"
	replica future moveBack: val.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 5/16/2006 13:28'!
moveDown: val
	"move in this direction by val units"
	replica future moveDown: val.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 5/7/2006 22:36'!
moveForward: val
	"move in this direction by val units"
	replica future moveForward: val.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 5/7/2006 22:36'!
moveLeft: val
	"move in this direction by val units"
	replica future moveLeft: val.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 5/7/2006 22:36'!
moveRight: val
	"move in this direction by val units"
	replica future moveRight: val.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 5/16/2006 13:28'!
moveUp: val
	"move in this direction by val units"
	replica future moveUp: val.! !

!TAvatarUser methodsFor: 'private events' stamp: 'ar 4/24/2006 16:21'!
myFarRef
	"temporary evilness while we're working towards Croquet/Tweak integration"
	^myFarRef! !

!TAvatarUser methodsFor: 'private events' stamp: 'ar 4/24/2006 16:21'!
myFarRef: farRef
	"temporary evilness while we're working towards Croquet/Tweak integration"
	myFarRef := farRef! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 1/28/2006 20:59'!
newPostcard

	^ newPostcard.! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 1/28/2006 20:59'!
newPostcard: npc

	newPostcard := npc.! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 1/29/2006 13:29'!
newTransform

	^ newTransform.! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 2/3/2006 00:26'!
newTransform: tfm

	newTransform := tfm.! !

!TAvatarUser methodsFor: 'private events' stamp: 'ar 4/24/2006 16:22'!
on: eventName notify: observer
	"Allow transparent forwarding of events from the replica for a simpler programming model"
	(replica isFarRef and:[replica isResolved]) ifTrue:[
		replica on: eventName notify: self.
	].
	^super on: eventName notify: observer.! !

!TAvatarUser methodsFor: 'private events' stamp: 'ar 4/24/2006 16:22'!
on: eventName unNotify: observer
	"Allow transparent forwarding of events from the replica for a simpler programming model"
	(replica isFarRef and:[replica isResolved]) ifTrue:[
		replica on: eventName unNotify: self.
	].
	^super on: eventName unNotify: observer! !

!TAvatarUser methodsFor: 'events' stamp: 'ar 2/2/2006 23:56'!
onSnapshot: postcard
	snapshots add: (TSnapshot new postcard: postcard).
! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:40'!
orientation

	^(replica get: #orientation).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:41'!
outVector

	^(replica get: #outVector).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:41'!
pitch

	^(replica get: #pitch).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 4/13/2006 20:42'!
pitch: pt

	replica future pitch:pt.! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:41'!
pitchYawRoll

	^(replica get: #pitchYawRoll).! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 5/28/2006 16:33'!
portalRay

	^ portalRay.! !

!TAvatarUser methodsFor: 'portals' stamp: 'das 9/11/2006 13:35'!
prepareRayTest

	| pos delta trans  |


	pos := (replica get: #translation).
"	lastPosition := (replica get: #lastPosition)."
lastPosition ifNil:[lastPosition := pos. ].
	(pos - lastPosition)length>0.0001 ifTrue:[
		portalRay resetTotal.
		delta := lastPosition-pos.
		(delta x = 0 and:[delta z = 0])ifTrue:[
			trans := Matrix4x4 at: delta up:(1@0@0) ]ifFalse:[
			trans := Matrix4x4 at: delta up:(0@1@0) ].
		trans translation: lastPosition.
		portalRay globalTransform: trans.
		portalRay maxDistance: (delta length max:0.5).
		portalRay doSelect: true. "turn this on only if we have a valid movement"
		lastPosition := pos.
		^ true.
	].
	portalRay doSelect: false. "there is no ray to test for - don't use the last one"
	^ false.! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:41'!
quaternion

	^(replica get: #quaternion).! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 2/2/2006 20:02'!
replica
	^replica! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 3/27/2006 15:48'!
replicaClass
	^replicaClass! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 4/18/2006 15:40'!
replicaClass: aClass

	replicaClass := aClass.! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:41'!
roll

	^(replica get: #roll).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 5/7/2006 22:18'!
sideVector

	^(replica get: #sideVector).! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 2/2/2006 23:54'!
snapshots
	^snapshots! !

!TAvatarUser methodsFor: 'accessing' stamp: 'ar 2/2/2006 23:54'!
snapshots: aCollection
	snapshots := aCollection! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 1/29/2006 14:13'!
speedControl: sc 

	speedControl := sc.
	replica future speedControl: sc.
! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 5/2/2006 11:43'!
speedControl: sc pointer: pointer redButton: redButton
	| |

	self speedControl: sc.
	
	self laserControl: pointer redButton: redButton! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 4/13/2006 16:10'!
startDriving

	driving := true.
	replica future startDriving.
	self signal: #startDriving.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 4/16/2006 22:14'!
stopDriving

	driving := false.
	replica future stopDriving.
! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 5/23/2006 13:28'!
takeMeThere: event

	"Go near whatever we're pointing to."

	| selection target ep |
	"Force pick, but restricted to the current space."
 
	harness ogl forcePick: true.
	harness ogl suppressPortals: true.
	harness render.
	ep := harness ogl eventPointer.
	harness ogl suppressPortals: false.

	ep ifNil: [
		^self.
	].
	target := ep selection frameID.

	"If we found something at the pick point, go to it."
	target ifNotNil: [
		selection := ep selection.
		replica future takeMeTo: selection orientation: ep globalOrientation.
	].



! !

!TAvatarUser methodsFor: 'portals' stamp: 'das 5/27/2006 14:15'!
testEnter: ogl portal: portal

	| pc vwPnt m1 m2 trans newSpace |
"this can only determine if a portal transition needs to occur. You can't actually do the portal transition here, because we are actually inside of the portal's Island. The actual transition is handled by the CroquetHarness if the newPostcard field of this object is not nil."
 	pc := portal postcardLink.
	portal = pc ifTrue:[ ^false].
	vwPnt := pc resolveViewpoint: ogl.	
	vwPnt ifNil:[ ^ false. ].
	portalRay doSelect ifFalse:[ ^ false. ].
	vwPnt send:[ :toPortal |
		toPortal root ifNil:[^ false]. "portal is closed"

		(portalRay pointerPick: portal boundSphere) ifTrue:[
			m1 := portal globalTransform.
			toPortal isPortal ifTrue:[ m2 := toPortal globalMatrixOut.]
			ifFalse:[ m2 := toPortal globalTransform.].
			trans := m2 composeWith: m1 orthoNormInverse.
			newTransform := trans composeWith: ogl avatar globalTransform.
			newSpace := toPortal root.
			newSpace send:[:ns |  newPostcard := ns postcard.].
			^ true.
		].
	].
	^ false.! !

!TAvatarUser methodsFor: 'accessing' stamp: 'das 6/7/2006 23:21'!
thirdPerson: delta
	"Set the viewpoint relative to the avatar... 0 is inside, 1 is just behind/above, 3 is far back"
	thirdPerson := delta max: 0.
! !

!TAvatarUser methodsFor: 'accessing' stamp: 'bgf 3/15/2006 12:02'!
thirdPersonLag: factor
	"Set the lag factor by which the camera chases the avatar."
	thirdPersonLag := factor max: 1.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 6/7/2006 23:21'!
toggleAvatarPerson
	thirdPerson := thirdPerson - 1.
	thirdPerson < 0 ifTrue:[thirdPerson := 3].
! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:37'!
translation

	^(replica get: #translation).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 5/1/2006 17:07'!
translation: trans

	replica future translation: trans.! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:42'!
upVector

	^(replica get: #upVector).! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 9/13/2006 00:20'!
updateCamera: aCamera
	"Update camera to follow avatar"
	| at ct trans bt weight tp2 sticky lockStep |

	replica ifNil:[^false].
	jumpCameraTrans ifNotNil:[
		aCamera localTransform: jumpCameraTrans.
		jumpCameraTrans := nil.
	].
	(viewAngle = aCamera viewAngle and:[viewAspect = aCamera aspect]) ifFalse:[
		viewAngle := aCamera viewAngle.
		viewAspect := aCamera aspect.
		replica ifNotNil:[replica future viewAngle: viewAngle aspect: viewAspect].
	].
	sticky := replica get: #sticky.
	(sticky or:[cameraBehind]) ifTrue:[
		"Put camera directly into the spot"
		^aCamera localTransform: (replica get: #globalTransform).
	].

	lockStep := (replica get:#lockStep).
	tp2 := thirdPerson squared.
	at := (replica get: #globalTransform).
	self signal:#pitchYawRoll with: at pitchYawRoll.
	ct := aCamera localTransform.
	weight := 1.0 / (thirdPersonLag max: 1).
	trans := Matrix4x4 identity. 
	trans translation: (Vector3 x:0.0 y:0.7 z: 4)*tp2.
	bt := at composeWith: trans. "This is the target position of the camera"
	lockStep ifTrue:[aCamera localTransform: bt. ] ifFalse:[
		trans := ct translation + ((bt translation - ct translation) * weight).
		at := at asQuaternion.
		ct := ct asQuaternion.
		ct := ct slerpTo: at at: weight.
		aCamera localTransform: (ct asMatrix4x4 translation:trans).
	].! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 4/4/2006 18:01'!
updateUser
	replica ifNil:[^nil].
	replica future driveStep.
	replica isResolved ifFalse:[^nil].
! !

!TAvatarUser methodsFor: 'accessing' stamp: 'bgf 7/19/2006 17:57'!
userID
	"This identifies (distinguishes) a user for the duration of a harness session."
	^ userID! !

!TAvatarUser methodsFor: 'accessing' stamp: 'bgf 7/19/2006 17:49'!
userID: aTObjectID
	userID := aTObjectID! !

!TAvatarUser methodsFor: 'private events' stamp: 'ar 4/24/2006 16:24'!
valueWithEvent: anEvent
	"Resignal an event from the replica"
	^self privateSignalEvent: (ScriptEvent new
		setSender: self selector: anEvent selector arguments: anEvent arguments
	)! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 3/20/2006 21:42'!
yaw

	^(replica get: #yaw).! !

!TAvatarUser methodsFor: 'proxy' stamp: 'das 4/13/2006 20:42'!
yaw: yw

	replica future yaw:yw.! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 6/7/2006 23:26'!
zoomIn

	thirdPerson := thirdPerson*0.9.
	thirdPerson <= 0.5 ifTrue:[thirdPerson := 0.0].
! !

!TAvatarUser methodsFor: 'driving' stamp: 'das 6/7/2006 23:27'!
zoomOut

	thirdPerson = 0.0 ifTrue:[thirdPerson := 0.9].
	thirdPerson := (thirdPerson*1.1) min: 31.6227766016838.
	
"1000 sqrt 31.6227766016838"! !

!TFrame class methodsFor: 'accessing' stamp: 'ar 6/20/2006 15:22'!
defaultObjectName
	"Answer the default object name for instances:
		TWindow defaultObjectName -> 'TWindow'
		TCube defaultObjectName -> 'TCube'
	"
	^self name asString! !

!TFrame class methodsFor: 'accessing' stamp: 'das 8/10/2006 10:26'!
eventBlockRecurse

	^ EventBlockRecurse.! !

!TFrame class methodsFor: 'accessing' stamp: 'das 1/29/2006 14:32'!
eventKeyboard

	^ EventKeyboard.! !

!TFrame class methodsFor: 'accessing' stamp: 'das 1/29/2006 14:32'!
eventNone

	^ EventNone.! !

!TFrame class methodsFor: 'accessing' stamp: 'das 1/29/2006 14:32'!
eventPointerDown

	^ EventPointerDown.! !

!TFrame class methodsFor: 'accessing' stamp: 'das 1/29/2006 14:32'!
eventPointerOver

	^ EventPointerOver.! !

!TFrame class methodsFor: 'class initialization' stamp: 'das 8/6/2006 22:05'!
initialize
	"TFrame initialize"
	EventNone := 0.
	EventKeyboard := 1.
	EventPointerDown := 2.
	EventPointerOver := 4.
	EventBlockRecurse := 32.

	"EventMaskBits map event names to their required eventMask bits.
	E.g., for getting notifications about the pointerOver event we need to set
	the EventPointerOver bit and in this dictionary provides that mapping."
	EventMaskBits := IdentityDictionary new.
	{
		{EventKeyboard. #keyDown. #keyStroke. #keyUp }.
		{EventPointerDown. #pointerDown. #pointerUp. #pointerMove }.
		{EventPointerOver. #pointerEnter. #pointerLeave. #pointerOver }.
	} do:[:spec|
		spec allButFirst do:[:evtName| EventMaskBits at: evtName put: spec first].
	].! !

!TFrame class methodsFor: 'instance creation' stamp: 'ar 1/30/2006 11:44'!
new
	protoData ifNil:[^self basicNew initialize].
	^TIslandCopyImporter new importData: protoData.! !

!TFrame class methodsFor: 'instance creation' stamp: 'ar 1/30/2006 11:48'!
prototype: aTFrame
	aTFrame ifNil:[^protoData := nil].
	protoData := TIslandCopier new export: aTFrame.
! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 7/8/2006 15:34'!
addChild: child

	frameChildren ifNil:[frameChildren := OrderedCollection new.].

"We can not add this frame until we have removed it from a previous parent."

	child isFrame ifTrue:[
		child parent ifNotNil:[child parent removeChild: child].
"Register the child if it can handle events. It remains its own object owner."
		child eventMask~=EventNone ifTrue:[child register.] 
		ifFalse:[ objectOwner eventMask ~= EventNone ifTrue:[
			objectOwner addOwnership: child.].
		].
"make sure child and all its subframes are informed of all viewers so they can be added to their tea parties"
		frameChildren add: child.
		child root: self root.
		child parent: self.
		child addLightFrames.
		child addPortalFrames.
		child addRayFrames.
		self signal: #structureChanged.
		child frameChanged.
		self root ifNotNil:[self root doStart:child. ].
	] ifFalse:[ child isPostcard ifTrue:[ frameChildren add: child. ].].! !

!TFrame methodsFor: 'frameManagement' stamp: ''!
addLightFrame: litefrm

	frameParent ifNotNil:[frameParent addLightFrame: litefrm.]! !

!TFrame methodsFor: 'frameManagement' stamp: 'DPR 1/9/2004 14:05'!
addLightFrames

	" When a frame is added as a child, any lights included in the heirarchy are reported to the base TRoom."
	self isLight ifTrue:[ self addLightFrame: self].
	frameChildren ifNotNil: [frameChildren do:[ :fc | fc addLightFrames ].].
	! !

!TFrame methodsFor: 'ownership' stamp: 'ar 1/27/2006 14:23'!
addOwnership: frame

	frame = self ifTrue:[ ^self ]. "we don't add ourselves to the list"
	frame eventMask~=EventNone ifTrue:[ ^self ]. "don't add components"
	ownedObjects ifNil:[ ownedObjects := Dictionary new.].
	ownedObjects at: frame objectID put: frame.
	frame forceObjectOwner: self.
	frame frameChildrenDo:[:child | self addOwnership: child. ].
! !

!TFrame methodsFor: 'frameManagement' stamp: ''!
addPortalFrame: prtlFrm

	frameParent ifNotNil:[frameParent addPortalFrame: prtlFrm.]! !

!TFrame methodsFor: 'frameManagement' stamp: 'DPR 1/9/2004 14:05'!
addPortalFrames

	" When a frame is added as a child, any portals included in the heirarchy are reported to the base TRoom."
	self isPortal ifTrue:[ self addPortalFrame: self].
	frameChildren ifNotNil: [frameChildren do:[ :fc | fc addPortalFrames ].].
	! !

!TFrame methodsFor: 'frameManagement' stamp: 'das 11/9/2002 15:06'!
addRayFrame: rayfrm

	frameParent ifNotNil:[frameParent addRayFrame: rayfrm.]! !

!TFrame methodsFor: 'frameManagement' stamp: 'DPR 1/9/2004 14:05'!
addRayFrames

	" When a frame is added as a child, any rays included in the heirarchy are reported to the base TSpace."
	self isRay ifTrue:[ self addRayFrame: self].
	frameChildren ifNotNil: [frameChildren do:[ :fc | fc addRayFrames ].].
	! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:28'!
addRotationAroundX: anAngle

	localTransform addRotationAroundX: anAngle.
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:29'!
addRotationAroundY: anAngle

	localTransform addRotationAroundY: anAngle.
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:29'!
addRotationAroundZ: anAngle

	localTransform addRotationAroundZ: anAngle.
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:12'!
addYaw: y
	self yaw: self yaw + y.! !

!TFrame methodsFor: 'Islands' stamp: 'das 12/13/2005 09:12'!
asFarRef

	^ Processor activeIsland asFarRef:self.! !

!TFrame methodsFor: 'accessing' stamp: 'das 4/25/2006 11:35'!
at: index

	^ frameChildren at: index.! !

!TFrame methodsFor: 'scripts' stamp: 'ar 10/3/2004 11:04'!
attachBehavior: aBehavior
	aBehavior target == self ifFalse:[^aBehavior attachTo: self].
	self myBehaviors: (self myBehaviors copyWith: aBehavior).! !

!TFrame methodsFor: 'scripts' stamp: 'ar 7/6/2006 12:13'!
attachScript: scriptCode named: scriptName owner: avatar
	"Attach the given script to the receiver"
	| mgr |
	mgr := self root scriptManager.
	mgr ifNotNil:[mgr attachScript: scriptCode named: scriptName to: self owner: avatar].! !

!TFrame methodsFor: 'accessing' stamp: 'das 1/21/2005 23:36'!
baseMaterial: mat
	frameChildren ifNil:[^self].
	frameChildren do:[:each| each material: mat].! !

!TFrame methodsFor: 'transform' stamp: 'das 8/28/2006 02:46'!
billboard: cam

	| mat pos out target dir sin angle |
	mat := self globalTransform.
	pos := mat a14 @ mat a34.
	out := (mat a13 @ mat a33) normalized.
	mat := cam globalTransform.
	target := mat a14 @ mat a34.
	dir := (target - pos) normalized.
	sin := dir crossProduct: out.
	angle := ((dir dotProduct: out)*0.9999999) arcCos radiansToDegrees.
	mat := Matrix4x4 identity.
	mat rotationAroundY: angle * sin sign.
	localTransform := localTransform composedWithLocal: mat.	
	globalTransform := globalTransform composedWithLocal: mat.
! !

!TFrame methodsFor: 'accessing' stamp: ''!
boundSphere

	" This will return a render bounds object if one exists. This is a TBoundSphere (location + radius), which is used to determine if a particular object is inside the clipping planes or for collision detection tests. Otherwise, the object will be assumed to have a local coordinate location of 0,0,0 and a radius of 1.0."
	^ nil.

! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 3/9/2004 13:23'!
boundSpheresDo: aBlock
	| bSphere |
	bSphere := self boundSphere.
	bSphere ifNotNil:[^aBlock value: bSphere].
	frameChildren ifNil:[^self].
	self frameChildren do:[:each| each boundSpheresDo: aBlock].
! !

!TFrame methodsFor: 'accessing' stamp: 'das 5/16/2005 09:08'!
boundingBox
	"Answer the bounding box for the receiver and its children"
	^ self boundingBox: Matrix4x4 identity.
! !

!TFrame methodsFor: 'accessing' stamp: 'das 5/22/2005 14:29'!
boundingBox: trans
	"Answer the bounding box for the receiver and its children"
	| box childBox |
	box := self frameBox.
	self frameChildrenDo:[:child|
		childBox := child boundingBox transformedBy: (trans * child localTransform).
		box := box merge: childBox.
	].
	^box! !

!TFrame methodsFor: 'accessing' stamp: 'das 7/9/2003 12:19'!
boundsChanged

	" Does nothing"
! !

!TFrame methodsFor: 'accessing' stamp: 'das 4/28/2002 23:40'!
boundsDepth: depth

	frameChildren ifNotNil:[
		frameChildren do:[:fc | fc boundsDepth: depth].].! !

!TFrame methodsFor: 'accessing' stamp: 'das 3/9/2004 15:21'!
changed: frame

	" Does nothing, just forward on to parent."
	self parent ifNotNil:[self parent changed: frame.]
! !

!TFrame methodsFor: 'hierarchy' stamp: ''!
child: index

	^frameChildren at: index.! !

!TFrame methodsFor: 'accessing' stamp: 'das 4/25/2006 17:51'!
childAt: index

	^ frameChildren at: index.! !

!TFrame methodsFor: 'accessing' stamp: 'das 3/9/2004 14:35'!
childChanged	
	
	self childChanged: self.! !

!TFrame methodsFor: 'accessing' stamp: 'das 3/9/2004 14:35'!
childChanged: frame

	" Does nothing, just forward on to parent."
	self parent ifNotNil:[self parent changed: frame.]
! !

!TFrame methodsFor: 'accessing' stamp: 'das 6/5/2005 22:48'!
clipPlane

	^Vector3 x: 0.0 y:0.0 z:-1.0.
! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 5/22/2005 14:29'!
collapse

"Forces all frames to identity transform while retaining translation, and pushes the previous transform to the frame children. This is used primarily for meshes and their supporting groups."
	| orient trans |

	trans := self translation.
	orient := self orientation.
	frameChildren ifNotNil:[ 
		frameChildren do:[ :fc |  
			fc localTransform: (orient * fc localTransform).
			fc collapse.
		].
	].
	self localTransform: Matrix4x4 identity.
	self translation: trans.
! !

!TFrame methodsFor: 'events' stamp: 'das 3/9/2004 13:25'!
collidesInto: aFrame
	"Answer whether I collide into the given frame"
	| mySphere itsSphere |
	mySphere := self compositeBoundSphere ifNil:[^false].
	itsSphere := aFrame compositeBoundSphere ifNil:[^false].
	^mySphere collideSphere: itsSphere! !

!TFrame methodsFor: 'enterExit' stamp: 'das 5/27/2006 16:47'!
collision: frame

	^ false.! !

!TFrame methodsFor: 'accessing' stamp: 'bgf 7/9/2006 13:40'!
colorize: color

	"Use the given color to paint the frame and its children 
       as both their ambient and diffuse colors."! !

!TFrame methodsFor: 'accessing' stamp: 'das 3/9/2004 13:25'!
compositeBoundSphere
	"Answer either my own or a composition of my children's bound spheres"
	| children bSphere |
	children := OrderedCollection new.
	self boundSpheresDo:[:bs|
		bSphere := bs union: bSphere.
		children add: bs.
	].
	children size = 0 ifTrue:[^nil].
	children size = 1 ifTrue:[^children first].
	bSphere children: children.
	bSphere frame: self.
	bSphere transform: self globalTransform.
	^bSphere
! !

!TFrame methodsFor: 'accessing' stamp: 'das 5/11/2005 19:30'!
costume

	^ self.! !

!TFrame methodsFor: 'accessing' stamp: 'das 8/27/2006 21:58'!
cullFace: bool

	frameChildren ifNotNil:[frameChildren do:[:fc | fc cullFace: bool]].! !

!TFrame methodsFor: 'accessing' stamp: 'bgf 7/9/2006 14:58'!
currentHiliteColor

	"If this frame is currently hilit, return the hilighting color in use;
       otherwise return nil."

	^ hiliteColor! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 5/11/2002 13:56'!
currentParent: fParent
	
	frameParent := fParent.! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 1/27/2005 06:22'!
delete
	self destroy.! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 12/13/2005 14:04'!
destroy

	frameParent ifNotNil:[frameParent removeChild: self].
! !

!TFrame methodsFor: 'scripts' stamp: 'ar 10/3/2004 11:04'!
detachBehavior: aBehavior
	aBehavior target == nil ifFalse:[
		aBehavior target == self ifFalse:[^self error:'Behavior is not owned by me'].
		^aBehavior detach].
	self myBehaviors: (self myBehaviors copyWithout: aBehavior).! !

!TFrame methodsFor: 'render' stamp: 'das 6/5/2005 15:38'!
disableClipPlane: ogl
	
	ogl glDisable: GLClipPlane0.
! !

!TFrame methodsFor: 'accessing' stamp: 'ar 10/10/2002 16:54'!
distanceTo: aFrame
	^(self globalPosition - aFrame globalPosition) length! !

!TFrame methodsFor: 'accessing' stamp: 'das 12/27/2002 17:46'!
do: block

	block value: self.
	frameChildren ifNotNil:[ 
		frameChildren do:[ :fc | fc do: block].].
! !

!TFrame methodsFor: 'accessing' stamp: 'das 2/2/2006 00:27'!
doForPick: block

"This is essentially a #do: method that stops at a TQuadTree instead of traversing its hierarchy as well. This is to prevent the #pick: methods from being called multiple times for the same TFrame."
	block value: self.
	frameChildren ifNotNil:[ 
		frameChildren do:[ :fc | fc doForPick: block].].
! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:53'!
doKeyDown: event

	(self eventMask anyMask: EventKeyboard) ifTrue:[
		event selection frame ifNil:[
			event selection frame: (self getOwnedObject: event selection frameID).
		].
		self keyDown: event.
		self signal: #keyDown with: event.
	] ifFalse:[
"otherwise recurse up the hierarchy"
		frameParent ifNotNil:[
			frameParent objectOwner doKeyDown: event.
		].
	].! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:54'!
doKeyStroke: event

	(self eventMask anyMask: EventKeyboard) ifTrue:[
		event selection frame ifNil:[
			event selection frame: (self getOwnedObject: event selection frameID).
		].
		self keyStroke: event.
		self signal: #keyStroke with: event.
	] ifFalse:[
"otherwise recurse up the hierarchy"
		frameParent ifNotNil:[
			frameParent objectOwner doKeyStroke: event.
		].
	].! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:54'!
doKeyUp: event

	(self eventMask anyMask: EventKeyboard) ifTrue:[
		event selection frame ifNil:[
			event selection frame: (self getOwnedObject: event selection frameID).
		].
		self keyUp: event.
		self signal: #keyUp with: event.
	] ifFalse:[
"otherwise recurse up the hierarchy"
		frameParent ifNotNil:[
			frameParent objectOwner doKeyUp: event.
		].
	].! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:50'!
doPointerDown: event

	(self eventMask anyMask: EventPointerDown) ifTrue:[
		(currentUser isNil or:[currentUser = event userID]) ifTrue:[
			event selection frame ifNil:[
				event selection frame: (self getOwnedObject: event selection frameID).
			].
			self pointerDown: event.
			currentUser := event userID.
			self signal: #pointerDown with: event.
		].
	] ifFalse:[
"now recurse up the hierarchy"
		frameParent ifNotNil:[
			frameParent objectOwner doPointerDown: event.
		].
	].! !

!TFrame methodsFor: 'events' stamp: 'das 8/6/2006 22:07'!
doPointerEnter: event

	(self eventMask anyMask: EventPointerOver) ifTrue:[
		currentUser ifNil:[
			event selection frame ifNil:[
				event selection frame: (self getOwnedObject: event selection frameID).
			].
			currentUser := event userID.
			self pointerEnter: event.
			self signal: #pointerEnter with: event.
		].
	].
	(self eventMask anyMask: EventBlockRecurse)ifFalse:[
"now (always) recurse up the hierarchy"
		frameParent ifNotNil:[
			frameParent objectOwner doPointerEnter: event.
		].
	].! !

!TFrame methodsFor: 'events' stamp: 'das 8/6/2006 22:07'!
doPointerLeave: event

	(self eventMask anyMask: EventPointerOver) ifTrue:[
		currentUser = event userID ifTrue:[
			event selection frame ifNil:[
				event selection frame: (self getOwnedObject: event selection frameID).
			].
			self signal: #pointerLeave with: event.
			self pointerLeave: event.
			currentUser := nil.
		].
	].
	(self eventMask anyMask: EventBlockRecurse)ifFalse:[
"now (always) recurse up the hierarchy"
		frameParent ifNotNil:[
			(frameParent objectOwner getAllEventsHandled anyMask: EventPointerOver) ifTrue:[
				frameParent objectOwner doPointerLeave: event.
			].
		].
	].! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:51'!
doPointerMove: event

	(self eventMask anyMask: EventPointerDown) ifTrue:[
		currentUser = event userID ifTrue:[
			event selection frame ifNil:[
				event selection frame: (self getOwnedObject: event selection frameID).
			].
			self pointerMove: event.
			self signal: #pointerMove with: event.
		].
	] ifFalse:[

"now recurse up the hierarchy"
		frameParent ifNotNil:[
			frameParent objectOwner doPointerMove: event.
		].
	].! !

!TFrame methodsFor: 'events' stamp: 'das 8/6/2006 22:07'!
doPointerOver: event

	(self eventMask anyMask: EventPointerOver) ifTrue:[
		currentUser = event userID ifTrue:[
			event selection frame ifNil:[
				event selection frame: (self getOwnedObject: event selection frameID).
			].
			self pointerOver: event.
			self signal: #pointerOver with: event.
		].
	].

	(self eventMask anyMask: EventBlockRecurse)ifFalse:[
	"now (always) recurse up the hierarchy"
		frameParent ifNotNil:[
			frameParent objectOwner doPointerOver: event.
		].
	].! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:52'!
doPointerUp: event

	(self eventMask anyMask: EventPointerDown) ifTrue:[
		currentUser = event userID ifTrue:[
			event selection frame ifNil:[
				event selection frame: (self getOwnedObject: event selection frameID).
			].
			self pointerUp: event.
			self signal: #pointerUp with: event.
		].
	] ifFalse:[

"now recurse up the hierarchy"
		frameParent ifNotNil:[
			frameParent objectOwner doPointerUp: event.
		].
	].! !

!TFrame methodsFor: 'render' stamp: 'bgf 7/9/2006 22:47'!
doRenderAlpha: ogl
	"This is used to render transparent ( alpha blended ) objects."

	| hilite |

	hilite := false.
	self currentHiliteColor ifNotNil: [
		hilite := ogl forceHilite: self currentHiliteColor.
		hilite ifTrue: [
			ogl pushFog.
		].
	].

	ogl transparency < 1.0 ifTrue:[
		self render: ogl.
	].
	self renderAlpha: ogl.

	hilite  ifTrue: [
		ogl forceHilite: nil.
		ogl popFog.
	].! !

!TFrame methodsFor: 'frameManagement' stamp: 'das 1/21/2006 16:43'!
doUpdate

	frameRoot ifNotNil:[
		frameRoot addUpdate: self.
	].! !

!TFrame methodsFor: 'events' stamp: 'tak 10/15/2004 16:57'!
dropFiles: aFileStream pointer: pointer 
	self
		signal: #dropFiles
		with: pointer
		with: aFileStream! !

!TFrame methodsFor: 'render' stamp: 'das 6/5/2005 22:48'!
enableClipPlane: ogl
	
"------ enableClipPlane is used to clip the TSpace to the front face of the portal. This ensures that objects don't get rendered in front of the portal, which would look bad and be confusing.------"
	| equation ov |

"------ We need to do it this way, because Squeak does not directly support doubles and gllClipPlane requires an array of same. ------"
	ov := self clipPlane.
	equation := ExternalData fromHandle: (ExternalAddress allocate: 8*4) type:ExternalType double.
	
	equation getHandle doubleAt: 1 put: ov x.
	equation getHandle doubleAt: 9 put: ov y.
	equation getHandle doubleAt: 17 put: ov z.
	equation getHandle doubleAt: 25 put: 0.0.

	ogl
		glPushMatrix;
		glMultTransposeMatrixf: self globalTransform;

		glClipPlane: GLClipPlane0 with: equation;


		glEnable: GLClipPlane0;

		glPopMatrix.
	equation free.! !

!TFrame methodsFor: 'events' stamp: 'das 5/21/2002 08:37'!
event2D: event2D! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:11'!
eventMask
	"Answer the receiver's event mask"
	^eventMask ifNil:[EventNone]

"There are three kinds of events - EventKeyboard, EventPointerOver, and EventPointerDown.  If this TFrame can handle any combination of events it must return the bitOr combination of the event types, hence, if you handle all three kinds of events, you return:

 ^ EventKeyboard bitOr: EventPointerOver bitOr: EventPointerDown. 

If you handle just the keyboard event, return:

 ^ EventKeyboard. 

The default is no event or EventNone which we return here."
! !

!TFrame methodsFor: 'events' stamp: 'bgf 8/8/2006 16:32'!
eventMask: newMask

	"The object owner magic in addChild 
	 sets up objectOwner links if the child has a 0 event mask.
	 So, when non-zeroing a mask, we have to undo that situation."

	(eventMask = EventNone  & (newMask ~= EventNone)) ifTrue: [
		objectOwner removeOwnership: self.
		"My kids may have been disowned while I had no mask...
	        so reclaim them."
		self frameChildrenDo: [:child | self addOwnership: child].
	]. 
	eventMask := newMask.
! !

!TFrame methodsFor: 'yellow-scripts-support' stamp: 'ar 9/8/2002 16:59'!
eventsGenerated
	"Answer a list of events generated by the receiver"
	^#(
		pointerDown pointerMove pointerUp
		pointerEnter pointerOver pointerLeave
		keyDown keystroke keyUp
		frameChanged
	).
! !

!TFrame methodsFor: 'accessing' stamp: 'das 3/9/2004 14:35'!
extent
	"Answer the bounding box for the receiver and its children"
	^ self boundingBox extent.
! !

!TFrame methodsFor: 'accessing' stamp: 'das 3/9/2004 14:35'!
extent: ext
	"Does nothing"! !

!TFrame methodsFor: 'accessing' stamp: 'das 8/15/2002 22:59'!
find: blockTest
	| list |
	blockTest ifNil:[^ nil].
	list := OrderedCollection new.
	self find: blockTest into: list.
	^ list.! !

!TFrame methodsFor: 'accessing' stamp: 'DPR 1/9/2004 14:05'!
find: blockTest into: list

	frameChildren ifNotNil:[ 
		frameChildren do:[ :fc | fc find: blockTest into: list].].
	(blockTest value: self) ifTrue:[ list add: self].
! !

!TFrame methodsFor: 'frameManagement' stamp: 'sn 2/26/2018 12:31'!
forceGlobalToLocal

"This is used by TMesh when the imported vertices are pre-transformed, and all of the nodes are in global coordinates and we want to put them back into their untransformed state. This occurs with the 3DS Max ASE files."

	frameChildren ifNotNil:[ frameChildren do:[ :fc | fc forceGlobalToLocal.]].
	frameParent ifNotNil:[
		self localTransform: (
			frameParent localTransform orthoNormInverse * self localTransform).
		].! !

!TFrame methodsFor: 'ownership' stamp: 'ar 1/16/2006 18:53'!
forceObjectOwner: oOwner

	objectOwner := oOwner.! !

!TFrame methodsFor: 'yellow-scripts' stamp: 'ar 7/21/2002 18:48'!
forwardBy: aDistance duration: aDuration style: aStyle rate: aSamplingRate
	^self move: #forward distance: aDistance duration: aDuration style: aStyle rate: aSamplingRate! !

!TFrame methodsFor: 'accessing' stamp: 'das 1/24/2005 00:57'!
frameBox
	"Answer the local bounding box of this frame"
	self boundSphere ifNotNil:[
		^ self boundSphere box.].
	^TBox origin: 0@0@0 corner: 0@0@0! !

!TFrame methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:41'!
frameChanged

	" This guarantees that if we make a change in a parent frame, all of the children are aware and can deal with it properly."
	frameChanged := true.
	frameChildren ifNotNil:[
		frameChildren do:[ :child | child frameChanged ].].
	myProperties ifNotNil:[self signal: #frameChanged]. "to inform interested parties"
! !

!TFrame methodsFor: 'accessing' stamp: ''!
frameChildren

	^ frameChildren.! !

!TFrame methodsFor: 'accessing' stamp: 'DPR 1/9/2004 14:05'!
frameChildrenDo: aBlock
	frameChildren ifNil:[^self].
	^frameChildren do: aBlock! !

!TFrame methodsFor: 'transform' stamp: 'das 10/3/2004 16:37'!
frameLookAt: frame

" used to put my lookAt vector into the frame of the argument"
	^ frame inverseGlobalOrientation localPointToGlobal: self lookAt negated.
! !

!TFrame methodsFor: 'accessing' stamp: 'das 4/15/2003 17:10'!
fullBright: bool

	frameChildren ifNotNil:[frameChildren do:[:fc | fc fullBright: bool]].! !

!TFrame methodsFor: 'events' stamp: 'ar 1/27/2006 14:23'!
getAllEventsHandled

" this finds the entire hierarchical event list. Even though this child may not handle an event, it's parent might"
	frameParent ifNotNil:[
		^self eventMask bitOr:frameParent objectOwner getAllEventsHandled.
	] ifNil:[ ^ self eventMask.].
	! !

!TFrame methodsFor: 'ownership' stamp: 'ar 1/17/2006 15:33'!
getOwnedObject: id

	id = objectID ifTrue:[^self].
ownedObjects ifNil:[^nil].
	^ id ifNotNil:[ownedObjects at:id ifAbsent:[nil]] ifNil:[nil].! !

!TFrame methodsFor: 'transform' stamp: 'das 10/14/2005 15:51'!
globalMatrixOut
" This simply reverses the direction of the matrix for TPortals (see TPortal>>#globalMatrixOut). Otherwise, we just use the globalTransform."

	^self globalTransform copy.
! !

!TFrame methodsFor: 'transform' stamp: 'sn 2/23/2018 12:44'!
globalOrientation
"This is used to return just the orientation part of the matrix.The translation part is 0.0."
	| mat |

	mat := self globalTransform shallowCopy.
	mat translationX: 0.0 y: 0.0 z: 0.0.
	^ mat.! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:16'!
globalPitch

	^self globalPitchYawRoll x! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:19'!
globalPitchYawRoll
	globalTransform ifNil:[^ 0.0@0.0@0.0].
	^globalTransform pitchYawRoll! !

!TFrame methodsFor: 'transform' stamp: 'ar 2/24/2006 11:28'!
globalPosition

	^ self globalTransform translation.! !

!TFrame methodsFor: 'transform' stamp: 'ar 2/24/2006 11:29'!
globalPosition: gp
	self translation: (frameParent ifNil:[gp] ifNotNil:[frameParent globalToLocal: gp]).
	! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:16'!
globalRoll
	^self globalPitchYawRoll z! !

!TFrame methodsFor: 'accessing' stamp: 'ar 10/8/2002 17:18'!
globalToLocal: anObject
	^self globalTransform inverseTransformation localPointToGlobal: anObject! !

!TFrame methodsFor: 'transform' stamp: 'ar 2/24/2006 11:27'!
globalTransform
	" This is the global transform of the frame - its location and orientation in global world coordinates. We only
	calculate this when we have to, which is only when a local coordinate frame has been changed AND when someone 	asks for it. "

	frameChanged ifTrue:[	
		frameParent ifNotNil: [ self globalTransform: 
			((frameParent globalTransform) * self localTransform)] ifNil:[
		self globalTransform:  self localTransform.].
	].
	^ globalTransform.
" NOT ^ self globalTransform - this causes bad recursion error!!"
					 
! !

!TFrame methodsFor: 'transform' stamp: 'ar 2/24/2006 11:27'!
globalTransform: gTrans
	
	frameChanged := false.
	globalTransform := gTrans.
	self globalTransformUpdate.
	^ globalTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:47'!
globalTransformUpdate

	" This is called when a new global transfom is calculated. It is designed to be extended by other frame sub-classes, such as TRigidBody, that would have variables that should be updated once when the frame gets changed."
	self boundSphere ifNotNil:[ self boundSphere transform: globalTransform.].
! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:16'!
globalYaw
	^self globalPitchYawRoll y! !

!TFrame methodsFor: 'events' stamp: 'das 5/21/2002 08:38'!
handlesEvent2D

	^ false.! !

!TFrame methodsFor: 'accessing' stamp: 'ar 9/7/2002 14:51'!
hasAlpha
	" Does this object have an alpha channel to render? Return true if it does."
	^ false.! !

!TFrame methodsFor: 'accessing' stamp: 'das 6/7/2006 21:21'!
hasAlpha: ogl

	^ self hasAlpha or:[ogl transparency < 1.0].! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 10/12/2004 07:53'!
hasAncestor: ancestor

"Determines if ancestor is one of our parents, parents parents, etc..."
	frameParent = ancestor ifTrue:[^ true].
	frameParent ifNil:[^ false].
	^ frameParent hasAncestor: ancestor.
! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 8/5/2006 16:53'!
hasAncestorOrMe: ancestor

"Determines if ancestor is one of our parents, parents parents, etc..."
	self = ancestor ifTrue:[^ true].
	^ self hasAncestor: ancestor.
! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 1/2/2003 17:57'!
hasChild: child

	frameChildren ifNotNil:[
		frameChildren do:[:fc | 
			fc = child ifTrue:[^ true ].].].
	^ false.
! !

!TFrame methodsFor: 'accessing' stamp: 'das 6/7/2006 21:22'!
hasOpaque

	^ true.! !

!TFrame methodsFor: 'accessing' stamp: 'das 6/7/2006 21:37'!
hasOpaque: ogl

	^ self hasOpaque and:[ogl transparency = 1.0].! !

!TFrame methodsFor: 'accessing' stamp: 'bgf 7/13/2006 00:11'!
hilite
	"Hilight this frame and its children.
	 The effect lasts until the state is cleared with 'frame unhilite.' "
 
	hiliteColor := Color white asVectorColor.! !

!TFrame methodsFor: 'accessing' stamp: 'bgf 7/14/2006 16:36'!
hilite: aColor
	"Hilight this frame and its children with a default hilite color.
	 The effect lasts until the state is cleared with 'frame unhilite.' "

	aColor ifNil: [ 
		self unhilite.			"self hilite:nil == self unhilite."
	] ifNotNil: [
	 	hiliteColor := aColor asVectorColor
	].! !

!TFrame methodsFor: 'accessing' stamp: 'das 4/30/2002 17:05'!
inertiaTensor
	
	^ nil.! !

!TFrame methodsFor: 'initialize' stamp: 'das 8/15/2002 13:43'!
initBounds
	frameChildren ifNotNil:[
		frameChildren do:[:fc | fc initBounds].].! !

!TFrame methodsFor: 'initialize' stamp: 'das 7/7/2006 19:38'!
initialize
	super initialize.
	frameChanged := true. 
	localTransform := Matrix4x4 identity.
	globalTransform := Matrix4x4 identity.
	frameChildren := nil.
	frameRoot := nil.
	eventMask := EventNone.
	self solid: true.
	self visible: true.
	objectOwner := self.
	objectID := TObjectID for: self.
	objectName := 'a', self class asString.
	self register.
	test := nil.
	^self
! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 2/26/2006 17:06'!
insertFrame: frm

	| parent |
" Insert this frame between the target frame and its parent. "
	
	self parent ifNotNil:[
		parent := self parent.
		parent removeChild: self.
		parent addChild: frm.
		].
	frm addChild: self.
! !

!TFrame methodsFor: 'transform' stamp: ''!
inverseGlobalOrientation

	^ self globalOrientation orthoNormInverse.! !

!TFrame methodsFor: 'transform' stamp: ''!
inverseGlobalTransform

	^ self globalTransform orthoNormInverse.! !

!TFrame methodsFor: 'testing' stamp: 'das 4/24/2006 21:29'!
isAvatar

	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'das 6/3/2002 12:49'!
isBrowser

	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'das 6/3/2002 12:49'!
isCamera

	^ false.! !

!TFrame methodsFor: 'hierarchy' stamp: 'DPR 1/9/2004 14:05'!
isChild: parent

	^ parent hasChild: self.
! !

!TFrame methodsFor: 'private' stamp: 'bgf 7/17/2006 12:32'!
isDropTarget: aPayload

	"Will this TFrame handle 'acceptDrop: aPayload' ?

	 This can be replaced in the fullness of time (once there is an Interface formalism
	 and an Object #supports: Interface method that works quickly).
	 For now, the method stands for 'self respondsTo:#acceptDrop:',
       and is needed because 'respondsTo:' is far too slow to use in interactive mouse loops."

	^ false! !

!TFrame methodsFor: 'testing' stamp: 'das 10/16/2002 00:12'!
isFrame

	^ true.! !

!TFrame methodsFor: 'accessing' stamp: 'das 1/26/2006 21:20'!
isFrameChanged

	^ frameChanged.! !

!TFrame methodsFor: 'testing' stamp: 'das 4/10/2003 23:12'!
isGroup

	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'das 4/10/2003 23:12'!
isLight

	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'das 4/10/2003 23:15'!
isMesh

	^ false.! !

!TFrame methodsFor: 'hierarchy' stamp: 'bgf 7/12/2006 22:44'!
isOrContains: aFrame

	"Is the given frame either myself or a descendant of myself?"

	^ (aFrame == self) or: [ aFrame hasAncestor: self ].
! !

!TFrame methodsFor: 'testing' stamp: 'das 4/10/2003 23:15'!
isPortal

	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'bgf 3/20/2006 21:56'!
isPostcard
	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'das 11/9/2002 15:03'!
isRay

	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'das 5/14/2002 10:40'!
isRigidBody

	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'das 5/1/2002 22:04'!
isSolid

	^ solid! !

!TFrame methodsFor: 'testing' stamp: 'das 6/3/2003 21:49'!
isSpace

	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'das 7/8/2002 15:43'!
isTexture

	^ false.! !

!TFrame methodsFor: 'testing' stamp: 'das 7/15/2002 20:45'!
isWindow

	^ false.! !

!TFrame methodsFor: 'Islands' stamp: 'das 1/22/2006 17:56'!
islandViolation: txt
"code flagged with this needs to be moved"! !

!TFrame methodsFor: 'toys' stamp: 'das 5/16/2005 08:49'!
jump: dist

	self translation: self translation + (Vector3 x: 0 y: dist z: 0).
! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:53'!
keyDown: event
! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:53'!
keyStroke: event
! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:54'!
keyUp: event
! !

!TFrame methodsFor: 'transform' stamp: 'bgf 6/27/2006 12:36'!
localCenter

	"Return a translation to the visual center of this object.
     By default, this is just the local translation.
     This is the basis of various Go-To operations,
     so objects whose visual center is different from the frame origin
     may override this."

	^ self translation! !

!TFrame methodsFor: 'accessing' stamp: 'ar 9/8/2002 15:59'!
localToGlobal: anObject
	^self globalTransform localPointToGlobal: anObject! !

!TFrame methodsFor: 'transform' stamp: 'das 5/10/2004 15:08'!
localTransform

	^localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:31'!
localTransform: m4x4

	localTransform := m4x4.	
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'accessing' stamp: 'ar 6/6/2002 16:59'!
locator
	"Return a locator telling us where to find this guy on the net"
	^nil! !

!TFrame methodsFor: 'accessing' stamp: 'bgf 3/8/2006 23:37'!
lookAt

	"Return a normalized global-space vector pointing along the orientation of this frame."

	^ self globalTransform column3 normalized.! !

!TFrame methodsFor: 'transform' stamp: 'das 6/6/2005 15:20'!
lookAt: frame up: u

	| at side m up |

"vector between frames in global coordinates"
	at := frame globalPosition - self globalPosition.
	at length = 0.0 ifTrue:[^ 0].
"remove my transform from the point"
	at:=frameParent orientation orthoNormInverse localPointToGlobal: at.
	at normalize.
	up := u.
	up ifNil:[
		at y abs = 1.0 ifFalse:[ up := Vector3 x: 0 y: 1 z: 0.] ifTrue:[
			up := Vector3 x:0 y:0 z:-1].].
	side := (at cross: up) normalized negated.
	up := (side cross: at) normalized negated.
	m := Matrix4x4 identity.
	m a11: side x.
	m a21: side y.
	m a31: side z.
	m a12: up x.
	m a22: up y.
	m a32: up z.
	m a13: at x.
	m a23: at y.
	m a33: at z.
	m translation: self translation.
	self localTransform: m.
! !

!TFrame methodsFor: 'accessing' stamp: 'das 7/31/2004 16:03'!
lookSide

	^ self globalTransform column1.! !

!TFrame methodsFor: 'accessing' stamp: 'das 7/31/2004 16:03'!
lookUp

	^ self globalTransform column2.! !

!TFrame methodsFor: 'accessing' stamp: 'ar 9/8/2002 18:42'!
material
	frameChildren isEmptyOrNil ifTrue:[^nil].
	^frameChildren anyOne material! !

!TFrame methodsFor: 'accessing' stamp: 'ar 9/8/2002 18:41'!
material: mat
	frameChildren ifNil:[^self].
	frameChildren do:[:each| each material: mat].! !

!TFrame methodsFor: 'accessing' stamp: 'ar 10/26/2002 23:04'!
materialAlpha: ignored
	"backstop"! !

!TFrame methodsFor: 'yellow-scripts' stamp: 'das 6/30/2005 12:20'!
moveBy: aDirection duration: aDuration style: aStyle rate: aSamplingRate
	| direction duration style rate endState styleFunc lastState startState deltaTime msecsDuration proportion newState startT progressT script |
	direction := aDirection ifNil:[self defaultDirection asMoveDirection * self defaultDistance].
	duration := aDuration ifNil:[self defaultDuration].
	style := aStyle ifNil:[self defaultInterpolationStyle].
	rate := aSamplingRate ifNil:[self defaultSamplingRate].

	"Get the ultimate state we're interested in"
	endState := direction.
	duration <= 0 ifTrue:[^self transformBy: (Matrix4x4 withOffset: endState)].
	"Compute msecs since that's what we'll be working on here"
	msecsDuration := duration * 1000.
	"Get the interpolation function (evaluating from 0-1 and returning values from 0-1)"
	styleFunc := style asScriptStyle.
	"Reset the interpolation state"
	lastState := startState := 0@0@0.
	"And go moving until we're out of time"
	rate := msecsDuration * rate // 1000.
	deltaTime := msecsDuration // rate.
	startT := self now.
	script := Processor activeProcess.
	1 to: rate do:[:i|
		progressT := (i*deltaTime) asFloat.
		(startT + progressT) waitUntil. "<-- here is where tea scheduling kicks in"
		"Compute the proportion of time that's over"
		proportion := progressT / msecsDuration asFloat.
		"Map it through the interpolation style"
		proportion := styleFunc value: proportion.
		"Compute the state (e.g., offset) at our new proportion"
		newState := startState interpolateTo: endState at: proportion.
		"Transform incrementally - this is so that we can superimpose animations."
		self transformBy: (Matrix4x4 withOffset: newState - lastState).
		"Remember last state for the next round"
		lastState := newState.
		"see if a stop was requested"
		script isStopRequested ifTrue:[^self].
	].
	(startT + msecsDuration) waitUntil. "<-- here is where tea scheduling kicks in"
	"Apply the remaining part of the transform (if any)"
	self transformBy: (Matrix4x4 withOffset: endState - lastState).
! !

!TFrame methodsFor: 'yellow-scripts' stamp: 'ar 10/12/2002 18:20'!
moveTo: aTarget duration: aDuration style: aStyle rate: aSamplingRate
	| target targetDir length |
	target := aTarget ifNil:[^self].
	targetDir := self globalToLocal: target globalPosition.
	targetDir y: 0.0.
	length := targetDir length.
	targetDir := targetDir / length.
	^self move: targetDir distance: length duration: aDuration style: aStyle rate: aSamplingRate! !

!TFrame methodsFor: 'yellow-scripts-support' stamp: 'ar 3/23/2006 19:42'!
myBehaviors
	^(myProperties ifNotNil:[myProperties behaviors]) ifNil:[#()]! !

!TFrame methodsFor: 'yellow-scripts-support' stamp: 'ar 3/23/2006 19:42'!
myBehaviors: aCollection
	self extraProperties behaviors: aCollection.
! !

!TFrame methodsFor: 'scripts' stamp: 'ar 3/23/2006 19:44'!
myPlayer
	^myProperties ifNotNil:[myProperties myPlayer]! !

!TFrame methodsFor: 'scripts' stamp: 'ar 3/23/2006 19:43'!
myPlayer: aPlayer
	self extraProperties player: aPlayer! !

!TFrame methodsFor: 'initialize' stamp: 'ar 3/23/2006 19:39'!
newExtraProperties
	^TFrameProperties new! !

!TFrame methodsFor: 'transform' stamp: 'DPR 1/9/2004 14:05'!
newRoot: newRoot newTransform: newTransform
	| oldRoot |
	oldRoot := self root.

	self signal: #spaceLeave with: oldRoot.
	oldRoot signal: #objectLeave with: self.

	self parent removeChild: self.
	newRoot addChild: self.
	self localTransform: newTransform.

	self signal: #spaceEnter with: newRoot.
	newRoot signal: #objectEnter with: self.! !

!TFrame methodsFor: 'events' stamp: 'ar 1/17/2006 16:40'!
now
	^Processor activeIsland time! !

!TFrame methodsFor: 'Islands' stamp: 'das 1/14/2006 18:03'!
objectID

	^ objectID.
	! !

!TFrame methodsFor: 'accessing' stamp: 'ar 2/22/2006 18:56'!
objectName

	^objectName ifNil:[self class defaultObjectName]! !

!TFrame methodsFor: 'accessing' stamp: 'ar 2/24/2006 16:08'!
objectName: oName
	self signal: #nameChanged.
	objectName := oName.
! !

!TFrame methodsFor: 'ownership' stamp: 'das 1/16/2006 08:26'!
objectOwner

"objectOwner should never be nil"
	^ objectOwner.! !

!TFrame methodsFor: 'accessing' stamp: 'das 2/20/2004 11:02'!
octreeBox
	| box childBox bs |

	box := TBox new.
	frameChildren ifNotNil:[
		frameChildren do:[ :fc | 	
			childBox := fc octreeBox.
			childBox ifNotNil:[
				box := box unionBox: childBox.].
			].
		].
	bs := self boundSphere.
	bs ifNotNil:[
		bs transform: self globalTransform.
		box growVertex: bs globalPosition.
		].
	box min x > box max x ifTrue:[^ nil ].
	^ box.! !

!TFrame methodsFor: 'accessing' stamp: 'das 6/23/2002 13:53'!
octreeRadius
	| rad childRad |

" Find the max bound sphere radius of the tree. This is used for quadtree/octree construction "
	self boundSphere ifNotNil:[ rad := self boundSphere radius.] ifNil:[ rad := 0 ].
	frameChildren ifNotNil:[
		frameChildren do:[ :fc | 	
			
			childRad := fc octreeRadius.
			rad := rad max: childRad.
			].
		].

	^ rad.! !

!TFrame methodsFor: 'accessing' stamp: 'das 6/24/2002 14:53'!
octreeSieve: octree
	| bs |

	bs := self boundSphere.
	bs ifNotNil:[ 		
		bs transform: self globalTransform.
		octree add: bs. 
		].
	frameChildren ifNotNil:[
		frameChildren do:[ :fc | 	fc octreeSieve: octree. ].
		].
! !

!TFrame methodsFor: 'accessing' stamp: 'das 11/7/2005 11:10'!
oglid

	^ oglid.! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:17'!
on: eventName notify: aClient
	"Overridden to update the eventMask as needed"
	| maskBit |
	maskBit := EventMaskBits at: eventName ifAbsent:[0].
	maskBit = 0 ifFalse:[self eventMask: (eventMask bitOr: maskBit)].
	^super on: eventName notify: aClient! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:39'!
on: eventName unNotify: aClient
	"Overridden to update the eventMask as needed"
	| maskBit newMask |
	super on: eventName unNotify: aClient.
	maskBit := EventMaskBits at: eventName ifAbsent:[0].
	maskBit = 0 ifFalse:["Reconstruct eventMask"
		newMask := 0.
		self myEventMap ifNotNilDo:[:map| map keysDo:[:evtName|
			newMask := newMask bitOr: (EventMaskBits at: evtName ifAbsent:[0])]].
		self eventMask: newMask.
	].! !

!TFrame methodsFor: 'transform' stamp: 'sn 2/23/2018 11:05'!
orientation
"This is used to return just the orientation part of the matrix.The translation part is 0.0."
	| mat |

	mat := localTransform shallowCopy.
	mat translationX: 0.0 y: 0.0 z: 0.0.
	^ mat.! !

!TFrame methodsFor: 'transform' stamp: 'sn 2/23/2018 12:44'!
orientation: mat
"This is used to set just the orientation part of the matrix. It keeps the translation parts intact."
	| trans |
	trans := self translation.
	localTransform := mat shallowCopy.
	self translation: trans.
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 1/27/2005 16:40'!
outVector

	^ self globalTransform column3.
	! !

!TFrame methodsFor: 'hierarchy' stamp: ''!
parent
	
	 ^frameParent.! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 12/13/2005 12:33'!
parent: fp
	
	 frameParent := fp.! !

!TFrame methodsFor: 'accessing' stamp: 'das 3/9/2004 15:22'!
parentChanged

	self parentChanged: self.
! !

!TFrame methodsFor: 'accessing' stamp: 'das 3/9/2004 15:22'!
parentChanged: frame

	" Does nothing, just forward on to children."
	frameChildren ifNotNil:[frameChildren do:[:fc | fc parentChanged: frame].].! !

!TFrame methodsFor: 'render' stamp: ''!
pick: pointer

^ false! !

!TFrame methodsFor: 'render' stamp: 'das 6/29/2002 13:12'!
pickFloor: pointer

^ self pick: pointer.! !

!TFrame methodsFor: 'render' stamp: 'bgf 8/9/2006 17:07'!
pickSelf: ogl

"This is used when we already (think we) know that we are just re-selecting an already selected object. This is used by TPointer>>#pointerPickCurrent"

	| frame |
	frame := self getOwnedObject: ogl eventPointer selection frameID.
	frame ifNotNil: [
		(ogl eventPointer pointerPick: frame boundSphere) ifFalse: [
			ogl eventPointer selectedGlobalPoint: 
				(self globalTransform localPointToGlobal: ogl eventPointer selectedPoint).
		].
	].! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:12'!
pitch

	^self pitchYawRoll x! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:12'!
pitch: p
	| angles |
	angles := self pitchYawRoll.
	angles x: p.
	self pitchYawRoll: angles.! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:11'!
pitchYawRoll
	^localTransform pitchYawRoll! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:11'!
pitchYawRoll: aVector
	| tfm |
	tfm := localTransform copy pitchYawRoll: aVector.
	self localTransform: tfm.! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:11'!
pitchYawRollBy: delta
	self pitchYawRoll: self pitchYawRoll + delta! !

!TFrame methodsFor: 'scripts' stamp: 'ar 3/23/2006 19:43'!
player
	^self myPlayer! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:50'!
pointerDown: event
! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:51'!
pointerEnter: event
! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:53'!
pointerLeave: event
! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:52'!
pointerMove: event 
! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:52'!
pointerOver: event
! !

!TFrame methodsFor: 'events' stamp: 'ar 7/5/2006 14:52'!
pointerUp: event 
! !

!TFrame methodsFor: 'copying' stamp: 'ar 10/2/2004 20:06'!
postCopy
	
	super postCopy.
	localTransform := localTransform copy.
	" Only make copies ofthe frame heirarchy. Everything else stays as-is."
	frameChildren ifNotNil:[
		frameChildren := frameChildren collect:[:fc|
			fc isTexture ifTrue:[fc] ifFalse: [fc copy]
		].
	].
	^self! !

!TFrame methodsFor: 'fileIn/Out' stamp: 'bgf 5/15/2006 14:34'!
postImportFromIslandCopier: importer
	| frame |
	super postImportFromIslandCopier: importer.
	"Create new object id after copying"
	objectID ifNil:[objectID := TObjectID for: self].
	self register.
	"Restore object ownership"
	ownedObjects ifNotNil:[
		ownedObjects keys do:[:id|
			frame := ownedObjects removeKey: id.
			ownedObjects at: frame objectID put: frame.
		].
	].! !

!TFrame methodsFor: 'accessing' stamp: 'sn 2/23/2018 22:21'!
postcard


	self register.
	^ TPostcard new routerAddress: nil id: Processor activeIsland id name: Processor activeIsland name viewpointID: objectID.
! !

!TFrame methodsFor: 'fileIn/Out' stamp: 'ar 1/23/2006 17:09'!
prepareToExportOnIslandCopier: exporter
	super prepareToExportOnIslandCopier: exporter.
	objectID := nil.
	globalTransform := nil.! !

!TFrame methodsFor: 'accessing' stamp: 'das 8/6/2006 14:36'!
printOn: aStream

"show the default - usually 'a ClassName'"
	super printOn: aStream.
"show the size of the frameChildren array"
	frameChildren ifNotNil:[
		aStream nextPutAll:'[';
			print:frameChildren size;
			nextPutAll:']'.
	].
"show the name of this object"
	objectName ifNotNil:[
		aStream nextPutAll:':';
		print: objectName.].! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 8/15/2002 13:40'!
prune
"
	frameChildren ifNotNil:[
		frameChildren do:[ :fc | fc prune ifTrue:[
			self removeChild: fc.].].
		frameChildren size = 0 ifTrue:[frameChildren := nil].
		].
	(self class = TFrame and:[ frameChildren = nil ])ifTrue:[^true].
	^ false."! !

!TFrame methodsFor: 'transform' stamp: 'das 5/10/2004 15:08'!
quaternion

	^ localTransform asQuaternion.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/2/2006 16:06'!
rearrange: ogl

	"Used by the TSystemSpace overlay world to reposition its children upon a resize event. Many TFrame objects can simply ignore this"! !

!TFrame methodsFor: 'Islands' stamp: 'bgf 5/14/2006 14:14'!
register

	"Register me in the island's nameMap under my objectID."

	^ Processor activeIsland register:self name: objectID.
	! !

!TFrame methodsFor: 'Islands' stamp: 'ar 2/22/2006 19:31'!
registerChildren
	"Register my children and give them a unique object name"
	| names newName pair |
	frameChildren ifNil:[^nil].
	names := Set new: frameChildren size.
	frameChildren do:[:each| 
		newName := each objectName.
		(names includes: newName) ifTrue:[
			[pair := newName stemAndNumericSuffix.
			newName := pair first, (pair last+1 max: 2) printString.
			names includes: newName] whileTrue.
			each objectName: newName.
		].
		names add: newName.
		each register
	].! !

!TFrame methodsFor: 'Islands' stamp: 'bgf 7/6/2006 14:10'!
registerGlobal: name

	| nm |
	"
		nm := Processor activeIsland globalNameAt:self ifAbsent:[nil].
		nm ifNotNil:[Processor activeIsland removeGlobal: nm].
	"
	^ Processor activeIsland at:name put: self.
	! !

!TFrame methodsFor: 'Islands' stamp: 'das 11/14/2005 12:51'!
registerName: name

	^ Processor activeIsland register:self name: name.
	! !

!TFrame methodsFor: 'transform' stamp: 'das 5/22/2005 14:30'!
relativeTransform: frame

	^ self globalTransform orthoNormInverse * frame globalTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/9/2004 14:35'!
releaseToRoot

	"Transfer the frame from the current parent to the root frame while keeping the pose in exactly the same orientation"

	| trans root |
	trans := self globalTransform.
	root := self root.
	self transferTo: root.
	self localTransform: trans.
! !

!TFrame methodsFor: 'hierarchy' stamp: ''!
removeAll
	| child |
	
	frameChildren ifNil:[^nil.].

	[frameChildren size = 0] whileFalse:[
		child := frameChildren at: 1.
		frameChildren removeAt: 1.
		child removeLightFrames.
		child removePortalFrames.
		].
	frameChildren := nil.
! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 8/11/2006 14:43'!
removeChild: child

	self root ifNotNil:[ self root doStop:child].
	objectOwner removeOwnership: child.
	child removeLightFrames.
	child removePortalFrames.	
	child removeRayFrames.
	child root: nil.
	frameChildren ifNotNil:[frameChildren remove: child ifAbsent:[].].
	child parent: nil.
	self signal: #structureChanged.! !

!TFrame methodsFor: 'frameManagement' stamp: ''!
removeLightFrame: litefrm

	frameParent ifNotNil:[frameParent removeLightFrame: litefrm.]! !

!TFrame methodsFor: 'frameManagement' stamp: 'DPR 1/9/2004 14:05'!
removeLightFrames

	" When a frame is disconnected from its parent, we need to remove the references to the lights in the TRoom."
	self isLight ifTrue:[self removeLightFrame: self.].
	frameChildren ifNotNil:[frameChildren do:[ :fc | fc removeLightFrames ].].
	! !

!TFrame methodsFor: 'ownership' stamp: 'ar 1/16/2006 19:12'!
removeOwnership: frame

	frame = self ifTrue:[ ^self ]. "we did not add ourselves to the dictionary"
	frame objectOwner ~= self ifTrue:[ ^self ]. "this frame is not in this dictionary "
	ownedObjects ifNotNil:[ownedObjects removeKey:frame objectID].
	frame forceObjectOwner: frame.   "set the objectOwner back to self"
	frame frameChildrenDo:[:child | self removeOwnership: child.].
! !

!TFrame methodsFor: 'frameManagement' stamp: 'das 7/11/2002 01:44'!
removePortalFrame: prtlFrm

	frameParent ifNotNil:[frameParent removePortalFrame: prtlFrm.]! !

!TFrame methodsFor: 'frameManagement' stamp: 'DPR 1/9/2004 14:05'!
removePortalFrames

	" When a frame is disconnected from its parent, we need to remove the references to the lights in the TRoom."
	self isPortal ifTrue:[self removePortalFrame: self.].
	frameChildren ifNotNil:[frameChildren do:[ :fc | fc removePortalFrames ].].
	! !

!TFrame methodsFor: 'frameManagement' stamp: 'das 11/9/2002 15:07'!
removeRayFrame: rayfrm

	frameParent ifNotNil:[frameParent removeRayFrame: rayfrm.]! !

!TFrame methodsFor: 'frameManagement' stamp: 'DPR 1/9/2004 14:05'!
removeRayFrames

	" When a frame is disconnected from its parent, we need to remove the references to the rays in the TSpace."
	self isRay ifTrue:[self removeRayFrame: self.].
	frameChildren ifNotNil:[frameChildren do:[ :fc | fc removeRayFrames ].].
	! !

!TFrame methodsFor: 'hierarchy' stamp: 'DPR 1/9/2004 14:05'!
removeSelf

	self parent ifNotNil:[
		self parent removeChild: self.].
! !

!TFrame methodsFor: 'render' stamp: 'das 8/27/2006 20:08'!
render: ogl

	self renderAxes: ogl.! !

!TFrame methodsFor: 'render' stamp: 'das 3/12/2004 13:36'!
render: ogl box: box
	|  |

	self render: ogl cube: box location: box center scale: 1.0.! !

!TFrame methodsFor: 'render' stamp: 'das 10/16/2004 15:08'!
render: ogl cube: cube location: loc scale: scale
	| dx dy dz x y z e |

" This message is used only to make sure a frame is actually drawing something. In this case, I draw a cube. "
	e := cube extent.
	dx := e x*scale/2.0.
	dy := e y*scale/2.0.
	dz := e z*scale/2.0.
	x := loc x.
	y := loc y.
	z := loc z.

	ogl glBegin: GLQuads;
			glNormal3f: 0.0 with: 0.0 with: 1.0;
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+(dx negated) with: y+(dy negated) with: z+dz;
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+dz;
			glTexCoord2f:1.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+dz;
			glTexCoord2f:0.0 with: 0.0;	glVertex3f: x+(dx negated) with: y+dy with: z+dz.

		dz := dz negated.
		ogl glNormal3f: 0.0 with: 0.0 with: -1.0;
			glTexCoord2f:1.0 with: 0.0;	glVertex3f: x+(dx negated) with: y+dy with: z+dz;
			glTexCoord2f:0.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+dz;
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+dz;
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+(dx negated) with: y+(dy negated) with: z+dz.

		dz := dz negated.

		ogl glNormal3f: 1.0 with: 0.0 with: 0.0;
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+dz;
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+(dz negated);
			glTexCoord2f:1.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+(dz negated);
			glTexCoord2f:0.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+dz.
		dx := dx negated.

			ogl glNormal3f: -1.0 with: 0.0 with: 0.0;
			glTexCoord2f:1.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+dz;
			glTexCoord2f:0.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+(dz negated);
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+(dz negated);
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+dz.
		dx := dx negated.

		ogl glNormal3f: 0.0 with: 1.0 with: 0.0;
			glTexCoord2f:1.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+dz;
			glTexCoord2f:0.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+(dz negated);
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+(dx negated) with: y+dy with: z+(dz negated);
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+(dx negated) with: y+dy with: z+dz.

		dy := dy negated.

		ogl	glNormal3f: 0.0 with: -1.0 with: 0.0;
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+(dx negated) with: y+dy with: z+dz;
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+(dx negated) with: y+dy with: z+(dz negated);
			glTexCoord2f:1.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+(dz negated);
			glTexCoord2f:0.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+dz.

	ogl glEnd.
! !

!TFrame methodsFor: 'render' stamp: 'das 3/12/2004 13:37'!
render: ogl cube: box scale: scale

	self renderCube: ogl location: 0.0@0.0@0.0 scale: scale.

! !

!TFrame methodsFor: 'render' stamp: 'das 5/16/2005 08:48'!
render: ogl sphere: bndSphr segments: segments
	| position radius ringSin ringCos pi2 seg2 vert rts rbs rtc rbc ax u du v dv |
	position := bndSphr localPosition.
	radius := bndSphr radius.
	seg2 := 1+ (segments * 2) .
	ringSin := FloatArray ofSize: seg2.
	ringCos := FloatArray ofSize: seg2.
	vert := Vector3 new.
	pi2 := Float pi *2.0.

	1 to: seg2-1 do:[ :index | 
		ax := ((index) * pi2)/ (seg2-1).
		ringSin at:index put: ax sin.
		ringCos at:index put: ax cos.].
	ringSin at: seg2 put: (ringSin at: 1).
	ringCos at: seg2 put: (ringCos at: 1).

	rts := 0.0.
	rtc := 1.0.
	rbs :=ringSin at: 1.
	rbc := ringCos at: 1.
	du := 1.0/(seg2-1).
	v := 0.0.
	dv := 1.0/segments.
	1 to: segments do:[ :iv |
		ogl glBegin: GLTriangleStrip.
		u := 0.0.
		1 to: seg2 do: [ :ih | 
			vert x: rts*(ringSin at: ih) y: rtc z: rts*(ringCos at: ih).
			ogl glNormal3fv: vert;
				glTexCoord2f: u with: v;
				glVertex3fv:  (vert * radius)+ position.

			vert x: rbs*(ringSin at: ih) y: rbc z: rbs*(ringCos at: ih).
			ogl glNormal3fv: vert;
				glTexCoord2f: u with: v+dv;
				glVertex3fv:  (vert * radius)+ position.

			u:= u+du.].
		v := v+dv.
		rts := rbs.
		rtc := rbc.
		rbs := ringSin at: iv+1.
		rbc := ringCos at: iv+1.
		ogl glEnd.].

! !

!TFrame methodsFor: 'render' stamp: 'das 3/12/2004 13:37'!
renderAlpha: ogl

	"This is used to render transparent ( alpha blended ) objects."! !

!TFrame methodsFor: 'render' stamp: 'das 12/2/2004 14:32'!
renderAxes: ogl

	ogl glDisable: GLLighting;
	" This is used to render solid objects (if any). "
		glColor3fv: #(1.0 0 0)asFloatArray;
		glBegin: GLLineStrip;
    
			glVertex3fv:#(0.0 0.0 0.0)asFloatArray;
    
			glVertex3fv:#(1.0 0.0 0.0)asFloatArray;
    	
			glVertex3fv:#(0.75 0.25 0.0)asFloatArray;
    
			glVertex3fv:#(0.75 -0.25 0.0)asFloatArray;
    
			glVertex3fv:#(1.0 0.0 0.0)asFloatArray;
    
			glVertex3fv:#(0.75 0.0 0.25)asFloatArray;
    
			glVertex3fv:#(0.75 0.0 -0.25)asFloatArray;
    
			glVertex3fv:#(1.0 0.0 0.0)asFloatArray;
    
		glEnd;
    

		glColor3fv: #(0.0 1.0 0)asFloatArray;
		glBegin: GLLineStrip;
    
			glVertex3fv:#(0.0 0.0 0.0)asFloatArray;
    
			glVertex3fv:#(0.0 1.0 0.0)asFloatArray;
    
			glVertex3fv:#(0.0 0.75 0.25)asFloatArray;
    
			glVertex3fv:#(0.0 0.75 -0.25)asFloatArray;
    
			glVertex3fv:#(0.0 1.0 0.0)asFloatArray;
    
			glVertex3fv:#(0.25 0.75 0.0)asFloatArray;
    
			glVertex3fv:#(-0.25 0.75 0.0)asFloatArray;
    
			glVertex3fv:#(0.0 1.0 0.0)asFloatArray;
    
		glEnd;
    

		glColor3fv: #(0.0 0.0 1.0)asFloatArray;
		glBegin: GLLineStrip;
    
			glVertex3fv:#(0.0 0.0 0.0)asFloatArray;
    
			glVertex3fv:#(0.0 0.0 1.0)asFloatArray;
    
			glVertex3fv:#(0.25 0.0 0.75)asFloatArray;
   
			glVertex3fv:#(-0.25 0.0 0.75)asFloatArray;
    
			glVertex3fv:#(0.0 0.0 1.0)asFloatArray;
    
			glVertex3fv:#(0.0 0.25 0.75)asFloatArray;
    
			glVertex3fv:#(0.0 -0.25 0.75)asFloatArray;
    
			glVertex3fv:#(0.0 0.0 1.0)asFloatArray;
    
		glEnd;
	glEnable: GLLighting.! !

!TFrame methodsFor: 'render' stamp: 'das 3/12/2004 13:37'!
renderCube: ogl location: loc scale: scale
	| dx dy dz x y z |

" This message is used only to make sure a frame is actually drawing something. In this case, I draw a cube. "
	dx := scale/2.0.
	dy := scale/2.0.
	dz := scale/2.0.
	x := loc x.
	y := loc y.
	z := loc z.

	ogl glBegin: GLQuads;
			glNormal3f: 0.0 with: 0.0 with: 1.0;
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+(dx negated) with: y+(dy negated) with: z+dz;
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+dz;
			glTexCoord2f:1.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+dz;
			glTexCoord2f:0.0 with: 0.0;	glVertex3f: x+(dx negated) with: y+dy with: z+dz.

		dz := dz negated.
		ogl glNormal3f: 0.0 with: 0.0 with: -1.0;
			glTexCoord2f:1.0 with: 0.0;	glVertex3f: x+(dx negated) with: y+dy with: z+dz;
			glTexCoord2f:0.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+dz;
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+dz;
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+(dx negated) with: y+(dy negated) with: z+dz.

		dz := dz negated.

		ogl glNormal3f: 1.0 with: 0.0 with: 0.0;
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+dz;
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+(dz negated);
			glTexCoord2f:1.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+(dz negated);
			glTexCoord2f:0.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+dz.
		dx := dx negated.

			ogl glNormal3f: -1.0 with: 0.0 with: 0.0;
			glTexCoord2f:1.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+dz;
			glTexCoord2f:0.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+(dz negated);
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+(dz negated);
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+dx with: y+(dy negated) with: z+dz.
		dx := dx negated.

		ogl glNormal3f: 0.0 with: 1.0 with: 0.0;
			glTexCoord2f:1.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+dz;
			glTexCoord2f:0.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+(dz negated);
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+(dx negated) with: y+dy with: z+(dz negated);
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+(dx negated) with: y+dy with: z+dz.

		dy := dy negated.

		ogl	glNormal3f: 0.0 with: -1.0 with: 0.0;
			glTexCoord2f:0.0 with:1.0;	glVertex3f: x+(dx negated) with: y+dy with: z+dz;
			glTexCoord2f:1.0 with:1.0;	glVertex3f: x+(dx negated) with: y+dy with: z+(dz negated);
			glTexCoord2f:1.0 with:0.0;	glVertex3f: x+dx with: y+dy with: z+(dz negated);
			glTexCoord2f:0.0 with: 0.0;	glVertex3f: x+dx with: y+dy with: z+dz.

	ogl glEnd.
! !

!TFrame methodsFor: 'render' stamp: 'das 3/12/2004 13:37'!
renderCube: ogl scale: scale

	self renderCube: ogl location: 0.0@0.0@0.0 scale: scale.

! !

!TFrame methodsFor: 'render' stamp: 'bgf 8/3/2006 21:27'!
renderFrame: ogl

" #renderFrame: must return the number of objects that were rendered - failure to do so will just crash."

	| count childCount ac distanceSquared hilite |

	count := 0.
	(visible or:[frameChildren notNil])ifFalse:[^ 0 ].
	
	ac := ogl camera.

	" apply the local transform to this matrix "
	frameChanged ifTrue:[
self islandViolation:'this code must be moved somewhere else'.
		self globalTransform: (frameParent globalTransform * localTransform).
	].
	ogl pushMatrix. 
	ogl glMultTransposeMatrixf: localTransform. 

	hilite := false.
	self currentHiliteColor ifNotNil: [
		hilite := ogl forceHilite: self currentHiliteColor.
		hilite ifTrue: [ ogl pushFog ].
	].

"------ is this inside the viewing pyramid?"
	(visible and:[ac testBounds: self boundSphere]) ifTrue:[
		ogl eventPointer ifNotNil:[
		(ogl forcePick or:[self objectOwner eventMask~=EventNone]) ifTrue:[
			self isPortal ifFalse:[
				(ogl eventPointer pointerPick: self boundSphere) ifTrue:[
					ogl eventPointer selectedCameraTransform: ac globalTransform copy.
					ac currentPortal ifNil:[ ogl eventPointer selectedPortal:nil point:nil].
				].
			].
		].
		].
		count := 1.
		distanceSquared :=((ac globalPosition-self globalPosition)squaredLength*ac frameScaleSquared).
		ogl distance: distanceSquared.

		(self hasOpaque:ogl)ifTrue:[
			self render: ogl.
		].

		(self hasAlpha:ogl) ifTrue: [ 
			self root addAlphaObject: self 
				transform: ogl peekMatrix
				distance: distanceSquared
				parent: frameParent
				transparency: ogl transparency.
		].
	].

	frameChildren ifNotNil:[
		" render all of the child frames "
		childCount := 0.
		frameChildren do:[:rFrame | childCount := childCount+(rFrame renderFrame: ogl).].
		childCount > 0 ifTrue:[count := count+childCount. ]
		].

	hilite ifTrue: [
		ogl forceHilite: nil.
		ogl popFog.
	].
	ogl popMatrix.
	^ count.


! !

!TFrame methodsFor: 'render' stamp: 'das 4/13/2006 20:11'!
renderOverlay: ogl overlay: ov

	| gTrans rval |
	ogl camera globalTransform: self globalMatrixOut.
	ogl eventPointer ifNotNil:[
		gTrans := ogl eventPointer globalTransform.
		ogl eventPointer globalTransform: self globalMatrixOut * ogl eventPointer localTransform.
	rval := ogl camera renderOverlay: ogl space: self root overlay: ov.
		ogl eventPointer globalTransform: gTrans.
	] ifNil: [	rval := ogl camera renderOverlay: ogl space: self root overlay: ov.].
	^ rval.! !

!TFrame methodsFor: 'render' stamp: 'ar 1/14/2006 17:27'!
renderView: ogl overlay: ov

	"ogl camera globalTransform: self globalMatrixOut."
	^ ogl camera renderView: ogl space: self root overlay: ov.
! !

!TFrame methodsFor: 'accessing' stamp: 'das 9/13/2006 14:36'!
resolveViewpoint: ogl
"allows a TFrame to stand in for a TPostcard"
	^ self.! !

!TFrame methodsFor: 'accessing' stamp: 'ar 7/22/2006 20:17'!
rigidBody
	"Answer my associated rigid body"
	^nil! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:12'!
roll
	^self pitchYawRoll z! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:13'!
roll: r
	| angles |
	angles := self pitchYawRoll.
	angles z: r.
	self pitchYawRoll: angles.! !

!TFrame methodsFor: 'yellow-scripts' stamp: 'ar 7/21/2002 19:34'!
rollBy: turns duration: duration style: style rate: samplingRate
	^self turn: #ccw turns: turns duration: duration style: style rate: samplingRate! !

!TFrame methodsFor: 'accessing' stamp: 'das 5/8/2005 14:57'!
root

	^ frameRoot.! !

!TFrame methodsFor: 'accessing' stamp: 'das 12/9/2005 11:43'!
root: rt

	frameRoot ifNotNil:[frameRoot removeFrame: self].
	frameRoot := rt.
	frameRoot ifNotNil:[frameRoot addFrame: self].
	frameChildren ifNotNil:[frameChildren do:[:fc | fc root: rt].].! !

!TFrame methodsFor: 'transform' stamp: 'das 5/22/2005 12:08'!
rotFromBallPoints: from to: to
 |  f t |
" This matrix needs to be transposed to convert from left handed to right (which is OpenGL's world)."
	f := from normalized.
	t := to normalized.
	f := (f+t) normalized.
	f := (f+t) normalized.
	^  (Quaternion 	a: (f x * t x) + (f y * t y) + (f z * t z )
				b: (f y * t z) - (f z * t y)
			  	c: (f z * t x) - (f x * t z)
			  	d: (f x * t y) - (f y * t x)
			  	).! !

!TFrame methodsFor: 'transform' stamp: 'das 5/22/2005 12:08'!
rotateBy: anAngle around: aVector3
	^self transformBy: (Quaternion angle: anAngle axis: aVector3) asMatrix4x4! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:30'!
rotation: anAngle around: aVector3

	| trans |
	trans := self translation.
	localTransform  := Matrix4x4 identity rotation: anAngle around: aVector3.
	self translation: trans.
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:30'!
rotation: anAngle aroundX: xValue y: yValue z: zValue

	| trans |
	trans := self translation.
	localTransform  := Matrix4x4 identity rotation: anAngle aroundX: xValue y: yValue z: zValue.
	self translation: trans.
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:29'!
rotationAroundX: anAngle

	| trans |
	trans := self translation.
	localTransform := Matrix4x4 identity rotationAroundX: anAngle.
	self translation: trans.
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:29'!
rotationAroundY: anAngle

	| trans |
	trans := self translation.
	localTransform  := Matrix4x4 identity rotationAroundY: anAngle.
	self translation: trans.
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:29'!
rotationAroundZ: anAngle

	| trans |
	trans := self translation.
	localTransform := Matrix4x4 identity rotationAroundZ: anAngle.
	self translation: trans.
	self frameChanged.
	^ localTransform.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/9/2004 14:36'!
scale
	
	^ 1.0.! !

!TFrame methodsFor: 'transform' stamp: 'das 7/15/2002 20:22'!
scale: scale

	self translation: scale * self translation.
	frameChildren ifNotNil:[
		frameChildren do:[ :fc | fc scale: scale.]].! !

!TFrame methodsFor: 'yellow-scripts-support' stamp: 'ar 3/23/2006 19:42'!
scriptBehaviors
	^self myBehaviors ifNil:[#()]! !

!TFrame methodsFor: 'scripts' stamp: 'ar 7/5/2006 14:22'!
scriptEventNames
	"Answer the names of events that script code can see"
	^#(
		pointerDown pointerMove pointerUp
		pointerEnter pointerOver pointerLeave
		keyDown keyStroke keyUp
	)! !

!TFrame methodsFor: 'accessing' stamp: 'ar 9/7/2002 19:16'!
scriptIcon
	^self class scriptIcon! !

!TFrame methodsFor: 'scripts' stamp: 'ar 7/5/2006 14:22'!
scriptManager
	"Get it from elsewhere"
	^nil! !

!TFrame methodsFor: 'accessing' stamp: 'das 3/9/2004 15:29'!
selected: fc

"Inform the parent frame that this frame has been selected"
	frameParent selected: fc.! !

!TFrame methodsFor: 'hierarchy' stamp: 'ar 6/6/2002 22:16'!
setFrameParent: fParent
	frameParent := fParent.! !

!TFrame methodsFor: 'accessing' stamp: ''!
showFrame

" TFrame >> showFrame recursively climbs the frame's tree and outputs it the Transcript."
	self showFrame: 0.! !

!TFrame methodsFor: 'accessing' stamp: ''!
showFrame: depth

	1 to: depth do:[ :i | Transcript show:'---|'].
	Transcript show: self objectName; cr.
	self frameChildren ifNotNil:[
		frameChildren do:[ : fc |
			fc showFrame: depth+1.].].! !

!TFrame methodsFor: 'scripts' stamp: 'ar 3/23/2006 19:43'!
signalEvent: anEvent
	"Signal the occurance of anEvent"
	| sndr |
	super signalEvent: anEvent.
	myProperties ifNotNil:[
		sndr := myProperties player ifNil:[^nil].
		anEvent setSender: sndr.
		sndr signalEvent: anEvent.
	].! !

!TFrame methodsFor: 'accessing' stamp: 'das 5/1/2002 22:03'!
solid

	^ solid.! !

!TFrame methodsFor: 'accessing' stamp: 'das 12/29/2003 17:14'!
solid: bool

	solid := bool.! !

!TFrame methodsFor: 'accessing' stamp: 'das 12/27/2002 17:52'!
solidTree: bool

	self do:[ :f | f solid: bool].
! !

!TFrame methodsFor: 'accessing' stamp: 'das 12/3/2002 13:53'!
solidVisibleTree: bool

	self solidTree: bool.
	self visibleTree: bool.! !

!TFrame methodsFor: 'accessing' stamp: 'das 9/13/2002 02:59'!
sphereTree

	| bs |
	
	bs := self boundSphere.
	bs ifNotNil:[
		frameChildren ifNotNil:[ frameChildren do:[ :fc | bs := bs union: (fc sphereTree).].].
		^ bs.
		].
	frameChildren ifNotNil:[
		frameChildren do:[ :fc |
			bs ifNil:[ bs := fc sphereTree. ] 
				ifNotNil: [bs := bs union: (fc sphereTree). ].
		].
	^ bs.
	].
	^ nil.! !

!TFrame methodsFor: 'events' stamp: 'das 12/13/2005 14:12'!
start

" Called when a TFrame is inserted into the hierarchy. Can be used to start sending future messages. "! !

!TFrame methodsFor: 'events' stamp: 'das 8/3/2006 18:58'!
stepTime

"default step time - used when we call #start"
	^100. ! !

!TFrame methodsFor: 'events' stamp: 'das 12/13/2005 14:12'!
stop

" Called when a TFrame is removed from the hierarchy. Can be used to stop sending future messages. "! !

!TFrame methodsFor: 'events' stamp: 'das 1/18/2005 12:37'!
task: task
"default behavior for handling tasks"
	^ task.! !

!TFrame methodsFor: 'accessing' stamp: 'das 7/10/2002 23:37'!
test

	^ test.! !

!TFrame methodsFor: 'accessing' stamp: 'das 7/10/2002 23:36'!
test: tst

	test := tst.! !

!TFrame methodsFor: 'events' stamp: 'das 11/7/2005 18:42'!
transcript: txt

	Transcript show: txt.! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 9/12/2006 18:44'!
transfer: frame

	frame transferTo: self.
! !

!TFrame methodsFor: 'hierarchy' stamp: 'das 9/13/2006 00:17'!
transferTo: frame

	| trns |
	trns :=  frame relativeTransform: self.
	self parent removeChild: self.
	self localTransform: trns.
	frame addChild: self.
! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:31'!
transformBy: m4x4

	localTransform := localTransform composedWithLocal: m4x4.	
	self frameChanged.
	^ localTransform.
! !

!TFrame methodsFor: 'transform' stamp: 'bgf 7/16/2006 20:59'!
translateIntoContactWith: aPoint alongNormal: norm

	"Adjust my frame translation to be 'in contact' with the given point, such that:
	 the given point touches my bounding box,
       and the normal from the point passes through my origin.
	 The point and normal are in my parent's frame of reference.

	 This is used for spatial drag and drop."

	| box offset |

	"For visually correct-looking results, we need to work with the bounding box,
      rather than the frame origin and boundSphere, since nontrivial models are
      rarely spheroid or balanced around the origin."

	box := self boundingBox transformedBy: (self orientation).

	offset := (
		(norm x > 0 ifTrue: [box min x negated] ifFalse: [box max x]) @
		(norm y > 0 ifTrue: [box min y negated] ifFalse: [box max y]) @
		(norm z > 0 ifTrue: [box min z negated] ifFalse: [box max z])).

	offset := offset * norm.   	"Offset is now the offset for the box center along the normal."
	
	self translation: aPoint - box center + (box center * norm) + offset.


! !

!TFrame methodsFor: 'transform' stamp: 'das 5/10/2004 15:10'!
translation

	^localTransform translation.! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:30'!
translation: aVector

	localTransform translation: aVector.
	self frameChanged.
	^ localTransform.
! !

!TFrame methodsFor: 'transform' stamp: 'das 3/29/2006 14:30'!
translationX: xValue y: yValue z: zValue
	
	localTransform translationX: xValue y: yValue z: zValue.
	self frameChanged.
	^ localTransform.


! !

!TFrame methodsFor: 'accessing' stamp: 'das 12/20/2002 13:36'!
transparency: trans
	
	frameChildren ifNotNil:[
		frameChildren do:[ :fc | fc transparency: trans].].! !

!TFrame methodsFor: 'accessing' stamp: 'bgf 7/9/2006 21:43'!
unhilite
	"Cancel hiliting of this frame, so it appears normally.  See #hilite: and #hilite ."
	hiliteColor := nil.! !

!TFrame methodsFor: 'transform' stamp: 'das 1/27/2005 16:41'!
upVector

	^ self globalTransform column2.

	! !

!TFrame methodsFor: 'accessing' stamp: ''!
visible

	^ visible.! !

!TFrame methodsFor: 'accessing' stamp: ''!
visible: bool

	visible := bool.! !

!TFrame methodsFor: 'accessing' stamp: 'das 10/7/2002 23:05'!
visibleTree

	self visible ifTrue:[^ true].
	frameChildren ifNotNil:[
		frameChildren do:[ :fc | fc visibleTree ifTrue:[^ true].].].
	^ false.! !

!TFrame methodsFor: 'accessing' stamp: 'das 12/27/2002 17:50'!
visibleTree: bool

	self do:[ :f | f visible: bool].
! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:13'!
yaw
	^self pitchYawRoll y! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:13'!
yaw: y
	| angles |
	angles := self pitchYawRoll.
	angles y: y.
	self pitchYawRoll: angles.! !

!TFrame methodsFor: 'transform' stamp: 'das 2/10/2003 12:14'!
yawTransform
	| v |
	v := localTransform row3.
	v x > 0 ifTrue:[self yaw: v z arcCos radiansToDegrees negated] ifFalse:
							[self yaw: v z arcCos radiansToDegrees.].! !

TFrame subclass: #TGroup
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TGroup commentStamp: 'das 10/10/2004 12:01' prior: 0!
TGroup is a TFrame which is always guaranteed to be invisible. It is used as a container and root object of a collection of child frames.

DAS!

TGroup subclass: #TAvatarCarry
	instanceVariableNames: 'originalTransparency eventPointer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TAvatarCarry commentStamp: 'das 6/8/2006 14:06' prior: 0!
This class simply turns off the transparency flag for the object that the avatar is carrying.!

!TAvatarCarry methodsFor: 'accessing' stamp: 'das 8/7/2006 09:02'!
eventPointer: ep

	eventPointer := ep.! !

!TAvatarCarry methodsFor: 'render' stamp: 'das 8/7/2006 09:02'!
renderFrame: ogl

	| trans rval |
	ogl eventPointer: eventPointer.
	trans := ogl transparency.
	ogl transparency: 1.0.
	rval := super renderFrame: ogl.
	ogl transparency: trans.
	^ rval.! !

TGroup subclass: #TAvatarReplica
	instanceVariableNames: 'downPointer footFrame height lastDownFrame lastDownPosition driving speedControl sticky controlPitch lastAliveTime aliveTimeout carrying viewAngle viewAspect gotoTransform gotoCount snapTrans costume head lastPosition lockStep lockLockStep userID originalTransforms'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TAvatarReplica commentStamp: 'das 1/29/2006 15:22' prior: 0!
See comment for TAvatarUser.!

!TAvatarReplica methodsFor: 'accessing' stamp: 'ar 2/1/2006 16:19'!
aliveTimeout
	^aliveTimeout! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'ar 2/1/2006 16:19'!
aliveTimeout: aNumber
	aliveTimeout := aNumber! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 8/30/2006 17:13'!
carrying: aFrame

	originalTransforms at: aFrame put: aFrame localTransform copy.

	aFrame parent ifNotNil:[aFrame parent removeChild: aFrame.].
	carrying ifNil:[carrying := TAvatarCarry new. self addChild: carrying.].
	carrying addChild: aFrame.
! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 6/8/2006 14:23'!
carryingDo: aBlock

	carrying ifNil:[^self].
	^carrying frameChildren do: aBlock
! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 8/30/2006 17:11'!
carryingDrop
	"Drop all the things I'm carrying... they pop back to their pickup locations."

	carrying ifNotNil:[
		carrying frameChildren do:[:each|
			each releaseToRoot.
			each localTransform: (originalTransforms at: each).
		].
		self removeChild: carrying.
		carrying := nil.
	].! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 8/30/2006 17:16'!
carryingRelease: aFrame

	"Release the given frame (which I am assumed to be carrying)
	 into the space; it retains its current global position in the space."

	carrying ifNotNil:[
		aFrame releaseToRoot.
		carrying frameChildren size = 0 ifTrue:[
			self removeChild: carrying.
			carrying := nil].].! !

!TAvatarReplica methodsFor: 'driving' stamp: 'das 1/29/2006 14:15'!
controlPitch: cp

	controlPitch:=cp.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/11/2006 08:00'!
costume
	^ costume! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 7/19/2006 14:22'!
costume: cstm 
	
	(costume notNil and: [costume ~= cstm]) ifTrue: [
		self removeChild: costume.
	].
	costume := cstm.
	costume ifNotNil: [
		self addChild: costume.
		costume solidTree: false.
	].! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/8/2006 10:33'!
costume: cstm head: hd

	self costume: cstm.
	self head: hd.

! !

!TAvatarReplica methodsFor: 'initialize' stamp: 'das 5/28/2006 16:11'!
destroy
	"Allow avatar tracking"
	self root signal: #avatarLeave with: self.
	^super destroy.! !

!TAvatarReplica methodsFor: 'driving' stamp: 'das 5/22/2006 13:03'!
doGoto
	"Move the camera to a specific pose (location and orientation) in the TSpace, via an interpolated rotation/translation. "
	| qFrom qTo fromLoc toLoc trans |
	gotoCount := gotoCount - 1.
	gotoCount <= 0 ifTrue:[
lastPosition := localTransform translation.
		self localTransform: gotoTransform.
		self yawTransform.
		^self].
	"Interpolate"
	qFrom := localTransform asQuaternion.
	qTo := gotoTransform asQuaternion.
	fromLoc := localTransform translation.
	toLoc := gotoTransform translation.
	fromLoc ~= toLoc ifTrue:[
		fromLoc := fromLoc  + ((toLoc - fromLoc)/gotoCount).].
	qFrom := qFrom slerpTo: qTo at: (1.0/gotoCount).
	trans := qFrom asMatrix4x4.
	trans translation: fromLoc.
	self localTransform: trans.
	self pitch: self pitch.
	(self future: 50) doGoto.! !

!TAvatarReplica methodsFor: 'events' stamp: 'das 8/11/2006 10:50'!
doPointerDown: event

	self pointerDown: event.! !

!TAvatarReplica methodsFor: 'events' stamp: 'das 8/11/2006 10:50'!
doPointerUp: event

	self pointerUp: event.! !

!TAvatarReplica methodsFor: 'driving' stamp: 'bgf 7/19/2006 12:31'!
drive

	sticky := false.
	controlPitch
		ifTrue: [self pitch: (speedControl y * 25 min: 75 max: -75)]
		ifFalse: [ self forwardOnY: (speedControl y) negated.].
	self addYaw: (speedControl x * 4).
! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 4/4/2006 18:02'!
driveStep
	self stillAlive.
	driving ifTrue:[self drive].
	sticky ifFalse:[self fall].
! !

!TAvatarReplica methodsFor: 'events' stamp: 'das 8/10/2006 10:18'!
eventMask

"this forces the avatar to be an object owner. This is because I need to test for 'avatar-ness' sometimes, and the only way to do that is if the avatar hierarchy is actually owned by someone."
	^ EventPointerOver bitOr: EventBlockRecurse.! !

!TAvatarReplica methodsFor: 'driving' stamp: 'sn 2/23/2018 12:44'!
fall

	| trans ht transFloor downPosition floor dist pos |
	floor := downPointer selectedFrame.
	dist :=  downPointer selectedDistance.
	pos := downPointer selectedFramePosition.
	downPointer resetTotal.
	floor ifNotNil:[
	floor isSolid ifTrue:[
		trans := self translation.
" ---- Calculate distance to the current floor."
		dist = height ifFalse:[
			ht := trans y + height - dist.
			trans y > ht ifTrue:[ trans y - ht > 0.5 ifTrue:[ ht := trans y - 0.5.].].
			trans y: ht].
" ---- Calculate the floor frame position in the floor's local frame. "
		transFloor := footFrame globalPosition-lastDownPosition.
		lastDownFrame ~= floor ifTrue:[
			floor addChild: footFrame.
			].
		downPosition := pos.
		downPosition y: (downPosition y-height).
		footFrame translation: downPosition.
		lastDownPosition := footFrame globalPosition.
		lastDownFrame := floor.
		trans x: trans x+transFloor x y: trans y z: trans z+ transFloor z.
		self translation = trans ifFalse: [
		"	lastPosition := trans."
			self translation: trans.
			self localTransform:  self localTransform shallowCopy. ]	"why clone?"	
		].
	].
! !

!TAvatarReplica methodsFor: 'driving' stamp: 'bgf 8/2/2006 18:53'!
forwardOnY: foy

	| v |
	v := localTransform lookSide .
	v := (v cross:(0@1@0))*foy.
	localTransform translation: (self root checkMotion: v forAvatar:self ).
	self pitch abs > 0.001 ifTrue:[self pitch: self pitch * 0.8].
	self roll abs > 0.001 ifTrue:[self roll:0.0].
! !

!TAvatarReplica methodsFor: 'driving' stamp: 'das 5/8/2006 08:22'!
goToSpaceHome
	"Go to the home position of the space we're in."

	| home |
	home := self root homeTransform: self.
	home ifNotNil: [
		self goto: home count: 16 sticky: false.
	].! !

!TAvatarReplica methodsFor: 'driving' stamp: 'das 5/8/2006 08:21'!
goto: transform count: count
	self goto: transform count: count sticky: true
! !

!TAvatarReplica methodsFor: 'driving' stamp: 'das 5/8/2006 08:21'!
goto: transform count: count sticky: bool
	" This is used to force the camera to move to a position/orientation. The final location is gotoTransform, and the number of interpolations is in gotoCount."
	sticky := bool.
	snapTrans := self localTransform.
	gotoTransform := transform.
	gotoCount := count.
	self doGoto.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/11/2006 08:01'!
head
	^ head! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/8/2006 10:32'!
head: hd

	head := hd.

! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 2/28/2006 18:25'!
headTransform

	^ head ifNil:[Matrix4x4 identity] ifNotNil:[head localTransform].! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/7/2006 16:32'!
heightAboveGround
	" The height that the avatar origin maintains above solid ground."

	^ height	! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/7/2006 16:32'!
heightAboveGround: h
	" The height that the avatar origin maintains above solid ground."

	height := h.
	! !

!TAvatarReplica methodsFor: 'initialize' stamp: 'bgf 8/30/2006 17:09'!
initialize
	| downMatrix |

	super initialize.
	height := 4.6.
	originalTransforms := Dictionary new.
	lastDownPosition := Vector3 new.
	footFrame := TGroup new.
	downMatrix := Matrix4x4 identity rotationAroundX: -90.
	downPointer := TRay new.
	downPointer localTransform: downMatrix.
	downPointer downRay: true.
	self addChild: downPointer.
	driving := false.
	speedControl := 0@0.
	controlPitch := false.
	sticky := false.
	aliveTimeout := 60000. "self-destruct after 60 seconds without activity."
	self stillAlive.
	lockStep := false.
	lockLockStep := false.
! !

!TAvatarReplica methodsFor: 'testing' stamp: 'das 4/24/2006 21:34'!
isAvatar

	^ true.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 9/19/2006 15:10'!
isCarrying: aFrame

	"Query: is the avatar currently carrying the given frame?"
	
	^ carrying notNil and: [carrying frameChildren includes:aFrame].
! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 7/19/2006 14:23'!
laser: las

	| carry |
	carry := TAvatarCarry new.
	carry addChild: las.
	self addChild: carry.
	las solidTree: false.
	! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 3/17/2006 20:35'!
lastPosition

	^ lastPosition.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 5/23/2006 10:18'!
lockLockStep: ls

	lockLockStep := ls.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 8/26/2006 18:39'!
lockStep
" force the camera to follow the avatar"
	| rval |
	rval := lockStep or:[lockLockStep].
	lockStep := false.
	^ rval.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 8/26/2006 18:40'!
lockStep: ls
"force the camera to follow the avatar with no histeresis."
	lockStep := ls.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 5/23/2006 11:08'!
lockYaw: y

	super yaw: y.
	lockStep := true.
! !

!TAvatarReplica methodsFor: 'driving' stamp: 'bgf 8/2/2006 18:56'!
moveBack: mb

	| v |
	v := localTransform lookSide .
	v := (v cross:(0@1@0))*mb.
	self translation: (self root  checkMotion: v forAvatar: self).
	lockStep := true.

! !

!TAvatarReplica methodsFor: 'driving' stamp: 'bgf 8/2/2006 18:57'!
moveDown: md

	self translation: (self root checkMotion: 0@md negated@0 forAvatar: self).
	lockStep := true.


! !

!TAvatarReplica methodsFor: 'driving' stamp: 'bgf 8/2/2006 18:58'!
moveForward: mf

	| v |
	v := localTransform lookSide .
	v := (v cross:(0@1@0))*mf negated.
	self translation:(self root checkMotion: v forAvatar: self).
	lockStep := true.


! !

!TAvatarReplica methodsFor: 'driving' stamp: 'bgf 8/2/2006 18:58'!
moveLeft: mb

	| v |
	v := localTransform lookAt.
	v := (v cross:(0@1@0))*mb.
	self translation: (self root checkMotion: v forAvatar:self).
	lockStep := true.


! !

!TAvatarReplica methodsFor: 'driving' stamp: 'bgf 8/2/2006 18:58'!
moveRight: mr

	| v |
	v := localTransform lookAt.
	v := (v cross:(0@1@0))*mr negated.
	self translation: (self root checkMotion: v forAvatar: self).
	lockStep := true.


! !

!TAvatarReplica methodsFor: 'driving' stamp: 'bgf 8/2/2006 18:59'!
moveUp: md

	self translation: (self root checkMotion: 0@md@0 forAvatar: self).
	lockStep := true.


! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 5/23/2006 11:06'!
pitch: p

	super pitch: p.
	costume ifNotNil:[
		costume pitch: p negated.
	].
	lockStep := true.! !

!TAvatarReplica methodsFor: 'render' stamp: 'jcg 10/20/2006 02:21'!
renderFrame: ogl

	| len lastTr rval ep|
	len := (self globalPosition - ogl camera globalPosition)squaredLength.
	lastTr := ogl transparency.
	len >10.0 ifTrue:[ogl transparency:1.0]ifFalse:[
		ogl transparency: len/25.0.].
	ep := ogl eventPointer.
	carrying ifNotNil:[carrying eventPointer: ep].
	ogl eventPointer:nil.
	rval := super renderFrame: ogl.
	ogl eventPointer: ep.
	carrying ifNotNil: [carrying eventPointer: nil].
	ogl transparency: lastTr.
	^ rval.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 5/29/2006 13:48'!
resetPitch

	costume ifNotNil:[
		costume pitch: self pitch negated.
	].
	lockStep := true.! !

!TAvatarReplica methodsFor: 'step' stamp: 'bgf 6/10/2006 22:09'!
selfDestruct
	"If we haven't gotten any messages in the last five seconds, self-destruct"
	| nowTime delta |
	nowTime := Processor activeIsland time.
	delta := nowTime - lastAliveTime.
	delta > aliveTimeout ifTrue:[^self destroy].
	(self future: 30000) selfDestruct.	 "Check every 30 seconds."! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 5/28/2006 19:27'!
showCostume: bool

	costume ifNotNil:[costume visibleTree: bool].! !

!TAvatarReplica methodsFor: 'driving' stamp: 'das 5/8/2006 08:22'!
snapBack
	self goto: snapTrans count: 6 sticky: false.
	snapTrans := nil.! !

!TAvatarReplica methodsFor: 'driving' stamp: 'das 1/28/2006 22:26'!
speedControl: sc 

	speedControl := sc.
! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 8/30/2006 16:51'!
startCarrying: aFrame

	"Add the frame to the avatar's carrying frame,
	 preserving its current global position."

	| xform |
	xform := aFrame globalTransform.
	self carrying: aFrame.
	aFrame localTransform: (self globalTransform orthoNormInverse composeWith: xform).! !

!TAvatarReplica methodsFor: 'driving' stamp: 'ar 2/2/2006 21:46'!
startDriving
	driving := true.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'ar 2/2/2006 21:34'!
sticky
	^sticky! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'ar 2/2/2006 21:34'!
sticky: aBool
	sticky := aBool! !

!TAvatarReplica methodsFor: 'step' stamp: 'ar 2/1/2006 16:18'!
stillAlive
	"Mark me as still being alive at this point"
	lastAliveTime := Processor activeIsland time.! !

!TAvatarReplica methodsFor: 'driving' stamp: 'ar 2/2/2006 21:46'!
stopDriving
	driving := false.! !

!TAvatarReplica methodsFor: 'initialize' stamp: 'bgf 3/27/2006 12:42'!
takeInitialPosition

	"Move the avatar to its containing space's initialAvatarTranslation 
	and initialAvatarYaw. if either or both are defined."

	| aSpace trans yaw |

	aSpace := self parent.
	trans := aSpace initialAvatarTranslation: self.
	trans ifNotNil: [
		self translation: trans.
	].

	yaw := aSpace initialAvatarYaw: self.
	yaw ifNotNil: [
		self yaw: yaw.
	].! !

!TAvatarReplica methodsFor: 'driving' stamp: 'bgf 4/13/2006 19:09'!
takeMeTo: selection orientation: orientation

	"Take the user towards the given selection.
	 Orientation is the orientation of the pointer at the time of the pick 
      that produced the selection."

	| frame distance xform y tran vector |

	frame := selection frame.
	frame ifNotNil: [
		"Rule out picks of the space itself, and of the avatar."
		(frame = self) ifTrue: [^self].
		(frame hasAncestor: self) ifTrue:[^ self]. 
		(frame root = self root) ifFalse:[^self]. 	
	].

	selection globalPoint ifNil: [ ^ self ].

	"Determine the destination.
	 We don't move all the way to the selection point.
	 Rather we move at most 4/5 of the way to it,
	 and we never move closer to it than our own height away.
	 Normal movement can be used for finer motion"

	distance := self translation distanceTo: selection globalPoint.
	distance := ((4 * distance) / 5) min: (distance - self extent y).

	"Limit the jump to 8x my height (empirical) to mediate against skybox jumps;
	 this is a workaround until there's a better way of making some things unpickable."

	distance := distance min: (self extent y * 8).

	(distance <= 0) ifTrue: [
		^ self
	].
	vector := (selection globalPoint - self translation) length: distance.   

	"For now, fix 'y' during these goTo: translations; this will be
      inappropriate for spaces without flat grounds."

	xform := orientation copy.
	y := self translation y.
	tran := self translation + vector.
	xform translation: (tran x)@y@(tran z).
	gotoTransform := xform.
	gotoCount := 6.
	self doGoto.
! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 3/17/2006 20:38'!
translation: tran

	lastPosition := localTransform translation.
	^super translation: tran.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/19/2006 17:57'!
userID

	"This identifies (distinguishes) a user for the duration of their harness session."
	^ userID! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/19/2006 17:51'!
userID: aTObjectID
	userID := aTObjectID! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'ar 2/2/2006 20:47'!
viewAngle
	^viewAngle! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'ar 2/2/2006 20:47'!
viewAngle: angle
	viewAngle := angle! !

!TAvatarReplica methodsFor: 'initialize' stamp: 'ar 2/2/2006 20:51'!
viewAngle: anAngle aspect: anAspect
	viewAngle := anAngle.
	viewAspect := anAspect.! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'ar 2/2/2006 20:47'!
viewAspect
	^viewAspect! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'ar 2/2/2006 20:47'!
viewAspect: aNumber
	viewAspect := aNumber! !

!TAvatarReplica methodsFor: 'accessing' stamp: 'das 5/23/2006 11:07'!
yaw: y

	super yaw: y.
! !

TGroup subclass: #TBillboard
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TBillboard commentStamp: 'das 4/3/2006 01:43' prior: 0!
An object that constantly faces the camera when rendering.!

!TBillboard methodsFor: 'render' stamp: 'das 6/6/2005 11:47'!
renderFrame: ogl 

	self lookAt: ogl camera up: nil.
	^super renderFrame: ogl.
! !

TGroup subclass: #TButton
	instanceVariableNames: 'enable frame frame2 action target action2 target2 animateTime color persist switch spin doSpin local deltaSpin stepping whileDown stepWhileDown whileDownTime pointerOnTop'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TButton commentStamp: '<historical>' prior: 0!
TButton is an active frame that tracks user events to initiate actions. When the user's pointer is over a child frame of a TButton it is highlighted. When it is clicked, a message can be sent to a target. TButton can have two different child frames to show a boolean state and generate two different messages as it toggles. The TButton itself has no representation, but expects to contain some kind of visible TFrame object that is selectable. This frame (or frames) is added using the #initializeWithFrame: or #initializeWithFrame:frame2: methods. (This should probably be changed to #contents:)

When you use the targetWhileDown:action: this will create a TButton that sends the action message to the target as long as the button is pressed.
DAS!

!TButton class methodsFor: 'initialize' stamp: 'das 8/16/2006 17:30'!
createNewWithFileName: fName scale: scl

	| tframe |
		tframe := (TLoad3DSMax new initializeWithFileName:  fName scale: scl) frame.
		tframe boundsDepth: 1.
		tframe initBounds.
"		TExporter export: tframe asBinary:url."
		tframe do:[ :tf |
			tf isMesh ifTrue:[ tf materialList do:[ :ml |
				ml ambientColor: #(0.2 0.5 1.0 0.8).
				ml diffuseColor:#(0.2 0.5 1.0 0.8).
				ml textureMode: GLModulate.
				].
			tf checkAlpha.
			].
		].
	^ tframe.
! !

!TButton class methodsFor: 'accessing' stamp: 'das 5/5/2006 14:46'!
downColor
	
	"^ #(0.1 0.7 0.5 0.8) asVectorColor."	"Strong blue-green"

	^ #(0.0 1.0 0.0 0.8) asVectorColor.
! !

!TButton class methodsFor: 'accessing' stamp: 'das 3/2/2006 16:23'!
downColorLite
	
	^ #(0.2 1.0 0.5 0.4) asVectorColor.! !

!TButton class methodsFor: 'accessing' stamp: 'das 5/8/2006 02:27'!
hiliteColor
	
	"^ #(0.3 0.6 0.9 0.9) asVectorColor."	"strong warm green"

	^ #(1.0 1.0 0.2 0.5) asVectorColor.! !

!TButton class methodsFor: 'accessing' stamp: 'das 3/2/2006 16:22'!
hiliteColorLite
	
	^ #(1.0 1.0 0.2 0.25) asVectorColor.! !

!TButton class methodsFor: 'accessing' stamp: 'bgf 3/17/2006 23:34'!
normColor
	
	^ #(0.32 0.44 0.54 0.8) asVectorColor.  " weak warm blue "

	"^ #(0.5 0.7 0.7 0.5) asVectorColor."! !

!TButton class methodsFor: 'accessing' stamp: 'das 3/2/2006 16:23'!
normColorLite
	
	^ #(0.5 0.7 0.7 0.25) asVectorColor.! !

!TButton class methodsFor: 'accessing' stamp: 'das 5/5/2006 14:48'!
overColor
	
	"^ #(0.1 0.4 0.9 0.8) asVectorColor."   "Strong blue."

	^ #(0.2 0.5 1.0 0.8) asVectorColor.! !

!TButton class methodsFor: 'accessing' stamp: 'das 3/2/2006 16:19'!
overColorLite
	
	^ #(0.2 0.5 1.0 0.4) asVectorColor.! !

!TButton methodsFor: 'accessing' stamp: 'das 10/17/2004 16:51'!
action

	^ action! !

!TButton methodsFor: 'accessing' stamp: 'ar 11/13/2006 14:48'!
action2

	^ action2! !

!TButton methodsFor: 'accessing' stamp: 'ar 11/13/2006 14:48'!
action2: act
	action2 := act.! !

!TButton methodsFor: 'accessing' stamp: 'ar 11/13/2006 14:48'!
action: act
	action := act.! !

!TButton methodsFor: 'accessing' stamp: 'das 9/17/2002 01:22'!
color: col
	color := col.
	frame colorize: color.
	frame2 ifNotNil:[frame2 colorize: color.].
! !

!TButton methodsFor: 'action' stamp: 'ar 1/17/2006 21:38'!
disable
	persist ifFalse:[
		animateTime := self now + self fadeDelay.
		self startStepping.
	].
! !

!TButton methodsFor: 'action' stamp: 'das 5/8/2006 03:21'!
enable

	self color: TButton overColor.
	switch ifNil:[ frame solidVisibleTree: true. ] ifNotNil:[
		frame solidVisibleTree: switch not.
		frame2 solidVisibleTree: switch.
		].
	animateTime := nil.
	persist ifFalse:[self startStepping].! !

!TButton methodsFor: 'events' stamp: 'das 8/6/2006 22:52'!
eventMask

	^ (EventPointerOver bitOr: EventPointerDown) bitOr: EventBlockRecurse. 
! !

!TButton methodsFor: 'accessing' stamp: 'das 9/11/2002 12:12'!
fadeDelay
	^2500.0  "2.5 seconds"! !

!TButton methodsFor: 'accessing' stamp: 'DPR 5/19/2004 16:25'!
fadeDelta
	^125.0 "1/8 second"! !

!TButton methodsFor: 'action' stamp: 'ar 10/5/2004 19:08'!
fire: aSelector target: aTarget with: optArg
	aTarget ifNil:[^nil].
	aSelector ifNil:[^nil].
	aSelector numArgs = 0 ifTrue:[^aTarget perform: aSelector].
	aSelector numArgs = 1 ifTrue:[^aTarget perform: aSelector with: optArg].
	self error:'Wrong number of args'! !

!TButton methodsFor: 'initialize' stamp: 'das 5/8/2006 03:21'!
initialize
	super initialize.
	stepping := false.
	persist := false.
	animateTime := nil.
	switch := nil.
	enable := false.
	spin := false.
	local := false.
	whileDown := false.
	pointerOnTop := false.! !

!TButton methodsFor: 'initialize' stamp: 'das 5/8/2006 03:24'!
initializeSwitchWithFrame: fr frame2: fr2

	frame := fr.
	frame2 := fr2.
	self addChild: frame.
	self addChild: frame2.
	self color: TButton normColor.
	frame solidVisibleTree: false.
	frame2 solidVisibleTree: false.
	switch := false.
	^self
! !

!TButton methodsFor: 'initialize' stamp: 'ar 1/17/2006 16:36'!
initializeWithFrame: fr

	frame := fr.
	self addChild: frame.
	self color: TButton normColor.
	frame solidVisibleTree: false.
	^self
! !

!TButton methodsFor: 'testing' stamp: 'das 10/4/2002 18:57'!
isSolid

	^ false.! !

!TButton methodsFor: 'testing' stamp: 'ar 1/18/2006 17:23'!
isSwitch
	^frame2 notNil! !

!TButton methodsFor: 'accessing' stamp: 'das 4/12/2003 12:39'!
local:bool

	local := bool.! !

!TButton methodsFor: 'accessing' stamp: 'das 9/11/2002 12:16'!
persist

	^ persist.! !

!TButton methodsFor: 'accessing' stamp: 'ar 1/17/2006 16:00'!
persist: bool

	persist := bool.
	persist ifTrue:[ 	
		self color: TButton overColor.
		self enable].
	persist ifFalse:[self disable].! !

!TButton methodsFor: 'events' stamp: 'das 8/26/2006 18:04'!
pointerDown: event 

	self color: TButton downColor.
	frame2 ifNotNil:[
		switch ifNotNil:[
			frame solidVisibleTree: switch not.
			frame2 solidVisibleTree: switch.
			] ifNil:[
			frame solidVisibleTree: false.
			frame2 solidVisibleTree: true.
			].
		].
	pointerOnTop := true.
	whileDown ifTrue:[self startWhileDown: event.].
	self pointerMove: event.! !

!TButton methodsFor: 'events' stamp: 'ar 1/27/2006 14:39'!
pointerEnter: event

	self enable.
	persist ifTrue:[self startStepping].
	self color: TButton hiliteColor.
	spin ifTrue:[
		deltaSpin := -1.
		doSpin := doSpin + deltaSpin.
	].! !

!TButton methodsFor: 'events' stamp: 'ar 1/27/2006 14:39'!
pointerLeave: event 

	self disable.
	self color: (persist ifTrue:[TButton overColor] ifFalse:[TButton normColor]).
	spin ifTrue:[
		doSpin > -180 ifTrue:[
			deltaSpin := deltaSpin negated.].
		].! !

!TButton methodsFor: 'events' stamp: 'das 5/8/2006 03:22'!
pointerMove: event

	| pointer |
	pointer :=  event makePointer.
	(pointer pointerPickTree: self) ifTrue:[ 
		pointerOnTop := true.
		self color: TButton downColor.
		switch ifNotNil:[
			frame solidVisibleTree: switch.
			frame2 solidVisibleTree: switch not.
		] ifNil:[
			frame2 ifNotNil:[
				frame solidVisibleTree: false.
				frame2 solidVisibleTree: true. 
			].
		].
	] ifFalse:[ 
		pointerOnTop := false.
		self color: TButton hiliteColor. 
		switch ifNotNil:[
			frame solidVisibleTree: switch not.
			frame2 solidVisibleTree: switch.
		] ifNil:[
			frame2 ifNotNil:[
				frame solidVisibleTree: true.
				frame2 solidVisibleTree: false. 
			].
		].
	].
! !

!TButton methodsFor: 'events' stamp: 'ar 1/27/2006 14:40'!
pointerOver: event
! !

!TButton methodsFor: 'events' stamp: 'das 5/8/2006 20:16'!
pointerUp: event
	| pointer |

	pointer := event makePointer.

	(pointer pointerPickTree: self) ifTrue:[
		switch ifNotNil:[
			self switch: switch not.
			switch 
				ifTrue:[self fire: action target: target with: event]
				ifFalse:[self fire: action2 target: target2 with: event]
		] 
		ifNil:[
			target ifNotNil:[self fire: action target: target with: event].
			frame solidVisibleTree: true.
			frame2 ifNotNil:[ frame2 solidVisibleTree: false.].
			].
		] ifFalse:[
			switch ifNotNil:[
				frame solidVisibleTree: switch not.
				frame2 solidVisibleTree: switch.
			]ifNil:[
				frame solidVisibleTree: true.
				frame2 ifNotNil:[ frame2 solidVisibleTree: false.].
			].
		].
	self color: (persist ifTrue:[TButton overColor] ifFalse:[TButton normColor]).
	pointerOnTop := false.
	whileDown ifTrue:[self stopWhileDown.].! !

!TButton methodsFor: 'accessing' stamp: 'ar 1/18/2006 17:22'!
scale: scl

	frame scale: scl.
	frame2 ifNotNil:[frame2 scale: scl].! !

!TButton methodsFor: 'accessing' stamp: 'das 4/11/2003 01:54'!
spin: bool

	spin := bool.
	doSpin := 0.! !

!TButton methodsFor: 'accessing' stamp: 'DPR 5/19/2004 16:31'!
spinDelta
	^125.0 "1/8 second"! !

!TButton methodsFor: 'stepping' stamp: 'ar 1/17/2006 16:33'!
startStepping
	stepping ifFalse:[
		stepping := true.
		self step].! !

!TButton methodsFor: 'stepping' stamp: 'das 8/26/2006 18:04'!
startWhileDown: event
	
	stepWhileDown := true.
	self stepWhileDown: event.! !

!TButton methodsFor: 'stepping' stamp: 'ar 1/17/2006 16:50'!
step

	"interpolate color"
	| timeLeft |
	persist ifFalse:[
		animateTime ifNotNil: [
			timeLeft := animateTime - self now.
			timeLeft <= 0 ifTrue: [
				stepping := false.
				animateTime := nil. 
				frame solidVisibleTree: false. 
				frame2 ifNotNil:[frame2 solidVisibleTree: false. ].
			] ifFalse: [
				color at: 4 put: timeLeft asFloat / self fadeDelay
			].
			self color: color.
		].
	].
	spin ifTrue:[ 
		doSpin ~= 0 ifTrue:[
			doSpin := doSpin + deltaSpin. 
			doSpin < -360 ifTrue:[doSpin := 0.].
			self rotationAroundY: doSpin.
		].
		doSpin = 0 ifTrue:[persist ifTrue:[stepping := false]].
	] ifFalse:[
		persist ifTrue:[stepping := false].
	].
	stepping ifTrue:[(self future: 100) step].! !

!TButton methodsFor: 'stepping' stamp: 'das 8/26/2006 18:48'!
stepWhileDown: event
	
	target ifNil:[ ^self ].
	action ifNil:[ ^self ].
	stepWhileDown ifTrue:[
	"only do this when the cursor is actually on the button."
	"	pointerOnTop ifTrue:["
		self fire: action target: target with: event.
	"]. "
		(self future: whileDownTime) stepWhileDown: event.
	].
! !

!TButton methodsFor: 'stepping' stamp: 'ar 1/17/2006 16:48'!
stepping
	^stepping! !

!TButton methodsFor: 'stepping' stamp: 'das 5/7/2006 20:55'!
stopWhileDown
	
	stepWhileDown := false.! !

!TButton methodsFor: 'accessing' stamp: 'das 10/4/2002 19:00'!
switch

	^ switch.! !

!TButton methodsFor: 'events' stamp: 'das 5/8/2006 03:23'!
switch: bool

	switch := bool.
	(frame visibleTree or:[frame2 visibleTree])ifTrue:[
		frame solidVisibleTree: switch not.
		frame2 solidVisibleTree: switch.
	].
! !

!TButton methodsFor: 'accessing' stamp: 'das 10/17/2004 16:40'!
target
	
	^ target.! !

!TButton methodsFor: 'action' stamp: 'das 10/4/2002 18:57'!
target2: trgt action2: act

	target2 := trgt.
	action2 := act.! !

!TButton methodsFor: 'action' stamp: 'das 10/4/2002 18:57'!
target: trgt action: act

	target := trgt.
	action := act.! !

!TButton methodsFor: 'action' stamp: 'das 5/8/2006 07:24'!
targetWhileDown: trgt action: act

	self targetWhileDown: trgt action: act repeatTime: 100
! !

!TButton methodsFor: 'action' stamp: 'das 5/7/2006 21:56'!
targetWhileDown: trgt action: act repeatTime: rt

	target := trgt.
	action := act.
	whileDown := true.
	whileDownTime := rt.! !

TGroup subclass: #TCamera
	instanceVariableNames: 'bounds viewAngle zNear zFar zScreen length clipPlanes clipPlanesTransform viewClip portalClip portalPlane currentPortal currentSpace texture killFrame frameScale frameScaleSquared frameScaleInverse subCornerVector cornerVector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TCamera commentStamp: '<historical>' prior: 0!
The TCamera is used to generate an image of the scene from a particular location in a TSpace. It does the majority of setting up OpenGL to render. 

The z value of the camera matrix actually points away from the direction of view. This seems to be a standard approach for OpenGL. 

instance variables:

bounds
bounds is the area that is rendered into. You can set this (#bounds:) to have an explicit hard value. This is a 2D bounds.

viewAngle
This is the view angle of the camera measured top to bottom. The default is 45 degrees, which means that the angle of the top and bottom edges of the rendered image are 45 degrees apart.

* zNear, zFar
The near and far clipping planes in the graphics engine. 

zScreen
This is the z distance from the camera that the virtual screen would be. That is, when you have an x,y location on the real screen, you can add the zScreen value (x,y,zScreen) and normalize which gives you an accurate pointing vector in 3D space.

length (deprecated)
Computed by initClipPlanes and used to render the test camera.

clipPlanes, clipPlanesTransform
clipPlanes are the four frustum clip planes defined in the cameras local orientation.
clipPlanesTransform are the four clipPlanes in the cameras global orientation frame. This is what is used to determine if objects are visible inside of the view frustum.

viewClip (deprecated)
This is used to test the visibility culling of the camera. It makes the clipping planes visible. This is obsolete, but it may be useful again if we ever start playing with the object clipping code again. 

portalClip
This is an additional clip plane defined by a portal. Objects on the near side of the portal need to be clipped away as well. This is a 4x4 transform.

portalPlane
This is a 3D vector calculated from the portalClip.

inPortal
A flag indicating whether we are rendering a portals contents from this camera or not. This is used to supress certain tests (such as finding floors), and to determine that the avatar should or should not be rendered.

currentSpace
The space we are currently inside of and rendering from. 

texture (deprecated)
Simply a texture that gets added to the default rendered camera. Not used anymore unless you really need to see the default camera. 

killFrame
A flag to indicate that the current rendered image is junk and should not be rendered. This suppresses a swapBuffers call.

cornerVector
This is a normalized vector pointing to the top right corner of the screen in 3-space. That is, width/2,height/2,zScreen. The cornerVector value is used as an easy way to find a corner of the screen in 3-space. It is computed from the bounds width, height, and zScreen values as follows:
	cornerVector := ((bounds width/2.0)@(bounds height/2.0)@(zScreen negated)) normalized.

DAS






 !

!TCamera methodsFor: 'accessing' stamp: 'ar 10/4/2004 13:46'!
aspect

	^(self bounds width/ self bounds height) asFloat.

! !

!TCamera methodsFor: 'accessing' stamp: 'das 8/7/2004 23:51'!
avatar

	^ nil.! !

!TCamera methodsFor: 'accessing' stamp: 'das 5/8/2005 18:00'!
bounds

	^ bounds.
! !

!TCamera methodsFor: 'accessing' stamp: 'das 7/8/2005 13:02'!
bounds:bnds

	| radians |

	bounds := bnds.
	radians := (viewAngle/2.0)*(Float pi/180.0).
	zScreen := (bounds height/2.0) / (radians tan).
	cornerVector := ((bounds width/2.0)@(bounds height/2.0)@(zScreen negated)) normalized.
	self reinitialize.! !

!TCamera methodsFor: 'accessing' stamp: 'das 5/16/2002 14:19'!
clipPlanes

	^ clipPlanes.! !

!TCamera methodsFor: 'accessing' stamp: 'das 5/16/2002 16:16'!
clipPlanes: cp
	
	cp ifNotNil:[
		clipPlanes := cp.].! !

!TCamera methodsFor: 'transform' stamp: 'das 7/31/2004 12:16'!
clipPlanesTransform

	^ clipPlanesTransform.! !

!TCamera methodsFor: 'transform' stamp: ''!
clipPlanesTransform: cpt

	clipPlanesTransform := cpt.! !

!TCamera methodsFor: 'accessing' stamp: 'das 6/3/2005 13:19'!
cornerVector
"The corner vector is a normalized vector pointing from the camera position to the top-left corner of the view bounds. This is useful for aligning buttons in the viewPortal."
	^ cornerVector.! !

!TCamera methodsFor: 'accessing' stamp: 'das 6/4/2005 14:33'!
currentPortal

	^ currentPortal.! !

!TCamera methodsFor: 'accessing' stamp: 'das 6/4/2005 14:33'!
currentPortal: cp

	currentPortal := cp.! !

!TCamera methodsFor: 'accessing' stamp: 'das 7/19/2004 19:05'!
currentSpace

	^currentSpace.! !

!TCamera methodsFor: 'accessing' stamp: 'das 7/19/2004 19:05'!
currentSpace: cs

	currentSpace := cs.! !

!TCamera methodsFor: 'accessing' stamp: 'das 8/1/2004 16:01'!
debugOverlay

	^ nil.! !

!TCamera methodsFor: 'transform' stamp: 'das 12/20/2003 15:08'!
doScale: scale

	frameChildren ifNotNil:[
		frameChildren do:[ :fc | fc scale: scale.]].
! !

!TCamera methodsFor: 'accessing' stamp: 'das 8/25/2004 13:43'!
exportOn: exporter
	self error:'Cannot export cameras'! !

!TCamera methodsFor: 'accessing' stamp: 'das 5/14/2005 15:09'!
frameScale

	^ frameScale.! !

!TCamera methodsFor: 'accessing' stamp: 'das 5/14/2005 15:07'!
frameScale: fs

	frameScale := fs.
	frameScaleSquared := fs*fs.
	frameScaleInverse := 1.0/fs.! !

!TCamera methodsFor: 'accessing' stamp: 'das 5/14/2005 15:09'!
frameScaleInverse

	^ frameScaleInverse.! !

!TCamera methodsFor: 'accessing' stamp: 'das 5/14/2005 15:13'!
frameScaleSquared

	^ frameScaleSquared.! !

!TCamera methodsFor: 'initialize' stamp: 'das 12/21/2003 12:08'!
frustumChanged! !

!TCamera methodsFor: 'transform' stamp: 'das 3/22/2006 15:30'!
globalPositionForce: gpf

	globalTransform translation:gpf.
	frameChanged := false.
! !

!TCamera methodsFor: 'accessing' stamp: 'ar 9/7/2002 14:51'!
hasAlpha

	^ viewClip.
! !

!TCamera methodsFor: 'initialize' stamp: 'das 5/27/2005 11:49'!
initClipPlanes

	| aspect |
	aspect := self aspect. 
	length := 1.0/(((viewAngle/2.0)*(Float pi/180.0))tan). 

	clipPlanes := Vector3Array ofSize: 4.
	
	clipPlanes at: 1 put: (Vector3 x: 0.0 y: length  z: 1)normalize negated.
	clipPlanes at: 2 put: (Vector3 x: 0.0 y: length negated z: 1)normalize negated.
	clipPlanes at: 3 put: (Vector3 x: length y:0.0 z: aspect)normalize negated.
	clipPlanes at: 4 put: (Vector3 x: length negated y:0.0 z: aspect)normalize negated.! !

!TCamera methodsFor: 'initialize' stamp: 'das 7/8/2005 13:10'!
initFrustum: ogl

	ogl initFrustum: viewAngle bounds: bounds zNear: zNear zFar: zFar.
	self frustumChanged.
! !

!TCamera methodsFor: 'initialize' stamp: 'das 7/7/2006 20:45'!
initialize

	super initialize.
	viewAngle := 45.0.
	zNear := 0.5.
	zScreen := 0.5. "give it any value - initFrustum gives it a real value"
	zFar := 1500.0.
	localTransform := Matrix4x4 identity.
	viewClip := true.
	currentPortal := nil.
	killFrame := false.
	self visible: false.
	portalClip := nil.
	self frameScale: 1.0.
! !

!TCamera methodsFor: 'testing' stamp: ''!
isCamera

	^ true.! !

!TCamera methodsFor: 'toys' stamp: 'das 5/16/2005 08:49'!
jump

	self translation: self translation + (Vector3 x: 0 y: 20 z: 0).! !

!TCamera methodsFor: 'accessing' stamp: 'ar 9/7/2002 14:51'!
killFrame

	^ killFrame.! !

!TCamera methodsFor: 'accessing' stamp: 'das 12/21/2003 13:45'!
killFrame: bool
"Flag to let the system know not to reveal this frame, either because it is dirty or incomplete. Supresses the swapBuffer call."
	killFrame := bool.! !

!TCamera methodsFor: 'transform' stamp: 'das 1/31/2006 21:27'!
localTransform: lt
	localTransform := lt.
	self globalTransform: lt.! !

!TCamera methodsFor: 'accessing' stamp: 'ar 10/10/2002 16:11'!
pitch

	^self pitchYawRoll x! !

!TCamera methodsFor: 'accessing' stamp: 'das 12/26/2003 12:22'!
pointer

	^ nil.! !

!TCamera methodsFor: 'accessing' stamp: 'das 2/17/2004 09:13'!
pointer: pntr

	^ nil.! !

!TCamera methodsFor: 'transform' stamp: 'das 12/11/2005 18:03'!
pointerXY: pXY

	| bCenter at up side cpointer |
" x and y are negated on purpose. Why? Because the z value should actually be negated as it is referring to a position in front of the camera (in the negative z direction) but anyone trying to do the math for TPointer will probably want to use a positive direction vector. OK? Get it? Essentially the math is wrong, but it is consistent. "
	bCenter := self bounds center.
	at := (Vector3
				x: bCenter x - pXY x
				y: pXY y - bCenter y
				z: zScreen ) normalized.
	up := Vector3 x: 0.0 y:1.0 z:0.0.
	side := (up cross: at) normalized.
	up := (at cross: side) normalized.

	cpointer := Matrix4x4 identity.
	cpointer at: 1 at: 1 put: side x.
	cpointer at: 1 at: 2 put: side y.
	cpointer at: 1 at: 3 put: side z.
	cpointer at: 2 at: 1 put: up x.
	cpointer at: 2 at: 2 put: up y.
	cpointer at: 2 at: 3 put: up z.
	cpointer at: 3 at: 1 put: at x.
	cpointer at: 3 at: 2 put: at y.
	cpointer at: 3 at: 3 put: at z.
	^ cpointer transposed.
! !

!TCamera methodsFor: 'accessing' stamp: 'das 2/17/2004 09:13'!
portalClip

	^ portalClip.! !

!TCamera methodsFor: 'accessing' stamp: 'das 5/23/2005 11:35'!
portalClip: pc

"This is an additional clip plane defined by a portal. Objects on the near side of the portal need to be clipped away as well. This is a 4x4 transform."

	portalClip := pc.
	portalClip ifNotNil:[
		portalPlane := (Vector3 x: portalClip a13  y:portalClip a23  z: portalClip a33) negated.].
! !

!TCamera methodsFor: 'render' stamp: 'das 7/28/2004 10:46'!
postRender! !

!TCamera methodsFor: 'render' stamp: 'das 7/28/2004 10:45'!
preRender! !

!TCamera methodsFor: 'transform' stamp: 'ar 2/23/2006 15:01'!
projectPoint: aVector3
	"Project the given point into normalized device coordinates (-1, +1)"
	^self projectionMatrix localPointToGlobal:
		(self localTransform orthoNormInverse localPointToGlobal: aVector3)! !

!TCamera methodsFor: 'transform' stamp: 'ar 2/23/2006 14:20'!
projectionMatrix
	"Answer the projection matrix for the receiver"
	| m sine cotangent deltaZ radians aspect |
	aspect := bounds width asFloat / bounds height.
	m := Matrix4x4 identity.
	radians := (viewAngle / 2.0)*(Float pi / 180.0).
	sine := radians sin.
	deltaZ := zFar - zNear.
	cotangent := radians cos / sine.
	m a11: cotangent / aspect.
	m a22: cotangent.
	m a33: ((zFar + zNear) negated) / deltaZ.
	m a34: -2 * zNear * zFar / deltaZ.
	m a43: -1.
	m a44: 0.
	^m! !

!TCamera methodsFor: 'initialize' stamp: 'das 6/4/2005 20:20'!
reinitialize

	viewClip := true.
	self initClipPlanes.
	currentPortal := nil.
	killFrame := false.
	portalClip := nil.
! !

!TCamera methodsFor: 'render' stamp: 'das 5/16/2005 08:50'!
render: ogl
	| vert origin scale sLength sAspect |

"This is deprecated code for sure, as there is no longer any way for a camera to render itself, but you never know if you will need to be able to visualize the viewport of a camera and this does just that."

	ogl glDisable: GLCullFace.
	origin := Vector3 new.
	vert := Vector3 new.

	scale := 1.

	sLength := scale * length negated.
	sAspect := scale * self aspect.

	ogl glBegin: GLTriangles.

	ogl glNormal3fv: ((clipPlanes at:1) negated).
	ogl glVertex3fv: origin.
	vert x: sAspect y: scale z: sLength.
	ogl glVertex3fv: vert.
	vert x: sAspect negated y:  scale z: sLength.
	ogl glVertex3fv: vert.

	ogl glNormal3fv: ((clipPlanes at:2) negated).
	ogl glVertex3fv: origin.
	vert x: sAspect negated y:  scale negated z: sLength.
	ogl glVertex3fv: vert.
	vert x: sAspect y: scale negated z: sLength.
	ogl glVertex3fv: vert.

	ogl glNormal3fv: ((clipPlanes at:3) negated).
	ogl glVertex3fv: origin.
	vert x: sAspect y: scale negated z: sLength.
	ogl glVertex3fv: vert.
	vert x: sAspect y:  scale z: sLength.
	ogl glVertex3fv: vert.

	ogl glNormal3fv: ((clipPlanes at:4) negated).
	ogl glVertex3fv: origin.
	vert x: sAspect negated y:  scale z: sLength.
	ogl glVertex3fv: vert.
	vert x: sAspect negated y: scale negated z: sLength.
	ogl glVertex3fv: vert.
	ogl glEnd.

	ogl setCull.
	texture ifNotNil:[
		texture enable: ogl.
		ogl glBegin: GLQuads;
				glNormal3f: 0.0 with: 0.0 with: -1.0;
				glTexCoord2f:0.0 with: 0.0;	
				glVertex3f: sAspect negated with: scale with: sLength;
				glTexCoord2f:1.0 with:0.0;	
				glVertex3f: sAspect with: scale with: sLength;
				glTexCoord2f:1.0 with:1.0;	
				glVertex3f: sAspect with: scale negated with: sLength;
				glTexCoord2f:0.0 with:1.0;	
				glVertex3f: sAspect negated with: scale negated with:sLength;
			 glEnd.
		texture disable: ogl.].



! !

!TCamera methodsFor: 'render' stamp: 'das 5/16/2005 08:50'!
renderAlpha: ogl
	| vert origin scale sLength sAspect |

	ogl glDisable: GLCullFace.
	origin := Vector3 new.
	vert := Vector3 new.
	
	scale := 20.

	sLength := scale * length negated.
	sAspect := scale * self aspect.

	ogl glBegin: GLTriangles.

	ogl glNormal3fv: (clipPlanes at:1) negated.
	ogl glVertex3fv: origin.
	vert x: sAspect y: scale z: sLength.
	ogl glVertex3fv: vert.
	vert x: sAspect negated y:  scale z: sLength.
	ogl glVertex3fv: vert.

	ogl glNormal3fv: (clipPlanes at:2) negated.
	ogl glVertex3fv: origin.
	vert x: sAspect negated y:  scale negated z: sLength.
	ogl glVertex3fv: vert.
	vert x: sAspect y: scale negated z: sLength.
	ogl glVertex3fv: vert.

	ogl glNormal3fv: (clipPlanes at:3) negated.
	ogl glVertex3fv: origin.
	vert x: sAspect y: scale negated z: sLength.
	ogl glVertex3fv: vert.
	vert x: sAspect y:  scale z: sLength.
	ogl glVertex3fv: vert.

	ogl glNormal3fv: (clipPlanes at:4) negated.
	ogl glVertex3fv: origin.
	vert x: sAspect negated y:  scale z: sLength.
	ogl glVertex3fv: vert.
	vert x: sAspect negated y: scale negated z: sLength.
	ogl glVertex3fv: vert.
	ogl glEnd.

	ogl setCull.


! !

!TCamera methodsFor: 'render' stamp: 'das 5/8/2006 07:16'!
renderOverlay: ogl space: space overlay: overlays

	| rval  lastTarget lastDistance |
	ogl glEnable: GLDepthTest.
	ogl glEnable: GLLighting.
	"------ set up the view frustum if necessary ------"
	ogl camera ~= self ifTrue:[
		ogl camera: self.
		self initFrustum: ogl. 
		].

	space renderClearBounds: ogl.
	ogl eventPointer ifNotNil:[
		lastTarget := ogl eventPointer selectedTarget.
		lastDistance := ogl eventPointer selectedDistance.
		ogl eventPointer selectedDistance: Float infinity.
		rval := space renderSpace: ogl.
		ogl eventPointer selectedTarget = lastTarget ifTrue:[
			ogl eventPointer selectedDistance: lastDistance.
			ogl eventPointer isOverlay: false.
		] ifFalse:[ ogl eventPointer isOverlay: true.].
	] ifNil:[
		rval := space renderSpace: ogl.
	].
	self globalTransform: Matrix4x4 identity.
	rval := rval + (self renderOverlay: ogl with: overlays).
	^ rval.! !

!TCamera methodsFor: 'render' stamp: 'das 1/19/2006 20:48'!
renderOverlay: ogl with: overlays

	| fw rval |
	fw := ogl forceWire.
	ogl forceWire: false.
	rval := 0.
	overlays ifNotNil:[
		overlays do:[:ov |
				"reset the buffers"
			ov ifNotNil:[
				ov send:[ :vwPnt |
					vwPnt rearrange: ogl.
					ogl glDepthMask: GLTrue;
						glDisable: GLBlend;
						glClear: (GLDepthBufferBit bitOr: GLStencilBufferBit).
				"and render each overlay"
					rval := rval + (vwPnt renderAsOverlay: ogl).
				].
			].
		].
	].
	ogl forceWire: fw.
	^ rval.! !

!TCamera methodsFor: 'render' stamp: 'ar 2/23/2006 14:28'!
renderView: ogl space: space overlay: overlays

	| rval |
	self preRender.
	ogl glEnable: GLDepthTest.
	ogl beginFrame.
	ogl glEnable: GLLighting.
	"------ set up the view frustum if necessary ------"
	ogl camera ~= self ifTrue:[
		ogl camera: self.
		self initFrustum: ogl. 
	].
	space renderClearAll: ogl.
	rval := space renderSpace: ogl.
	self globalTransform: Matrix4x4 identity.
	rval := rval + (self renderOverlay: ogl with: overlays).
	ogl endFrame.
	self postRender.
	^ rval.! !

!TCamera methodsFor: 'accessing' stamp: 'ar 10/10/2002 16:11'!
roll
	^self pitchYawRoll z! !

!TCamera methodsFor: 'accessing' stamp: 'ar 10/10/2002 16:18'!
roll: r
	| angles |
	angles := self pitchYawRoll.
	angles z: r.
	self pitchYawRoll: angles.! !

!TCamera methodsFor: 'transform' stamp: 'DPR 1/9/2004 14:05'!
scale: scale

	self doScale: scale.
! !

!TCamera methodsFor: 'accessing' stamp: 'das 12/13/2002 10:23'!
sideVector

	^ self globalTransform column1
	! !

!TCamera methodsFor: 'accessing' stamp: 'das 3/2/2006 17:17'!
subBounds:bnds

	| radians va zs |
	bnds ifNil:[subCornerVector := nil. ^ self].
	va := (bnds height asFloat/bounds height)*viewAngle.
	radians := (va/2.0)*(Float pi/180.0).
	zs := (bnds height/2.0) / (radians tan).
	subCornerVector := ((bnds width/2.0)@(bnds height/2.0)@(zs negated)) normalized.
! !

!TCamera methodsFor: 'accessing' stamp: 'das 3/7/2006 13:35'!
subCornerVector

	^ subCornerVector ifNil:[cornerVector].
! !

!TCamera methodsFor: 'render' stamp: 'ar 10/25/2002 23:03'!
teleportTo: newTransform in: newSpace
	"Should be done better. Oh well..."
	self localTransform: newTransform.
	self enter: newSpace.! !

!TCamera methodsFor: 'render' stamp: 'das 5/14/2005 15:08'!
testBounds: bnds
	| position pposition d cp fsi |

	bnds ifNil: [^ true].
	position := bnds globalPosition - self globalPosition.
"------- test to the near and far clip planes -------"
	d := (self lookAt dot: position) negated.
	fsi := frameScaleInverse.
	d + bnds radius < (zNear*fsi) ifTrue:[^ false].
	d - bnds radius > (zFar*fsi) ifTrue:[^false].
	1 to: 4 do:[:i|
		cp := clipPlanesTransform at: i.
		(0 > ((cp dot: position)+bnds radius)) ifTrue:[ ^ false.]].

	" portalClip is the clipping plane of the portal if we happen to be rendering inside of one. "
	portalClip ifNotNil:[
		pposition := bnds globalPosition - portalClip translation.
		0 > ((portalPlane dot: pposition)+ bnds radius) ifTrue:[ ^ false].
		].
^ true! !

!TCamera methodsFor: 'render' stamp: 'das 5/23/2005 11:35'!
testSphere: pos radius: radius
	| pposition position plane d  fsi |

	position := pos - self globalPosition.
"------- test to the near and far clip planes -------"
	plane := self lookAt.
	d := ( plane dot: position ) negated.
	fsi := frameScaleInverse.
	d + radius < (zNear*fsi) ifTrue:[^ false].
	d - radius > (zFar*fsi) ifTrue:[^false].
"------- test to the four clip planes -------"
	1 to: 4
		do: [:index | 0 > (((clipPlanesTransform at:index)
								dot: position) + radius)
				ifTrue: [^ false]].
	" portalClip is the clipping plane of the portal if we happen to be rendering inside of one. "
	portalClip ifNotNil:[
		pposition := position - portalClip translation.
		plane := Vector3 x: portalClip a13 negated y:portalClip a23 negated z: portalClip a33 negated.
		0 > ((plane dot: pposition)+ radius) ifTrue:[^ false].
	].
	^ true! !

!TCamera methodsFor: 'accessing' stamp: ''!
texture

	^ texture.! !

!TCamera methodsFor: 'accessing' stamp: ''!
texture: txtr

	texture := txtr.! !

!TCamera methodsFor: 'transform' stamp: 'das 3/22/2006 13:58'!
transformClipPlanes
	|  trans |
	clipPlanesTransform := Array new: 4.
	trans := self globalOrientation.
	1 to: 4 do: [:index | 
		clipPlanesTransform at: index put: (trans localPointToGlobal: (clipPlanes at: index))].

	! !

!TCamera methodsFor: 'accessing' stamp: 'das 12/13/2002 10:22'!
upVector

	^ self globalTransform column2.

	! !

!TCamera methodsFor: 'transform' stamp: 'das 12/20/2003 14:58'!
updateTransform! !

!TCamera methodsFor: 'accessing' stamp: ''!
viewAngle

	^ viewAngle.
! !

!TCamera methodsFor: 'accessing' stamp: 'das 12/21/2003 12:39'!
viewAngle: va

	viewAngle := va.
! !

!TCamera methodsFor: 'accessing' stamp: ''!
viewClip

	^ viewClip.! !

!TCamera methodsFor: 'accessing' stamp: ''!
viewClip: bool

	viewClip := bool.! !

!TCamera methodsFor: 'accessing' stamp: 'ar 10/10/2002 16:12'!
yaw
	^self pitchYawRoll y! !

!TCamera methodsFor: 'accessing' stamp: 'ar 10/10/2002 16:18'!
yaw: y
	| angles |
	angles := self pitchYawRoll.
	angles y: y.
	self pitchYawRoll: angles.! !

!TCamera methodsFor: 'accessing' stamp: ''!
zFar

	^ zFar.! !

!TCamera methodsFor: 'accessing' stamp: 'das 12/21/2003 12:39'!
zFar: zf

	zFar := zf.
! !

!TCamera methodsFor: 'accessing' stamp: 'das 12/21/2003 12:39'!
zNear

	^ zNear.! !

!TCamera methodsFor: 'accessing' stamp: 'das 12/21/2003 12:39'!
zNear: zn

	zNear := zn.
! !

!TCamera methodsFor: 'accessing' stamp: 'das 5/20/2002 18:06'!
zScreen

	^ zScreen.! !

!TCamera methodsFor: 'accessing' stamp: 'das 7/31/2004 12:16'!
zScreen: zs

	zScreen := zs.! !

!TGroup methodsFor: 'accessing' stamp: 'das 8/26/2006 18:56'!
colorize: col

	"Apply this color to all children of the group."

	frameChildren ifNotNil:[frameChildren do:[ :fc | fc colorize: col].].! !

!TGroup methodsFor: 'initialize' stamp: 'bgf 7/9/2006 13:42'!
initialize
	super initialize.
	visible := false.! !

!TGroup methodsFor: 'testing' stamp: 'das 4/10/2003 23:13'!
isGroup

	^ true.! !

!TGroup methodsFor: 'accessing' stamp: 'das 9/8/2002 22:40'!
visibleTree: bool

	"Skip yourself, just do the children."
	frameChildren ifNotNil:[
		frameChildren do:[ :fc | fc visibleTree: bool.].].! !

TGroup subclass: #TGroupNoClip
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TGroupNoClip methodsFor: 'as yet unclassified' stamp: 'das 7/10/2006 12:40'!
enableClipPlane: ogl
! !

TGroup subclass: #TLaser
	instanceVariableNames: 'shaft point selected arrowLength radius length color glowMaterial glowPointMaterial spell attachment'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TLaser commentStamp: 'das 10/10/2004 12:04' prior: 0!
TLaser is used by the TUserCamera and TAvatar to show other users what this user is currently pointing at or manipulating. This is just a visible representation of the TPointer and does nothing on it's own.

DAS !

!TLaser methodsFor: 'accessing' stamp: 'das 3/9/2004 13:59'!
addAttachment: att

	attachment := att.
	self addChild: att.! !

!TLaser methodsFor: 'accessing' stamp: 'DPR 5/29/2004 11:39'!
color
	^color.! !

!TLaser methodsFor: 'accessing' stamp: 'bgf 7/19/2006 15:10'!
color: col

	color := col copy.
	color at: 4 put: 0.2.

	glowMaterial ambientColor: color.
	glowMaterial diffuseColor: color.
	glowPointMaterial ambientColor: color.
	glowPointMaterial diffuseColor: color.

	shaft material: glowMaterial.
	point material: glowPointMaterial.

	! !

!TLaser methodsFor: 'accessing' stamp: 'ar 1/14/2006 18:50'!
glowPointTexture
	^glowPointMaterial texture! !

!TLaser methodsFor: 'accessing' stamp: 'ar 1/14/2006 18:50'!
glowPointTexture: aTexture
	glowPointMaterial texture: aTexture! !

!TLaser methodsFor: 'accessing' stamp: 'ar 1/14/2006 18:47'!
glowTexture
	^glowMaterial texture! !

!TLaser methodsFor: 'accessing' stamp: 'ar 1/14/2006 18:47'!
glowTexture: aTexture
	glowMaterial texture: aTexture! !

!TLaser methodsFor: 'accessing' stamp: 'das 7/5/2006 13:28'!
initialize

	| glow |
	super initialize.
	arrowLength := 0.5.
	radius := 0.025.
	length := 1.0.
	color := #(5.0 0.0 0.1 0.35).

	shaft := TGroup new.
	shaft rotationAroundX: -90.0.
	self addChild: shaft.

	point := TGroup new.
	point rotationAroundX: -90.0.
	point translationX: 0.0 y:0.0 z:arrowLength-length.
	self addChild: point.

	glowMaterial := TMaterial new.
	glowMaterial ambientColor: color.
	glowMaterial diffuseColor: color.
	glowMaterial textureMode: GLModulate.
	glowMaterial cullFace: false.


	glowPointMaterial := TMaterial new.
	glowPointMaterial ambientColor: color.
	glowPointMaterial diffuseColor: color.
	glowPointMaterial textureMode: GLModulate.
	glowPointMaterial cullFace: false.

	1 to: 4 do:[ :i |
	glow := TTriangle new.
	glow v1: (Vector3 x:0 y: arrowLength z:0)
			   v2: (Vector3 x:radius*5 y: -0.1 z: 0)
			   v3: (Vector3 x:radius*-5 y: -0.1 z: 0).
	glow material: glowPointMaterial.
	glow rotationAroundY: 45.0 * i.
	point addChild: glow.

	glow := TRectangle new.
	glow extent: (radius*6)@(length - arrowLength).
	glow location: (Vector3 x:0 y: (length-arrowLength)/2.0 z:0.0).
	glow material: glowMaterial.
	glow rotationAroundY: 45.0 * i.
	shaft addChild: glow.].

	self solidTree: false.
	spell := nil.
	^self! !

!TLaser methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:56'!
length: ln 
	ln > 0 ifTrue:[
	length := ln.
	point translationX: 0.0 y: 0.0 z: arrowLength - length.
	shaft frameChildren do:[ :glow |
		glow extent: radius * 6 @ (0.2 + length - arrowLength).
		glow location: (Vector3 x: 0 y: length - arrowLength / 2.0 z: 0.0).
		].].
	spell ifNotNil:[spell translation:(Vector3 x:0 y: 0 z: length negated).].
! !

!TLaser methodsFor: 'accessing' stamp: 'das 3/9/2004 14:00'!
removeAttachment

	self removeChild: attachment.
	attachment := nil.
! !

!TLaser methodsFor: 'accessing' stamp: 'das 9/14/2002 00:17'!
scale: scale
	super scale: scale.
	arrowLength := scale * arrowLength.
	radius := scale * radius.
	length := scale * length.! !

TGroup subclass: #TLight
	instanceVariableNames: 'type local position ambientColor diffuseColor specularColor spotDirection spotExponent spotCutoff renderObject distance'
	classVariableNames: ''
	poolDictionaries: 'OpenGLConstants'
	category: 'Croquet-Objects'!

!TLight commentStamp: 'das 10/10/2004 12:19' prior: 0!
TLight

- Spaces. Each space can contain any number of lights, but only the closest eight are actually enabled, with the furthest two linearly fading depending upon distance ratios. The limit of eight lights is a hard coded issue with OpenGL. The lights and their frames are stored in an array in the TSpace object. When a TSpace is rendered, it activates the lights first.

- Visibility. Lights are render objects that can be visible or not. This is to help in placing them. A visible light is represented either with a TSphere for positional, a TCylinder for spot, or two TCylinders for directional. These objects are rendered with an alpha value in the color of the light. 

- Locality. A light can be applied only to a frame and its children by setting local to true. Otherwise, it is global to the containing space.

- Attenuation. This assumes constant attenuation for performance reasons.


Things to do:

	- Need to utilize material types for rendering light objects.

type - the kind of light this is.
	#directional - light is positioned at infinity, rays are parallel.
	#point - light is positioned at a specific point location and radiating from this.
	#spot - point light with a focus in a specific direction - a spot light
local - (I don't remember what this is for)
position - OpenGL position flag.
ambientColor - 
diffuseColor - 
specularColor -		light type
spotDirection - 
spotExponent -
spotCutoff -
renderObject - object to use when we render the light. The actual object is determined by the light type.
distance - used to determine the distance from the camera for sorting.

DAS!

!TLight class methodsFor: 'as yet unclassified' stamp: 'das 4/1/2006 14:43'!
makeLights: space

	| light |

"warm light in northwest corner."
	light := TLight new.
	light ambientColor: #(0.3 0.4 0.5 0.6).
	light diffuseColor: #(0.48 0.35 0.4 0.6).
	light specularColor: #(0.3 0.23 0.2 0.5).
	light rotationAroundY: -130.
	light addRotationAroundX: 120.
	space addChild: light.

"cool light diagonally opposite."
	light := TLight new.
	light ambientColor: #(0.1 0.1 0.1 0.1).
	light diffuseColor: #(0.4 0.35 0.48 0.6).
	light specularColor: #(0.2 0.23 0.3 0.5).
	light rotationAroundY: 132.
	light addRotationAroundX: 120.
	space addChild: light.

" fainter green light in third corner."
	light := TLight new.
	light ambientColor: #(0.1 0.1 0.1 0.1).
	light diffuseColor: #(0.1 0.23 0.1 0.4).
	light specularColor: #(0.3 0.2 0.3 0.3).
	light rotationAroundY: 42.
	light addRotationAroundX: 120.
	space addChild: light.

"and another opposite it"
	light := TLight new.
	light ambientColor: #(0.1 0.1 0.1 0.1).
	light diffuseColor: #(0.1 0.23 0.1 0.4).
	light specularColor: #(0.4 0.3 0.2 0.2).
	light rotationAroundY: -42.
	light addRotationAroundX: 120.
	space addChild: light.
! !

!TLight methodsFor: 'accessing' stamp: ''!
ambientColor

	^ ambientColor.! !

!TLight methodsFor: 'accessing' stamp: 'ar 2/7/2006 17:35'!
ambientColor: color

	ambientColor := color asVectorColor.! !

!TLight methodsFor: 'accessing' stamp: ''!
diffuseColor

	^ diffuseColor.! !

!TLight methodsFor: 'accessing' stamp: 'ar 2/7/2006 17:35'!
diffuseColor: color

	diffuseColor := color asVectorColor.! !

!TLight methodsFor: 'accessing' stamp: 'das 9/4/2002 14:53'!
distance

	^ distance.! !

!TLight methodsFor: 'accessing' stamp: 'das 9/4/2002 14:53'!
distance: dist

	distance := dist.! !

!TLight methodsFor: 'render' stamp: 'das 8/23/2006 15:06'!
enable: ogl index: lNum

" lNum initial value must be 1."
	| glLight |

	glLight := GLLight0 + lNum - 1.
	ogl glPushMatrix;

		glMultMatrixf: self globalTransform transposed;
		glLightfv: glLight with: GLPosition with: position.
	type = #spot ifTrue:[ 
		ogl glLightfv: glLight with: GLSpotDirection with: spotDirection;
			glLightf: glLight with: GLSpotExponent with: spotExponent;
			glLightf: glLight with: GLSpotCutoff with: spotCutoff.
		] ifFalse:[
		ogl glLightf: glLight with: GLSpotCutoff with: 180.0.
		"ogl glLightfv: glLight with: GLSpotDirection with: #(0 0 0)asFloatArray;
			glLight: glLight with: GLSpotExponent with: 0.0."
		].
	ogl glLightfv: glLight with: GLAmbient with: ambientColor;
		glLightfv: glLight with: GLDiffuse with: diffuseColor;
		glLightfv: glLight with: GLSpecular with: specularColor;
		glEnable: glLight;
		glPopMatrix.

! !

!TLight methodsFor: 'render' stamp: 'das 3/12/2004 13:39'!
enable: ogl index: lNum scaled: scale

" lNum initial value must be 1."
	| glLight |

	glLight := GLLight0 + lNum - 1.
	ogl glPushMatrix;
		glMultMatrixf: self globalTransform transposed;
		glLightfv: glLight with: GLPosition with: position;
		glLightfv: glLight with: GLAmbient with: (ambientColor*scale);
		glLightfv: glLight with: GLDiffuse with: (diffuseColor*scale);
		glLightfv: glLight with: GLSpecular with: (specularColor*scale);
		glEnable: glLight;
		glPopMatrix.

! !

!TLight methodsFor: 'accessing' stamp: 'das 9/4/2002 14:53'!
global
	^ local not.! !

!TLight methodsFor: 'accessing' stamp: 'das 9/4/2002 14:53'!
global: boolGlobal
	local := boolGlobal not.! !

!TLight methodsFor: 'events' stamp: 'das 6/13/2002 13:56'!
handlesKeyboard: ptr

	^true.! !

!TLight methodsFor: 'initialize' stamp: 'das 8/23/2006 13:06'!
initialize
	"default is a directional light - if w is non-zero then it is a positional"
	super initialize.	
	position      := #(0.0 -1.0  0.0  0.0) asFloatArray.
	ambientColor  := #(1.0  1.0  1.0  1.0)  asVectorColor.
	diffuseColor  := #(1.0  1.0  1.0  1.0)  asVectorColor.
	specularColor := #(1.0  1.0  1.0  1.0)  asVectorColor.
	spotDirection := #(0.0  0.0 -1.0) asFloatArray.
	spotExponent  := 1.0.
	spotCutoff    := 22.5.
	type := #directional.
	local := false.
	^self
! !

!TLight methodsFor: 'accessing' stamp: 'hrs 5/16/2006 09:48'!
initializeRenderObjectIfNeeded
	visible ifTrue: [self renderObject].! !

!TLight methodsFor: 'testing' stamp: 'das 6/13/2002 13:55'!
isLight

	^ true.! !

!TLight methodsFor: 'events' stamp: 'ar 1/27/2006 14:42'!
keyDown: evt

	| c |
	c := evt keyValue asCharacter.
	c = $p ifTrue:[ self type: #point.].
	c = $d ifTrue:[ self type: #directional.].
	c = $s ifTrue:[ self type: #spot.].
! !

!TLight methodsFor: 'events' stamp: 'ar 1/27/2006 14:42'!
keyStroke: evt

	| c |
	c := evt keyValue asCharacter.
	c = $p ifTrue:[ self type: #point.].
	c = $d ifTrue:[ self type: #directional.].
	c = $s ifTrue:[ self type: #spot.].
! !

!TLight methodsFor: 'accessing' stamp: ''!
local
	^ local! !

!TLight methodsFor: 'accessing' stamp: ''!
local: boolLocal
	local := boolLocal.! !

!TLight methodsFor: 'render' stamp: 'hrs 5/16/2006 09:49'!
render: ogl

	ogl glPushMatrix.

	ogl 
		glRotatef: 270.0 
		with: 1.0
		with: 0.0
		with: 0.0.	

"	local ifTrue:[]."

	ogl glPopMatrix.! !

!TLight methodsFor: 'render' stamp: 'das 3/12/2004 13:39'!
renderAlpha: ogl

	"self render: space."
! !

!TLight methodsFor: 'render' stamp: 'das 7/5/2006 13:23'!
renderDirectional

	| part |

	" This renders a visual representation of a directional light. This is used to orient the light."

	renderObject := TGroup new.

	part := TCylinder new.
	part topRadius: 0.0.
	part baseRadius: 0.5.
	part height: 0.5.	
	renderObject addChild: part.

	part := TCylinder new.
	part baseRadius: 0.25.
	part topRadius: 0.25.
	part height: 0.5.
	part translationX: 0.0 y: -0.5 z:0.0.
	renderObject addChild: part.

	self addChild: renderObject.! !

!TLight methodsFor: 'accessing' stamp: 'hrs 5/16/2006 09:46'!
renderObject
	^renderObject ifNil:[
		type = #point ifTrue:[ self renderPoint].	
		type = #directional ifTrue:[ self renderDirectional].
		type = #spot ifTrue:[ self renderSpot].
		renderObject.
		]. ! !

!TLight methodsFor: 'render' stamp: 'das 7/5/2006 13:26'!
renderPoint

	" This renders a visual representation of a point light source. This is used to orient the light."

	renderObject := TSphere new.
	renderObject radius: 0.5.
	self addChild: renderObject.

! !

!TLight methodsFor: 'render' stamp: 'das 7/5/2006 13:23'!
renderSpot

	| part |

	" This renders a visual representation of a spot light. This is used to orient the light."

	renderObject := TGroup new.

	part := TCylinder new.
	part topRadius: 0.5.
	part baseRadius: 0.25.
	part height: 0.75.		
	renderObject addChild: part.

	part := TCylinder new.
	part baseRadius: 0.25.
	part topRadius: 0.25.
	part height: 0.5.
	part translationX: 0.0 y: -0.5 z:0.0.
	renderObject addChild: part.

	self addChild: renderObject.
! !

!TLight methodsFor: 'accessing' stamp: ''!
specularColor

	^ specularColor.! !

!TLight methodsFor: 'accessing' stamp: 'ar 2/7/2006 17:35'!
specularColor: color

	specularColor := color asVectorColor.! !

!TLight methodsFor: 'accessing' stamp: 'das 8/23/2006 13:05'!
spotCutoff
	^ spotCutoff asFloat.! !

!TLight methodsFor: 'accessing' stamp: 'das 8/23/2006 14:00'!
spotCutoff: sc
	spotCutoff := sc asFloat.! !

!TLight methodsFor: 'accessing' stamp: ''!
spotDirection

	^ spotDirection.! !

!TLight methodsFor: 'accessing' stamp: ''!
spotDirection: sd

	spotDirection := sd.! !

!TLight methodsFor: 'accessing' stamp: ''!
spotExponent
	^ spotExponent.! !

!TLight methodsFor: 'accessing' stamp: 'das 8/23/2006 13:05'!
spotExponent: se
	spotExponent := se asFloat.! !

!TLight methodsFor: 'accessing' stamp: ''!
type

	^ type.! !

!TLight methodsFor: 'accessing' stamp: 'hrs 5/16/2006 09:49'!
type: tp

	| |
	
	tp = type ifTrue:[ ^ self ].
	type := tp.
"	pos := self positionV."

	tp = #directional ifTrue:[
		position := #(0 -1.0 0 0) asFloatArray.
		] 
		ifFalse:[
	tp = #point ifTrue:[
		position := #(0 0 0 1) asFloatArray.
		] ifFalse:[
	tp = #spot ifTrue:[
		position := #(0 -1.0 0 1) asFloatArray.
	]]].

"	self positionV: pos."
	self removeAll.
	renderObject := nil.
	self initializeRenderObjectIfNeeded.
	! !

!TLight methodsFor: 'accessing' stamp: 'hrs 5/16/2006 09:50'!
visible: aBoolean
	super visible: aBoolean.
	self initializeRenderObjectIfNeeded.
! !

TGroup subclass: #TMultiCostume
	instanceVariableNames: 'costumes costume'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TMultiCostume methodsFor: 'accessing' stamp: 'das 8/27/2006 13:25'!
addCostume: cstm

	costumes ifNil:[costumes:= OrderedCollection new. ].
	costumes add: cstm.
	cstm solidTree: false.
	costume ifNil:[self selectCostume: 1].
	! !

!TMultiCostume methodsFor: 'accessing' stamp: 'das 8/26/2006 18:52'!
colorize: col

	costumes do:[:cos | cos colorize:col.].! !

!TMultiCostume methodsFor: 'accessing' stamp: 'das 5/26/2006 11:12'!
selectCostume: index

	costumes ifNotNil:[
		costumes size >= index ifTrue:[
			costume ~= (costumes at: index) ifTrue:[
				costume ifNotNil:[self removeChild: costume.].
				costume := costumes at: index.
				self addChild: costume.
			].
		].
	].! !

!TMultiCostume methodsFor: 'accessing' stamp: 'das 8/27/2006 14:49'!
spin360
	
	self spin360:360.! !

!TMultiCostume methodsFor: 'accessing' stamp: 'das 8/27/2006 14:53'!
spin360: angle

	angle <= 0 ifTrue:[^self].
	self addRotationAroundY:5.
	(self future:100)spin360:angle-5.! !

!TMultiCostume methodsFor: 'accessing' stamp: 'das 5/28/2006 19:15'!
visibleTree: bool

	costumes do:[:cs | cs visibleTree: bool].! !

TGroup subclass: #TRandTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TRandTest methodsFor: 'initialize' stamp: 'das 11/3/2005 12:24'!
initialize
	super initialize.
	(self future: 1000)step.
! !

!TRandTest methodsFor: 'step' stamp: 'das 12/2/2005 13:36'!
step

"	Transcript show: 100 atRandom; show: ' - '; show: Processor activeIsland random count; cr.
"	(self future: 1000)step.! !

TGroup subclass: #TScrollBox3D
	instanceVariableNames: 'sides contents box matNorm matOver matHilite matDown selectedFrame selectedPoint selectedCorner scaleContents basePosition baseLength baseScale lastSelected shiftPressed selectedAction changed rotateAround deferScale newScale lastSpin rotPoint rotateDirection radius boxPoint sideScale'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TScrollBox3D commentStamp: '<historical>' prior: 0!
TScrollBox3D

This is used to manipulate a 3D portal, as well as acting as a base class for the 3D TEditBox.!

TScrollBox3D subclass: #TEditBox3D
	instanceVariableNames: 'editPortal ghostFrame ghostParent ghostChild'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TEditBox3D methodsFor: 'events' stamp: 'das 2/9/2005 00:46'!
delete

	super delete.
	editPortal ifNotNil:[editPortal clearSelection.].! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/3/2006 11:12'!
duplicate

	| copy |
	copy := contents copy.
	copy localTransform: Matrix4x4 identity.
	copy translation: (copy translation + (2@2@2)).
	contents addChild: copy.
	TEditBox3D new initializeWithContents: copy editPortal: editPortal.

! !

!TEditBox3D methodsFor: 'initialize' stamp: 'das 3/10/2004 22:16'!
initialize

	super initialize.
	self deferScale: false.
! !

!TEditBox3D methodsFor: 'initialize' stamp: 'das 8/3/2006 11:04'!
initializeWithContents: conts editPortal: ep

	contents := conts.
	ep ifNotNil:[
		editPortal := ep.
		editPortal selection: self.
		ghostFrame := editPortal ghost.
	].
	self initializeWithContents: conts.
! !

!TEditBox3D methodsFor: 'initialize' stamp: 'das 8/3/2006 11:14'!
intoBox

	ghostFrame ifNil:[^super intoBox.].
	ghostParent := ghostParent ifNil:[ TGroup new ].
	ghostParent localTransform: contents parent globalTransform copy.
	ghostFrame addChild: ghostParent.
	ghostChild := ghostChild ifNil:[TGroup new].
	ghostChild localTransform: contents localTransform copy.
	ghostParent addChild: ghostChild.

	box := self getBox: contents.
	self localTransform: ghostChild localTransform copy.
	self translation: self translation + (self orientation localPointToGlobal:box center).
	ghostChild localTransform: Matrix4x4 identity.
	ghostChild translation: (box center negated).
	ghostChild insertFrame: self.
	! !

!TEditBox3D methodsFor: 'events' stamp: 'das 8/5/2006 15:59'!
keyDown: pointer

	| c |
	c := pointer event2D keyCharacter.
	c = $d ifTrue:[self delete.].
	c = $@ ifTrue:[self duplicate.].
	pointer event2D keyValue= 28 ifTrue:[ self selectLeftSibling. ]. "left arrow"
	pointer event2D keyValue= 29 ifTrue:[ self selectRightSibling.]. "right arrow"
	pointer event2D keyValue= 30 ifTrue:[ self selectParent. ]. "up arrow"
	pointer event2D keyValue= 31 ifTrue:[ self selectChild. ]. "down arrow"
	^ true.! !

!TEditBox3D methodsFor: 'events' stamp: 'das 8/5/2006 15:59'!
keyStroke: pointer

	| c |
	c := pointer event2D keyCharacter.
	c = $d ifTrue:[self delete.].
	pointer event2D keyValue= 28 ifTrue:[ self selectLeftSibling. ]. "left arrow"
	pointer event2D keyValue= 29 ifTrue:[ self selectRightSibling.]. "right arrow"
	pointer event2D keyValue= 30 ifTrue:[ self selectParent. ]. "up arrow"
	pointer event2D keyValue= 31 ifTrue:[ self selectChild. ]. "down arrow"
	^ true.! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/18/2006 16:03'!
moveBox: event

		self translation: self translation +
			(self orientation localPointToGlobal: (selectedFrame orientation localPointToGlobal: event selectedPoint-selectedPoint)).

	ghostFrame ifNotNil:[
		contents localTransform: (self localTransform composeWith: ghostChild localTransform).
		].
! !

!TEditBox3D methodsFor: 'initialize' stamp: 'das 8/10/2005 11:30'!
outOfBox

	ghostFrame ifNil:[^super outOfBox.].

	self removeChild: ghostChild.
	ghostFrame removeChild: ghostParent.
! !

!TEditBox3D methodsFor: 'initialize' stamp: 'das 8/12/2005 18:39'!
resetBox

	self outOfBox.
	self intoBox.! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/3/2006 11:02'!
resizeBox: pointer
"This needs to be fixed - since the plane does not go through the center of the cube, there is a limit to how small this can get in any one motion. The advantage is this keeps us from getting too small."
	| len1 len2  norm sp sc|

	norm := (pointer camera frameLookAt: self) negated.
	sp := selectedFrame localTransform localPointToGlobal: selectedPoint.
	(pointer frame: self pickPlane: sp normal: norm) ifTrue:[
		len1 := sp length.
		len2 := pointer selectedPoint length.
		len1 = 0.0 ifTrue:[^ false.].
		sc := len2/len1.
		self scale: sc.
		scaleContents := scaleContents * sc.
		selectedPoint := (selectedFrame localTransform orthoNormInverse localPointToGlobal: pointer selectedPoint).
		^ true.].
	^ false.! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/5/2006 15:59'!
rotateBoxX: pointer

	super rotateBoxX: pointer.
	ghostFrame ifNotNil:[
		contents localTransform: (self localTransform composeWith: ghostChild localTransform).
		].
! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/5/2006 15:59'!
rotateBoxY: pointer

	super rotateBoxY: pointer.
	ghostFrame ifNotNil:[
		contents localTransform: (self localTransform composeWith: ghostChild localTransform).
		].
! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/5/2006 15:59'!
rotateBoxZ: pointer

	super rotateBoxZ: pointer.
	ghostFrame ifNotNil:[
		contents localTransform: (self localTransform composeWith: ghostChild localTransform).
		].
! !

!TEditBox3D methodsFor: 'accessing' stamp: 'das 1/25/2005 16:37'!
scale: sc

	box scale: sc.
	self updateBox.
	self parentChanged: contents.
! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/5/2006 15:59'!
scaleBox: pointer

	| delta delta2 c1 c2 |
	delta := pointer selectedPoint - selectedPoint.
	delta2 := delta * selectedCorner.
"delta is currently defined in terms of the selectedFrame - need to convert to this frame"
	delta := selectedFrame orientation localPointToGlobal: delta.
	selectedFrame translation: selectedFrame translation + delta.
	delta2 := selectedFrame orientation abs localPointToGlobal: delta2.
	c1 := (self getBox: contents) center.
	contents extent: contents extent+delta2.
	c2 := (self getBox: contents) center.
	contents translation: contents translation + c1-c2 + (self orientation localPointToGlobal:(delta/2.0)).

! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/3/2006 11:02'!
selectChild
	| eb sel |

	self parent ifNil: [^nil.].

	contents frameChildren ifNotNil:[
		contents frameChildren size > 0 ifTrue:[
			eb:= TEditBox3D new.
			sel := contents frameChildren at: 1.
			self delete.
			eb initializeWithContents: sel editPortal: editPortal.
			editPortal ifNotNil:[editPortal selection: eb.].
			].
		].
! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/3/2006 11:03'!
selectLeftSibling
	|  siblings previous eb |

	self parent ifNil: [^nil.].

	siblings := self parent frameChildren.

	siblings size > 1 ifTrue:[
		previous := siblings previous: self.
		self delete.
		eb := TEditBox3D new. 
		eb initializeWithContents: previous editPortal: editPortal.
		editPortal ifNotNil:[editPortal selection: eb.].
		].
! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/3/2006 11:03'!
selectParent
	| par eb |
	self parent ifNotNil: [ "we are the root"
		self parent ~= self root ifTrue:[ " don't select the root, as that would be really hard to edit"
		par := self parent.
		self delete.
		eb := TEditBox3D new.
		eb initializeWithContents: par editPortal: editPortal.
		editPortal ifNotNil:[editPortal selection: eb.].
		].
	].
		! !

!TEditBox3D methodsFor: 'actions' stamp: 'das 8/3/2006 11:03'!
selectRightSibling
	|  siblings next eb |
	self parent ifNil: [^nil.].

	siblings := self parent frameChildren.

	siblings size > 1 ifTrue:[
		next := siblings next: self.
		self delete.
		eb := TEditBox3D new.
		eb initializeWithContents: next editPortal: editPortal.
		editPortal ifNotNil:[editPortal selection: eb.].
		].
! !

!TEditBox3D methodsFor: 'accessing' stamp: 'das 8/3/2006 11:03'!
selectedPoint: sp

	selectedPoint := sp.! !

!TEditBox3D methodsFor: 'accessing' stamp: 'das 1/24/2005 00:46'!
task: task

	^ nil.! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 8/9/2003 11:46'!
calcRotateDimension

	| dir  |

	dir := (selectedFrame translation cross: boxPoint) abs .
	radius := boxPoint length.
	dir x > dir y ifTrue:[
		dir x > dir z ifTrue:[ selectedAction := #rotateBoxX:. ] ifFalse:[  selectedAction := #rotateBoxZ:]
	]ifFalse:[
		dir y > dir z ifTrue:[selectedAction := #rotateBoxY:. ] ifFalse:[ selectedAction := #rotateBoxZ:]
	].
! !

!TScrollBox3D methodsFor: 'accessing' stamp: 'das 8/5/2003 16:10'!
changed

	changed := true.! !

!TScrollBox3D methodsFor: 'events' stamp: 'das 10/21/2003 16:05'!
child: frame
	super changed: frame.
	changed := true.! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'das 7/31/2005 16:43'!
contents

	^ contents.! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'das 4/1/2006 11:38'!
contents: conts

	| extent |
	contents := conts.
	box := self getBox: contents.
	self intoBox.
	extent := box extent.
	box min: (extent/-2.0).
	box max: (extent/2.0).
	self addChild: conts.
	scaleContents := 1.0.
	selectedAction := nil.
	changed := true.
	self updateBox.
! !

!TScrollBox3D methodsFor: 'accessing' stamp: 'das 8/1/2003 22:00'!
deferScale: bool

	deferScale := bool.! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 1/24/2005 11:06'!
delete

	self outOfBox.
! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 3/21/2006 14:56'!
doCorner: pointer

	(pointer frame: selectedFrame pickPlane: selectedPoint normal: (Vector3 x:0 y:0 z: 1)) 
	ifTrue:[
		shiftPressed ifTrue:[^ self scaleBox: pointer.]
		ifFalse:[^ self moveBox: pointer.].
		].
	^ false.! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 7/31/2003 21:51'!
doEdge: pointer

	shiftPressed ifTrue:[^ false.]
	ifFalse:[^ self rotateBox: pointer.].! !

!TScrollBox3D methodsFor: 'events' stamp: 'das 3/21/2006 14:51'!
eventMask

	^ (EventPointerDown bitOr: EventPointerOver).! !

!TScrollBox3D methodsFor: 'accessing' stamp: 'das 7/22/2003 15:04'!
extent

	contents ifNotNil:[ ^ contents extent ].
	^ nil.! !

!TScrollBox3D methodsFor: 'accessing' stamp: 'das 8/10/2005 11:44'!
extent: ext

	contents ifNotNil:[ 
		contents extent: ext.
		box := self getBox: contents.].
	changed := true.! !

!TScrollBox3D methodsFor: 'accessing' stamp: 'das 8/18/2006 22:35'!
getBox: cont

	^ contents boundingBox copy.! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'bgf 3/26/2006 17:10'!
initialize

	| txtr |
	super initialize.

	sideScale := 10.0.
	txtr := TTexture new
				initializeWithFileName: 'box.png'
				mipmap: true
				shrinkFit: false.

	matHilite := TMaterial new.
	matHilite ambientColor: TButton hiliteColor.
	matHilite diffuseColor: TButton hiliteColor.
"	matHilite cullFace: false."
	matHilite texture: txtr.
	matHilite allowWireFrame: false.

	matOver := TMaterial new.
	matOver ambientColor: TButton overColor.
	matOver diffuseColor: TButton overColor.
"	matOver cullFace: false."
	matOver texture: txtr.
	matOver allowWireFrame: false.

	matDown := TMaterial new.
	matDown ambientColor: TButton downColor.
	matDown diffuseColor: TButton downColor.
"	matDown cullFace: false."
	matDown texture: txtr.
	matDown allowWireFrame: false.

	matNorm := TMaterial new.
	matNorm ambientColor: TButton normColor.
	matNorm diffuseColor: TButton normColor.
"	matNorm cullFace: false."
	matNorm texture: txtr.
	matNorm allowWireFrame: false.
	
	deferScale := false.
	self makeSides.
	changed := true.
! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'das 3/21/2006 15:18'!
initializeWithContents: conts

	| extent |
	contents := conts.
	box := self getBox: contents.
	self intoBox.
	extent := box extent.
	box min: (extent/-2.0).
	box max: (extent/2.0).
	self addChild: conts.
	scaleContents := 1.0.
	selectedAction := nil.
	changed := true.
	self updateBox.
! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'das 3/21/2006 15:05'!
intoBox

	box := self getBox:contents.
	self localTransform: contents localTransform.
	self translation: self translation + (self orientation localPointToGlobal:box center).
	contents localTransform: Matrix4x4 identity.
	contents translation: (box center negated).
	contents insertFrame: self.! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'das 7/5/2006 14:14'!
makeSides
	| s g |
	sides := OrderedCollection ofSize: 6.
	1 to: 6 do:[: groupIndex |
		g := TGroup new.
		1 to: 8 do:[:index |
		
			s := TRectangle new.
			s materialAndTexture: matNorm.
			s solid: false.
			s normal: nil. " this allows the user to select from both sides"
			index <= 4 ifTrue:[ s objectName: #doCorner:.]
			ifFalse:[s objectName: #doEdge:.].
			g addChild: s.
			].
		sides at: groupIndex put: g.
		self addChild: g.
		].
	(sides at: 1)rotationAroundX:-90.
	(sides at: 2)rotationAroundX: 90.
	(sides at: 3)rotationAroundY:-90.
	(sides at: 4)rotationAroundY: 90.
	(sides at: 5)rotationAroundY:180.
! !

!TScrollBox3D methodsFor: 'accessing' stamp: 'das 7/22/2003 17:04'!
material: mat

	" do nothing - this is my material - you can't set it externally"
! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 7/18/2006 19:29'!
moveBox: pointer

	| delta |
	delta := pointer selectedPoint - selectedPoint.
"delta is currently defined in terms of the selectedFrame - need to convert to this frame"
	delta := selectedFrame orientation localPointToGlobal: delta.
	selectedFrame translation: selectedFrame translation + delta.
	contents offset: (contents offset - (delta/contents scale)).
	^ true.
! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'das 8/18/2006 22:19'!
outOfBox

	| p |
	contents localTransform: (self localTransform composeWith: contents localTransform).
	p := self parent.
	p ifNotNil:[
		p removeChild: self.
		p addChild: contents.
	] ifNil:[self removeChild: contents.].! !

!TScrollBox3D methodsFor: 'events' stamp: 'das 8/6/2006 12:32'!
pointerDown: event

	| trans |
	shiftPressed := event shiftPressed.
	selectedFrame := event selection frame parent.
	selectedFrame = self ifTrue:[selectedFrame := nil. ^self].
	(selectedFrame hasAncestorOrMe: contents) ifTrue:[selectedFrame := nil. ^self].

	trans := event selection frame translation.
	selectedPoint := event selection point + trans.
	selectedAction := event selection frame objectName.
	boxPoint := (selectedFrame localTransform localPointToGlobal: selectedPoint).	
	selectedAction = #doEdge: ifTrue:[ rotPoint := nil. self calcRotateDimension.]
	ifFalse:[
		selectedCorner := (trans x/trans x abs)@(trans y/trans y abs)@0.0.
	].

	basePosition := selectedFrame frameChildren collect:[:fc | fc translation.].
	baseLength := selectedPoint length.
	baseScale := contents scale.
	newScale := 1.0.
	selectedFrame frameChildren do:[:fc | fc material: matDown.].
! !

!TScrollBox3D methodsFor: 'events' stamp: 'das 8/6/2006 12:30'!
pointerEnter: event

	selectedFrame := event selection frame parent.
	selectedFrame = self ifTrue:[selectedFrame := nil. ^self].
	(selectedFrame hasAncestorOrMe: contents) ifTrue:[selectedFrame := nil. ^self].

	selectedFrame frameChildren do:[:fc | fc material: matHilite.].
! !

!TScrollBox3D methodsFor: 'events' stamp: 'das 8/6/2006 12:31'!
pointerLeave: event

	selectedFrame ifNil:[^self].
	selectedFrame frameChildren do:[:fc | fc material: matNorm.].
	selectedFrame := nil.! !

!TScrollBox3D methodsFor: 'events' stamp: 'das 8/5/2006 16:46'!
pointerMove: event

	| pointer |
	selectedFrame ifNotNil:[
		pointer := event makePointer.
		selectedAction ifNotNil:[ 
			self perform: selectedAction with: pointer. 
			self parentChanged: contents.
		].
	].! !

!TScrollBox3D methodsFor: 'events' stamp: 'das 8/5/2006 16:45'!
pointerOver: event

	selectedFrame = event selection frame ifFalse:[
		self pointerLeave: event.
		self pointerEnter: event.].
! !

!TScrollBox3D methodsFor: 'events' stamp: 'das 8/6/2006 12:20'!
pointerUp: event

	selectedFrame ifNotNil:[
		selectedFrame frameChildren do:[:fc | fc material: matOver.].
		deferScale ifTrue:[newScale ~= 1.0 ifTrue:[
			contents scale: baseScale *newScale.
	 		box := self getBox: contents.
		]].
		self parentChanged: contents.
		changed := true.
		selectedAction := nil.
		self updateBox.
	].
	selectedFrame := nil.! !

!TScrollBox3D methodsFor: 'render' stamp: 'das 1/25/2005 14:22'!
renderFrame: ogl parent: parent root: root

	changed ifTrue:[self updateBox. changed := false.].
	^ super renderFrame: ogl parent: parent root: root.! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'das 3/21/2006 15:05'!
resetBox

	contents localTransform: (self localTransform composeWith: contents localTransform).

	box := self getBox:contents.
	self localTransform: contents localTransform.
	self translation: self translation + (self orientation localPointToGlobal:box center).
	contents localTransform: Matrix4x4 identity.
	contents translation: (box center negated).
! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 3/21/2006 15:02'!
rotateBox: pointer

	| dir |
	dir := (selectedFrame translation cross: boxPoint) abs .
	radius := boxPoint length.
	dir x > dir y ifTrue:[
		dir x > dir z ifTrue:[ self rotateBoxX: pointer. ] ifFalse:[ self rotateBoxZ: pointer]
	]ifFalse:[
		dir y > dir z ifTrue:[ self rotateBoxY: pointer. ] ifFalse:[ self rotateBoxZ: pointer]
	].
! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 9/4/2006 18:52'!
rotateBoxX: pointer

	| len delta deltaLoc spto spfrom theta trans |
	" Rotate around the x-axis "

	len := 100.
	(pointer frame: self pickRightCylinderFrom: ( Vector3 x: len  y: 0 z: 0)
		to:( Vector3 x: len negated y: 0 z: 0) radius: radius) 
	ifFalse:["we are not on the cylinder anymore...."
		delta := 2*radius.
		(pointer framePointer row2 dot: (pointer framePosition))  > 0 
			ifFalse:[  delta := delta negated.].
		deltaLoc := pointer framePointer row2 * delta.
		(pointer frame: self pickRightCylinderFrom: (Vector3 x: len  y: deltaLoc y z: deltaLoc z)
			to: ( Vector3 x: len negated y: deltaLoc y z: deltaLoc z) radius: radius)
				ifFalse:[lastSpin := nil. ^ false.].
		spto := pointer selectedPoint copy- deltaLoc.
		spto x: 0
			y: spto y negated
			z: spto z negated.
			]
	ifTrue: [spto := pointer selectedPoint copy.].
	rotPoint ifNil:[
		rotPoint := spto. 
		(rotPoint - boxPoint) length >0.01 ifTrue:[rotateDirection := -1.0.] ifFalse:[rotateDirection := 1.0.].
	].
	spfrom := rotPoint copy.
	spfrom x: 0.0.
	spfrom normalize.
	spto x: 0.0.
	spto normalize.
	theta := ((spfrom y * spto z) - (spto y * spfrom z))"/selectedRadiusSquared".
	theta < -1.0 ifTrue:[theta:= -1.0].
	theta > 1.0 ifTrue:[theta := 1.0].
	theta := theta arcSin radiansToDegrees/2.0.
	lastSpin := Matrix4x4 identity rotationAroundX: theta.

	trans := self translation.
	self translationX: 0.0 y:0.0 z:0.0.
	self localTransform: (self localTransform composeWith: lastSpin).
	self translation: trans.
	^ true.
! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 9/4/2006 18:52'!
rotateBoxY: pointer

	| trans spfrom spto deltaLoc theta len delta |
		" Rotate around the y-axis "

		len := 100.0.
		(pointer frame: self pickRightCylinderFrom: ( Vector3 x: 0.0 y: len z: 0.0)
				to: (Vector3 x: 0.0 y: len negated z: 0.0) radius: radius)
			ifFalse: [ "we are not on the cylinder anymore...."
			delta := 2*radius.
			(pointer framePointer row1 dot: (pointer framePosition)) > 0
				ifFalse:[ delta := delta negated ].
			deltaLoc := pointer framePointer row1 * delta.
			(pointer frame: self pickRightCylinderFrom: (Vector3 x: deltaLoc x y: len z: deltaLoc z )
				to: (Vector3 x: deltaLoc x y: len negated z: deltaLoc z) radius: radius)
				ifFalse:[lastSpin := nil. ^ false.].
			spto := pointer selectedPoint copy- deltaLoc.
			spto x: spto x negated
				y: 0
				z: spto z negated.
			]
	ifTrue: [spto := pointer selectedPoint copy.].
	rotPoint ifNil:[
		rotPoint := spto. 
		(rotPoint - boxPoint) length >0.01 ifTrue:[rotateDirection := -1.0.] ifFalse:[rotateDirection := 1.0.].
	].

	spfrom := rotPoint copy.
	spfrom y: 0.0.
	spfrom normalize.
	spto y: 0.0.
	spto normalize.
	theta := ((spfrom z * spto x) - (spto z * spfrom x))"/selectedRadiusSquared".
	theta < -1.0 ifTrue:[theta:= -1.0].
	theta > 1.0 ifTrue:[theta := 1.0].
	theta := theta arcSin radiansToDegrees/2.0.
	lastSpin := Matrix4x4 identity rotationAroundY: theta.

	trans := self translation.
	self translationX: 0.0 y:0.0 z:0.0.
	self localTransform: (self localTransform composeWith: lastSpin).
	self translation: trans.
	^ true.
! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 9/4/2006 19:07'!
rotateBoxZ: pointer

	| delta trans spfrom spto deltaLoc theta len |
		" Rotate around the z-axis "

		len := 100.0.
		(pointer frame: self pickRightCylinderFrom: (Vector3 x: 0.0 y: 0.0 z: len)
				to: (Vector3 x: 0.0 y: 0.0 z: len negated) radius: radius)
			ifFalse: [ "we are not on the cylinder anymore...."
			delta := 2*radius.
			(pointer framePointer row2 dot: (pointer framePosition))  > 0 
				ifFalse:[  delta := delta negated. ].
			deltaLoc := pointer framePointer row2 * delta.
			(pointer frame: self pickRightCylinderFrom: (Vector3 x: deltaLoc x y: deltaLoc y z: len)
				to: (Vector3 x: deltaLoc x y: deltaLoc y z: len negated) radius: radius)
				ifFalse:[lastSpin := nil. ^ false.].
			spto := pointer selectedPoint copy- deltaLoc.
			spto x: spto x negated
				y: spto y negated
				z: 0.
			]
	ifTrue: [spto := pointer selectedPoint copy.].

	rotPoint ifNil:[
		rotPoint := spto. 
		(rotPoint - boxPoint) length >0.01 ifTrue:[rotateDirection := -1.0.] ifFalse:[rotateDirection := 1.0.].
	].
	spfrom := rotPoint copy.
	spfrom z: 0.0.
	spfrom normalize.
	spto z: 0.0.
	spto normalize.
	theta := ((spfrom x * spto y) - (spto x * spfrom y))"/selectedRadiusSquared".
	theta < -1.0 ifTrue:[theta:= -1.0].
	theta > 1.0 ifTrue:[theta := 1.0].
	theta := theta arcSin radiansToDegrees/2.0.
	lastSpin := Matrix4x4 identity rotationAroundZ: theta.

	trans := self translation.
	self translationX: 0.0 y:0.0 z:0.0.
	self localTransform: (self localTransform composeWith: lastSpin).
	self translation: trans.
	^ true.
! !

!TScrollBox3D methodsFor: 'actions' stamp: 'das 3/21/2006 15:03'!
scaleBox: pointer

	| mn fc sp mnBase scl |
	sp := pointer selectedPoint abs.	
	selectedPoint := selectedPoint abs.
	mn := sp x min: sp y.
	mnBase := selectedPoint x min: selectedPoint y.
	scl := mn/mnBase.
	newScale := scl * newScale.
	newScale < 0.1 ifTrue:[newScale := 0.1].
	selectedPoint := sp.
	fc := selectedFrame frameChildren.
	1 to: 8 do:[:index | (fc at: index) translation: (basePosition at: index) *newScale.].
	deferScale ifFalse:[contents scale: baseScale *newScale].
	^ true.
! !

!TScrollBox3D methodsFor: 'accessing' stamp: 'bgf 3/26/2006 17:09'!
sideScale: factor

	"Set the scaling factor applied to the control handles.
      10.0 is the default value; larger values make smaller handles."

	factor <= 0 ifTrue: [
		self error: 'TScrollBox3D control scale must be positive.'.
	] ifFalse: [
		sideScale := factor.
	].! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'das 7/31/2005 16:44'!
target

	^ contents.! !

!TScrollBox3D methodsFor: 'initialize' stamp: 'das 8/18/2006 22:20'!
updateBox

	| mn mx corner extent center delta |
	self resetBox.
	box := self getBox: contents.
	mn := box min.
	mx := box max.

	corner := (mx-mn).
	extent := (box extent x + box extent y + box extent z)/2.5/sideScale.
"	extent := box extentMax/scale."

	corner := 1.01*(corner-extent)/2.0.
	corner := corner max: extent@extent@extent.
	mx := mx max: (box center + ((extent@extent@extent)*1.5)).
	mn := mn min: (box center -((extent@extent@extent)*1.5)).
	delta := 1.01*(mx-mn/2.0).
	center := box center+(box center negated).
	(sides at: 1) translation:  center+(0@delta y@0).
	(sides at: 2) translation:  center-(0@delta y@0).
	(sides at: 3) translation:  center-(delta x@0@0).
	(sides at: 4) translation:  center+(delta x@0@0).
	(sides at: 5) translation:  center-(0@0@delta z).
	(sides at: 6) translation:  center+(0@0@delta z).

{ 	corner x@corner z@0. 
	corner x@corner z@0. 
	corner z@corner y@0. 
	corner z@corner y@0.
	corner x@corner y@0.
	corner x@corner y@0. }
	doWithIndex:[ :crnr : index| | group |
		group := sides at: index.
		group frameChildren do:[ :fc | fc extent: extent@extent.].
		(group frameChildren at: 1) translation: (crnr x@crnr y@0).
		(group frameChildren at: 2) translation: (crnr x@crnr y negated@0).
		(group frameChildren at: 3) translation: (crnr x negated@crnr y@0).
		(group frameChildren at: 4) translation: (crnr x negated@crnr y negated@0).

		(group frameChildren at: 5) translation: (crnr x @ 0 @ 0).
		(group frameChildren at: 6) translation: (0 @ crnr y @ 0).
		(group frameChildren at: 7) translation: (crnr x negated@ 0 @ 0).
		(group frameChildren at: 8) translation: (0 @ crnr y negated @ 0).
		].
! !

TGroup subclass: #TSkyBox
	instanceVariableNames: 'dalpha'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TSkyBox commentStamp: 'das 10/10/2004 14:08' prior: 0!
TSkyBox is used to render a slowly rotating sky in the world. It is actually a collection of five fullbright textures on a cube (except for the bottom).

dalpha - rotational distance every step.

DAS!

!TSkyBox methodsFor: 'initialize' stamp: 'ar 2/7/2006 17:05'!
initializeWithFileName: fname

	|  txtr mat scale shrink sc dy |
	sc := 10.
	scale :=128 * sc.
	dy := 0.
	shrink := 8 * sc.
	mat := TMaterial new.
	mat fullBright: true.
	mat ambientColor: #(1.0 1.0 1.0 1.0).
	mat diffuseColor: #(1.0 1.0 1.0 1.0).
	mat emissiveColor: #(1.0 1.0 1.0 1.0).

	txtr := TTexture
				new initializeWithFileName: 'SkyBox', fname, 'BK.bmp'
				mipmap: false
				shrinkFit: false.
	txtr extent: scale@scale.
	txtr translationX: 0.0 y:dy z: (scale-shrink) negated/2.0.
	txtr material: mat.
	self addChild: txtr.

	txtr := TTexture
				new initializeWithFileName: 'SkyBox', fname, 'LF.bmp'
				mipmap: false
				shrinkFit: false.
	txtr extent: scale@scale.
	txtr rotationAroundY: 90.
	txtr translationX: (scale-shrink)/2.0 negated y:dy z: 0.0.
	txtr material: mat.
	self addChild: txtr.

	txtr := TTexture
				new initializeWithFileName: 'SkyBox', fname, 'RT.bmp'
				mipmap: false
				shrinkFit: false.
	txtr extent: scale@scale.
	txtr rotationAroundY: -90.
	txtr translationX: (scale-shrink)/2.0 y:dy z: 0.0.
	txtr material: mat.
	self addChild: txtr.

	txtr := TTexture
				new initializeWithFileName: 'SkyBox', fname, 'FR.bmp'
				mipmap: false
				shrinkFit: false.
	txtr extent: scale@scale.
	txtr rotationAroundY: 180.
	txtr translationX: 0.0 y:dy z: (scale-shrink)/2.
	txtr material: mat.
	self addChild: txtr.


	txtr := TTexture
				new initializeWithFileName: 'SkyBox', fname, 'UP.bmp'
				mipmap: false
				shrinkFit: false.
	txtr extent: scale@scale.
	txtr rotationAroundX: 90.
	txtr translationX: 0.0 y: ((scale-shrink)/2) + (dy)z: 0.0.
	txtr material: mat.
	self addChild: txtr.
	self solidTree: false.
	dalpha := 0.02.
! !

!TSkyBox methodsFor: 'stepping' stamp: 'ar 1/15/2006 17:56'!
step
	"rotation is dalpha units per 20 milliseconds, and we assume start at zero clock time" 
	self rotationAroundY: (IslandTime millisecondClockValue asFloat/20.0) * dalpha.
	(self future: 20) step! !

TGroup subclass: #TSpinner
	instanceVariableNames: 'contents spin lastSpin matNorm matOver matDown selectedPoint selectedFrame tinyCube stepping cameraNorm selectedRadiusSquared'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

TSpinner subclass: #TDragger
	instanceVariableNames: 'test2'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TDragger methodsFor: 'initialize' stamp: 'das 1/31/2006 23:07'!
initMaterial! !

!TDragger methodsFor: 'events' stamp: 'das 1/31/2006 23:04'!
pointerEnter: event! !

!TDragger methodsFor: 'events' stamp: 'das 1/31/2006 23:04'!
pointerLeave: event! !

!TDragger methodsFor: 'events' stamp: 'das 1/31/2006 22:26'!
pointerMove: event

	event shiftPressed ifTrue:[^ self doSpin:event.].
	self doDrag: event.

! !

!TDragger methodsFor: 'events' stamp: 'ar 1/15/2006 16:51'!
pointerOver: event! !

!TSpinner methodsFor: 'accessing' stamp: 'ar 1/15/2006 16:48'!
contents

	^ contents.! !

!TSpinner methodsFor: 'accessing' stamp: 'ar 1/16/2006 19:20'!
contents: cnt

	contents ifNotNil:[
		self removeChild: contents.].
	contents := cnt.
	matNorm ifNotNil:[ contents material: matNorm. ].
	self addChild: contents.! !

!TSpinner methodsFor: 'actions' stamp: 'das 1/31/2006 22:07'!
doDrag: event

	| pointer delta |
	pointer := event makePointer.
	(pointer frame: self pickPlane: selectedPoint normal: cameraNorm) ifTrue:[
		delta := selectedPoint - pointer selectedPoint.
		self translation: (self translation - (self orientation localPointToGlobal: delta)).
	]. 
! !

!TSpinner methodsFor: 'actions' stamp: 'das 2/3/2006 21:47'!
doSpin: event

	| pointer |
	pointer := event makePointer.
	pointer frame: self pickSphere: Vector3 new radiusSquared: selectedRadiusSquared.
	pointer selectedPoint ifNotNil:[
		lastSpin := (self rotFromBallPoints: selectedPoint to: pointer selectedPoint) asMatrix4x4.
	].
	lastSpin ifNotNil:[self orientation: (self localTransform composeWith: lastSpin).].

! !

!TSpinner methodsFor: 'events' stamp: 'ar 1/27/2006 14:22'!
eventMask

	^ EventPointerOver bitOr: EventPointerDown.! !

!TSpinner methodsFor: 'initialize' stamp: 'ar 2/7/2006 17:00'!
initMaterial

	| |
	matNorm := TMaterial new.
	matNorm ambientColor: #(0.5 0.7 0.7 0.5).
	matNorm diffuseColor: #(0.5 0.7 0.7 0.5).

	matOver := TMaterial new.
	matOver ambientColor: #(0.2 0.5 1.0 0.8).
	matOver diffuseColor: #(0.2 0.5 1.0 0.8).

	matDown := TMaterial new.
	matDown ambientColor: #(0.2 1.0 0.5 1.0).
	matDown diffuseColor: #(0.2 1.0 0.5 1.0).

! !

!TSpinner methodsFor: 'initialize' stamp: 'das 7/5/2006 13:20'!
initialize
	| matTiny |

	super initialize.
	spin := 0.
	stepping := false.
	self initMaterial.

	matTiny := TMaterial new.
	matTiny ambientColor: #(1.0 1.0 0.2 1.0) .
	matTiny diffuseColor: #(1.0 1.0 0.2 1.0).

	tinyCube := TCube new.
	tinyCube extentX: 0.05 y: 0.05 z: 0.05.
	tinyCube visible: false.
	tinyCube material: matTiny.
	self addChild: tinyCube.
	^self
! !

!TSpinner methodsFor: 'accessing' stamp: ''!
matDown
	
	^ matDown.! !

!TSpinner methodsFor: 'accessing' stamp: ''!
matDown: md
	
	matDown:= md.! !

!TSpinner methodsFor: 'accessing' stamp: ''!
matNil

	matOver := nil.
	matDown := nil.
	matNorm := nil.! !

!TSpinner methodsFor: 'accessing' stamp: ''!
matNorm
	
	^ matNorm.! !

!TSpinner methodsFor: 'accessing' stamp: ''!
matNorm: mn
	
	matNorm := mn.! !

!TSpinner methodsFor: 'accessing' stamp: ''!
matOver
	
	^ matOver.! !

!TSpinner methodsFor: 'accessing' stamp: ''!
matOver: mo
	
	matOver:= mo.! !

!TSpinner methodsFor: 'events' stamp: 'das 1/31/2006 22:05'!
pointerDown: event

	selectedPoint := event selection point.
	tinyCube translation: selectedPoint.
	tinyCube visible: true.

	matDown ifNotNil:[
		contents material: matDown.
		].
	cameraNorm := self inverseGlobalOrientation localPointToGlobal: 
		event selection cameraTransform lookAt negated.
	selectedRadiusSquared := selectedPoint dot: selectedPoint.
	self spin: nil.
! !

!TSpinner methodsFor: 'events' stamp: 'ar 1/27/2006 14:39'!
pointerEnter: event

	matOver ifNotNil:[
		contents material: matOver.].! !

!TSpinner methodsFor: 'events' stamp: 'ar 1/27/2006 14:40'!
pointerLeave: event
	matNorm ifNotNil:[
		contents material: matNorm.].! !

!TSpinner methodsFor: 'events' stamp: 'das 1/31/2006 22:09'!
pointerMove: event

	event shiftPressed ifTrue:[^ self doDrag:event.].
	self doSpin: event.

! !

!TSpinner methodsFor: 'events' stamp: 'ar 1/15/2006 16:39'!
pointerOver: event
! !

!TSpinner methodsFor: 'events' stamp: 'ar 1/27/2006 14:38'!
pointerUp: event

	matNorm ifNotNil:[contents material: matNorm.].
	tinyCube visible: false.
	lastSpin ifNotNil:[
		(lastSpin squaredErrorDistanceTo: Matrix4x4 identity)<0.001 ifTrue:[lastSpin := nil].
	].
	self spin: lastSpin.
	lastSpin :=  nil.! !

!TSpinner methodsFor: 'accessing' stamp: ''!
spin
	^spin! !

!TSpinner methodsFor: 'accessing' stamp: 'ar 1/15/2006 16:42'!
spin: aNumber
	spin := aNumber.
	stepping ifFalse:[
		stepping := true.
		self step.
	].! !

!TSpinner methodsFor: 'stepping' stamp: 'das 1/26/2006 09:57'!
step
	| trans transform |
	spin ifNil:["stop stepping"
		stepping := false.
		^self].
	trans := self translation.
	self translationX: 0.0 y:0.0 z:0.0.
	transform := self localTransform.
	transform := transform composeWith: spin.
	100 atRandom = 1 ifTrue:[transform normalize]. "renormalize once in a while"
	self localTransform: transform.
	self translation: trans.
	(self future: 50) step.! !

TFrame subclass: #TMaterial
	instanceVariableNames: 'ambientColor diffuseColor specularColor emissiveColor shininess transparency renderObject cullFace flipFace fullBright texture allowWireFrame saveWire subMaterialList scale boundSphere'
	classVariableNames: 'DefaultAmbient DefaultDiffuse DefaultEmissive DefaultShininess DefaultSpecular DefaultWhite OGLDefaultAmbient OGLDefaultDiffuse OGLDefaultEmissive OGLDefaultShininess OGLDefaultSpecular'
	poolDictionaries: 'OpenGLConstants'
	category: 'Croquet-Objects'!

!TMaterial commentStamp: 'das 6/7/2006 23:14' prior: 0!
TMaterial is used to maintain and set material properties of objects. It can also be rendered itself to aid the design of the material. It displays the canonical teapot when used this way.

ambientColor - 
diffuseColor -
specularColor -
emissiveColor -
shininess - 
transparency -
renderObject - object to use if we use the TMaterial in a render hierarchy
cullFace - determine if we should cull the back faces of the object being rendered.
flipFace - determine if we should just render the back faces or not
fullBright - ignore lighting, just use full bright lighting
texture - texture to be used as part of the material
textureMode - the OpenGL GlTexture mode.
subMaterialList - list of materials, typically used by the TMeshes.
uvScale - scaling of the TTexture
scale - scale of the material when used as a rendered object
boundSphere - bound of the rendered object

DAS
!

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/7/2006 23:11'!
defaultAmbientColor

	^ DefaultAmbient copy.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/7/2006 23:11'!
defaultDiffuseColor

	^ DefaultDiffuse copy.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/8/2006 16:37'!
defaultEmissiveColor

	^ DefaultEmissive copy.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/8/2006 16:38'!
defaultOGLAmbientColor

	^ OGLDefaultAmbient copy.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/8/2006 16:38'!
defaultOGLDiffuseColor

	^ OGLDefaultDiffuse copy.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/8/2006 16:38'!
defaultOGLEmissiveColor

	^ OGLDefaultEmissive copy.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/8/2006 16:38'!
defaultOGLShininess

	^ OGLDefaultShininess.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/8/2006 16:38'!
defaultOGLSpecularColor

	^ OGLDefaultSpecular copy.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/7/2006 23:12'!
defaultShininess

	^ DefaultShininess.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/7/2006 23:12'!
defaultSpecularColor

	^ DefaultSpecular copy.! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/8/2006 16:37'!
enableDefault:ogl

	| trans |
	ogl transparency = 1.0 ifTrue:[
		ogl glMaterialfv: GLFront with: GLSpecular with: DefaultSpecular;
			glMaterialfv: GLFront with: GLDiffuse with: DefaultDiffuse;
			glMaterialfv: GLFront with: GLAmbient with: DefaultAmbient;
			glMaterialfv: GLFront with: GLEmission with: DefaultEmissive;
			glMaterialfv: GLFront with: GLShininess with: DefaultShininess.
	]ifFalse:[
		trans := #(1 1 1 1)  asFloatArray.
		trans at: 4 put: ogl transparency.
		ogl glMaterialfv: GLFront with: GLSpecular with: DefaultSpecular*trans;
			glMaterialfv: GLFront with: GLDiffuse with: DefaultDiffuse*trans;
			glMaterialfv: GLFront with: GLAmbient with: DefaultAmbient*trans;
			glMaterialfv: GLFront with: GLEmission with: DefaultEmissive*trans;
			glMaterialfv: GLFront with: GLShininess with: DefaultShininess.
	].
! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/8/2006 16:39'!
enableOGLDefault:ogl

	| trans |
	ogl transparency = 1.0 ifTrue:[
		ogl glMaterialfv: GLFront with: GLSpecular with: OGLDefaultSpecular;
			glMaterialfv: GLFront with: GLDiffuse with: OGLDefaultDiffuse;
			glMaterialfv: GLFront with: GLAmbient with: OGLDefaultAmbient;
			glMaterialfv: GLFront with: GLEmission with: OGLDefaultEmissive;
			glMaterialfv: GLFront with: GLShininess with: OGLDefaultShininess.
	]ifFalse:[
		trans := #(1 1 1 1)  asFloatArray.
		trans at: 4 put: ogl transparency.
		ogl glMaterialfv: GLFront with: GLSpecular with: OGLDefaultSpecular*trans;
			glMaterialfv: GLFront with: GLDiffuse with: OGLDefaultDiffuse*trans;
			glMaterialfv: GLFront with: GLAmbient with: OGLDefaultAmbient*trans;
			glMaterialfv: GLFront with: GLEmission with: OGLDefaultEmissive*trans;
			glMaterialfv: GLFront with: GLShininess with: OGLDefaultShininess.
	].
! !

!TMaterial class methodsFor: 'class initialization' stamp: 'das 6/8/2006 16:46'!
initialize
	"TMaterial initialize"
	DefaultAmbient := #(0.5 0.5 0.5 1.0) asVectorColor.
	DefaultDiffuse := #(0.5 0.5 0.5 1.0) asVectorColor.
	DefaultSpecular := #(0.0 0.0 0.0 1.0) asVectorColor.
	DefaultEmissive :=  #(0.0 0.0 0.0 1.0) asVectorColor.
	DefaultShininess := 0.0.
	DefaultWhite := #(1 1 1)  asVectorColor.

	OGLDefaultAmbient := #(0.2 0.2 0.2 1.0) asVectorColor.
	OGLDefaultDiffuse := #(0.8 0.8 0.8 1.0) asVectorColor.
	OGLDefaultSpecular := #(0.0 0.0 0.0 1.0) asVectorColor.
	OGLDefaultEmissive :=  #(0.0 0.0 0.0 1.0) asVectorColor.
	OGLDefaultShininess := 0.0.
! !

!TMaterial methodsFor: 'accessing' stamp: 'das 1/25/2005 14:42'!
allowWireFrame: bool

	allowWireFrame := bool.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
ambientColor

	^ ambientColor.! !

!TMaterial methodsFor: 'accessing' stamp: 'das 7/10/2006 15:26'!
ambientColor: ac

	| acol |
	acol := ac.
	acol class = Color ifTrue:[acol := acol asVectorColor].
	acol class = TranslucentColor ifTrue:[ acol := acol asVectorColor].
	acol size = 3 ifTrue:[acol := acol@ambientColor alpha].
	ambientColor := acol asVectorColor.	! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 1/19/2006 10:33'!
boundSphere

	boundSphere ifNil:[
		boundSphere := TBoundSphere localPosition: 0@0@0 radius: scale.
		boundSphere frame: self.
	].
	^ boundSphere.! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 2/7/2006 17:35'!
color: aColor
	ambientColor := aColor  asVectorColor.
	diffuseColor := ambientColor copy.
! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 8/10/2006 18:38'!
colorize: col
	"Preserve alpha when colorizing material"
	| ac dc |
	ac := col asVectorColor copy.
	ac alpha: ambientColor alpha.
	dc := col asVectorColor copy.
	dc alpha: diffuseColor alpha.
	self ambientColor: ac.
	self diffuseColor: dc.
! !

!TMaterial methodsFor: 'accessing' stamp: ''!
cullFace

	^ cullFace.! !

!TMaterial methodsFor: 'accessing' stamp: 'das 7/22/2003 16:08'!
cullFace: bool

	cullFace := bool.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
diffuseColor

	^ diffuseColor.! !

!TMaterial methodsFor: 'accessing' stamp: 'das 7/10/2006 15:25'!
diffuseColor: dc

	| dcol |
	dcol := dc.
	dcol class = Color ifTrue:[ dcol := dcol asVectorColor].
	dcol class = TranslucentColor ifTrue:[ dcol := dcol asVectorColor].
	dcol size = 3 ifTrue:[dcol := dcol@diffuseColor alpha].
	diffuseColor := dcol asVectorColor.! !

!TMaterial methodsFor: 'render' stamp: 'ar 9/9/2006 16:51'!
disable: ogl

	ogl forceHilite ifNotNil: [
		ogl glMaterialfv: GLFront with: GLAmbient with: DefaultAmbient.
		ogl glMaterialfv: GLFront with: GLDiffuse with: DefaultDiffuse.
		ogl glMaterialfv: GLFront with: GLEmission with:  DefaultEmissive.
		"(Specular was already cleared to default, by the enable: path.)"
	] ifNil: [
		fullBright ifTrue:[
			ogl glEnable: GLLighting.
			ogl popFog.
		] ifFalse:[
		"	ogl glPopAttrib."
			ogl glMaterialfv: GLFront with: GLSpecular with: DefaultSpecular.
			ogl glMaterialfv: GLFront with: GLDiffuse with: DefaultDiffuse.
			ogl glMaterialfv: GLFront with: GLAmbient with: DefaultAmbient.
			ogl glMaterialfv: GLFront with: GLEmission with:  DefaultEmissive.
			ogl glMaterialfv: GLFront with: GLShininess with: DefaultShininess.
		].
	].

	texture ifNotNil:[texture disable: ogl].

	flipFace ifTrue:[ ogl flipFace.].
	ogl setCull.
	allowWireFrame ifFalse:[
		ogl forceWire: saveWire.
	].
! !

!TMaterial methodsFor: 'accessing' stamp: ''!
emissiveColor

	^ emissiveColor.! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 2/7/2006 17:35'!
emissiveColor: ec

	emissiveColor := ec asVectorColor.! !

!TMaterial methodsFor: 'render' stamp: 'ar 9/9/2006 16:49'!
enable: ogl

	| trans transx hilite |

	hilite := ogl forceHilite.
	hilite ifNotNil: [
		"This is not a 'hot' code path since highlighting is a special situation,
	       so we can take enough care to create a nice appearance."
		transx := transparency * ogl transparency.
		ogl glMaterialfv: GLFront with: GLDiffuse with: ((diffuseColor+hilite)/2 alpha:transx).
		ogl glMaterialfv: GLFront with: GLAmbient with: ((ambientColor+hilite)/2 alpha:transx).
		ogl glMaterialfv: GLFront with: GLEmission with: ((diffuseColor+hilite)/2 alpha:transx).  "Glow."
		ogl glMaterialfv: GLFront with: GLSpecular with: DefaultSpecular. "No need to clear this after."
	] ifNil: [
		fullBright ifTrue:[
			ogl glDisable: GLLighting.
			ogl glColor4fv: diffuseColor.
			ogl pushFog.
			"ogl glColor3fv: DefaultWhite."
		] ifFalse:[
			transx := transparency * ogl transparency.
		"	ogl glPushAttrib: GLCurrentBit."
			transx = 1.0 ifTrue:[
				ogl glMaterialfv: GLFront with: GLSpecular with: specularColor;
					glMaterialfv: GLFront with: GLDiffuse with: diffuseColor;
					glMaterialfv: GLFront with: GLAmbient with: ambientColor;
					glMaterialfv: GLFront with: GLEmission with: emissiveColor;
					glMaterialfv: GLFront with: GLShininess with: shininess.
			] ifFalse:[
				trans := #(1 1 1 1)  asFloatArray.
				trans at: 4 put: transx.
				ogl glMaterialfv: GLFront with: GLSpecular with: specularColor*trans;
					glMaterialfv: GLFront with: GLDiffuse with: diffuseColor*trans;
					glMaterialfv: GLFront with: GLAmbient with: ambientColor*trans;
					glMaterialfv: GLFront with: GLEmission with: emissiveColor*trans;
					glMaterialfv: GLFront with: GLShininess with: shininess.
			].
		].
	].
	texture ifNotNil:[texture enable: ogl].
	cullFace ifFalse:[ ogl glDisable: GLCullFace.].
	flipFace ifTrue:[ ogl flipFace. ].
	allowWireFrame ifFalse:[
		saveWire := ogl forceWire.
		ogl forceWire: false.
	].! !

!TMaterial methodsFor: 'accessing' stamp: ''!
flipFace

	^ flipFace.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
flipFace: bool

	flipFace := bool.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
fullBright

	^ fullBright.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
fullBright: bool

	fullBright := bool.! !

!TMaterial methodsFor: 'accessing' stamp: 'das 12/11/2004 12:39'!
hasAlpha

	(diffuseColor at: 4) < 1.0 ifTrue: [^ true].
	transparency < 1.0 ifTrue:[^ true].
	^ false.! !

!TMaterial methodsFor: 'initialize' stamp: 'ar 9/9/2006 16:50'!
initialize
	super initialize.	
	self ambientColor:  DefaultAmbient copy.
	self diffuseColor:  DefaultDiffuse copy.
	self specularColor: DefaultSpecular copy.
	self emissiveColor: DefaultEmissive copy.
	shininess := DefaultShininess.
	cullFace := true.
	flipFace := false.
	fullBright := false.
	visible := true.
	scale := 1.0.
	transparency := 1.0.
	allowWireFrame := true.
	^self
! !

!TMaterial methodsFor: 'initialize' stamp: 'das 6/8/2006 16:41'!
oglDefault

	self ambientColor:  OGLDefaultAmbient copy.
	self diffuseColor:  OGLDefaultDiffuse copy.
	self specularColor: OGLDefaultSpecular copy.
	self emissiveColor: OGLDefaultEmissive copy.
	shininess := OGLDefaultShininess.
! !

!TMaterial methodsFor: 'copying' stamp: 'DPR 1/9/2004 14:05'!
postCopy

	super postCopy.
	ambientColor := ambientColor copy.
	diffuseColor := diffuseColor copy.
	specularColor := specularColor copy.
	emissiveColor := emissiveColor copy.
	subMaterialList ifNotNil:[
		subMaterialList := subMaterialList collect:[:mat| mat copy].
	].
	^self! !

!TMaterial methodsFor: 'printing' stamp: 'das 7/14/2006 15:55'!
printOn: aStream 

	super printOn: aStream. 
	texture ifNotNil:[
		aStream nextPut: $[. 
		texture printOn: aStream.
		aStream nextPut: $].
	].
	subMaterialList ifNotNil:[ 
		aStream nextPut:$[.
		subMaterialList printOn: aStream.
		aStream nextPut: $].
	].
! !

!TMaterial methodsFor: 'initialize' stamp: 'oxe 12/10/2004 11:52'!
randomNice

	self randomNiceWithFloor: 0.5.! !

!TMaterial methodsFor: 'initialize' stamp: 'ar 2/7/2006 17:04'!
randomNiceWithFloor: floor

	| prg r g b |
	prg := Random new.


	self shininess: 1000000.0.
	self ambientColor:  #(0.1 0.1 0.1 1.0).
	self emissiveColor: #(0.0 0.0 0.0 1.0).

	r := prg next * (1.0 - floor) + floor.
	g := prg next * (1.0 - floor) + floor.
	b := prg next * (1.0 - floor) + floor.
	self diffuseColor:  {r.   g.   b.   1.0}.

	r := 1.0 - r.
	g := 1.0 - g.
	b := 1.0 - b.
	self specularColor: {r.   g.   b.   1.0}.
! !

!TMaterial methodsFor: 'render' stamp: 'das 7/5/2006 13:20'!
render: ogl

	self hasAlpha ifFalse:[
		ogl glPushMatrix.

		renderObject ifNil: [ renderObject := TCube new. renderObject scale: scale.].
		self enable: ogl.
		renderObject render: ogl.
		ogl glPopMatrix.].! !

!TMaterial methodsFor: 'render' stamp: 'das 7/5/2006 13:20'!
renderAlpha: ogl

	ogl glPushMatrix.

	renderObject ifNil: [ renderObject := TCube new. ].
	self enable: ogl.
	renderObject render: ogl.
	self disable: ogl.
	ogl glPopMatrix.! !

!TMaterial methodsFor: 'accessing' stamp: 'das 1/16/2006 08:32'!
scale: scl

	super scale: scl.
	scale := scl.
	boundSphere ifNotNil:[
		boundSphere := TBoundSphere localPosition:0@0@0 radius: scale.
	].! !

!TMaterial methodsFor: 'accessing' stamp: ''!
shininess
	
	^ shininess.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
shininess: s

	shininess := s.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
specularColor

	^ specularColor.! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 2/7/2006 17:35'!
specularColor: sc

	specularColor := sc asVectorColor.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
subMaterialList

	^ subMaterialList.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
subMaterialList: sml

	subMaterialList := sml.! !

!TMaterial methodsFor: 'accessing' stamp: ''!
texture

	^ texture.! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:44'!
texture: txtr

	texture := txtr.! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:50'!
textureMode
	"OBSOLETE. Remove this."
	^texture ifNil:[GLModulate] ifNotNil:[texture textureMode].! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:50'!
textureMode: tm
	texture ifNotNil:[texture textureMode: tm]! !

!TMaterial methodsFor: 'accessing' stamp: 'das 12/20/2002 12:22'!
transparency

	^ transparency.! !

!TMaterial methodsFor: 'accessing' stamp: 'das 12/20/2002 12:22'!
transparency: trans

	transparency := trans.! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:46'!
uvAngle
	"OBSOLETE. Remove this."
	^texture ifNil:[0.0] ifNotNil:[texture uvAngle]! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:46'!
uvAngle: uva
	"OBSOLETE. Remove this."
	texture ifNotNil:[texture uvAngle: uva]! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:46'!
uvOffset
	"OBSOLETE. Remove this."
	^texture ifNil:[0.0@0.0] ifNotNil:[texture uvOffset]! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:47'!
uvOffset: uvo
	"OBSOLETE. Remove this."
	texture ifNotNil:[texture uvOffset: uvo].! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:48'!
uvScale 
	"OBSOLETE. Remove this."
	^texture ifNil:[1.0@1.0] ifNotNil:[texture uvScale]! !

!TMaterial methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:48'!
uvScale: uvs
	"OBSOLETE. Remove this."
	texture ifNotNil:[texture uvScale: uvs]! !

TFrame subclass: #TMesh
	instanceVariableNames: 'materialList alias faceGroups faceBuffers vertexBuffer normalBuffer colorBuffer texCoordBuffer indexBuffer alpha opaque boundsChanged boundSphere boundsDepth boundMaterial mode rigidBody'
	classVariableNames: ''
	poolDictionaries: 'OpenGLConstants'
	category: 'Croquet-Objects'!

!TMesh commentStamp: 'das 10/10/2004 12:47' prior: 0!
TMesh
Almost all simple arrays are 0 based when indexed by other arrays. Not the materialList.

materialList - this is an OrderedCollection of materials that are used in the mesh. This is what the first element of the faceGroups refers to. This must always be a list.
alias - you can ignore this for now. It is important when we start doing mesh based transforms, because it tells you where the aliased vertices are.
vertices - 0 index based list of 3D vertices.
vtxNormals - the normal 3D vector for the given vertex.
alpha - checks the materials referenced by the facegroups for alpha values. Calculated for you based upon the materialList.
opaque  - same as alpha, but inverted.
textureUV - the 2D u,v coordinates of the texture at the associated vertex.
faceGroups - an array of materialList index (indexing starting at 1), and vertex face index arrays (index starting at 0). These are simply interleaved. 
boundsChanged - calculated for you at construction time. 
boundSphere - calculated for you.
boundsDepth - part of construction and used to determine the depth of the hierarchy.
boundMaterial - only used if you need to see the bound spheres or bound cubes.

These are the analogs to the equivalent TPrimitives. There are two sets because a TMesh may have some materials that have alpha components and some that don't. These are rendered at different times, so they need separate display lists.

materialList - the list of materials used by the different face sets
vertices - the vertices of the mesh
alias - aliased vertex sets
vtxNormals - normals at each vertex
textureUV - the texture uv location at each vertex
faceGroups - collection of face groups (a face group is an ordered list of indices into the vertex arrays)
alpha - flag indicates if this mesh has any alpha values
opaque - flag indicates if mesh has no alpha values
boundsChanged - indicates the boundaries of the mesh have changed
boundSphere - the bounding sphere of the mesh, typically a hierarchy
boundsDepth - depth of the bound sphere hierarchy
boundMaterial - material to render the bound spheres with
cachingEnabled - currently not used, but specifies the OpenGL caching
cachingAlphaEnabled - ""

DAS!

!TMesh class methodsFor: 'instance creation' stamp: 'sn 2/24/2018 16:24'!
initializeWithVertices: v alias: a norms: n textureUV: tuv faceGroups: fg material: ml 

	"self deprecatedExplanation: 'Use #triMeshWithVertices:... or #quadMeshWithVertices:...'."
	^ self 
		triMeshWithVertices: v 
		alias: a 
		norms: n 
		textureUV: tuv 
		faceGroups: fg 
		material: ml ! !

!TMesh class methodsFor: 'instance creation' stamp: 'jcg 6/4/2006 00:13'!
pointCloudWithVertices: v alias: a norms: n textureUV: tuv faceGroups: fg material: ml 
	
	^ super new 
			initializeWithVertices: v 
			alias: a 
			norms: n 
			textureUV: tuv 
			faceGroups: fg 
			material: ml 
			mode: GLPoints 


	! !

!TMesh class methodsFor: 'instance creation' stamp: 'jcg 6/4/2006 00:13'!
quadMeshWithVertices: v alias: a norms: n textureUV: tuv faceGroups: fg material: ml 
	
	^ super new 
			initializeWithVertices: v
			alias: a 
			norms: n 
			textureUV: tuv 
			faceGroups: fg 
			material: ml 
			mode: GLQuads 


	! !

!TMesh class methodsFor: 'instance creation' stamp: 'jcg 6/4/2006 00:14'!
triMeshWithVertices: v alias: a norms: n textureUV: tuv faceGroups: fg material: ml 
	
	^ super new 
			initializeWithVertices: v 
			alias: a 
			norms: n 
			textureUV: tuv 
			faceGroups: fg 
			material: ml 
			mode: GLTriangles 


	! !

!TMesh class methodsFor: 'instance creation' stamp: 'jcg 6/4/2006 00:14'!
triStripsWithVertices: v alias: a norms: n textureUV: tuv faceGroups: fg material: ml 
	
	^ super new 
			initializeWithVertices: v 
			alias: a 
			norms: n 
			textureUV: tuv 
			faceGroups: fg 
			material: ml 
			mode: GLTriangleStrip 


	! !

!TMesh methodsFor: 'accessing' stamp: 'das 5/18/2006 12:02'!
boundMaterial: bm

	boundMaterial:= bm.
	self checkAlpha.! !

!TMesh methodsFor: 'render' stamp: 'das 7/10/2003 21:27'!
boundSphere

	boundsChanged ifTrue:[
		self initBounds. 
		boundsChanged := false.
		].
	^ boundSphere.! !

!TMesh methodsFor: 'accessing' stamp: ''!
boundsChanged
	boundsChanged := true.! !

!TMesh methodsFor: 'accessing' stamp: 'das 4/25/2002 10:12'!
boundsDepth

	^ boundsDepth.
! !

!TMesh methodsFor: 'accessing' stamp: 'das 5/11/2002 15:05'!
boundsDepth: depth

	boundsDepth := depth.
	frameChildren ifNotNil:[ frameChildren do:[ :fc | fc boundsDepth: depth].].
	self boundsChanged.
! !

!TMesh methodsFor: 'initialize' stamp: 'das 7/5/2006 14:29'!
checkAlpha

	opaque := false.
	alpha := false.
	1 to: faceGroups size by: 2 do:[:i |
		((materialList at:(faceGroups at: i)) hasAlpha) ifTrue: [ self setAlpha: true. ] ifFalse:[ self opaque: true.].
		].
	boundMaterial ifNotNil:[self setAlpha: true].! !

!TMesh methodsFor: 'accessing' stamp: 'jcg 5/23/2005 16:09'!
collapse
"Forces all frames to identity transform while retaining translation, and pushes the previous transform to the frame children. This is used primarily for meshes and their supporting groups."
	| orient verts norms |

	orient := self orientation.
	(verts := self vertices) withIndexDo: [:v :ind |
		verts at: ind put: (orient localPointToGlobal: v)].
	(norms := self normals) withIndexDo: [:n :ind |
		norms at: ind put: (orient localPointToGlobal: n)].
	super collapse
! !

!TMesh methodsFor: 'collision' stamp: 'das 5/11/2002 15:07'!
collideFloor: floor

	(self boundSphere globalPosition y)-floor > boundSphere radius ifTrue:[ ^ nil ].

	^ boundSphere collideFloor: floor transform: self globalTransform.! !

!TMesh methodsFor: 'collision' stamp: 'das 5/8/2002 12:29'!
collidePlane: normal offset: offset

	| tNormal tOffset |

" Simple test in world coordinates:"
	(boundSphere globalPosition dot: normal)-offset > boundSphere radius ifTrue:[ ^ nil ].

"Put the global coordinates of the plane into the local coordinate frame."
	tNormal := self globalOrientation orthoNormInverse localPointToGlobal: normal.
	tOffset := (tNormal dot: (self globalTransform orthoNormInverse localPointToGlobal: (normal*offset))).

	^ boundSphere collidePlane: tNormal offset: tOffset.! !

!TMesh methodsFor: 'accessing' stamp: 'ar 8/10/2006 18:39'!
colorize: col

	self materialList do:[ :ml |
		ml colorize: col.
		ml textureMode: GLModulate.
	].
	self checkAlpha.
	frameChildren ifNotNil:[frameChildren do:[ :fc | fc colorize: col].].! !

!TMesh methodsFor: 'accessing' stamp: 'ar 6/30/2006 14:49'!
colors
	^colorBuffer! !

!TMesh methodsFor: 'accessing' stamp: 'ar 6/30/2006 14:49'!
colors: colorArray
	colorBuffer := colorArray.! !

!TMesh methodsFor: 'accessing' stamp: 'das 8/27/2006 21:58'!
cullFace: bool

	super fullBright: bool.
	materialList ifNotNil:[ materialList do:[ :ml | ml cullFace: bool]].! !

!TMesh methodsFor: 'initialize' stamp: 'das 7/4/2006 10:18'!
faceGroups

	^ faceGroups.
! !

!TMesh methodsFor: 'initialize' stamp: 'ar 3/22/2006 13:37'!
faceGroups: fg

	faceGroups := fg.
	faceBuffers := ((2 to: fg size by: 2) collect: [:ind | (fg at: ind)]).! !

!TMesh methodsFor: 'accessing' stamp: 'ar 6/23/2002 02:07'!
faceGroupsDo: aBlock
	1 to: faceGroups size by: 2 do:[ :i | aBlock value: (faceGroups at: i+1)].! !

!TMesh methodsFor: 'accessing' stamp: 'sn 2/26/2018 12:32'!
forceGlobalToLocal

"This is used by TMesh where we assume that the imported vertices are pre-transformed and we want to put them back into their untransformed state. Only TMesh will probably need this, but TMeshes can have TGroups which don't know what to do."
	| invMat invOrient verts norms |

	frameChildren ifNotNil:[ frameChildren do:[ :fc | fc forceGlobalToLocal.]].
	invMat := self localTransform orthoNormInverse.
	invOrient := self orientation orthoNormInverse.
	(verts := self vertices) withIndexDo: [:v :ind |
		verts at: ind put: (invMat localPointToGlobal: v)].
	(norms := self normals) withIndexDo: [:n :ind |
		norms at: ind put: (invOrient localPointToGlobal: n)].
	frameParent ifNotNil:[
		self localTransform: (
			frameParent localTransform orthoNormInverse composeWith: self localTransform).
		].
! !

!TMesh methodsFor: 'accessing' stamp: 'das 4/18/2003 19:04'!
frameBox
	"Answer the local bounding box of this frame"
	boundsChanged ifTrue:[self initBounds].
	boundSphere ifNil:[^super frameBox].
	^boundSphere box ifNil:[super frameBox]! !

!TMesh methodsFor: 'accessing' stamp: 'das 4/15/2003 17:12'!
fullBright: bool

	super fullBright: bool.
	materialList ifNotNil:[ materialList do:[ :ml | ml fullBright: bool]].! !

!TMesh methodsFor: 'render' stamp: 'das 6/7/2006 21:32'!
hasAlpha

	^ alpha.! !

!TMesh methodsFor: 'render' stamp: 'das 6/7/2006 21:35'!
hasOpaque

	^ opaque.! !

!TMesh methodsFor: 'initialize' stamp: 'jcg 5/23/2005 16:57'!
initBounds

	boundsDepth = 0 
		ifTrue:[boundSphere := TBoundSphere mtfBall: self vertices. ]
		ifFalse:[
			faceGroups size = 2 
				ifTrue:[	
					boundSphere := TBoundSphere 
											calcTree: self vertices 
											faces: (faceGroups at: 2) 
											depth: boundsDepth. ] 
				ifFalse:[
					boundSphere := TBoundSphere mtfBall: self vertices.
					1 to: faceGroups size by: 2 do:[ :i || child |
						child := TBoundSphere 
										calcTree: self vertices 
										faces: (faceGroups at: i+1) 
										depth: boundsDepth-1.
						boundSphere addChild: child.].
			].].
	boundSphere frame: self.
	boundsChanged := false.! !

!TMesh methodsFor: 'initialize' stamp: 'ar 10/28/2006 13:37'!
initializeWithVertices: v alias: a norms: n textureUV: tuv faceGroups: fg material: ml 
	^self initializeWithVertices: v 
			alias: a 
			norms: n 
			textureUV: tuv 
			faceGroups: fg 
			material: ml 
			mode: GLTriangles 
! !

!TMesh methodsFor: 'initialize' stamp: 'das 7/12/2006 22:05'!
initializeWithVertices: v alias: a norms: n textureUV: tuv faceGroups: fg material: ml mode: md 
	| mat |
	alias := a.
	mode := md.
	self 
		vertices: v;
		normals: n;
		texCoords: tuv;
		faceGroups: fg.
	alpha := false.
	opaque := false.
	mat := ml.
	mat ifNil:[ 
		mat := TMaterial new.
		mat ambientColor: #(1.0 0.2 0.2 0.5).
		mat diffuseColor: #(1.0 0.2 0.2 0.5).
	].
	mat subMaterialList ifNotNil:[ materialList := mat subMaterialList. ]
	ifNil:[
		materialList := OrderedCollection new.
		materialList add: mat.
		1 to: faceGroups size by: 2 do:[ :i |
			faceGroups at: i put: 1. 
		].
	].

	1 to: faceGroups size by: 2 do:[:i |
		faceGroups at: i put: 1+(((faceGroups at: i)-1)\\materialList size).
		((materialList at:(faceGroups at: i)) hasAlpha) 
			ifTrue:[self setAlpha: true] 
			ifFalse:[ self setOpaque: true].
	].
	self optimizeFaceGroups.

	boundsDepth := 2. "must always be a mimimum of 1 - the top sphere, and at least one recursion."
	self boundsChanged.
	^self
! !

!TMesh methodsFor: 'testing' stamp: 'das 4/10/2003 23:15'!
isMesh

	^ true.! !

!TMesh methodsFor: 'accessing' stamp: 'ar 10/26/2002 20:19'!
material
	^materialList first! !

!TMesh methodsFor: 'accessing' stamp: 'jcg 6/11/2006 16:55'!
material: aMaterial
	materialList at: 1 put: aMaterial.
	self checkAlpha.! !

!TMesh methodsFor: 'accessing' stamp: 'ar 10/26/2002 20:19'!
materialList

	^ materialList.! !

!TMesh methodsFor: 'accessing' stamp: 'ar 10/26/2002 20:19'!
materialList: ml

	materialList := ml.
	self checkAlpha.! !

!TMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:03'!
mode

	^ mode.! !

!TMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:06'!
modeName
	mode = GLPoints ifTrue:[^ #GLPoint].
	mode = GLQuads ifTrue:[^ #GLQuads].
	mode = GLTriangles ifTrue:[^ #GLTriangles].
	mode = GLTriangleStrip ifTrue:[^ #GLTriangleStrip].! !

!TMesh methodsFor: 'accessing' stamp: 'ar 3/22/2006 13:36'!
normals
	^normalBuffer! !

!TMesh methodsFor: 'accessing' stamp: 'ar 3/22/2006 13:37'!
normals: normalArray
	normalBuffer := normalArray.! !

!TMesh methodsFor: 'accessing' stamp: 'das 7/10/2002 17:16'!
opaque

	^ opaque.! !

!TMesh methodsFor: 'accessing' stamp: 'das 7/10/2002 17:16'!
opaque: bool

	opaque := bool.! !

!TMesh methodsFor: 'initialize' stamp: 'das 7/12/2006 22:04'!
optimizeFaceGroups
	"Optimize the face groups to minimize state changes"
	| matList faces index |
	matList := IdentityDictionary new.
	1 to: faceGroups size by: 2 do:[:i|
		index := faceGroups at: i.
		faces := faceGroups at: i+1.
		matList at: index put: ((matList at: index ifAbsent:[#()]) copyWith: faces).
	].
	index := 0.
	matList keysAndValuesDo:[:mIndex :fList|
		fList do:[:fGroup|
			faceGroups at: (index := index+1) put: mIndex.
			faceGroups at: (index := index+1) put: fGroup.
		].
	].
	faceBuffers := ((2 to: faceGroups size by: 2) collect: [:ind | (faceGroups at: ind)]).! !

!TMesh methodsFor: 'accessing' stamp: 'abc 7/31/2002 17:11'!
pick: pointer

" We are here because we already picked the top level boundSphere, so we know that is true. Now what we need to test against the hierarchy, which will return a list of faces if true - otherwise, it will return nil."

	^ boundSphere pickChildren: pointer.! !

!TMesh methodsFor: 'copying' stamp: 'ar 1/19/2006 14:14'!
postCopy

	super postCopy.
	materialList ifNotNil:[
		materialList := materialList collect:[:mat| mat copy].
	].
	self vertices: self vertices copy.
	self boundsChanged.
	^ self.! !

!TMesh methodsFor: 'render' stamp: 'das 6/7/2006 21:40'!
render: ogl

	^self renderPrimitive: ogl alpha: false.
! !

!TMesh methodsFor: 'render' stamp: 'das 6/7/2006 22:23'!
renderAlpha: ogl

	self renderPrimitive: ogl alpha: true.
	boundMaterial ifNotNil:[
		boundMaterial enable:ogl.
		boundSphere renderTriangles: ogl.
		boundMaterial disable: ogl
	].
! !

!TMesh methodsFor: 'render' stamp: 'das 7/7/2006 00:09'!
renderPrimitive: ogl alpha: alphaPass
	| lastMat nextMat |
	lastMat := nil.

	ogl beginVertices: self vertices normals: self normals colors: self colors texCoords: self texCoords.

	1 to: faceGroups size by: 2 do:[ :i |
		nextMat := materialList at: (faceGroups at: i).
		nextMat hasAlpha = alphaPass ifTrue:[
			nextMat == lastMat ifFalse:[
				lastMat ifNotNil:[lastMat disable: ogl].
				nextMat enable: ogl.
				lastMat := nextMat.
			].
			ogl drawIndexed: mode faces: (faceBuffers at: i+1//2).
		].
	].
	lastMat ifNotNil:[lastMat disable: ogl].
	ogl end.

"	boundSphere renderBox: self."! !

!TMesh methodsFor: 'accessing' stamp: 'ar 7/22/2006 20:17'!
rigidBody
	"Answer my associated rigid body"
	^rigidBody! !

!TMesh methodsFor: 'accessing' stamp: 'ar 7/22/2006 20:17'!
rigidBody: aBody
	"Set my associated rigid body"
	rigidBody := aBody! !

!TMesh methodsFor: 'accessing' stamp: 'das 5/29/2006 13:21'!
scale: scale
	super scale: scale.
	scale isNumber ifTrue:[
		vertexBuffer *= scale.
		self scaleBounds: scale.
	] ifFalse:[
		1 to: vertexBuffer size do:[:i | vertexBuffer at: i put:((vertexBuffer at:i)*scale).].
		self boundsChanged.
	].! !

!TMesh methodsFor: 'accessing' stamp: 'yo 9/24/2003 11:29'!
scaleBounds: scale

	boundSphere ifNil: [self boundsChanged. ^ self].
	boundSphere scale: scale.
! !

!TMesh methodsFor: 'accessing' stamp: 'das 6/7/2006 21:20'!
setAlpha: bool

	alpha := bool.! !

!TMesh methodsFor: 'accessing' stamp: 'das 6/27/2006 13:10'!
setOpaque: bool

	opaque := bool.! !

!TMesh methodsFor: 'accessing' stamp: 'ar 3/22/2006 13:36'!
texCoords
	^texCoordBuffer! !

!TMesh methodsFor: 'accessing' stamp: 'ar 3/22/2006 13:36'!
texCoords: texCoordArray
	texCoordBuffer := texCoordArray.! !

!TMesh methodsFor: 'accessing' stamp: 'das 8/4/2004 23:17'!
transparency: trans

	self materialList do:[ :ml |
		ml transparency: trans.
		].
	self checkAlpha.
	frameChildren ifNotNil:[frameChildren do:[ :fc | fc transparency: trans].].! !

!TMesh methodsFor: 'enumerating' stamp: 'ar 7/22/2006 14:27'!
triangleFacesDo: aBlock
	"Evaluate aBlock with all the triangles defined in the receiver"
	| group fA fB fC fD |
	mode == nil ifTrue:[^nil]. "undefined"
	mode == GLPoints ifTrue:[^nil]. "point cloud"
	mode == GLLines ifTrue:[^nil]. "lines"

	1 to: faceGroups size by: 2 do:[:gIndex|
		group := faceGroups at: gIndex+1.
		mode == GLTriangles ifTrue:[
			1 to: group size by: 3 do:[:i|
				fA := group at: i.
				fB := group at: i+1.
				fC := group at: i+2.
				aBlock value: fA value: fB value: fC.
			].
		].
		mode == GLTriangleFan ifTrue:[
			group size > 0 ifTrue:[fA := group at: 1].
			3 to: group size do:[:i|
				fB := group at: i-1.
				fC := group at: i.
				aBlock value: fA value: fB value: fC.
			].
		].
		mode == GLTriangleStrip ifTrue:[
			group size > 2 ifTrue:[
				fA := group at: 1.
				fB := group at: 2.
			].
			3 to: group size do:[:i|
				fC := group at: i.
				"Ensure correct face orientation by flipping all even faces"
				i odd 
					ifTrue:[aBlock value: fA value: fB value: fC]
					ifFalse:[aBlock value: fA value: fC value: fB].
				fA := fB.
				fB := fC.
			].
		].
		mode == GLQuads ifTrue:[
			1 to: group size by: 4 do:[:i|
				fA := group at: i.
				fB := group at: i+1.
				fC := group at: i+2.
				fD := group at: i+3.
				aBlock value: fA value: fB value: fC.
				aBlock value: fA value: fC value: fD.
			].
		].
		mode == GLQuadStrip ifTrue:[
			group size > 2 ifTrue:[
				fA := group at: 1.
				fB := group at: 2.
			].
			4 to: group size by: 2 do:[:i|
				fC := group at: i-1.
				fD := group at: i.
				aBlock value: fA value: fB value: fC.
				aBlock value: fA value: fC value: fD.
				fA := fD.
				fB := fC.
			].
		].
	].! !

!TMesh methodsFor: 'enumerating' stamp: 'ar 7/22/2006 19:58'!
triangleVerticesDo: aBlock
	"Like triangleFacesDo: - just passes the vertices directly (instead of vertex index)"
	self triangleFacesDo:[:fA :fB :fC|
		aBlock value: (vertexBuffer at: 1+fA)
				value: (vertexBuffer at: 1+fB)
				value: (vertexBuffer at: 1+fC).
	].! !

!TMesh methodsFor: 'initialize' stamp: 'ar 6/30/2006 14:50'!
validateFaceGroups
	"Check to see if all face groups are valid. For debugging only."
	| vtxSize face |
	vtxSize := self vertices size.
	self normals ifNotNil:[
		self normals size = vtxSize ifFalse:[self error:'normal size mismatch'].
	].
	self texCoords ifNotNil:[
		self texCoords size = vtxSize ifFalse:[self error:'texture size mismatch'].
	].
	self colors ifNotNil:[
		self colors size = vtxSize ifFalse:[self error:'color size mismatch'].
	].
	self faceGroupsDo:[:fg|
		fg basicSize \\ 3 = 0 ifFalse:[self error:'Face group not divisable by 3'].
		1 to: fg basicSize do:[:i|
			face := fg basicAt: i.
			(face >= 0 and:[face < vtxSize]) ifFalse:[self error:'Vertex out of range'].
		].
	].! !

!TMesh methodsFor: 'accessing' stamp: 'ar 3/22/2006 13:36'!
vertices
	^ vertexBuffer! !

!TMesh methodsFor: 'accessing' stamp: 'ar 3/22/2006 13:36'!
vertices: vertexArray
	vertexBuffer := vertexArray.! !

TMesh subclass: #TTriangle
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TTriangle commentStamp: 'das 10/10/2004 14:46' prior: 0!
TTriangle is a TPrimitive used to render triangles.

This creates and renders your simple OpenGL triangle. Usually used for testing, but you have to start somewhere...

This was really the first object ever generated in Croquet - pay no attention to the TFlag and the TTeapot. They are lieing.

DAS!

!TTriangle methodsFor: 'initialize' stamp: 'jcg 6/10/2006 13:22'!
initBounds

	| d1 d2 d3 c1 c2 c3 c radius center v1 v2 v3 |

	v1 := vertexBuffer first.
	v2 := vertexBuffer second.
	v3 := vertexBuffer third.

" Graphics Gems 1 - page 22"

	d1 := (v2-v1) dot: (v2-v1).
	d2 := (v3-v2) dot: (v1- v2).
	d3 := (v1-v3) dot: (v2-v3).
	c1 := d2*d3.
	c2 := d3*d1.
	c3 := d1*d2.
	c := c1+c2+c3.

	radius := ((d1+d2)*(d2+d3)*(d3+d1)/c)sqrt * 0.5.
	center := (((c2+c3)*v1) + ((c3+c1)*v2) + ((c1+c2)*v3))/(2*c).
	boundSphere := TBoundSphere localPosition: center radius: radius.
	boundSphere frame: self.

! !

!TTriangle methodsFor: 'initialize' stamp: 'das 7/5/2006 17:04'!
initialize
	super initialize.
	mode := GLTriangles.
	vertexBuffer := Vector3Array ofSize: 3.
	normalBuffer := Vector3Array ofSize: 3.
	texCoordBuffer := Vector2Array ofSize: 3.
	texCoordBuffer 
		at: 1 put: (Vector2 u: 0.0 v: 1.0);
		at: 2 put: (Vector2 u: 0.0 v: 0.0);
		at: 3 put: (Vector2 u: 1.0 v: 0.0).		
	self faceGroups: {1. #(0 1 2) asIntegerArray}.
	self materialList: {TMaterial new}.! !

!TTriangle methodsFor: 'render' stamp: 'jcg 6/10/2006 13:18'!
pick: pointer
	^ pointer 
		pickTriangle: normalBuffer first 
		tri: vertexBuffer first 
		tri: vertexBuffer second 
		tri: vertexBuffer third.
	
		
! !

!TTriangle methodsFor: 'render' stamp: 'jcg 6/10/2006 13:42'!
v1: v1 v2: v2 v3: v3
	| normal |
	normal := (v3-v2 cross: v1-v2) normalized.
	normalBuffer atAllPut: normal.
	vertexBuffer
			at: 1 put: v1;
			at: 2 put: v2;
			at: 3 put: v3.
	self boundsChanged! !

TMesh subclass: #TUpdatingMesh
	instanceVariableNames: 'changed'
	classVariableNames: ''
	poolDictionaries: 'OpenGLConstants'
	category: 'Croquet-Objects'!

!TUpdatingMesh commentStamp: 'jcg 6/10/2006 07:57' prior: 0!
This is a mesh that updates itself updates its vertices, etc. in response to various events.  Each subclass provides a different method for generating its geometry.!

TUpdatingMesh subclass: #TCube
	instanceVariableNames: 'extent location'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TCube commentStamp: 'jcg 6/3/2006 12:39' prior: 0!
"TCube generates and renders a simple cube. It is one of the simplest objects that we have."
	-- DAS

TCube2 is a mesh-based replacement for TCube... using meshes instead of display lists will greatly simplify rendering with shaders.

Eventually we might want to have one set of geometry shared by all cubes (use the transform matrix to differentiate TCubes), but it's not worth the time right now.

'uvScale' is a hack until we start using the texture matrix.!

!TCube class methodsFor: 'instance creation' stamp: 'jcg 6/3/2006 10:47'!
initializeWithVertices: v alias: a norms: n textureUV: tuv faceGroups: fg material: ml 
	self shouldNotImplement.
	^ self new.! !

!TCube class methodsFor: 'instance creation' stamp: 'jcg 6/3/2006 10:47'!
new
	^ self newWithMaterial: TMaterial new! !

!TCube class methodsFor: 'instance creation' stamp: 'jcg 6/4/2006 23:48'!
newWithMaterial: aTMaterial

	| faces normals n texCoords |
	faces :=
		{
			1.	#(0 1 2 3) asIntegerArray.
			1.	#(4 5 6 7) asIntegerArray.
			1.	#(8 9 10 11) asIntegerArray.
			1.	#(12 13 14 15) asIntegerArray.
			1.	#(16 17 18 19) asIntegerArray.
			1.	#(20 21 22 23) asIntegerArray.
		}.

	normals := Vector3Array new: 24.
	n := 0@0@1.
	1 to: 4 do: [:i | normals at: i put: n].
	n := 0@0@-1.
	5 to: 8 do: [:i | normals at: i put: n].
	n := 1@0@0.
	9 to: 12 do: [:i | normals at: i put: n].
	n := -1@0@0.
	13 to: 16 do: [:i | normals at: i put: n].
	n := 0@1@0.
	17 to: 20 do: [:i | normals at: i put: n].
	n := 0@-1@0.
	21 to: 24 do: [:i | normals at: i put: n].

	texCoords := Vector2Array new: 24.
	1 to: 21 by: 4 do: [:i | texCoords at: i put: (Vector2 x: 1.0 y: 0.0)].
	2 to: 22 by: 4 do: [:i | texCoords at: i put: (Vector2 x: 0.0 y: 0.0)].
	3 to: 23 by: 4 do: [:i | texCoords at: i put: (Vector2 x: 0.0 y: 1.0)].
	4 to: 24 by: 4 do: [:i | texCoords at: i put: (Vector2 x: 1.0 y: 1.0)].

	^ self
		quadMeshWithVertices: (Vector3Array new: 24)
		alias: nil
		norms: normals
		textureUV: texCoords
		faceGroups: faces
		material: aTMaterial
			! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/8/2006 23:14'!
boxTop: top bottom: bottom

	self extent: (top - bottom).
	self location: (bottom + top)/2.0.
	self boundsChanged.

! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/3/2006 09:38'!
colorize: col

	| mat |
	mat := TMaterial new.
	mat ambientColor: col.
	mat diffuseColor: col.
	mat textureMode: GLModulate.

	self material: mat.
! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/3/2006 09:38'!
extent

	^ extent.
! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/8/2006 23:14'!
extent: vect3

	extent := vect3.
	self boundsChanged.
! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/8/2006 23:14'!
extentX: ex

	extent x: ex.
	self boundsChanged.
! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/8/2006 23:14'!
extentX: x y: y z: z

	extent x:x y:y z:z.
	self boundsChanged.
! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/8/2006 23:14'!
extentY: ey

	extent y: ey.
	self boundsChanged.

! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/8/2006 23:14'!
extentZ: ez

	extent z: ez.
	self boundsChanged.
! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/3/2006 10:42'!
frameBox
	"Answer the local bounding box of this frame"
	self flag: #jcg.  "Shouldn't this take 'location' into account?"

	^TBox origin: extent/-2.0 corner: extent/2.0.! !

!TCube methodsFor: 'initialize' stamp: 'jcg 6/3/2006 10:43'!
initBounds

	boundSphere := TBoundSphere localPosition: location radius: (extent length)/2.
	boundSphere frame: self.

! !

!TCube methodsFor: 'initialize' stamp: 'das 7/5/2006 14:32'!
initialize
	super initialize.
	extent := Vector3 x: 1.0 y: 1.0 z: 1.0.
	location := Vector3 x: 0.0 y:0.0 z:0.0.
	self initBounds.
	self materialList: {TMaterial new}.
! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/3/2006 10:42'!
location

	^ location.! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/3/2006 10:42'!
location: loc

	location := loc.
	self boundsChanged.

! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/3/2006 10:42'!
locationX: x y: y z: z

	location x: x y: y z: z.
	self boundsChanged.
! !

!TCube methodsFor: 'render' stamp: 'bgf 7/15/2006 03:35'!
pick: pointer

	| base |
	base := 1.
	1 to: 6 do: [:i |	
		(pointer 
			pickQuad: (normalBuffer at: base)
			q1: (vertexBuffer at: base)
			q2: (vertexBuffer at: base+1)
			q3: (vertexBuffer at: base+2)
			q4: (vertexBuffer at: base+3)) ifTrue: [
		^ true].
		base := base+4.].
	^ false

 	"*** The above is a slighty CSE-optimized version of this original: 
	1 to: 6 do: [:i |
		(pointer 
			pickQuad: (normalBuffer at: 4*(i-1)+1)
			q1: (vertexBuffer at: 4*(i-1)+1)
			q2: (vertexBuffer at: 4*(i-1)+2)
			q3: (vertexBuffer at: 4*(i-1)+3)
			q4: (vertexBuffer at: 4*(i-1)+4)) ifTrue: [^ true]].
	^ false
	*** "! !

!TCube methodsFor: 'accessing' stamp: 'jcg 6/10/2006 08:07'!
scale: scl
	extent := extent * scl.
	self translation: scl * self translation.
	self boundsChanged.
! !

!TCube methodsFor: 'initialize' stamp: 'bgf 7/13/2006 14:24'!
updateMesh

	| dx dy dz x y z tempVertices groups |

	mode := GLQuads.
	vertexBuffer := Vector3Array ofSize: 24.
	normalBuffer := Vector3Array ofSize: 24.
	texCoordBuffer := Vector2Array ofSize: 24.
	
	dx := extent x/2.0.
	dy := extent y/2.0.
	dz := extent z/2.0.
	x := location x.
	y := location y.
	z := location z.

	tempVertices := Vector3Array new: 8.
	tempVertices at: 1 put: (Vector3 x: x+(dx negated) y: y+(dy negated) z: z+dz).
	tempVertices at: 2 put: (Vector3 x: x+dx y: y+(dy negated) z: z+dz).
	tempVertices at: 3 put: (Vector3 x: x+dx y: y+ dy z: z+dz).
	tempVertices at: 4 put: (Vector3 x: x+(dx negated) y: y+dy z: z+dz).

	dz := dz negated.

	tempVertices at: 5 put: (Vector3 x: x+(dx negated) y: y+(dy negated) z: z+dz).
	tempVertices at: 6 put: (Vector3 x: x+dx y: y+(dy negated) z: z+dz).
	tempVertices at: 7 put: (Vector3 x: x+dx y: y+ dy z: z+dz).
	tempVertices at: 8 put: (Vector3 x: x+(dx negated) y: y+dy z: z+dz).

	groups := #((1 2 3 4) (8 7 6 5) (2 6 7 3) (1 4 8 5) (4 3 7 8) (1 5 6 2)).
	groups withIndexDo: [:gr :i | 
		1 to: 4 do: [:j | vertexBuffer at: (i-1)*4+j put: (tempVertices at: (gr at: j))]].

	#((0 0 1) (0 0 -1) (1 0 0) (-1 0 0) (0 1 0) (0 -1 0)) withIndexDo:[ :norm :i |
		1 to: 4 do: [:j | normalBuffer at:(i-1)*4+j put: (Vector3 from: norm).].].

	1 to: 6 do:[ :i |
		#((0 0) (0 1.0) (1.0 1.0) (1.0 0)) withIndexDo:[ :norm :j | 
			texCoordBuffer at: (i-1)*4+j put: (Vector2 from: norm).
		].
	].
	self faceGroups:{1. #(0 1 2 3  4 5 6 7  8 9 10 11  12 13 14 15  16 17 18 19  20 21 22 23) asIntegerArray}.
	changed := false
	

! !

TUpdatingMesh subclass: #TCylinder
	instanceVariableNames: 'baseRadius topRadius height slices stacks capTop capBase capFaceGroups capNormals'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TCylinder commentStamp: 'das 10/10/2004 11:24' prior: 0!
TCylinder generates an OGL generalized cylinder (cone). The base of the cylinder is at 0@0@0, the top is at 0@height@0.

baseRadius - the radius of the bottom part of the cylinder.
topRadius - the radius of the top part of the cylinder.
height - the height from the base to the top of the cylinder.
slices - the number of slices around the center of the cylinder.
stacks - the number of stacks from bottom to top.
capped - flag indicating whether the cylinder is capped top and bottom.

DAS!

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:50'!
baseRadius

	^ baseRadius.! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/10/2006 12:06'!
baseRadius: br

	baseRadius := br.
	self boundsChanged.

! !

!TCylinder methodsFor: 'accessing' stamp: 'das 7/7/2006 19:10'!
capped: bool
	
	capTop := bool.
	capBase := bool.

	changed := true.! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:52'!
extent

	^ (self radius*2.0)@height@(self radius*2.0).! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/10/2006 12:06'!
extent: ext

	| sc rad |
	rad := (topRadius max: baseRadius) * 2.0.
	rad := (ext x- rad)abs > (ext z-rad)abs ifTrue:[ext x/2.0] ifFalse:[ext z/2.0].
	
	sc := rad / (topRadius max: baseRadius).
	topRadius := topRadius * sc.
	baseRadius := baseRadius * sc.
	height := ext y.
	self boundsChanged.
! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:55'!
frameBox

	| r |
	r := baseRadius> topRadius ifTrue:[baseRadius] ifFalse:[topRadius].
	^ TBox min:(r negated@0.0@r negated) max:(r@height@r).! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:55'!
height

	^ height.! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/10/2006 12:07'!
height: ht

	height := ht.
	self boundsChanged.
! !

!TCylinder methodsFor: 'initialize' stamp: 'jcg 6/4/2006 13:50'!
initBounds

	| location radius h2 |
	h2 := height/2.0.
	h2 := h2*h2.
	baseRadius < topRadius ifTrue:[radius := topRadius*topRadius.] 
							ifFalse:[radius := baseRadius * baseRadius.].
	radius := (radius+h2) sqrt.
	location := Vector3 x:0.0 y: height/2.0 z:0.0.
	boundSphere := TBoundSphere localPosition: location radius: radius.
	boundSphere frame: self.

! !

!TCylinder methodsFor: 'initialize' stamp: 'das 7/7/2006 19:11'!
initialize
	super initialize.
	mode := GLTriangleStrip.
	baseRadius := 0.5.
	topRadius := 0.0. "default is a cone"
	height := 1.0.
	slices := 8.
	stacks := 4.
	capTop := true.
	capBase := true.
	self materialList: {TMaterial new}.! !

!TCylinder methodsFor: 'render' stamp: 'das 7/7/2006 19:12'!
pick: pointer

	(pointer pickCylinderFrom: Vector3 new radius: baseRadius 
		to: (Vector3 x:0.0 y: height z: 0.0) radius: topRadius) ifTrue:[^ true] ifFalse:[
			capBase ifTrue:[
				baseRadius > 0.0 ifTrue:[
					(pointer pickCircle: 0@0@0 normal:0@-1@0 radius: baseRadius)ifTrue:[^true].].
				].
			capTop ifTrue:[
				topRadius > 0.0 ifTrue:[
					(pointer pickCircle: 0@height@0 normal:0@1@0 radius: topRadius)ifTrue:[^true]].
				].
			^ false.].! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:55'!
radius

	^ baseRadius max: topRadius.
! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:55'!
radius: rad

	self topRadius: rad.
	self baseRadius: rad.! !

!TCylinder methodsFor: 'render' stamp: 'bgf 8/10/2006 21:38'!
renderPrimitive: ogl alpha: alphaPass
	| savedFaceGroups savedNormals |

	super renderPrimitive: ogl alpha: alphaPass.
	
	capFaceGroups ifNil: [^ self].
	"Hackery to render caps"
	[
		savedFaceGroups := faceGroups.
		savedNormals := normalBuffer.
		self faceGroups: capFaceGroups.
		normalBuffer := capNormals.
		mode := GLTriangleFan.
		super renderPrimitive: ogl alpha: alphaPass
	]	 
	ensure: 
	[
		normalBuffer := savedNormals.
		self faceGroups: savedFaceGroups.
		mode := GLTriangleStrip.
	]! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/10/2006 12:06'!
scale: scl

	self translation: (self translation * scl).
	baseRadius := baseRadius * scl.
	topRadius := topRadius * scl.
	height := height * scl.
	self boundsChanged.! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:56'!
slices

	^ slices.! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:56'!
slices: sl

	slices := sl.
	changed := true! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:56'!
stacks
	^ stacks.! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:56'!
stacks: st
	
	stacks := st.
	changed := true! !

!TCylinder methodsFor: 'accessing' stamp: 'das 7/24/2006 14:57'!
topGlobalPosition

"the global position of the top of the cylinder"
	^ self globalTransform localPointToGlobal:0@height@0.! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/4/2006 13:56'!
topRadius

	^ topRadius.! !

!TCylinder methodsFor: 'accessing' stamp: 'jcg 6/10/2006 12:05'!
topRadius: tr

	topRadius := tr.
	self boundsChanged.! !

!TCylinder methodsFor: 'initialize' stamp: 'bgf 8/10/2006 21:55'!
updateMesh
	| halfPi slicePercentage sin cos stackPercentage radius ind stream |

	"For tex-coords to interpolate properly, the first/last vertices of each ring must be duplicates (with differing tex-coords)."
	vertexBuffer := Vector3Array ofSize: (slices+1)*stacks.
	normalBuffer := Vector3Array ofSize: (slices+1)*stacks.
	texCoordBuffer := Vector2Array ofSize: (slices+1)*stacks.

	"Compute geometry."
	halfPi := Float pi *2.0.
	1 to: slices+1 do: 
		[:sliceInd |
		slicePercentage := ((sliceInd-1) / slices) asFloat.
		sin := (slicePercentage*halfPi) sin.
		cos := (slicePercentage*halfPi) cos.
		1 to: stacks do: 
			[:stackInd |
			stackPercentage := ((stackInd-1)/(stacks-1)) asFloat.
			radius := baseRadius interpolateTo: topRadius at: stackPercentage.
			ind := (sliceInd-1)*stacks + stackInd.
			vertexBuffer at: ind put: (sin*radius)@(stackPercentage*height)@(cos*radius).
			normalBuffer at: ind put: sin@0@cos.
			texCoordBuffer at: ind put: (Vector2 u: slicePercentage v: stackPercentage)]].

	"Compute face groups"
	faceGroups := OrderedCollection new.
	0 to: stacks-2 do: 
		[:stackInd |
		faceGroups add: 1.  "material index"
		stream := WriteStream on: IntegerArray new.
		"stream nextPut: stackInd+1.
		stream nextPut: stackInd."
		0 to: slices do: 
			[:sliceInd |
			stream nextPut: (sliceInd*stacks)+stackInd+1.
			stream nextPut: (sliceInd*stacks)+stackInd].
		faceGroups add: stream contents].

	self faceGroups: faceGroups.  "init faceBuffers"

	"Compute cap face groups, if necessary."
	capFaceGroups := nil.
	capBase ifTrue: 
	[
		capFaceGroups := OrderedCollection new.
		capFaceGroups add: 1.
		stream := WriteStream on: IntegerArray new.
		slices to: 0 by: -1 do: [:sliceInd | stream nextPut: (sliceInd*stacks)].
		capFaceGroups add: stream contents.
	].
	capTop ifTrue:
	[
		capFaceGroups ifNil:[capFaceGroups := OrderedCollection new.].
		capFaceGroups add: 1.
		stream := WriteStream on: IntegerArray new.
		0 to: slices do: [:sliceInd | stream nextPut: (sliceInd*stacks)+stacks-1].
		capFaceGroups add: stream contents.
	].

	capNormals := nil.
	(capTop | capBase) ifTrue: [
		capNormals := normalBuffer copy.
		0 to: slices do: [ :index|
			capNormals at: (index*stacks+1) put: 0@-1@0.  "base"
			capNormals at: ((index*stacks)+stacks) put: 0@1@0.  "top"
		].
	].
	changed := false. ! !

TUpdatingMesh subclass: #TRectangle
	instanceVariableNames: 'extent location flip aspect aspectFixed norm divisions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TRectangle commentStamp: 'das 10/10/2004 14:02' prior: 0!
TRectangle is a TPrimitive used to render rectangles.

DAS!

TRectangle subclass: #TButtonHolder
	instanceVariableNames: 'buttonWidth radio radioButton'
	classVariableNames: ''
	poolDictionaries: 'OpenGLConstants'
	category: 'Croquet-Objects'!

!TButtonHolder commentStamp: 'das 1/22/2005 13:10' prior: 0!
TButtonHolder contains and positions TButtons in a 2D surface. This is used by the pop-up window and the Wicket CAD system. The TButtonHolder can be set into radio button mode by sending #radioOn. This sets it such that only one switch button can be selected at a time. 

DAS!

!TButtonHolder methodsFor: 'accessing' stamp: 'das 5/5/2006 16:43'!
addChild: bttn

	| size box scale |
	box := bttn boundingBox.
	scale := 0.8*buttonWidth/(box extentMax).
	bttn scale: scale.

	super addChild: bttn.
	size := (frameChildren size)sqrt*buttonWidth.
	self extent: (size@size).
	radio ifTrue:[
		bttn target: self action: #radio:.
		bttn target2: self action2: #radio:.
		frameChildren size = 1 ifTrue:[
			bttn switch: true.
			radioButton := bttn.
		].
	].! !

!TButtonHolder methodsFor: 'accessing' stamp: 'das 5/5/2006 16:43'!
buttonWidth: bw

	buttonWidth := bw.! !

!TButtonHolder methodsFor: 'accessing' stamp: 'das 2/8/2005 16:50'!
doRadio: bttn

	frameChildren do:[:fc |
		fc switch:fc = bttn.
		fc switch ifTrue:[self radioButton: fc].
		].! !

!TButtonHolder methodsFor: 'accessing' stamp: 'das 5/5/2006 16:43'!
extent: ext

	| dx dy ddx ddy topCorner box across |

	frameChildren ifNil:[^super extent: ext].
	dx := ext x/buttonWidth. 
	dx < 1 ifTrue: [ dx := 1.].
	dy := frameChildren size / dx.
	(dx*dy) < frameChildren size ifTrue:[dy := dy + buttonWidth.].
	extent := (dx asFloat * buttonWidth)@(dy asFloat*buttonWidth).
	topCorner := (extent x/-2.0 + (buttonWidth/2.0))@(extent y/2.0 - (buttonWidth/2.0)).
	across := (dx+0.5) asInteger.
	frameChildren doWithIndex:[ :fc :index |
		ddx := (index-1\\across)*buttonWidth.
		ddy := (index-1//across)*buttonWidth negated.
		box := fc boundingBox.
		fc translationX: topCorner x + ddx-box center x y: topCorner y + ddy-box center y z:0.0.
		].
! !

!TButtonHolder methodsFor: 'initialize' stamp: 'das 5/5/2006 16:44'!
initialize
	super initialize.
	extent := 1.0@1.0.
	self solid: false.
	buttonWidth := 1.0.
	radio := false.
	^self
! !

!TButtonHolder methodsFor: 'render' stamp: 'das 1/21/2005 23:35'!
material: mat

	"don't allow materials to be set via this, because TWindows will do the wrong thing. 
	use #baseMaterial: instead."! !

!TButtonHolder methodsFor: 'render' stamp: 'das 3/9/2004 15:14'!
pick: pointer

	^ false.
"	^super pick: pointer."! !

!TButtonHolder methodsFor: 'accessing' stamp: 'das 8/5/2006 15:58'!
radio: ptr

	self doRadio: ptr selectedObject.
! !

!TButtonHolder methodsFor: 'accessing' stamp: 'das 1/22/2005 12:48'!
radioButton

	^ radioButton.! !

!TButtonHolder methodsFor: 'accessing' stamp: 'das 5/5/2006 16:43'!
radioButton: rb

	radioButton := rb.! !

!TButtonHolder methodsFor: 'accessing' stamp: 'das 5/5/2006 16:44'!
radioOn

" this sets the button holder in radio mode. This means that only one button in its group can be selected at a time. You can retrieve the current selected button via #radioButton."
	radio := true.
	frameChildren ifNotNil:[
		frameChildren doWithIndex:[:bttn :index |
			bttn target: self action: #radio:.
			bttn target2: self action2: #radio:.
			index = 1 ifTrue:[
				bttn switch: true.
				radioButton := bttn.
			].
		].
	].
! !

!TButtonHolder methodsFor: 'render' stamp: 'das 1/21/2005 23:34'!
render: ogl

	"do nothing."
! !

!TButtonHolder methodsFor: 'render' stamp: 'das 3/12/2004 13:34'!
renderAlpha: ogl

	"do nothing."
	super renderAlpha: ogl.! !

TRectangle subclass: #TPortal
	instanceVariableNames: 'postcard outVector blocked locator cameraDistance ghost overlays'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TPortal commentStamp: '<historical>' prior: 0!
TPortal is used to connect and see between TSpace objects. A TPortal has a link to another TPortal which in turn is embedded in another space. If the TPortal is linked back to itself, which is the default, it is treated as a mirror.

toPortal (deprecated - replaced with viewPoint).
viewPoint - the viewPoint frame or portal that this portal is connected to. This can actually be any kind of TFrame, or it could even be a TSnapShot. The default viewPoint is self which is a mirror.
outVector - normal vector for TPortal
lastCameraPosition - last camera position used by testEnter:. If we are in front of the portal, and then behind it, we enter the portal.
blocked - boolean determines if use can enter the portal
locator - address to load contents of portal from
cameraDistance - used to sort the portals back to front.
camera - every portal has its own camera when rendering through. 
overlaySpace - this is a TSpace with a few subtle differences. 
viewPoint - use this when the frame you want to link to is in the same Island.
DAS!

!TPortal methodsFor: 'accessing' stamp: 'das 7/8/2005 15:55'!
addOverlay: ov

	overlays := overlays, {ov}.

! !

!TPortal methodsFor: 'accessing' stamp: 'das 4/29/2002 19:29'!
blocked

	^ blocked.! !

!TPortal methodsFor: 'accessing' stamp: 'das 4/29/2002 19:29'!
blocked: bl

	blocked := bl.! !

!TPortal methodsFor: 'accessing' stamp: 'das 10/28/2005 11:11'!
bounds: bnds
	
" force extent to match the screen. Use extent width as base-line, only change height. "

	self extent: self extent x@ (self extent x * bnds height/bnds width).
! !

!TPortal methodsFor: 'accessing' stamp: 'das 5/27/2005 10:08'!
cameraDistance

" used by TSpace to sort the portals back to front"
	^ cameraDistance.! !

!TPortal methodsFor: 'accessing' stamp: 'das 5/27/2005 10:08'!
cameraDistance: cd

" used by TSpace to sort the portals back to front"
	cameraDistance := cd.! !

!TPortal methodsFor: 'accessing' stamp: 'das 6/5/2005 22:48'!
clipPlane

	^ outVector.! !

!TPortal methodsFor: 'accessing' stamp: 'das 1/16/2005 14:30'!
ghost

	^ ghost.! !

!TPortal methodsFor: 'accessing' stamp: 'das 1/16/2005 14:30'!
ghost: g

	ghost := g.! !

!TPortal methodsFor: 'accessing' stamp: 'das 10/14/2005 15:50'!
globalMatrixOut
" This simply reverses the direction of the matrix. An example of use is for TPortals, where the global matrix used for rendering the outside of the portal is the reverse of the exiting render, hence we need to flip the matrix around to get what we want."

	| gt |
	gt := self globalTransform copy.
	gt a11: 0.0 - gt a11.
	gt a21: 0.0 - gt a21.
	gt a31: 0.0 - gt a31.
	gt a13: 0.0 - gt a13.
	gt a23: 0.0 - gt a23.
	gt a33: 0.0 - gt a33.
	^ gt

! !

!TPortal methodsFor: 'accessing' stamp: 'ar 9/7/2002 14:52'!
hasAlpha

	^ true.! !

!TPortal methodsFor: 'transform' stamp: 'bgf 3/15/2006 17:38'!
initClipPlanes: camTrans mirror: mirror
	| clip dx dy trans v1 v2 v3 v4 |

"calculate the four camera clipping planes based upon the location of the portal relative to the camera. camTrans is the cameras global transform matrix. If any part of the portal is behind the camera, revert to the current clipping planes."

	clip := Vector3Array ofSize: 4.
	
	dx := self extent x/2.
	dy := self extent y/2.
	trans := camTrans orthoNormInverse * self globalTransform.

	v1 := trans localPointToGlobal:(Vector3 x: location x + dx negated y: location y + dy z: location z).
	v1 z > 0.0 ifTrue:[^ nil.].
	v2 := trans localPointToGlobal:(Vector3 x: location x + dx negated y: location y + dy negated z:location z).
	v2 z > 0.0 ifTrue:[^ nil.].
	v3 := trans localPointToGlobal:(Vector3 x: location x + dx y: location y + dy negated z: location z).
	v3 z > 0.0 ifTrue:[ ^ nil.].
	v4 := trans localPointToGlobal:(Vector3 x: location x + dx y: location y + dy z: location z).
	v4 z > 0.0 ifTrue:[ ^ nil.].
	
	mirror ifFalse:[
		clip at: 1 put: ((v2 cross: v1) safelyNormalized).
		clip at: 2 put: ((v3 cross: v2) safelyNormalized).
		clip at: 3 put: ((v4 cross: v3) safelyNormalized).
		clip at: 4 put: ((v1 cross: v4) safelyNormalized).
	] ifTrue: [
		clip at: 1 put: ((v1 cross: v2) safelyNormalized).
		clip at: 2 put: ((v2 cross: v3) safelyNormalized).
		clip at: 3 put: ((v3 cross: v4) safelyNormalized).
		clip at: 4 put: ((v4 cross: v1) safelyNormalized).
	].
	^ clip.! !

!TPortal methodsFor: 'initialize' stamp: 'ar 2/7/2006 17:05'!
initialize
 
	| mat |
	super initialize.
	outVector := Vector3 x: 0.0 y:0.0 z:1.0.
	mat := TMaterial new.
	mat ambientColor:#(0.5 0.5 0.5 0.0).
	mat diffuseColor:#(0.5 0.5 0.5 0.0).
	super material: mat.
	postcard := self.
	self initBounds.
	self objectName:'portal'copy.
	blocked := false. 
	ghost := nil.
	overlays := {}.
"This is just to have something when I test."
"	space := TSpace new.
	postcard := TFrame new.
	space addChild: postcard."
	^self
! !

!TPortal methodsFor: 'testing' stamp: 'das 5/28/2006 16:58'!
isNotPortal3D

	^ true.! !

!TPortal methodsFor: 'testing' stamp: ''!
isPortal

	^ true.! !

!TPortal methodsFor: 'testing' stamp: 'das 5/28/2006 16:55'!
isPortal3D

	^ false.! !

!TPortal methodsFor: 'testing' stamp: 'das 4/29/2002 21:48'!
isSolid 

	^ false.! !

!TPortal methodsFor: 'testing' stamp: 'das 1/27/2005 16:57'!
isTaskPortal

	^ false.! !

!TPortal methodsFor: 'accessing' stamp: 'das 5/23/2005 11:36'!
lookAt

	| mat |
	mat := self globalTransform.
	^ Vector3 x:  mat a13 y: mat a23 z: mat a33.! !

!TPortal methodsFor: 'enterExit' stamp: 'das 8/28/2006 02:33'!
lookAtCamera: cam

	| delta |
	delta := cam globalTransform translation - self globalTransform translation.
	(delta dot: self lookAt)>0 ifTrue:[
		self cameraDistance: delta squaredLength.
		^ true.
	].
	^ false.! !

!TPortal methodsFor: 'accessing' stamp: ''!
material: mat! !

!TPortal methodsFor: 'accessing' stamp: ''!
outVector

	^ outVector.! !

!TPortal methodsFor: 'accessing' stamp: ''!
outVector: ov

	outVector := ov.! !

!TPortal methodsFor: 'render' stamp: 'das 9/9/2006 15:25'!
paintPortal: ogl color: clr

" set the color of the portal background "
	ogl glColor3fv: clr asVector3.! !

!TPortal methodsFor: 'render' stamp: 'das 3/13/2006 21:02'!
portalPick: eventPointer

		| saveSelected portalDistance portalPoint |
		saveSelected := eventPointer copiedSelection.
		
		(eventPointer pointerPick: self boundSphere) ifTrue:[
				portalDistance := eventPointer selectedDistance.
				eventPointer minDistance: portalDistance.
				portalPoint := eventPointer selectedPoint.
				eventPointer selected: saveSelected. "restore the current selections"
				eventPointer selectedPortal: self point: portalPoint. "if we picked a portal"
			] ifFalse:[ 
				eventPointer pointerDown ifTrue:[
					eventPointer selectedPortal = self ifTrue:[
						(eventPointer frame: self pickPlane: (Vector3 new) normal: outVector)
						 ifTrue:[ 
							portalPoint := eventPointer selectedPoint.
							eventPointer selected: saveSelected.
							eventPointer selectedPortalPoint: portalPoint.
						].
					].
				]. 
				portalDistance := Float infinity.
				eventPointer doSelect: false. "don't select since pointer is outside portal"
				eventPointer selected: saveSelected.
			].
	^ portalDistance.! !

!TPortal methodsFor: 'accessing' stamp: 'das 1/19/2006 11:07'!
postcardLink
	
	^ postcard.
! !

!TPortal methodsFor: 'accessing' stamp: 'das 1/19/2006 11:06'!
postcardLink: pc

	pc ifNil:[postcard :=  self] ifNotNil:[ postcard := pc. ].
! !

!TPortal methodsFor: 'render' stamp: 'das 1/19/2006 14:21'!
rearrange: ogl

" this allows a portal to rearrange itself. Called when a portal is about to be rendered as an overlay. Default to allow the space to rearrange its contents if necessary."

	| vwPnt |
	vwPnt := postcard resolveViewpoint: ogl.
	vwPnt ifNotNil:[
		vwPnt root rearrange: ogl.
	].! !

!TPortal methodsFor: 'accessing' stamp: 'das 1/6/2005 22:02'!
reentrant

"Used to indicate that the same portal can view itself in the same space. This is usually the case for normal portals but not for task and filter portals"

	^ true.! !

!TPortal methodsFor: 'render' stamp: 'das 5/14/2005 15:30'!
render: ogl

	" Don't do anything ."! !

!TPortal methodsFor: 'render' stamp: 'bgf 10/3/2006 17:24'!
render: ogl depth: depth
	| trans m2 toSpace gt portalClip clipPlanes renderedObjects ac saveDoSelect saveMinDistance portalDistance vwPnt |

" This method is the recursive portal renderer. If you want the camera entering check out TPortal>>testEnter: ac."

" ------- We keep track of which portal we are currently working in to avoid a tight recursion error,where the portal re-enters itself to render. It is (somewhat) OK to re-enter from another portal. This is equivalent to a mirror in a mirror situation. Here we need to keep a count of the re-entries.------"
	postcard ifNotNil:[vwPnt := postcard resolveViewpoint: ogl.] ifNil:[ vwPnt := self].
	vwPnt ifNil:[ vwPnt := self viewPoint. ]. " make it a mirror for now..."
	vwPnt isFarRef ifTrue: [
		"PC #resolveViewpoint: gives a far ref; we need a near ref for rendering"
		vwPnt wait.
		vwPnt := vwPnt value.	
	].
	toSpace := vwPnt root.
	toSpace ifNil:[^0].

	ogl enterPortal2D.

" ----- Check if the postcard is installed into a TSpace, otherwise we can't render anything"
	toSpace isSpace ifFalse:[ vwPnt := self. toSpace := self root. ]. 
" ------ For rendering the portal ONLY - we turn off depth testing. We just want to write into the stencil buffer here and set the color of the toSpace. Depth testing is turned back on before we recurse into the connected portal.------"
	gt := self globalTransform. 
	depth = 1 ifTrue:[ogl glEnable: GLStencilTest.].
	ogl glPushMatrix. 
	ogl glMultTransposeMatrixf: gt.
	ogl glDisable: GLDepthTest. 

" ------ Prepare to paint the portal with the color of the TSpace ------ "
	self paintPortal: ogl color: toSpace color asVector3.

" ------ Render the stencil here. ------ "
	self renderStencil: ogl depth: depth.

	m2 := vwPnt globalMatrixOut.

" ------- For simplicity - and because I can't think of a good reason not to do this - if a portal points back to itself, it is considered to be a mirror. Note that we flip the directions of the polyon tests as well. This is restored at the end. We are inverting the x-coordinate of the matrix here. ------" 
	self = vwPnt ifTrue:[
		m2 a11: m2 a11 negated.
		m2 a21: m2 a21 negated.
		m2 a31: m2 a31 negated.
		ogl flipFace.].

	ac := ogl camera.
	portalClip :=ac portalClip.
	ac portalClip: m2.

" ------ This math is simply: 

			m1 = trans*m2
			m1 * m2**-1 = trans * m2 * m2**-1
			m1 * m2**-1 = trans. 

 We then replace the transform matrix of the TCamera with the appropriate transform, which places the TCamera into the correct position in the new space.------"
	clipPlanes := ac clipPlanes.
	ac clipPlanes: (self initClipPlanes: ac globalTransform mirror: ogl isMirror).
	self = vwPnt ifTrue:[ ogl mirrorFlip.].
	trans := m2 * gt orthoNormInverse.
	ac globalTransform: (trans * ac globalTransform).
	ogl eventPointer ifNotNil:[
		saveMinDistance := ogl eventPointer minDistance.
		saveDoSelect := ogl eventPointer doSelect.
		portalDistance := self portalPick: ogl eventPointer.
		ogl eventPointer globalTransform: (ac globalTransform * ogl eventPointer localTransform).
		].

" ------ Render TSpace here!! -------"
	renderedObjects := 1 + (toSpace renderSpace: ogl port: vwPnt depth: depth+1 ghostFrame: ghost). 
	ac clipPlanes: clipPlanes.
	self = vwPnt ifTrue:[ogl flipFace. ogl mirrorFlip.].
	ogl glPopMatrix.

" ------ Just in case we jumped into a portal, restore the stencil depth ------"
	ogl glStencilFunc: GLLequal with: depth-1 with: -1.
	ogl glStencilOp: GLKeep with: GLKeep with: GLKeep.

"------ Render the front face invisibly to set the z-buffer to ensure nothing bleeds into the portal space from our space. ------"			
	self renderPortal: ogl.

	depth = 1 ifTrue:[ogl glDisable: GLStencilTest.].
	ac portalClip: portalClip.
	ogl eventPointer ifNotNil:[
		ogl eventPointer doSelect: saveDoSelect.
		ogl eventPointer minDistance: saveMinDistance.
		ogl eventPointer maxDistance: (ogl eventPointer maxDistance min: portalDistance).
		].

	ogl leavePortal2D.

	^ renderedObjects.
! !

!TPortal methodsFor: 'render' stamp: 'das 5/14/2005 15:30'!
renderAlpha: ogl

	" Don't do anything ."! !

!TPortal methodsFor: 'render' stamp: 'das 2/1/2006 19:56'!
renderAsOverlay: ogl

	| rval cp vwPnt |

	postcard ~= self ifTrue:[
		cp := ogl camera currentPortal.
		ogl camera currentPortal: self.
		vwPnt := postcard resolveViewpoint: ogl.
		rval := vwPnt send:[:vp | vp renderOverlay: ogl overlay: overlays.].
		ogl camera currentPortal: cp.
		].
	^ rval.! !

!TPortal methodsFor: 'render' stamp: 'das 7/10/2006 12:02'!
renderPortal: ogl

	| rval saveForceWire |

"#renderPortal: simply sets up a stencil in the shape of the portal we are about to render through. This is where it happens."
	ogl glPushMatrix.
	saveForceWire := ogl forceWire.
	ogl glPolygonOffset: 2.1 with: 4.0.
	ogl glEnable: GLPolygonOffsetFill.
	ogl glColorMask: 0 with: 0 with: 0 with: 0.
	ogl glMultTransposeMatrixf: self globalTransform.
	ogl forceWire: false.
	rval := self renderPrimitive: ogl alpha: true.
	ogl forceWire: saveForceWire.
	ogl glColorMask:GLTrue with: GLTrue with:GLTrue with:GLTrue.
	ogl glDisable: GLPolygonOffsetFill.
	ogl glPopMatrix.
	^ rval.! !

!TPortal methodsFor: 'render' stamp: 'das 7/10/2006 12:02'!
renderStencil: ogl depth: depth

	| saveForceWire |
	ogl glDisable: GLLighting.
	ogl glStencilFunc:GLEqual with:depth-1 with: -1.
	ogl glStencilOp:GLKeep with:GLKeep with:GLIncr.

	saveForceWire := ogl forceWire.
	ogl forceWire:false.
	self renderPrimitive: ogl alpha: true.
	ogl forceWire: saveForceWire.

	ogl glStencilFunc: GLEqual with: depth with: -1.
	ogl glStencilOp: GLKeep with: GLKeep with: GLKeep.

	ogl glEnable: GLDepthTest.
	ogl glEnable: GLLighting.
	ogl glPopMatrix.

	ogl glPushMatrix.
! !

!TPortal methodsFor: 'render' stamp: 'das 1/18/2006 20:47'!
renderView: ogl with: systemOverlays

	| rval vwpnt |
	postcard ifNil:[^0].
	postcard ~= self ifTrue:[
		vwpnt := postcard resolveViewpoint: ogl.
		vwpnt ifNil:[^ 0 ].
		ogl camera currentPortal: self.
		rval := vwpnt renderView:ogl overlay:(overlays,systemOverlays).
		ogl camera currentPortal: nil.
	].
	^ rval.! !

!TPortal methodsFor: 'accessing' stamp: 'das 1/19/2006 13:39'!
resolveViewpoint: ogl

" This is used when we render via a portal and this TFrame is used as the target viewpoint. See TPostCard for more info"
	^ self.! !

!TPortal methodsFor: 'accessing' stamp: 'das 1/18/2005 12:39'!
task
"default to nil task, so no task is generated for normal portals"
	^ nil.! !

!TPortal methodsFor: 'enterExit' stamp: 'bgf 7/5/2006 02:49'!
testEnter: ogl avatar: avatar

	| pc vwPnt m1 m2 trans newSpace |
"this can only determine if a portal transition needs to occur. You can't actually do the portal transition here, because we are actually inside of the portal's Island. The actual transition is handled by the CroquetHarness if the newPostcard field of this object is not nil."

	blocked ifTrue:[^ false ].  "You can only look, not pass through."

 	pc := postcard.
	self = pc ifTrue:[ ^false].
	vwPnt := pc resolveViewpoint: ogl.	
	vwPnt ifNil:[ ^ false. ].
	avatar portalRay doSelect ifFalse:[ ^ false. ].
	vwPnt send:[ :toPortal |
		toPortal root ifNil:[^ false]. "portal is closed"

		(avatar portalRay pointerPick: self boundSphere) ifTrue:[
			m1 := self globalTransform.
			toPortal isPortal ifTrue:[ m2 := toPortal globalMatrixOut.]
			ifFalse:[ m2 := toPortal globalTransform.].
			trans := m2 * m1 orthoNormInverse.
			avatar newTransform: (trans * ogl avatar globalTransform).
			avatar jumpCameraTrans: (trans * ogl camera globalTransform).
			newSpace := toPortal root.
			newSpace send:[:ns |  avatar newPostcard: ns postcard.].
			^ true.
		].
	].
	^ false.! !

!TPortal methodsFor: 'accessing' stamp: 'das 9/13/2006 14:34'!
viewPoint

	^ self.
! !

TPortal subclass: #TPortalCube
	instanceVariableNames: 'cube viewPoint'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TPortalCube methodsFor: 'initialize' stamp: 'das 7/10/2006 11:43'!
boundSphere

	^ cube boundSphere.! !

!TPortalCube methodsFor: 'accessing' stamp: 'das 8/18/2006 22:29'!
boundingBox

	^ cube boundingBox.
! !

!TPortalCube methodsFor: 'accessing' stamp: 'das 8/28/2006 02:58'!
extent

	^ cube extent.! !

!TPortalCube methodsFor: 'accessing' stamp: 'das 8/28/2006 02:58'!
extent: ext

	cube extent: ext.! !

!TPortalCube methodsFor: 'initialize' stamp: 'das 7/10/2006 11:38'!
hasAlpha

	^ cube hasAlpha.! !

!TPortalCube methodsFor: 'initialize' stamp: 'das 7/10/2006 11:38'!
hasOpaque

	^ cube hasOpaque.! !

!TPortalCube methodsFor: 'initialize' stamp: 'das 9/13/2006 01:05'!
initialize

	cube := TCube new.
	cube extent:2@2@2.
	super initialize.
	cube material color: #(0.4 0.5 0.7 0.25)asVectorColor.
	cube material specularColor: #(0.9 0.9 0.9 1.0)asVectorColor.
	cube checkAlpha.
	viewPoint := TGroupNoClip new.
	self addChild: viewPoint.! !

!TPortalCube methodsFor: 'accessing' stamp: 'das 8/28/2006 02:31'!
lookAtCamera: cam

	| delta |
	delta := cam globalTransform translation - self globalTransform translation.
	self cameraDistance: delta squaredLength.
	^ true.
! !

!TPortalCube methodsFor: 'initialize' stamp: 'das 7/10/2006 13:08'!
pick: ptr

	^ false! !

!TPortalCube methodsFor: 'initialize' stamp: 'das 7/10/2006 13:20'!
portalPick:eventPointer

	^ Float infinity.! !

!TPortalCube methodsFor: 'render' stamp: 'das 9/9/2006 15:57'!
render: ogl depth: depth


	| rval mx lTrans |
	lTrans := localTransform.
	ogl forceWire: ogl forceWire not.
	mx := ogl maxPortalDepth.
	ogl maxPortalDepth: 1.
	rval := super render: ogl depth: depth .
	ogl forceWire: ogl forceWire not.
	ogl maxPortalDepth: mx.
	localTransform := lTrans.
	^ rval.! !

!TPortalCube methodsFor: 'render' stamp: 'das 9/13/2006 01:05'!
renderAlpha: ogl
	
	| saveFW |
	saveFW := ogl forceWire.
	ogl forceWire: false.
	cube renderPrimitive: ogl alpha: true.
	ogl forceWire: saveFW.! !

!TPortalCube methodsFor: 'render' stamp: 'das 7/10/2006 12:54'!
renderPrimitive: ogl alpha: alphaPass

	| saveFW |
	saveFW := ogl forceWire.
	ogl forceWire: false.
	cube renderPrimitive: ogl alpha: alphaPass.
	ogl forceWire: saveFW.
! !

!TPortalCube methodsFor: 'render' stamp: 'das 7/10/2006 12:23'!
resolveViewpoint: ogl

" This is used when we render via a portal and this TFrame is used as the target viewpoint. See TPostCard for more info"
	^ viewPoint.! !

TPortal subclass: #TPortalWire
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TPortalWire methodsFor: 'initialize' stamp: 'das 9/13/2006 03:07'!
pick: ptr

	^ false! !

!TPortalWire methodsFor: 'initialize' stamp: 'das 9/13/2006 03:07'!
portalPick:eventPointer

	^ Float infinity.! !

!TPortalWire methodsFor: 'render' stamp: 'das 9/13/2006 03:07'!
render: ogl depth: depth


	| rval mx lTrans |
	lTrans := localTransform.
	ogl forceWire: ogl forceWire not.
	mx := ogl maxPortalDepth.
	ogl maxPortalDepth: 1.
	rval := super render: ogl depth: depth .
	ogl forceWire: ogl forceWire not.
	ogl maxPortalDepth: mx.
	localTransform := lTrans.
	^ rval.! !

!TPortalWire methodsFor: 'render' stamp: 'das 9/13/2006 03:09'!
renderAlpha: ogl
	
	| saveFW |
	saveFW := ogl forceWire.
	ogl forceWire: false.
	super renderPrimitive: ogl alpha: true.
	ogl forceWire: saveFW.! !

!TPortalWire methodsFor: 'render' stamp: 'das 9/13/2006 03:08'!
renderPrimitive: ogl alpha: alphaPass

	| saveFW |
	saveFW := ogl forceWire.
	ogl forceWire: false.
	super renderPrimitive: ogl alpha: alphaPass.
	ogl forceWire: saveFW.
! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:36'!
aspect

	^ aspect.! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:37'!
aspect: asp

	aspect := asp.! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:37'!
aspectFixed
	^ aspectFixed.
! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:37'!
aspectFixed: bool
	aspectFixed := bool.
! !

!TRectangle methodsFor: 'accessing' stamp: 'das 7/4/2006 17:30'!
divisions: div

	div isPoint ifTrue:[divisions := div max:(1@1).] ifFalse:[divisions := div@div max:(1@1)].
! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:37'!
extent

	^ extent.
! !

!TRectangle methodsFor: 'accessing' stamp: 'das 8/22/2006 15:37'!
extent: ext

	aspectFixed ifTrue:[
	 (extent x/extent y)abs >= 1 ifTrue:[
			extent := (ext x) @ (ext x * aspect).] 
		ifFalse:[
			extent := ext y/aspect @ ext y.].
		]
	ifFalse: [	extent := ext.].
	self boundsChanged.

! !

!TRectangle methodsFor: 'accessing' stamp: 'ar 9/25/2006 15:22'!
extent: bounds aspect: ratio

	"Fit me within the given extent bounds, respecting the aspect ratio."

	| bh bw |

	(bounds x = 0) ifTrue: [
		bh := 0.
		bw := 0.
	] ifFalse: [
		(ratio > (bounds y / bounds x)) ifTrue: [
			bh := bounds y.
			bw := bounds y / ratio.
		] ifFalse: [
			bw := bounds x.
			bh := bounds x * ratio.
		].
	].

	self aspect: ratio.
	extent := bw@bh.
	self boundsChanged.

! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:37'!
extentRect: extRect

	self extent: (extRect extent).
	self location: (extRect center - (extRect extent/2.0))@0.0.
! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:39'!
flip: bool
	flip == bool ifTrue: [^ self].
	(flip := bool)
		ifFalse: [norm := Vector3 x: 0.0 y: 0.0 z: 1.0]
		ifTrue: [norm := Vector3 x: 0.0 y: 0.0 z: -1.0].
	changed := true.! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:40'!
frameBox
	"Answer the local bounding box of this frame"
	^TBox origin: (Vector3 x: extent x/-2.0 y: extent y/-2.0 z:0.0) 
		corner: (Vector3 x: extent x/2.0 y: extent y/2.0 z:0.0) .! !

!TRectangle methodsFor: 'initialize' stamp: 'jcg 6/10/2006 14:40'!
initBounds
	boundSphere := TBoundSphere localPosition: location radius: extent asPoint r / 2.
	boundSphere frame: self! !

!TRectangle methodsFor: 'initialize' stamp: 'das 7/5/2006 14:31'!
initialize
	super initialize.
	location := Vector3 x: 0.0 y:0.0 z:0.0.
	extent := Point x: 1 y: 1. " x-y extent of the rect."
	self flip: false.
	norm := 0@0@1.
	aspectFixed := false.
	aspect := 1.0.
	divisions := 1@1.
	self initBounds.
	self materialList: {TMaterial new}.! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:40'!
location

	^ location.! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:40'!
location: loc

	location := loc.
	self boundsChanged.! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 16:14'!
normal: n

	norm := n.! !

!TRectangle methodsFor: 'render' stamp: 'jcg 6/10/2006 14:45'!
pick: pointer
	| dx dy |

	dx := self extent x/2.
	dy := self extent y/2.

	flip ifFalse:[
	^ pointer pickQuad: norm
		q1: (Vector3 x: location x + dx negated y: location y + dy z: location z)
		q2: (Vector3 x: location x + dx negated y: location y + dy negated z: location z)
		q3: (Vector3 x: location x + dx y: location y + dy negated z: location z)
		q4: (Vector3 x: location x + dx y: location y + dy z: location z).
		]
	ifTrue:[
	^ pointer pickQuad: norm
		q1: (Vector3 x: location x + dx y: location y + dy z: location z)
		q2: (Vector3 x: location x + dx y: location y + dy negated z: location z)
		q3: (Vector3 x: location x + dx negated y: location y + dy negated z: location z)
		q4: (Vector3 x: location x + dx negated y: location y + dy z: location z).]
! !

!TRectangle methodsFor: 'accessing' stamp: 'jcg 6/10/2006 14:40'!
scale: scl

	self translation: self translation * scl.
	self extent: extent * scl.! !

!TRectangle methodsFor: 'initialize' stamp: 'das 7/4/2006 14:08'!
update11

	| dx dy x y z vert texCoord newNorm |
	mode := GLQuads.

	vertexBuffer := Vector3Array ofSize: 4.
	normalBuffer := Vector3Array ofSize: 4.
	texCoordBuffer := Vector2Array ofSize: 4.

	dx := self extent x/2.0.
	dy := self extent y/2.0.
	x := location x.
	y := location y.
	z := location z.

	vert := Vector3 new.
	texCoord := Vector2 new.

	vert x: x-dx y: y+dy z: z.
	texCoord u: 0.0 v: 0.0.
	vertexBuffer at: 1 put: vert.
	texCoordBuffer at: 1 put: texCoord.

	vert x: x-dx y: y-dy z: z.
	texCoord u: 0.0 v: 1.0.
	vertexBuffer at: 2 put: vert.
	texCoordBuffer at: 2 put: texCoord.

	vert x: x+dx y: y-dy z: z.
	texCoord u: 1.0 v: 1.0.
	vertexBuffer at: 3 put: vert.
	texCoordBuffer at: 3 put: texCoord.

	vert x: x+dx y: y+dy z: z.
	texCoord u: 1.0 v: 0.
	vertexBuffer at: 4 put: vert.
	texCoordBuffer at: 4 put: texCoord.

	flip 
		ifFalse:
			[newNorm := 0@0@1.
			self faceGroups: {1. #(0 1 2 3) asIntegerArray}]
		ifTrue:
			[newNorm := 0@0@-1.
			self faceGroups: {1. #(3 2 1 0) asIntegerArray}].
	normalBuffer atAllPut: newNorm.

	"If 'norm' is nil, it is intentional... it means that the rectangle can be picked from both sides. Preserve this property."
	norm ifNotNil: [norm := newNorm].

	changed := false! !

!TRectangle methodsFor: 'accessing' stamp: 'das 7/4/2006 17:00'!
updateMesh
	| nverts |

	nverts := (divisions x+1)*(divisions y+1).
	vertexBuffer := Vector3Array ofSize: nverts.
	normalBuffer := Vector3Array ofSize: nverts.
	texCoordBuffer := Vector2Array ofSize: nverts.

	divisions = (1@1) ifTrue:[ self update11.] ifFalse:[
		divisions y > divisions x ifTrue:[self updateY]ifFalse:[self updateX].
	].
	changed := false.

! !

!TRectangle methodsFor: 'initialize' stamp: 'das 7/5/2006 10:04'!
updateX

	| dx dy x y z vert texCoord newNorm deltax deltay fg strip |
	mode := GLTriangleStrip.

	dx := self extent x/2.0.
	dy := self extent y/2.0.
	deltax := self extent x/divisions x asFloat.
	deltay := self extent y/divisions y asFloat.

	x := location x.
	y := location y.
	z := location z.

	vert := Vector3 new.
	texCoord := Vector2 new.

	0 to: divisions y do:[ :yIndex |
		0 to: divisions x do:[ :xIndex |
			
			vert x: (xIndex*deltax-dx+x)
				y: (yIndex*deltay-dy+y)
				z: z.
			texCoord u: xIndex/divisions x asFloat v: divisions y - (yIndex/divisions y) asFloat.

			vertexBuffer at: divisions x + 1 * yIndex + xIndex + 1 put: vert.
			texCoordBuffer at: divisions x + 1 * yIndex + xIndex + 1 put: texCoord.
		].
	].

	fg := OrderedCollection new.
	flip 
		ifFalse:[
			newNorm := 0@0@1.
			0 to: divisions y-1 do:[ :yIndex |
				strip := OrderedCollection new.
				0 to: divisions x do:[ :xIndex |
					strip add:divisions x + 1 *(yIndex+1) + xIndex.
					strip add:divisions x + 1 * yIndex + xIndex.
				].
			fg add:1; add: strip asIntegerArray.
			]
		]
		ifTrue:[
			newNorm := 0@0@-1.
			0 to: divisions y-1 do:[ :yIndex |
				strip := OrderedCollection new.
				0 to: divisions x do:[ :xIndex |
					strip add:divisions x + 1 * yIndex + xIndex.
					strip add:divisions x + 1 *(yIndex+1) + xIndex.
				].
			fg add: 1; add: strip asIntegerArray.
			].
		].

	self faceGroups: fg.
	normalBuffer atAllPut: newNorm.

	"If 'norm' is nil, it is intentional... it means that the rectangle can be picked from both sides. Preserve this property."
	norm ifNotNil: [norm := newNorm].

	changed := false! !

!TRectangle methodsFor: 'initialize' stamp: 'das 7/5/2006 10:02'!
updateY

	| dx dy x y z vert texCoord newNorm deltax deltay fg strip |
	mode := GLTriangleStrip.

	dx := self extent x/2.0.
	dy := self extent y/2.0.
	deltax := self extent x/divisions x asFloat.
	deltay := self extent y/divisions y asFloat.

	x := location x.
	y := location y.
	z := location z.

	vert := Vector3 new.
	texCoord := Vector2 new.

	0 to: divisions x do:[ :xIndex |
		0 to: divisions y do:[ :yIndex |
			
			vert x: (xIndex*deltax-dx+x)
				y: (yIndex*deltay-dy+y)
				z: z.
			texCoord u: xIndex/divisions x asFloat v: divisions y - (yIndex/divisions y) asFloat.

			vertexBuffer at: divisions y + 1 * xIndex + yIndex + 1 put: vert.
			texCoordBuffer at: divisions y + 1 * xIndex + yIndex + 1 put: texCoord.
		].
	].

	fg := OrderedCollection new.
	flip 
		ifFalse:[
			newNorm := 0@0@1.
			0 to: divisions x-1 do:[ :xIndex |
				strip := OrderedCollection new.
				0 to: divisions y do:[ :yIndex |
					strip add:divisions y + 1 * xIndex + yIndex.
					strip add:divisions y + 1 *(xIndex+1) + yIndex.
				].
			fg add: 1; add: strip asIntegerArray.
			]
		]
		ifTrue:[
			newNorm := 0@0@-1.
			0 to: divisions x-1 do:[ :xIndex |
				strip := OrderedCollection new.
				0 to: divisions y do:[ :yIndex |
					strip add:divisions y + 1 *(xIndex+1) + yIndex.
					strip add:divisions y + 1 * xIndex + yIndex.
				].
			fg add: 1; add: strip asIntegerArray.
			].
		].

	self faceGroups: fg.
	normalBuffer atAllPut: newNorm.

	"If 'norm' is nil, it is intentional... it means that the rectangle can be picked from both sides. Preserve this property."
	norm ifNotNil: [norm := newNorm].

	changed := false! !

TRectangle subclass: #TTexture
	instanceVariableNames: 'tform baseScale isPremultiplied textureMode uvScale uvOffset uvAngle textureMatrix'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TTexture commentStamp: '<historical>' prior: 0!
This is used to create a GL texture from a form, either directly or from a file.

This is dependent upon the TForm class, which handles the actual bits.

tform - this hold both a Squeak Form and the OpenGL GLID for the texture in OpenGL memory.
baseScale - 
uvScale - the uv values can be modified for certain objects like the TPrimitives and the TMesh/TLoad3DSMax. Do that here.
uvOffset - this is used by the TMesh and TLoad3DSMax objects.
mode - used when we render the texture by itself, though it can be used by anything.
uvAngle - again used by the 3DSMax importer for the TMesh objects.

isPremultiplied: if true, indicates that the texture data has premultiplied alpha,
so rendering is done with GLBlendFunc GLOne with: GLOneMinusSrcAlpha.

DAS
!

TTexture subclass: #TLogoBug
	instanceVariableNames: 'corner'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TLogoBug commentStamp: 'das 8/22/2006 16:09' prior: 0!
TLogoBug is a texture that is placed into one of the corners:
#topLeft
#topRight
#bottomLeft
#bottomRight

I still need to complete this.!

!TLogoBug methodsFor: 'accessing' stamp: 'das 8/22/2006 09:31'!
corner: crnr

	corner := crnr.! !

!TLogoBug methodsFor: 'initialize' stamp: 'das 8/22/2006 15:02'!
initialize

	super initialize.
	corner := #topRight. "default corner"
	self material color:(VectorColor r:1 g:1 b:1 a:0.99).
	self material fullBright: true.
	self checkAlpha.! !

!TLogoBug methodsFor: 'initialize' stamp: 'das 8/27/2006 17:59'!
rearrange: ogl

	| cornerVector width |
	
	cornerVector := ogl camera cornerVector.
	width := cornerVector x/2.
	self extent:width@width.
	self translation: cornerVector-(width/1.8@(width/2.8) @0).
! !

!TTexture methodsFor: 'accessing' stamp: 'das 6/30/2004 23:29'!
bMipmap

	^ tform bMipmap.! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 17:02'!
computeTextureMatrix
	(uvAngle = 0.0 and:[uvOffset = 0.0 and:[uvScale x = 1.0 and:[uvScale y = 1.0]]])
		ifTrue:[^textureMatrix := nil].
	textureMatrix := Matrix4x4 identity.
	textureMatrix rotationAroundZ: uvAngle.
	textureMatrix scaling: (Vector3 x: uvScale x y: uvScale y z: 1.0).
	textureMatrix translation: (Vector3 x: uvOffset x y: uvOffset y z: 0.0).
	^textureMatrix! !

!TTexture methodsFor: 'copying' stamp: 'das 7/8/2002 15:15'!
copy

	^ self shallowCopy.! !

!TTexture methodsFor: 'render' stamp: 'ar 9/9/2006 16:51'!
disable: ogl
	"TTexture overrides parent implementation (since this is a port of old code to a new framework), but eventually should work within the framework."

	(tform notNil & ogl changeTexture) ifFalse:[^false].
	isPremultiplied ifTrue: [
		ogl glPopAttrib.
	].
	textureMatrix ifNotNil: [
		ogl glTexEnvi: GLTextureEnv with: GLTextureEnvMode with: GLModulate;
			glMatrixMode:GLTexture;
			glPopMatrix;
			glMatrixMode: GLModelview.
	].
	ogl glDisable: GLTexture2d.
! !

!TTexture methodsFor: 'render' stamp: 'ar 9/9/2006 16:49'!
enable: ogl
	"TTexture overrides parent implementation (since this is a port of old code to a new framework), but eventually should work within the framework."

	(tform notNil and: [ogl installTexture: tform]) ifTrue:[
		ogl glEnable: GLTexture2d.
	
		isPremultiplied ifTrue: [
			ogl glPushAttrib: GLColorBufferBit.
			ogl glBlendFunc: GLOne with: GLOneMinusSrcAlpha.
		].
		ogl glTexParameteri: GLTexture2d with: GLTextureWrapS with:GLRepeat;
			glTexParameteri: GLTexture2d with: GLTextureWrapT with:GLRepeat.

		ogl glTexEnvi: GLTextureEnv with: GLTextureEnvMode with: textureMode.

		textureMatrix ifNotNil:[
			ogl glMatrixMode:GLTexture;
				glPushMatrix;
				glLoadTransposeMatrixf: textureMatrix;
				glMatrixMode: GLModelview.
		].
	].
! !

!TTexture methodsFor: 'accessing' stamp: 'das 5/24/2004 13:09'!
extension
	
	^ tform extension.! !

!TTexture methodsFor: 'accessing' stamp: 'das 1/26/2006 11:43'!
extent: ext

	super extent: ext.
	baseScale := (extent x/tform extent x) asFloat.! !

!TTexture methodsFor: 'accessing' stamp: 'das 7/1/2004 10:18'!
fileName

	^ tform fileName.! !

!TTexture methodsFor: 'initialize' stamp: 'das 5/24/2004 11:13'!
flipVertically
	"Flip the texture vertically"
	tform form flipVertically.
	tform updateAll.
! !

!TTexture methodsFor: 'initialize' stamp: 'ar 9/9/2006 17:11'!
initialize
	super initialize.
	textureMode := GLModulate.
	uvScale := 1@1.
	uvAngle := 0.
	uvOffset := 0@0.
	textureMatrix := nil.
! !

!TTexture methodsFor: 'initialize' stamp: 'das 7/8/2006 12:57'!
initializeOpaque: aBool extent: targetExtent
| mat |
	self 
		initializeWithForm: (Form extent: targetExtent depth: 32) 
		mipmap: false "aCroquetWorld ogl autoMipmap"
		shrinkFit: false. "the form needs to grow"
	self uvScale: (tform extent/tform extent) asFloatPoint.
	mat := TMaterial new.
	aBool ifTrue:[
		mat ambientColor: #(1.0 1.0 1.0 1).
		mat diffuseColor: #(1.0 1.0 1.0 1).
	] ifFalse:[
		mat ambientColor: #(1.0 1.0 1.0 0.9).
		mat diffuseColor: #(1.0 1.0 1.0 0.9).
	].
	mat emissiveColor: #(1.0 1.0 1.0 1.0).
	super material: mat.
	^self.! !

!TTexture methodsFor: 'initialize' stamp: 'das 1/5/2006 16:02'!
initializeWithFileName: fName

	^self initializeWithFileName: fName mipmap: true shrinkFit: false.
! !

!TTexture methodsFor: 'initialize' stamp: 'das 1/5/2006 16:02'!
initializeWithFileName: fName mipmap: bMM

	^self initializeWithFileName: fName mipmap: bMM shrinkFit: false.

! !

!TTexture methodsFor: 'initialize' stamp: 'ar 11/18/2004 19:17'!
initializeWithFileName: fName mipmap: bMM shrinkFit: bSF 

	^self initializeWithFileName: fName mipmap: bMM shrinkFit: bSF extension: nil.

! !

!TTexture methodsFor: 'initialize' stamp: 'sn 2/25/2018 16:15'!
initializeWithFileName: fName mipmap: bMM shrinkFit: bSF extension: ext

	tform := TFormManager default addFileNamed: fName mipmap: bMM shrinkFit: bSF extension: ext.
	tform ifNil:[tform := TForm defaultTForm thumb. Transcript cr; show:'TForm File: <'; show: fName; show:'> was not found. Using default texture.' ]. 
	
	self setup.
! !

!TTexture methodsFor: 'initialize' stamp: 'das 5/24/2004 11:06'!
initializeWithForm: baseForm mipmap: bMM shrinkFit: bSF 

	self initializeWithForm: baseForm mipmap: bMM shrinkFit: bSF extension: nil.
! !

!TTexture methodsFor: 'initialize' stamp: 'das 7/8/2004 18:46'!
initializeWithForm: baseForm mipmap: bMM shrinkFit: bSF extension: ext

	tform := TForm fromForm: baseForm mipmap: bMM shrinkFit: bSF extension: ext.
	self setup.! !

!TTexture methodsFor: 'initialize' stamp: 'das 1/2/2006 16:57'!
initializeWithTForm: tf

	tform := tf.
	self setup.! !

!TTexture methodsFor: 'accessing' stamp: 'ar 10/1/2004 15:04'!
invalidate: aRect

	tform invalidate: aRect.! !

!TTexture methodsFor: 'accessing' stamp: 'bgf 6/23/2006 14:25'!
isPremultiplied

	"Is this texture deemed to contain premultiplied color values
	 (usually as a result of supersampling without normalization,
      such as in a dynamically generated image) ?

	(The value is specified via #isPremultiplied:flag,
	it is not computed automatically.)."

	^ isPremultiplied! !

!TTexture methodsFor: 'accessing' stamp: 'bgf 6/23/2006 14:20'!
isPremultiplied: bFlag

	"Specify whether this texture contains premultiplied color values
	 (usually as a result of supersampling without normalization,
      such as in a dynamically generated image). 

	 Default is false."

	isPremultiplied := bFlag! !

!TTexture methodsFor: 'accessing' stamp: 'das 7/1/2004 09:51'!
isStatic
	^ tform isStatic.! !

!TTexture methodsFor: 'testing' stamp: 'das 7/8/2002 15:43'!
isTexture

	^ true.! !

!TTexture methodsFor: 'accessing' stamp: 'das 7/1/2004 10:18'!
magFilter
	^ tform magFilter.! !

!TTexture methodsFor: 'accessing' stamp: 'das 7/1/2004 10:18'!
minFilter
	^ tform minFilter.! !

!TTexture methodsFor: 'accessing' stamp: 'bgf 3/16/2006 17:50'!
originalAspect
	"The aspect ratio of the image material or other form content
	before being transformed into powers of two."

	^ tform extent y / tform extent x asFloat ! !

!TTexture methodsFor: 'accessing' stamp: 'bgf 3/16/2006 17:47'!
originalExtent
	"The extent of the image material or other form content
	before being transformed into powers of two."

	^ tform extent! !

!TTexture methodsFor: 'fileIn/Out' stamp: 'ar 7/12/2006 15:35'!
postImportFromIslandCopier: importer
	super postImportFromIslandCopier: importer.
	tform ifNotNil:[
		"Ensure that we only have one copy of the target tform in the current island"
		| formKey existing |
		formKey := tform sha asSymbol.
		existing := self island at: formKey.
		existing ifNil:[self island at: formKey put: (existing := tform)].
		tform := existing.
	].! !

!TTexture methodsFor: 'accessing' stamp: 'das 6/21/2003 20:47'!
scale

	^ baseScale.! !

!TTexture methodsFor: 'accessing' stamp: 'das 5/24/2004 11:12'!
scale: sc

	baseScale := baseScale * sc.
	self translation: (self translation * sc).
	extent := tform extent * baseScale. 	"save the original value if we have to render this object."
	self boundsChanged.
! !

!TTexture methodsFor: 'initialize' stamp: 'ar 9/9/2006 16:35'!
setup

	isPremultiplied := false.
	baseScale := 0.02.
	extent := tform extent * baseScale.
	self aspectFixed: true.
	self aspect: (extent y/extent x) asFloat.
	self texture: self.
"	extent := extent * baseScale." 	"save the original value if we have to render this object."
	self objectName: tform url copy.
! !

!TTexture methodsFor: 'accessing' stamp: 'das 7/1/2004 09:32'!
shrinkFit

^ tform bShrinkFit.! !

!TTexture methodsFor: 'accessing' stamp: 'das 7/1/2004 09:32'!
shrinkFit: bSF

 tform shrinkFit: bSF.! !

!TTexture methodsFor: 'accessing' stamp: 'jcg 6/1/2005 07:13'!
target

	^ tform target.! !

!TTexture methodsFor: 'accessing' stamp: 'das 7/8/2004 08:04'!
targetExtent

	^ tform extent.! !

!TTexture methodsFor: 'accessing' stamp: 'das 7/8/2004 10:22'!
targetForm
	^tform form.! !

!TTexture methodsFor: 'accessing' stamp: 'das 5/24/2004 11:12'!
textureForm
	^tform form.! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:42'!
textureMatrix
	"Answer the OpenGL texture matrix (derived from the uv* set of attributes)"
	^textureMatrix! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:37'!
textureMatrix: aMatrix4x4
	"Set the OpenGL texture matrix (derived from the uv* set of attributes)"
	textureMatrix := aMatrix4x4! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:36'!
textureMode
	"Answer the OpenGL blend mode used for this texture"
	^textureMode! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:36'!
textureMode: aBlendMode
	"Set the OpenGL blend mode used for this texture"
	textureMode := aBlendMode! !

!TTexture methodsFor: 'accessing' stamp: 'ar 10/1/2004 15:13'!
tform
	^tform! !

!TTexture methodsFor: 'render' stamp: 'das 5/24/2004 14:32'!
upload

	tform updateAll.! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:33'!
uvAngle
	"Answer the angle of u,v coordinates"
	^uvAngle! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:42'!
uvAngle: degrees
	"Set the angle of u,v coordinates"
	uvAngle := degrees.
	self computeTextureMatrix.! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:33'!
uvOffset
	"Answer the offset for u,v coordinates."
	^uvOffset! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 17:01'!
uvOffset: aPoint
	"Set the offset for u,v coordinates."
	uvOffset := aPoint asPoint.
	self computeTextureMatrix.! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 16:34'!
uvScale
	"Answer the scale for texture coordinates"
	^uvScale! !

!TTexture methodsFor: 'accessing' stamp: 'ar 9/9/2006 17:01'!
uvScale: aPoint
	"Set the scale for texture coordinates"
	uvScale := aPoint asPoint.
	self computeTextureMatrix.! !

TUpdatingMesh subclass: #TSphere
	instanceVariableNames: 'segments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TSphere commentStamp: 'jcg 6/10/2006 07:36' prior: 0!
TSphere2 is a mesh-based replacement for the TSphere primitive.  From the TSphere class comment:

"TSphere is a TPrimitive used to render a sphere.

segments - the number of slices of the sphere around the axis and the number of slices perpendicular to it.

DAS"!

!TSphere methodsFor: 'initialize' stamp: 'jcg 6/10/2006 12:16'!
initBounds
	^ self "we reuse the same boundSphere, and manipulate it directly."! !

!TSphere methodsFor: 'initialize' stamp: 'das 7/5/2006 14:31'!
initialize
	super initialize.
	mode := GLTriangleStrip.
	
	boundSphere := TBoundSphere localPosition: (Vector3 new x:0.0 y:0.0 z:0.0) radius: 1.0.
	boundSphere frame: self.
	segments := 10.
	self materialList: {TMaterial new}.! !

!TSphere methodsFor: 'accessing' stamp: 'jcg 6/10/2006 07:36'!
localPosition

	^ boundSphere localPosition.! !

!TSphere methodsFor: 'accessing' stamp: 'jcg 6/10/2006 07:37'!
localPosition: pval

	boundSphere localPosition: pval.
	self boundsChanged.
! !

!TSphere methodsFor: 'accessing' stamp: 'das 9/6/2006 19:51'!
pick: pointer

" We are here because we already picked the top level boundSphere, so we know that is true. We just need to figure out where we clicked on the sphere."
	^ pointer pickSphere: 0@0@0 radiusSquared: boundSphere radiusSquared.
! !

!TSphere methodsFor: 'accessing' stamp: 'jcg 6/10/2006 07:37'!
radius
	
	^ boundSphere radius! !

!TSphere methodsFor: 'accessing' stamp: 'jcg 6/10/2006 07:37'!
radius: rval

	boundSphere radius: rval.
	self boundsChanged.
! !

!TSphere methodsFor: 'accessing' stamp: 'jcg 6/10/2006 07:37'!
segments
	^ segments! !

!TSphere methodsFor: 'accessing' stamp: 'jcg 6/10/2006 07:39'!
segments: segs
	segments := segs.
	changed := true.! !

!TSphere methodsFor: 'initialize' stamp: 'ar 7/22/2006 19:49'!
updateMesh
	| position radius ringSin ringCos pi2 seg2 vert norm texCoord rts rbs rtc rbc ax u du v dv count stream |

	position := boundSphere localPosition.
	radius := boundSphere radius.
	seg2 := 1+ (segments * 2) .
	ringSin := FloatArray ofSize: seg2.
	ringCos := FloatArray ofSize: seg2.
	vert := Vector3 new.
	norm := Vector3 new.
	texCoord := Vector2 new.
	pi2 := Float pi *2.0.

	1 to: seg2-1 do:[ :index | 
		ax := ((index) * pi2)/ (seg2-1).
		ringSin at:index put: ax sin.
		ringCos at:index put: ax cos.].
	ringSin at: seg2 put: (ringSin at: 1).
	ringCos at: seg2 put: (ringCos at: 1).

	rts := 0.0.
	rtc := 1.0.
	rbs :=ringSin at: 1.
	rbc := ringCos at: 1.
	du := 1.0/(seg2-1).
	v := 0.0.
	dv := 1.0/segments.

	vertexBuffer := Vector3Array ofSize: (segments+1)*seg2.
	normalBuffer := Vector3Array ofSize: (segments+1)*seg2.
	texCoordBuffer := Vector2Array ofSize: (segments+1)*seg2.
	count := 1.

	"Create geometry."
	1 to: segments+1 do:[ :iv |
		u := 0.0.
		1 to: seg2 do: [ :ih | 
			norm x: rts*(ringSin at: ih) y: rtc z: rts*(ringCos at: ih).
			vert := (norm*radius) + position.
			texCoord u: u v: v.
			vertexBuffer at: count put: vert.
			normalBuffer at: count put: norm.
			texCoordBuffer at: count put: texCoord.
			count := count + 1.
			u:= u+du.].
		v := v+dv.
		rts := rbs.
		rtc := rbc.
		rbs := ringSin at: iv+1.
		rbc := ringCos at: iv+1.].

	"Create face groups."
	faceGroups := OrderedCollection new.
	0 to: segments-1 do: [:iv |
		faceGroups add: 1.	"material index"
		stream := WriteStream on: IntegerArray new.
		0 to: seg2-1 do: [:ih |
			stream nextPut: iv*seg2 + ih.
			stream nextPut: (iv+1)*seg2 + ih].
		faceGroups add: stream contents].

	self faceGroups: faceGroups.  "init faceBuffers"

	changed := false.
	

! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:54'!
boundSphere
	changed ifTrue: [self updateMesh].
	^ super boundSphere.! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'jcg 6/11/2006 17:03'!
boundsChanged
	super boundsChanged.

	"I'm not sure of the intended semantics of 'frameChanged'... TPrimitive set it in #boundsChanged, and I've decided to do the same here (it fixes a bug)."
	self flag: #jcg.
	frameChanged := true.

	changed := true.! !

!TUpdatingMesh methodsFor: 'initialize' stamp: 'das 7/4/2006 16:55'!
checkAlpha
	changed ifTrue: [self updateMesh].	"ensure that we have 'faceGroups' to check"
	super checkAlpha.! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:55'!
collapse
	changed ifTrue: [self updateMesh].
	super collapse! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:55'!
faceGroupsDo: aBlock
	changed ifTrue: [self updateMesh].
	super faceGroupsDo: aBlock
! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:55'!
frameBox
	changed ifTrue: [self updateMesh].
	^ super frameBox.! !

!TUpdatingMesh methodsFor: 'initialize' stamp: 'das 7/5/2006 14:31'!
initialize
	changed := false.
	super initialize.
	changed := true.
	boundsChanged := true.
! !

!TUpdatingMesh methodsFor: 'initialize' stamp: 'das 7/4/2006 16:55'!
initializeWithVertices: v alias: a norms: n textureUV: tuv faceGroups: fg material: ml mode: md 
	"Basically ignores the superclass implementation, since #update recomputes geometry."
	super 
		initializeWithVertices: v 
		alias: a 
		norms: n 
		textureUV: tuv 
		faceGroups: fg 
		material: ml 
		mode: md..
	self updateMesh.
	! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'bgf 7/8/2006 20:12'!
material: mat

	"Assign my material.    If the incoming material has a texture,
       use that as my texture.  Otherwise, if I already had a texture,
       use my old texture. "

	| txtr |

	"Hang on to the texture of my old material."
	txtr :=  (materialList at:1)texture.
	super material: mat copy.

	"If the new material didn't come with a texture, use the old one."
	(materialList at: 1) texture ifNil: [
		(materialList at: 1)texture: txtr.
	].
	self checkAlpha.! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/5/2006 14:10'!
materialAndTexture: mat

	super material: mat copy.
	self checkAlpha.
! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:55'!
normals
	changed ifTrue: [self updateMesh].
	^ super normals.! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:55'!
renderPrimitive: ogl alpha: alphaPass
	changed ifTrue: [self updateMesh].
	super renderPrimitive: ogl alpha: alphaPass! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:55'!
scale: scale

	super scale: scale.
	changed ifTrue: [self updateMesh].
! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:56'!
texCoords
	changed ifTrue: [self updateMesh.].
	^ super texCoords.! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/7/2006 20:29'!
texture

	^ (materialList at:1)texture.
! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/5/2006 10:24'!
texture: txtr

	(materialList at:1)texture: txtr.
! !

!TUpdatingMesh methodsFor: 'initialize' stamp: 'das 7/4/2006 16:54'!
updateMesh
	"Subclass should override this.  Make sure that the statement 'changed := false' appears at the end of the update."
	self subclassResponsibility.! !

!TUpdatingMesh methodsFor: 'initialize' stamp: 'das 7/4/2006 16:56'!
validateFaceGroups
	changed ifTrue: [self updateMesh].
	super validateFaceGroups.! !

!TUpdatingMesh methodsFor: 'accessing' stamp: 'das 7/4/2006 16:56'!
vertices
	changed ifTrue: [self updateMesh].
	^ super vertices.! !

TFrame subclass: #TOctree
	instanceVariableNames: 'subTrees'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TOctree methodsFor: 'initialize' stamp: 'das 8/1/2006 17:50'!
vertices: verts faces: faces

	| box |
	box := TBox new.
	verts do:[:v | box growVertex:v].! !

!TOctree methodsFor: 'initialize' stamp: 'das 8/1/2006 17:52'!
vertices: verts faces: faces depth: d

	| box |
	box := TBox new.
	verts do:[:v | box growVertex:v].
	1 to: faces size by:3 do:[: findex |
		
	].
	
	! !

TFrame subclass: #TParticle
	instanceVariableNames: 'size pPosition pVelocity pAcceleration pLifetime pColor pDeltaColor positionRange velocityRange accelerationRange lifetimeRange maxTime lastStepTime material boundsChanged boundSphere boundBox deltaTime billBoard startColor endColor tar tvr lengthArray lastGlobalPosition attrArray tpr pLastPosition globalVelocity attractors doStep'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TParticle commentStamp: '<historical>' prior: 0!
TParticle

This is a very simple particle system.

size - the length or the particle array, determines max number of particles.
pPosition - B3DVector3Array of particle positions
pVelocity - B3DVector3Array of particle velocities
pAcceleration - B3DVector3Array of particle accelerations
pLifetime - number of seconds (floating point) of life that each particle has left
positionRange - the TBox within which new particles get created (see TBox >> #atRandom:)
velocityRange - the 3D velocity range within which the new particles start life.
accelerationRange - the 3D acceleration range applied to each particle on creation.
lifetimeRange - the lifespan range of each particle.
lastTime - for #the stepAt: method, used to determine time between this and the last cycle.
material - what color are the particles.

To use the particle system you do the following:

	ps _ TParticle initialize: ogl size: 1000
	ps setPositionRangeMin:(-0.1@-0.1@-0.1) max: (0.1@0.1@0.1).
	ps setVelocityRangeMin:(-1.2@6.4@-1.2) max:(1.2@9.6@1.2).
	ps setAccelerationRangeMin:(0@-10@0) max:(0@-8@0).
	ps setLifetimeRange: (1500 to: 2000).

These values are called inside of the initialize method, but this is how you would set your one.

If you want the particle system to move, call  self startStepping.

DAS!

!TParticle class methodsFor: 'instance creation' stamp: 'ar 9/30/2004 12:39'!
initializeWithSize: sz
	^ self new initializeWithSize: sz! !

!TParticle methodsFor: 'accessing' stamp: 'das 10/16/2005 23:10'!
addAttractor: attr

	attractors add: attr.	
	attrArray ifNil:[
		attrArray := Vector3Array new: size.
		lengthArray := FloatArray new: size.
		].
! !

!TParticle methodsFor: 'accessing' stamp: 'das 2/10/2003 12:26'!
billBoard: bool

	billBoard := bool.! !

!TParticle methodsFor: 'initialize' stamp: 'DPR 1/9/2004 14:05'!
boundSphere
	| vbox bb |
	boundsChanged ifTrue:[
		boundBox := positionRange copy.
		vbox := velocityRange copy.
		lifetimeRange max timesRepeat:[
			bb := boundBox addBox: (vbox * 0.001).
			boundBox := boundBox unionBox: bb.
			vbox := vbox addBox: (accelerationRange * 0.001).
			].
		boundSphere := TBoundSphere localPosition: boundBox center radius: (boundBox diagonal/2.0).
		boundSphere frame: self.
		boundsChanged := false.
		].
	^ boundSphere.! !

!TParticle methodsFor: 'initialize' stamp: 'das 2/10/2003 06:49'!
boundsChanged

	boundsChanged := true.! !

!TParticle methodsFor: 'render' stamp: 'das 10/16/2005 23:34'!
determinePosition
	| msecs tt |
	msecs := (tt := self now asFloat) - (lastStepTime asFloat).
	msecs = 0 ifTrue: [^self].
	deltaTime := msecs.
	lastStepTime := tt.
	lastGlobalPosition ifNotNil:[
		globalVelocity := (self globalPosition - lastGlobalPosition)/msecs * 1000.0.
		].
	lastGlobalPosition := self globalPosition.
! !

!TParticle methodsFor: 'stepping' stamp: 'das 3/9/2004 14:06'!
doAttractors: secs

	| loc v n |

	attractors do:[:attr |  
		attr hard ifTrue:[
			loc:=attr globalPosition.
			1 to: size do:[:index | attrArray at: index put: loc.].
			attrArray -= pLastPosition.
			attrArray squaredLengthInto: lengthArray.
			attrArray divideByArray: (lengthArray sqrted).
			1 to: size do:[:index | 
				(lengthArray at: index)<attr radiusSquared ifTrue:[ 
					v := pVelocity at: index.			
					n := attrArray at: index.
					pVelocity at:index put: v-(2*n*(v dot: n)).
					attrArray at: index put: (0.0@0.0@0.0). "This sets force inside sphere to 0"
				].
			].
			attrArray divideByArray: lengthArray.
			attrArray *= (attr mass*secs).
			pVelocity += attrArray.
		]ifFalse:[
			loc:=attr globalPosition.
			1 to: size do:[:index | attrArray at: index put: loc.].
			attrArray -= pLastPosition.
			attrArray squaredLengthInto: lengthArray.
			attrArray divideByArray: (lengthArray sqrted).
			1 to: size do:[:index | 
				(lengthArray at: index)<attr radiusSquared ifTrue:[ 
				((lengthArray at: index))<(attr radiusSquared) ifTrue:[pLifetime at: index put:-1.]
				ifFalse:[
					v := pVelocity at: index.			
					n := attrArray at: index.
					pLastPosition at:index put: loc-(n*attr radius).
					attrArray at: index put: (0.0@0.0@0.0). 
				].
				].
			].
			attrArray divideByArray: lengthArray.
			attrArray *= (attr mass*secs).
			pVelocity += attrArray.
		].
	].
! !

!TParticle methodsFor: 'render' stamp: 'das 7/30/2006 18:31'!
doRenderColor: ogl
	
	| secs transfer go |

	doStep ifFalse:[^ self].

	material enable: ogl.

	ogl glPointSize: 2.0. 
	ogl glPushMatrix.
	ogl glMultMatrixf: self globalTransform orthoNormInverse transposed.
	pLifetime -= deltaTime.
	secs := deltaTime/1000.0.
	secs := secs min: 0.1.
	attractors size > 0 ifTrue:[self doAttractors: secs].
	pPosition replaceFrom: 1 to: pPosition size with: pLastPosition startingAt: 1.
	pPosition += (pVelocity * secs).
	pVelocity += (pAcceleration * secs).
	go := self globalOrientation.
	tpr := go localBoxToGlobal: positionRange.
	tvr := go localBoxToGlobal: velocityRange.
	tar := go localBoxToGlobal: accelerationRange.
	pColor -= pDeltaColor*deltaTime.
	ogl glLineWidth: 2.0.
	ogl glBegin: GLLines.
	1 to: size do:[:index |
		ogl	glColor3fv: (pColor at: index);
			glVertex3fv: (pPosition at: index);
			glVertex3fv:(pLastPosition at: index).
		].
	ogl glEnd.
"Transcript show: (self globalOrientation orthoNormInverse localPointToGlobal:globalVelocity); cr.
"	ogl glLineWidth: 2.0.
	material disable: ogl.
	transfer := pPosition.
	pPosition := pLastPosition.
	pLastPosition := transfer.
	ogl glPopMatrix.
! !

!TParticle methodsFor: 'render' stamp: 'das 7/30/2006 18:31'!
doRenderMonochrome: ogl
	
	doStep ifFalse:[^ self].

	material enable: ogl.

"	ogl glPointSize: 2.0. "
"	ogl	glBegin: GLPoints."
	ogl glPushMatrix.
	ogl glMultMatrixf: self globalTransform orthoNormInverse transposed.
"	ogl glLineWidth: 2.0."
	material enable: ogl.
	ogl glColor3fv: material ambientColor.
ogl glBegin: GLLines.
	1 to: size do:[:index |
		ogl	
			glVertex3fv: (pPosition at: index);
			glVertex3fv:(pLastPosition at: index).
		].
ogl glEnd.
	material disable: ogl.
	ogl glPopMatrix.
! !

!TParticle methodsFor: 'accessing' stamp: 'das 7/30/2006 18:31'!
doStep

	^ doStep.! !

!TParticle methodsFor: 'accessing' stamp: 'das 2/10/2003 23:08'!
endColor

	^ endColor.! !

!TParticle methodsFor: 'accessing' stamp: 'ar 2/7/2006 17:45'!
endColor: col

"This requires either a Vector4 or a Color4 "
	endColor := col asVectorColor.
	self updateVars.! !

!TParticle methodsFor: 'render' stamp: 'das 6/5/2003 00:10'!
hasAlpha

	^ true.! !

!TParticle methodsFor: 'accessing' stamp: 'das 3/9/2004 14:11'!
hasAttractor: attr

	attractors do:[:a | a = attr ifTrue:[^ true]].	
	^ false.! !

!TParticle methodsFor: 'initialize' stamp: 'das 7/30/2006 18:29'!
initialize

	super initialize.
	doStep := false.! !

!TParticle methodsFor: 'initialize' stamp: 'das 7/30/2006 18:28'!
initializeWithSize: sz

	size := sz.
	pPosition := Vector3Array new: sz.
	pLastPosition := Vector3Array new: sz.
	pVelocity := Vector3Array new: sz.
	pAcceleration := Vector3Array new: sz.
	pLifetime := FloatArray new: sz.
	pColor := nil.
	pDeltaColor := nil.
	startColor := VectorColor r:1.0  g:1.0 b:1.0 a:0.9.
	endColor := nil. "default is to not change the color of particles over their lifetimes."
	material := TMaterial new.
	material ambientColor: startColor.
	material diffuseColor: startColor.
	material fullBright: true.

	self setPositionRangeMin:(-0.1@-0.1@-0.1) max: (0.1@0.1@0.1).
	self setVelocityRangeMin:(-1.2@6.4@-1.2) max:(1.2@9.6@1.2).
	self setAccelerationRangeMin:(0@-10@0) max:(0@-8@0).
	self setLifetimeRange: (1000 to: 1500).
	maxTime := lifetimeRange max.

	deltaTime := 0.
	billBoard := true.
	self updateVars.
	self boundsChanged.
	globalVelocity := 0@0@0.
	lastStepTime := 0.0.
	lastGlobalPosition := nil.
	attractors := OrderedCollection new.
	^self
! !

!TParticle methodsFor: 'accessing' stamp: 'das 12/21/2002 22:15'!
material

	^ material.! !

!TParticle methodsFor: 'accessing' stamp: 'das 12/21/2002 22:14'!
material: mat

	material := mat.! !

!TParticle methodsFor: 'render' stamp: 'ar 2/24/2006 11:29'!
regenerate: index

	| lt |
	lt := lifetimeRange atRandom asFloat.
	pLifetime at: index put: lt.
	pPosition at: index put: (tpr atRandom: 1000)+self globalPosition.
	pLastPosition at: index put: (pPosition at: index).
	pVelocity at: index put: (tvr atRandom: 1000)+globalVelocity.
	pAcceleration at: index put: (tar atRandom: 1000).
	pColor ifNotNil:[
		pColor at: index put: startColor.
		pDeltaColor at: index put: (startColor - endColor)/lt.
		].

"
	a := Float pi*(500-(1000 atRandom))/500.0.

	x := a cos.
	z := a sin.

	a := Float pi*(500-(1000 atRandom))/1000.0.
	x2 := a cos.
	gp := (attractors at: 1) globalPosition.
	r:=1.0+(attractors at:1)radius.
	x:= x*x2*r.
	y := a sin*r.
	z:= z*x2*r.
	pPosition at: index put:gp+(x@y@z).
"
	! !

!TParticle methodsFor: 'accessing' stamp: 'das 3/9/2004 14:13'!
removeAttractor: attr

	attractors remove: attr.	
! !

!TParticle methodsFor: 'render' stamp: 'das 1/31/2006 20:54'!
render: space


	self hasAlpha ifFalse:[
		pColor
			ifNil:[self doRenderMonochrome: space]
			ifNotNil:[self doRenderColor: space]
		].
	
! !

!TParticle methodsFor: 'render' stamp: 'das 1/31/2006 20:57'!
renderAlpha: ogl


	self determinePosition.	
	self doRenderMonochrome: ogl
	! !

!TParticle methodsFor: 'accessing' stamp: 'das 12/21/2002 21:56'!
setAccelerationRangeMin: min max: max

	accelerationRange := TBox min: min max: max.! !

!TParticle methodsFor: 'accessing' stamp: 'das 2/12/2003 23:32'!
setLifetimeFrom: from to: to

	lifetimeRange := from to: to.
	1 to: size do:[:index | pLifetime at: index put: (lifetimeRange max atRandom).].
	self updateVars.
! !

!TParticle methodsFor: 'accessing' stamp: 'das 2/12/2003 23:32'!
setLifetimeRange: range

	lifetimeRange := range.
	1 to: size do:[:index | pLifetime at: index put: (lifetimeRange max atRandom).].
	self updateVars.
! !

!TParticle methodsFor: 'accessing' stamp: 'das 12/21/2002 21:56'!
setPositionRangeMin: min max: max

	positionRange := TBox min: min max: max.! !

!TParticle methodsFor: 'accessing' stamp: 'das 12/21/2002 21:56'!
setVelocityRangeMin: min max: max

	velocityRange := TBox min: min max: max.! !

!TParticle methodsFor: 'stepping' stamp: 'das 8/14/2006 12:34'!
start

	doStep ifFalse:[
		doStep := true.
		self step.
	].! !

!TParticle methodsFor: 'accessing' stamp: 'das 2/10/2003 23:08'!
startColor

	^ startColor.! !

!TParticle methodsFor: 'accessing' stamp: 'ar 2/7/2006 17:45'!
startColor: col

"This requires either a Vector4 or a Color4 "
	startColor := col asVectorColor.
	self updateVars.! !

!TParticle methodsFor: 'stepping' stamp: 'das 8/14/2006 12:33'!
step
	| secs go transfer |

	doStep ifTrue:[
		self determinePosition.
		deltaTime := 60.0.
		lastGlobalPosition ifNotNil:[
			globalVelocity := (self globalPosition - lastGlobalPosition)/deltaTime*1000.0.
			].
		lastGlobalPosition := self globalPosition.

		pLifetime -= deltaTime.
		secs := deltaTime/1000.0.
		secs := secs min: 0.1.
		attractors size > 0 ifTrue:[self doAttractors: secs].
		pPosition replaceFrom: 1 to: pPosition size with: pLastPosition startingAt: 1.
		pPosition += (pVelocity * secs).
		pVelocity += (pAcceleration * secs).
		pColor ifNotNil:[	pColor -= (pDeltaColor*deltaTime).].
		go := self globalOrientation.
		tpr := go localBoxToGlobal: positionRange.
		tvr := go localBoxToGlobal: velocityRange.
		tar := go localBoxToGlobal: accelerationRange.

		1 to: size do:[:index |
			(pLifetime at: index) < 0.0 ifTrue:[self regenerate: index.].
			].
		transfer := pPosition.
		pPosition := pLastPosition.
		pLastPosition := transfer.
		(self future: self stepTime) step.
	].
	
! !

!TParticle methodsFor: 'stepping' stamp: 'das 7/30/2006 18:29'!
stop

	doStep := false.
! !

!TParticle methodsFor: 'accessing' stamp: 'das 10/16/2005 23:19'!
updateVars

	| |
	maxTime := lifetimeRange max.
	endColor ifNotNil:[
		pColor ifNil:[ pColor := VectorColor4Array new: size.
					pDeltaColor := VectorColor4Array new: size.
				].
		].
	material ambientColor: startColor.
	material diffuseColor: startColor.
! !

TParticle subclass: #TParticleTxtr
	instanceVariableNames: 'textures texture pTexture rectangle'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TParticleTxtr commentStamp: 'das 10/10/2004 12:57' prior: 0!
TParticleTxtr is a textured particle.

DAS!

!TParticleTxtr methodsFor: 'initialize' stamp: 'das 7/5/2006 13:25'!
addTexture: txtr

	texture := txtr.
	rectangle := TRectangle new extent: txtr extent.
"	pTexture ifNil:[ textures := OrderedCollection new.
					pTexture := OrderedCollection ofSize: size.
					1 to: size do:[:index | pTexture at: index put: txtr.].
					].
	textures add: txtr."! !

!TParticleTxtr methodsFor: 'render' stamp: 'das 1/31/2006 21:29'!
doRenderColor: ogl
	| mat yaw |

	ogl glMultMatrixf: self globalTransform orthoNormInverse transposed.
	mat := Matrix4x4 identity.
	material enable: ogl.
	texture enable: ogl.
	billBoard ifTrue:[
		yaw := ogl camera globalYaw.
		mat rotationAroundY: yaw.
		].
	1 to: size do:[:index |
"		pt := pTexture at: index."
		mat translation: (pPosition at: index).
		ogl glPushMatrix. 
		ogl glMultMatrixf: mat transposed. 
		ogl glColor4fv: (pColor at: index).
		rectangle render: ogl.
		ogl glPopMatrix.
	].
	texture disable: ogl.
	material disable: ogl.
! !

!TParticleTxtr methodsFor: 'render' stamp: 'das 1/31/2006 20:59'!
doRenderColorX: ogl
	| mat yaw |

	mat := Matrix4x4 identity.
	material enable: ogl.
	texture enable: ogl.
	billBoard ifTrue:[
		yaw := self globalYaw - ogl camera globalYaw.
		mat rotationAroundY: yaw negated.
		].
	1 to: size do:[:index |
"		pt := pTexture at: index."
		mat translation: (pPosition at: index).
		ogl glPushMatrix. 
		ogl glMultMatrixf: mat transposed. 
		ogl glColor4fv: (pColor at: index).

		rectangle render: ogl.
		ogl glPopMatrix.
	].
	texture disable: ogl.
	material disable: ogl.
! !

!TParticleTxtr methodsFor: 'render' stamp: 'ar 3/8/2006 19:02'!
doRenderMonochrome: ogl
	| pt mat yaw  |

	endColor ifNotNil:[^ self doRenderColor: ogl].

	mat := Matrix4x4 identity.
	material enable: ogl.
	billBoard ifTrue:[
		yaw := self globalYaw - ogl camera globalYaw.
		mat rotationAroundY: yaw negated.
		].
	1 to: size do:[:index |
		pt := pTexture at: index.

		mat translation: (pPosition at: index).
		ogl glPushMatrix; 
			glMultMatrixf: mat transposed. 
		pt renderPrimitive: ogl.
		ogl glPopMatrix.
		].
	material disable: ogl.
! !

!TParticleTxtr methodsFor: 'render' stamp: 'das 3/9/2004 14:15'!
hasAlpha

	^ true.! !

TFrame subclass: #TPortal3D
	instanceVariableNames: 'postcard inPortal scale transform clipBox boundSphere boundsChanged cube blocked'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TPortal3D commentStamp: 'das 6/5/2006 20:35' prior: 0!
TPortal3D are a kind of 3D portal. They render a miniature version of a space (or a subset of it). The content of this portal can be manipulated normally.


toSpace - the space we are rendering inside of the TPortal3D
inPortal - used to avoid recursion for now. At some point, we can use this as a recursion counter instead.
scale - scale at which we render the contents of the portal. This can be anything, even a number larger than 1.0. It is initially 0.05.
scaleInverse - the 1.0/scale inverse saved for performance.
transform - the offset position and orientation of the space that we render from. That is, this offset value represents where in the space we render at the center point of the TPortal3D
clipBox - the five sided box that contains the TPortal3D and clips it. I would use six sides, but I need one for the 2D portals. 
boundSphere - the usual suspect
bounsChanged
cube - a subtle cube mapped over the TPortal3D to indicate its presence. 

DAS!

!TPortal3D methodsFor: 'accessing' stamp: 'das 5/28/2006 20:09'!
blocked: bool

	blocked := bool.! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 6/19/2003 22:50'!
boundSphere

	boundsChanged ifTrue:[
		self initBounds. 
		boundsChanged := false.
		].
	^ boundSphere.! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 7/22/2003 20:54'!
boundingBox

	^ clipBox.
! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 5/28/2006 17:26'!
colorize: col

"do nothing"! !

!TPortal3D methodsFor: 'render' stamp: 'bgf 8/3/2006 19:27'!
disableClipPlanes: ogl

	ogl inPortal2D ifFalse:[
		ogl glDisable: GLClipPlane0.
	].
	ogl glDisable: GLClipPlane1.
	ogl glDisable: GLClipPlane2.
	ogl glDisable: GLClipPlane3.
	ogl glDisable: GLClipPlane4.
	ogl glDisable: GLClipPlane5.
! !

!TPortal3D methodsFor: 'render' stamp: 'bgf 8/3/2006 19:27'!
enableClipPlanes: ogl
	
"------ enableClipPlanes is used to clip the TSpace inside the 3DPortal to a sub-cube inside the portal. This ensures that objects don't get rendered outside of the portal, which would look bad and be confusing.------"
	| equation |

"------ We need to do it this way, because Squeak does not directly support doubles and gllClipPlane requires an array of same. ------"

	equation := ExternalData fromHandle: (ExternalAddress allocate: 8*4) type:ExternalType double.
	
	equation getHandle doubleAt: 1 put:  1.0.
	equation getHandle doubleAt: 9 put: 0.0.
	equation getHandle doubleAt: 17 put: 0.0.
	equation getHandle doubleAt: 25 put: clipBox max x.
	ogl glClipPlane: GLClipPlane1 with: equation.
	ogl glEnable: GLClipPlane1.

	equation getHandle doubleAt: 1 put: -1.0.
	equation getHandle doubleAt: 9 put: 0.0.
	equation getHandle doubleAt: 17 put: 0.0.
	equation getHandle doubleAt: 25 put: clipBox min x negated.
	ogl glClipPlane: GLClipPlane2 with: equation.
	ogl glEnable: GLClipPlane2.

	equation getHandle doubleAt: 1 put: 0.0.
	equation getHandle doubleAt: 9 put: 0.0.
	equation getHandle doubleAt: 17 put: 1.0.
	equation getHandle doubleAt: 25 put: clipBox max z.
	ogl glClipPlane: GLClipPlane3 with: equation.
	ogl glEnable: GLClipPlane3.

	equation getHandle doubleAt: 1 put: 0.0.
	equation getHandle doubleAt: 9 put: 0.0.
	equation getHandle doubleAt: 17 put: -1.0.
	equation getHandle doubleAt: 25 put: clipBox min z negated.
	ogl glClipPlane: GLClipPlane4 with: equation.
	ogl glEnable: GLClipPlane4.

	equation getHandle doubleAt: 1 put: 0.0.
	equation getHandle doubleAt: 9 put: -1.0.
	equation getHandle doubleAt: 17 put: 0.0.
	equation getHandle doubleAt: 25 put: clipBox min y negated.
	ogl glClipPlane: GLClipPlane5 with: equation.
	ogl glEnable: GLClipPlane5.

	"Use clip plane 0 for the bottom, but only if it's available."
	ogl inPortal2D ifFalse:[
		equation getHandle doubleAt: 1 put: 0.0.
		equation getHandle doubleAt: 9 put: 1.0.
		equation getHandle doubleAt: 17 put: 0.0.
		equation getHandle doubleAt: 25 put: clipBox max y.
		ogl glClipPlane: GLClipPlane0 with: equation.
		ogl glEnable: GLClipPlane0.
	].

	equation free.! !

!TPortal3D methodsFor: 'events' stamp: 'das 3/21/2006 15:17'!
eventMask

	^ EventKeyboard bitOr:(EventPointerDown bitOr: EventPointerOver).! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 7/24/2003 14:50'!
extent

	| ext |
	ext := clipBox extent.
	^ ext x@ ext y.
! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 3/26/2006 13:16'!
extent: ext

	| clip clipN |
	clip := ext x/2.0.
	clipN := clip negated.
	self translation:( Vector3 x:0 y:0 z:0 ).
	clipBox := TBox min: (Vector3 x:clipN y:clipN z:clipN) max: (Vector3 x:clip y: clip z: clip).
	cube extent: clipBox extent.
	! !

!TPortal3D methodsFor: 'transform' stamp: 'das 5/28/2006 17:50'!
globalTransformUpdate

 	cube globalTransform: self globalTransform.
! !

!TPortal3D methodsFor: 'render' stamp: 'das 3/27/2006 11:26'!
hasAlpha

	^ true.! !

!TPortal3D methodsFor: 'initialize' stamp: 'das 5/28/2006 17:35'!
initBounds

	boundSphere := TBoundSphere localPosition: clipBox center radius: (clipBox diagonal)/2.
	boundSphere frame: self.
	cube initBounds.

! !

!TPortal3D methodsFor: 'initialize' stamp: 'das 7/5/2006 13:21'!
initialize

	| txtr mat |
	super initialize.
	inPortal := false.
	cube := TCube new.
"	self addChild: cube."
	cube solid: false.
	txtr := TTexture
				new initializeWithFileName: 'box.png'
				mipmap: true
				shrinkFit: false.

	mat := TMaterial new.
	mat ambientColor: #(0.5 0.7 0.7 0.1).
	mat diffuseColor: #(0.5 0.7 0.7 0.1).
	mat cullFace: false.
	cube material: mat.
	cube texture: txtr.
	scale := 0.1. "0.05."
	self objectName:'Portal3D'.
	boundsChanged := true.
	transform := Matrix4x4 identity.
	self extent:(Vector3 x:4 y:4 z:4).
	blocked := false.
	^self
! !

!TPortal3D methodsFor: 'testing' stamp: 'das 5/28/2006 16:58'!
isNotPortal3D

	^ false.! !

!TPortal3D methodsFor: 'testing' stamp: 'das 5/28/2006 16:48'!
isPortal

	^ true.! !

!TPortal3D methodsFor: 'testing' stamp: 'das 5/28/2006 16:55'!
isPortal3D

	^ true.! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 5/28/2006 17:25'!
material: mat

"do nothing"! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 7/18/2006 19:20'!
offset 


	^(transform orientation localPointToGlobal:transform translation)negated.
! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 7/18/2006 19:20'!
offset: os

	| offset |
	offset := transform orientation orthoNormInverse localPointToGlobal:os negated.
	transform translation: offset.
! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 5/28/2006 16:28'!
postcardLink
	
	^ postcard.
! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 3/20/2006 18:37'!
postcardLink: pcl

	postcard := pcl.! !

!TPortal3D methodsFor: 'render' stamp: 'das 7/19/2006 13:30'!
render: ogl
| rval ac currentCameraTrans currentCameraPos cpt toSpace currentPointerTrans currentScale copySelection currentPointer saveEventPointer ttrans offset |

	rval := 0.

"Can't recurse into TPortal3D objects because of the limitation in number of clipping planes. Maybe if I figure out a different way to do it later..."
	ogl inPortal3D ifTrue:[^ rval].
	postcard ifNil:[^ 0 ].
	toSpace := postcard resolveViewpoint: ogl.
	toSpace ifNil:[^ 0 ].
	inPortal ifTrue:[^ 0 ].
	ogl inPortal3D: true.
	inPortal := true.
	ogl glPushMatrix. 
	self enableClipPlanes: ogl.
	ac := ogl camera.
"	ip := ac inPortal.
	ac inPortal: true."
	cpt := ac clipPlanesTransform.
	currentCameraTrans := ac globalTransform copy.
	currentCameraPos := ac globalPosition copy.
	currentScale := ac frameScale.
	ttrans := transform copy.
	offset := ttrans translation.
	ttrans translation: offset * scale.
	ac  globalTransform: (self globalTransform *  ttrans orientation * ttrans translationMatrix)orthoNormInverse * ac globalTransform.
	ac globalPositionForce: ac globalPosition/scale.
	ac transformClipPlanes.
	currentPointer := ogl eventPointer.
	currentPointer ifNotNil:[
		copySelection := currentPointer selection copy.
		currentPointerTrans := currentPointer globalTransform copy.
		ogl eventPointer globalTransform: (ac globalTransform composeWith: currentPointer localTransform).
		currentPointer frameScale: scale.
	"	ac downPointer setAutomatic: false."
	].
	

	ac frameScale: scale.
	ogl glScalef: scale with: scale with: scale.
	ogl glMultTransposeMatrixf: transform orientation. 
	ogl glTranslatef: offset x with: offset y with: offset z.
	ogl forcePick ifTrue:[saveEventPointer := ogl eventPointer. ogl eventPointer:nil].
	toSpace renderSubspace: ogl. ">>>>>>>>>>>>RENDER CALL IS HERE!!!!!!!!!!<<<<<<<<<<"
	ogl forcePick ifTrue:[ogl eventPointer:saveEventPointer].
	ac globalTransform: currentCameraTrans.
	ac globalPositionForce: currentCameraPos.
	ac clipPlanesTransform: cpt.
	self disableClipPlanes: ogl.
	ogl glPopMatrix.

	ogl inPortal3D: false.
	ac frameScale: currentScale.
	inPortal := false.
" XXXXXXXXXXXXXXXX
Need a way to figure out if this is a viewPortal or not. If so do NOT do this 

	ogl camera currentPortal ifNotNil:[
		ogl camera currentPortal enableClipPlane: ogl.]. 
 XXXXXXXXXXXXXXXXX"
	currentPointer ifNotNil:[
		currentPointer globalTransform: currentPointerTrans.
		currentPointer frameScale: currentScale.
		copySelection target ~= currentPointer selection target ifTrue:[
			currentPointer pointerDown ifFalse:[
			(clipBox pointInBox: currentPointer selectedGlobalPoint+offset*scale)ifFalse:[
				"selected point ain't in the box"
				currentPointer selection: copySelection.
			].
			].
			"ac downPointer setAutomatic: true."
		].
	].
"	ac inPortal: ip."
	
	^ rval.! !

!TPortal3D methodsFor: 'render' stamp: 'das 5/30/2006 02:35'!
renderAlpha: ogl

	cube renderAlpha: ogl.! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 5/31/2003 09:10'!
scale

	^ scale.! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 3/20/2006 22:01'!
scale: scl

	scale := scl.
! !

!TPortal3D methodsFor: 'enterExit' stamp: 'das 9/6/2006 01:29'!
testEnter: ogl avatar: avatar

	| pc vwPnt trans otrans ctrans ttrans relativeTrans |
"this can only determine if a portal transition needs to occur. You can't actually do the portal transition here, because we are actually inside of the portal's Island. The actual transition is handled by the CroquetHarness if the newPostcard field of this object is not nil."
	blocked ifTrue:[ ^false ].
	pc := postcard.
	pc ifNil:[ ^false ].
	self = pc ifTrue:[ ^false ].
	vwPnt := pc resolveViewpoint: ogl.	
	vwPnt ifNil:[ ^ false. ].
	avatar portalRay doSelect ifFalse:[ ^ false. ].
	cube test: #cubeTest.
	vwPnt send:[ :toPortal |
		toPortal root ifNil:[^ false]. 
		(avatar portalRay pointerPick: cube boundSphere) ifTrue:[
			ttrans := transform copy.
			ttrans translation: ttrans translation * scale.
			otrans:=(self globalTransform * ttrans orientation * ttrans translationMatrix)orthoNormInverse.
			trans := otrans * avatar globalTransform.
			trans translation: trans translation/scale.
			avatar newTransform: trans.
relativeTrans := ogl camera relativeTransform: avatar.
			ctrans := otrans * ogl camera globalTransform.
			ctrans translation: ctrans translation/scale.
			avatar jumpCameraTrans: ctrans.
			ogl camera globalTransform: ctrans.
			ogl camera  globalPositionForce: ogl camera globalPosition.
avatar newTransform: ogl camera globalTransform * relativeTrans.
			toPortal root send:[:sp| avatar newPostcard: sp postcard.].
			^ true.
		].
	].
	^ false.! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 6/6/2006 13:12'!
transform

	^ transform.! !

!TPortal3D methodsFor: 'accessing' stamp: 'das 6/6/2006 13:12'!
transform: trans

	transform := trans.! !

TFrame subclass: #TPrimitive
	instanceVariableNames: 'texture material boundsChanged boundSphere'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TPrimitive commentStamp: 'das 10/10/2004 13:32' prior: 0!
TPrimitive is the base class of a lot of the simple classes such as TCube, TSphere, TRectangle. It is primarily used to help manage graphics caching.

texture - texture applied to the opaque object
textureAlpha - texture applied to the translucent object
material - material applied to the opaque object
materialAlpha - material applied to the alpha object
boundsChanged - if the boundSphere needs to be recomputed
boundSphere - minimal sphere that contains the entire object
glListID - OpenGL ID of the cached display list
oglInstance - the OpenGL instance value
cachingEnabled - flag indicating same

DAS
!

TPrimitive subclass: #TMeshSimple
	instanceVariableNames: 'mode vertices faces normals colors texCoords pointSize box boundsDepth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TMeshSimple methodsFor: 'accessing' stamp: 'das 7/1/2006 11:07'!
boundsDepth: bd

	boundsDepth := bd.
	self boundsChanged.
! !

!TMeshSimple methodsFor: 'initialize' stamp: 'das 6/30/2006 08:08'!
clear

	vertices := nil.
	normals := nil.
	colors := nil.! !

!TMeshSimple methodsFor: 'accessing' stamp: 'das 6/30/2006 08:08'!
colors: anObject
	"Set the value of colors"

	colors := anObject! !

!TMeshSimple methodsFor: 'accessing' stamp: 'das 6/30/2006 08:13'!
faces: f

	faces := f.! !

!TMeshSimple methodsFor: 'initialize' stamp: 'das 8/1/2006 17:44'!
initBounds

	boundsDepth = 0 
		ifTrue:[boundSphere := TBoundSphere mtfBall: vertices. ]
		ifFalse:[
			boundSphere := TBoundSphere  calcTree: vertices 
											faces: faces
											depth: boundsDepth. 
		].
	boundSphere frame: self.
	boundsChanged := false.! !

!TMeshSimple methodsFor: 'initialize' stamp: 'das 7/1/2006 11:07'!
initialize

	super initialize.

	self clear.

	pointSize := 2.
	mode := GLTriangles.
	boundsDepth := 0.! !

!TMeshSimple methodsFor: 'initialize' stamp: 'das 6/30/2006 08:08'!
normalizeDown: aFloatArray

	"assumes values >= 0.
     scales the array s.t. the largest value is 1.0"

	| biggest targ |

	targ := 0.0.

	aFloatArray size < 1 ifTrue: [^nil].

	biggest := 1.0.

	aFloatArray do: [:n|
		n < biggest ifTrue: [
			biggest := n.
			].
		].

   biggest = 0.0 ifTrue: [
		1 to: aFloatArray size do: [:n|
			aFloatArray at: n put: 0.0.
			].		
		]
	ifFalse: [
		1 to: aFloatArray size do: [:n|
			aFloatArray at: n put: ((aFloatArray at: n) + (targ-biggest)).
			].
"		aFloatArray *= (1.0 / biggest)."
		].


	! !

!TMeshSimple methodsFor: 'initialize' stamp: 'das 6/30/2006 08:08'!
normalizeUp: aFloatArray

	"assumes values >= 0.
     scales the array s.t. the largest value is 1.0"

	| biggest targ |

	targ := 1.0.

	aFloatArray size < 1 ifTrue: [^nil].

	biggest := 0.0.

	aFloatArray do: [:n|
		n > biggest ifTrue: [
			biggest := n.
			].
		].

   biggest = 0.0 ifTrue: [
		1 to: aFloatArray size do: [:n|
			aFloatArray at: n put: 0.0.
			].		
		]
	ifFalse: [
		1 to: aFloatArray size do: [:n|
			aFloatArray at: n put: ((aFloatArray at: n) + (targ-biggest)).
			].
"		aFloatArray *= (1.0 / biggest)."
		].


	! !

!TMeshSimple methodsFor: 'accessing' stamp: 'das 6/30/2006 08:08'!
normals: x

	normals := x.! !

!TMeshSimple methodsFor: 'render' stamp: 'das 7/1/2006 11:13'!
pick: pointer

" We are here because we already picked the top level boundSphere, so we know that is true. Now what we need to test against the hierarchy, which will return a list of faces if true - otherwise, it will return nil."

	^ boundSphere pickChildren: pointer.! !

!TMeshSimple methodsFor: 'accessing' stamp: 'das 6/30/2006 08:08'!
pointSize: x
	pointSize := x.
"	self resetCaching."! !

!TMeshSimple methodsFor: 'initialize' stamp: 'das 6/30/2006 08:08'!
provideColors

	"make an array of colors to match the vertices"

	| c o blorp |
	self colors: (Vector3Array new: vertices size withAll: #(1 0 0)asFloatArray).

	c := Vector4 new.
	
	1 to: vertices size do: [:n|
		blorp := (vertices at: n) z > 0.
		blorp := false not.
		blorp ifTrue: [
		    o := Vector4 x: 0.6 y: 0.6 z: 0.6 w: 0.0.
			c := (vertices at: n) normalized + o.
			self normalizeUp: c.
			]
		ifFalse: [
		    o := Vector4 x: 0.0 y: 0.0 z: 0.0 w: 0.0.
			c := (vertices at: n) normalized + o.
			self normalizeDown: c.
			].
		colors at: n put: c.
		].


	"self resetCaching."! !

!TMeshSimple methodsFor: 'render' stamp: 'das 6/30/2006 12:00'!
renderPrimitive: ogl

	vertices ifNil: [^super renderPrimitive: ogl].

	normals ifNil: [
		ogl glDisable: GLLighting.
		].
	ogl drawIndexed: mode
		faces: faces
		vertices: vertices
		normals:normals
		colors: colors
		texCoords: texCoords.
	ogl glEnable: GLLighting.
! !

!TMeshSimple methodsFor: 'initialize' stamp: 'das 6/30/2006 08:08'!
sphericalShell: numPoints

	self sphericalShell: numPoints outerRadius: 1.0 innerRadius:  0.9.! !

!TMeshSimple methodsFor: 'initialize' stamp: 'das 6/30/2006 08:08'!
sphericalShell: numPoints outerRadius: or innerRadius: ir
	
	| good pt vs irsq orsq ptsq rg m1 or2 ns |

	self clear.

	vs := Vector3Array new: numPoints.
	ns := Vector3Array new: numPoints.
	pt := Vector3      new.
	rg := Random new.
	m1 := Vector3 x: 0.5 y: 0.5 z: 0.5.
	
	irsq := ir * ir.
	orsq := or * or.
	or2 := or * 2.0.

	good := 0.
	[good < numPoints] whileTrue: [
		pt x: rg next
		   y: rg next
		   z: rg next.
		pt -= m1.
		pt *= or2.
		ptsq := pt squaredLength.
		((ptsq > irsq) and: [ptsq < orsq]) ifTrue: [	
			good := good + 1.
			vs at: good put: pt.
			ns at: good put: pt normalized.
			].
		].

	self vertices: vs.
	self normals:  ns.

	self provideColors.

	self boundsChanged.
	! !

!TMeshSimple methodsFor: 'accessing' stamp: 'das 6/30/2006 08:08'!
vertices: anObject
	"Set the value of vertices"

	vertices := anObject.

	self boundsChanged.! !

TPrimitive subclass: #TPointCloud
	instanceVariableNames: 'vertices normals colors pointSize box'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TPointCloud methodsFor: 'initialize' stamp: 'das 6/26/2006 03:10'!
clear

	vertices := nil.
	normals := nil.
	colors := nil.! !

!TPointCloud methodsFor: 'accessing' stamp: 'das 6/26/2006 03:22'!
colors: anObject
	"Set the value of colors"

	colors := anObject! !

!TPointCloud methodsFor: 'initialize' stamp: 'das 6/26/2006 03:12'!
initBounds

	vertices ifNil: [
		boundSphere := TBoundSphere new.
		boundSphere localPosition: Vector3 new.
		boundSphere radius: 1.0.
		boundSphere frame: self.
		]
	ifNotNil: [
		Transcript show: 'Calculating bounding box point by point....';cr.
		box := TBox new.
		vertices do: [:v|
			box growVertex: v
			].
		vertices do: [:v|
			v -= box center.
			].
		box := TBox new.
		vertices do: [:v|
			box growVertex: v
			].

		boundSphere := TBoundSphere new.
		boundSphere localPosition: box center.
		boundSphere radius: (box radius * 0.5).
		boundSphere frame: self.
		boundSphere box: box.
		
		].
! !

!TPointCloud methodsFor: 'initialize' stamp: 'das 6/26/2006 03:11'!
initialize

	super initialize.

	self clear.

	pointSize := 2.! !

!TPointCloud methodsFor: 'initialize' stamp: 'das 6/26/2006 03:18'!
normalizeDown: aFloatArray

	"assumes values >= 0.
     scales the array s.t. the largest value is 1.0"

	| biggest targ |

	targ := 0.0.

	aFloatArray size < 1 ifTrue: [^nil].

	biggest := 1.0.

	aFloatArray do: [:n|
		n < biggest ifTrue: [
			biggest := n.
			].
		].

   biggest = 0.0 ifTrue: [
		1 to: aFloatArray size do: [:n|
			aFloatArray at: n put: 0.0.
			].		
		]
	ifFalse: [
		1 to: aFloatArray size do: [:n|
			aFloatArray at: n put: ((aFloatArray at: n) + (targ-biggest)).
			].
"		aFloatArray *= (1.0 / biggest)."
		].


	! !

!TPointCloud methodsFor: 'initialize' stamp: 'das 6/26/2006 03:17'!
normalizeUp: aFloatArray

	"assumes values >= 0.
     scales the array s.t. the largest value is 1.0"

	| biggest targ |

	targ := 1.0.

	aFloatArray size < 1 ifTrue: [^nil].

	biggest := 0.0.

	aFloatArray do: [:n|
		n > biggest ifTrue: [
			biggest := n.
			].
		].

   biggest = 0.0 ifTrue: [
		1 to: aFloatArray size do: [:n|
			aFloatArray at: n put: 0.0.
			].		
		]
	ifFalse: [
		1 to: aFloatArray size do: [:n|
			aFloatArray at: n put: ((aFloatArray at: n) + (targ-biggest)).
			].
"		aFloatArray *= (1.0 / biggest)."
		].


	! !

!TPointCloud methodsFor: 'accessing' stamp: 'das 6/26/2006 03:23'!
normals: x

	normals := x.! !

!TPointCloud methodsFor: 'accessing' stamp: 'das 6/26/2006 03:23'!
pointSize: x
	pointSize := x.
"	self resetCaching."! !

!TPointCloud methodsFor: 'initialize' stamp: 'das 6/26/2006 03:18'!
provideColors

	"make an array of colors to match the vertices"

	| c o blorp |
	self colors: (Vector3Array new: vertices size withAll: #(1 0 0)asFloatArray).

	c := Vector4 new.
	
	1 to: vertices size do: [:n|
		blorp := (vertices at: n) z > 0.
		blorp := false not.
		blorp ifTrue: [
		    o := Vector4 x: 0.6 y: 0.6 z: 0.6 w: 0.0.
			c := (vertices at: n) normalized + o.
			self normalizeUp: c.
			]
		ifFalse: [
		    o := Vector4 x: 0.0 y: 0.0 z: 0.0 w: 0.0.
			c := (vertices at: n) normalized + o.
			self normalizeDown: c.
			].
		colors at: n put: c.
		].


	"self resetCaching."! !

!TPointCloud methodsFor: 'render' stamp: 'das 6/30/2006 07:39'!
renderPrimitive: ogl

	vertices ifNil: [^super renderPrimitive: ogl].

	ogl glPointSize: pointSize. 

	ogl glEnableClientState: GLVertexArray.

	normals ifNil: [
		ogl glDisable: GLLighting.
		]
	ifNotNil: [
		ogl glEnableClientState: GLNormalArray.
		ogl glNormalPointer:         GLFloat with: 0 with: normals.
		].

	ogl glVertexPointer: 3 with: GLFloat with: 0 with: vertices.
	colors ifNotNil: [
		ogl glEnableClientState: GLColorArray.
		ogl glEnable: GLColorMaterial.
		ogl glColorPointer:  3 with: GLFloat with: 0 with: colors.
		].

	ogl glDrawArrays: GLPoints with: 0 with: vertices size.

	"clean up"
	ogl glDisableClientState: GLColorArray.
	ogl glDisableClientState: GLNormalArray.
	ogl glDisableClientState: GLVertexArray.
	ogl glColor3fv: #(1 1 1)asFloatArray.
	ogl glEnable: GLLighting.
	ogl glDisable: GLColorMaterial.

! !

!TPointCloud methodsFor: 'initialize' stamp: 'das 6/26/2006 03:20'!
sphericalShell: numPoints

	self sphericalShell: numPoints outerRadius: 1.0 innerRadius:  0.9.! !

!TPointCloud methodsFor: 'initialize' stamp: 'das 6/26/2006 03:19'!
sphericalShell: numPoints outerRadius: or innerRadius: ir
	
	| good pt vs irsq orsq ptsq rg m1 or2 ns |

	self clear.

	vs := Vector3Array new: numPoints.
	ns := Vector3Array new: numPoints.
	pt := Vector3      new.
	rg := Random new.
	m1 := Vector3 x: 0.5 y: 0.5 z: 0.5.
	
	irsq := ir * ir.
	orsq := or * or.
	or2 := or * 2.0.

	good := 0.
	[good < numPoints] whileTrue: [
		pt x: rg next
		   y: rg next
		   z: rg next.
		pt -= m1.
		pt *= or2.
		ptsq := pt squaredLength.
		((ptsq > irsq) and: [ptsq < orsq]) ifTrue: [	
			good := good + 1.
			vs at: good put: pt.
			ns at: good put: pt normalized.
			].
		].

	self vertices: vs.
	self normals:  ns.

	self provideColors.

	self boundsChanged.
	! !

!TPointCloud methodsFor: 'accessing' stamp: 'das 6/26/2006 03:23'!
vertices: anObject
	"Set the value of vertices"

	vertices := anObject.

	self boundsChanged.! !

!TPrimitive methodsFor: 'render' stamp: 'das 3/12/2004 13:57'!
boundSphere
	| |
	boundsChanged ifTrue:[
		self initBounds. 
		boundsChanged := false.
	"	ts := TSphere initialize: ogl.
		ts translation: boundSphere localPosition.
		ts radius: boundSphere radius.
		ts material: (TMaterial initialize: ogl).
		ts boundSphere frame: nil.
		ts segments: 4.
		self removeAll.
		self addChild: ts."
		].
	^ boundSphere.! !

!TPrimitive methodsFor: 'accessing' stamp: 'jcg 6/12/2006 22:25'!
boundsChanged

	boundsChanged := true.
	frameChanged := true.
! !

!TPrimitive methodsFor: 'accessing' stamp: 'das 6/8/2006 21:05'!
colorize: col

	| mat |
	mat := TMaterial new.
	mat ambientColor: col.
	mat diffuseColor: col.
	mat textureMode: GLModulate.

	material ifNotNil:[
		mat emissiveColor: material emissiveColor.
		mat specularColor: material specularColor.
		mat shininess: material shininess.
		mat texture: material texture.
		].
	self material: mat.
! !

!TPrimitive methodsFor: 'copying' stamp: 'das 7/10/2002 12:16'!
copy

	| c |
	c := super copy.
	c boundsChanged.
	^ c.! !

!TPrimitive methodsFor: 'render' stamp: 'ar 1/16/2006 17:13'!
hasAlpha
	" Does this object have an alpha channel to render? Return true if it does."
	material ifNil:[^false].
	^material hasAlpha! !

!TPrimitive methodsFor: 'render' stamp: 'das 6/7/2006 22:21'!
hasOpaque
	" Does this object have an opaque channel to render? Return true if it does."
	material ifNil:[^true].
	^material hasAlpha not! !

!TPrimitive methodsFor: 'initialize' stamp: ''!
initBounds! !

!TPrimitive methodsFor: 'initialize' stamp: 'ar 1/16/2006 17:12'!
initialize
	super initialize.
	boundsChanged := true.
	^self
! !

!TPrimitive methodsFor: 'accessing' stamp: 'ar 1/16/2006 17:13'!
material
	^ material! !

!TPrimitive methodsFor: 'accessing' stamp: 'jcg 6/12/2006 22:25'!
material: mat
	material := mat.
! !

!TPrimitive methodsFor: 'render' stamp: 'das 6/8/2006 17:02'!
render: ogl

	material ifNotNil:[material enable: ogl]
		ifNil:[self root defaultMaterial enable:ogl.].
	texture ifNotNil:[texture enable: ogl].
	self renderPrimitive: ogl.
	texture ifNotNil:[texture disable: ogl].
	material ifNotNil:[material disable: ogl]
		ifNil:[self root defaultMaterial disable: ogl].
	
! !

!TPrimitive methodsFor: 'render' stamp: 'das 6/7/2006 22:23'!
renderAlpha: ogl

	material ifNotNil:[material enable: ogl. ].
	texture ifNotNil:[texture enable: ogl].
	self renderPrimitive: ogl.
	texture ifNotNil:[texture disable: ogl.].
	material ifNotNil:[material disable: ogl.].
! !

!TPrimitive methodsFor: 'render' stamp: 'das 3/12/2004 13:57'!
renderPrimitive: ogl

	super render: ogl! !

!TPrimitive methodsFor: 'accessing' stamp: 'das 4/18/2006 14:42'!
texture

	^ texture ifNil:[material ifNotNil:[material texture]].! !

!TPrimitive methodsFor: 'accessing' stamp: 'jcg 6/12/2006 22:25'!
texture: txtr
	texture := txtr.
! !

!TPrimitive methodsFor: 'accessing' stamp: 'jcg 6/12/2006 22:26'!
transparency: trans
	super transparency: trans.
	material ifNotNil:[material transparency: trans].
! !

!TPrimitive methodsFor: 'accessing' stamp: 'das 4/18/2006 14:47'!
uvAngle

	material ifNotNil:[^ material uvAngle].
	texture ifNotNil:[^ texture uvAngle].
	^ 0.! !

!TPrimitive methodsFor: 'accessing' stamp: 'das 4/18/2006 14:47'!
uvOffset

	material ifNotNil:[^ material uvOffset].
	texture ifNotNil:[^ texture uvOffset].
	^ 0@0.! !

!TPrimitive methodsFor: 'accessing' stamp: 'bgf 6/8/2006 17:30'!
uvScale
	texture ifNotNil:[^ texture uvScale].
	material ifNotNil:[^ material uvScale].
	^ 1.0@1.0.! !

TFrame subclass: #TQuadTree
	instanceVariableNames: 'quadCenter quadSize quadCorner inBox outBox radius center spheres depth qtTL qtTR qtBL qtBR boundSphere quadOn inReinstall'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TQuadTree commentStamp: 'das 10/10/2004 13:49' prior: 0!
TQuadTree

This is a spatial "loose" quadtree. It is used for fast visibility and collision detection.

A loose octree/quadtree is one where an object is contained in only one octree cell. This cell is smallest cell that can completely contain the object, and the one that contains the "center" of the object. In our case, the one that contains the center of the bound sphere. The object is allowed to overlap onto adjacent cells, thus when we test for collisions, we need to compare to the current cell and all of its adjacents. The advantage is significantly simpler and faster bookkeeping.

To use the TQuadTree just use the initialize method:

#initializeWithSpace: space frame: frame.

This will figure out the proper size of the quadtree from the elements inside of frame and it will place the TBoundSpheres into their proper slots. Then just add the TQuadTree to another frame, usually the TSpace, and you are done.

TRay tests only work if the TRay is a downRay.

The TQuadTree  is a hierarchical structure made up of smaller TQuadTrees. If a TQuadTree is empty, it's value is set to nil in its containing TQuadTree.
quadCenter - the center of this quad.
quadSize - the 2D x,z size of this  quad
quadCorner - the location of the corner
inBox - box containing the centers of objects in this quad
outbox -box containing the entire objects in this quad
radius - radius of the quad
center - center of the quad
spheres - collection of all the bound spheres in this quad
depth - maximum depth of quad tree
qtTL - quad tree Top Left 
qtTR - quad tree Top Right
qtBL - quad tree Bottom Left
qtBR - quad tree Bottom Right
boundSphere - bound sphere of the quad
quadOn - flag for rendering. If false, just render normally.
 
DAS!

!TQuadTree methodsFor: 'accessing' stamp: 'das 6/29/2002 16:10'!
add: bdSph

" Adds the boundSphere to the octree/quadtree."

	self add: bdSph depth: depth.! !

!TQuadTree methodsFor: 'accessing' stamp: 'das 2/17/2004 16:08'!
add: bndSph depth: dpth

" Adds the boundSphere to the octree/quadtree."

	| sphLoc |

	sphLoc := bndSph globalPosition.
	inBox growVertex: sphLoc.
	outBox := outBox unionBox: 
		(inBox unionSphere:bndSph globalPosition radius: bndSph radius).
	radius := outBox diagonal/2.0.
	center := outBox center.
"------ This is as deep as we go - add the sphere -------"
	dpth = 0 ifTrue:[ 
		self addSphere: bndSph.
 		^ true. 
		].

"------ Will the sphere fit into this quad? ------"
	bndSph radius > (quadSize/2.0) ifTrue:[ 
 		self addSphere: bndSph. 
		^ true.	
		].

"------ Then which quad should it go into? ------ "

	sphLoc x < center x ifTrue:[
		sphLoc z < center z ifTrue:[ 
			qtTL ifNil:[ qtTL :=TQuadTree new initializeWithCorner: quadCorner size: quadSize depth: depth.].
			^ qtTL add: bndSph depth: (dpth-1).
			].
		qtBL ifNil:[ 
			qtBL := TQuadTree new initializeWithCorner: (quadCorner+(Vector3 x:0 y:0 z:quadSize)) size: quadSize depth: depth.].
			^ qtBL add: bndSph depth: (dpth-1).
		].
	sphLoc z < center z ifTrue:[
		qtTR ifNil:[ 
			qtTR :=TQuadTree new initializeWithCorner: (quadCorner+(Vector3 x:quadSize y:0 z:0)) size: quadSize depth: depth.].
		^ qtTR add: bndSph depth: (dpth-1).
		].
	qtBR ifNil:[
		qtBR :=TQuadTree new initializeWithCorner: (quadCorner+(Vector3 x:quadSize y:0 z:quadSize)) size: quadSize depth: depth.].
	^ qtBR add: bndSph depth: (dpth-1).
	! !

!TQuadTree methodsFor: 'accessing' stamp: 'das 6/22/2002 01:13'!
addSphere: bs

	spheres ifNil:[spheres := OrderedCollection new.].
	spheres add: bs.
! !

!TQuadTree methodsFor: 'accessing' stamp: 'das 6/29/2002 14:30'!
boundSphere

	^ boundSphere.! !

!TQuadTree methodsFor: 'accessing' stamp: 'das 2/2/2006 00:17'!
doForPick: block

	block value: self.
! !

!TQuadTree methodsFor: 'accessing' stamp: 'das 7/11/2003 11:32'!
globalTransformUpdate

	self reinstall.! !

!TQuadTree methodsFor: 'render' stamp: 'ar 9/7/2002 14:52'!
hasAlpha

	^ false.! !

!TQuadTree methodsFor: 'initialize' stamp: 'das 7/31/2006 19:33'!
initialize

	super initialize.
	inReinstall := false.
	quadOn := true.
	visible := false.
! !

!TQuadTree methodsFor: 'initialize' stamp: 'sn 2/23/2018 12:44'!
initializeWithCorner: crnr size: sz depth: dpth

	|  cornerMax |

	self initialize.
	quadSize := sz/2.0.
	quadCorner := crnr shallowCopy.
	quadCorner y: Float infinity.
	cornerMax := crnr + sz.
	cornerMax y: Float infinity negated.
	inBox := TBox new. "min: corner max:cornerMax."
	outBox := TBox new.
	quadCenter := quadCorner + (quadSize/2.0).
	radius := 0.
	self visible: false.
	depth := dpth.
	self quadOn: true.
	^self
! !

!TQuadTree methodsFor: 'initialize' stamp: 'ar 1/16/2006 19:19'!
initializeWithFrame: frame

	self initialize.
	self addChild: frame.
	self visible: false.
	self quadOn: true.
	^self! !

!TQuadTree methodsFor: 'initialize' stamp: 'das 7/31/2006 19:34'!
initializeWithSpace: space frame: frame

	space addChild: self.
	self addChild: frame.
	frameChildren do:[ :fc | fc frameChanged.].
	self reinstall.! !

!TQuadTree methodsFor: 'accessing' stamp: 'das 6/29/2002 16:52'!
outBox

	^ outBox.! !

!TQuadTree methodsFor: 'render' stamp: 'das 2/2/2006 00:21'!
pickFloor: pointer

	^ self quadPickFloor: pointer location: pointer framePosition.	

! !

!TQuadTree methodsFor: 'accessing' stamp: 'das 12/3/2002 10:09'!
quadOn

	^ quadOn.! !

!TQuadTree methodsFor: 'accessing' stamp: 'das 12/3/2002 10:09'!
quadOn: bool

	"Turn the quadtree rendering on/off"
	quadOn := bool.! !

!TQuadTree methodsFor: 'render' stamp: 'das 1/2/2003 10:33'!
quadPickFloor: pointer location: location

	| d rval |
	
	(outBox pointOverBox: location) ifFalse:[^ false.].
	rval := false.
	spheres ifNotNil:[
		spheres do:[ : sp |
			sp frame isSolid ifTrue:[
				d := (sp globalPosition - pointer globalPosition) abs.
				( d x < sp radius and: [d z < sp radius]) ifTrue:[
					(pointer pickDownSphere: sp) ifTrue:[
						(pointer pointerPickFloor: sp frame) ifTrue:[rval := true].
						].
					].
				].
			].
		].
		
	qtTL ifNotNil:[(qtTL quadPickFloor: pointer location: location) ifTrue:[rval := true].].
	qtBL ifNotNil:[(qtBL quadPickFloor: pointer location: location) ifTrue:[rval := true].].
	qtTR ifNotNil:[(qtTR quadPickFloor: pointer location: location) ifTrue:[rval := true].].
	qtBR ifNotNil:[(qtBR quadPickFloor: pointer location: location) ifTrue:[rval := true].].
	^ rval.! !

!TQuadTree methodsFor: 'initialize' stamp: 'das 10/20/2004 23:52'!
reinstall

	| frame rad ext |
	inReinstall ifTrue:[^ self]. "don't recurse"
	inReinstall := true.
	frame := frameChildren at: 1.

	rad := frame octreeRadius.
	inBox := frame octreeBox.
	outBox := TBox new.
	quadCenter := inBox center.
	qtTL:= nil.
	qtTR:= nil.
	qtBL:= nil.
	qtBR:= nil.
	spheres := nil.
"----- size is 1/2 of the length of a quadtree side. -----"
	ext := inBox extent.
	quadSize := ext x > ext z ifTrue:[ ext x ] ifFalse:[ext z].
	"size := size > ext y ifTrue:[ size ] ifFalse:[ext y]. don't need for quadtrees"
	quadSize := quadSize/2.0.
	quadSize < rad ifTrue:[ quadSize := rad.].
	quadCorner := quadCenter - quadSize.
	radius := 0.0.
	depth := 5.
	self visible: false.

	frameChildren do:[ :fc |
			fc octreeSieve: self.].
	boundSphere := TBoundSphere localPosition: outBox center radius: outBox diagonal/2.0.
	boundSphere frame: self.

	inReinstall := false.! !

!TQuadTree methodsFor: 'render' stamp: 'das 3/12/2004 14:00'!
renderAlpha: ogl
"	| mat |

	ogl glDisable: GLCullFace.
	mat := TMaterial initialize: cWorld.
	mat ambientColor: #(0.7  0.2 0.1 0.3)asFloatArray.
	mat diffuseColor: #(0.7  0.2 0.1 0.3)asFloatArray.
	mat enable: ogl.
	ogl renderBox: box.
	qtTL ifNotNil:[ qtTL renderAlpha: ogl.].
	qtTR ifNotNil:[ qtTR renderAlpha: ogl.].
	qtBL ifNotNil:[ qtBL renderAlpha: ogl.].
	qtBR ifNotNil:[ qtBR renderAlpha: ogl.].
	ogl setCull."! !

!TQuadTree methodsFor: 'render' stamp: 'das 7/9/2006 17:51'!
renderFrame: ogl
	| count |

	false ifFalse:[^ super renderFrame: ogl]. "Traditional rendering method - used for testing and picking"
	" something of a problem - Quadtrees do not lend themselves to standard picking using this method."

	ogl pushMatrix.
	count := self renderTree: ogl root: self root.
	ogl popMatrix.
"------ This is just test code to see where the quads really are. ------"
"	self alpha ifTrue: [ 
		space addAlpha: 
		(TRenderAlpha object: self 
			transform: space currentTransform clone
			distance: (ogl activeCamera globalPosition - space currentTranslation ) squaredLength
			parent: space currentParent).]."

	^ count.
! !

!TQuadTree methodsFor: 'render' stamp: 'das 7/7/2006 15:34'!
renderFrame: ogl parent: parent root: root
	| count |
	quadOn ifFalse:[^ super renderFrame: ogl parent: parent root: root.]. "Traditional rendering method - used for testing and picking"
	" something of a problem - Quadtrees do not lend themselves to standard picking using this method."

	ogl pushMatrix.
	count := self renderTree: ogl root: root.
	ogl popMatrix.
"------ This is just test code to see where the quads really are. ------"
"	self alpha ifTrue: [ 
		space addAlpha: 
		(TRenderAlpha object: self 
			transform: space currentTransform clone
			distance: (ogl activeCamera globalPosition - space currentTranslation ) squaredLength
			parent: space currentParent).]."

	^ count.
! !

!TQuadTree methodsFor: 'render' stamp: 'das 6/7/2006 21:55'!
renderTree: ogl root: root
	| count globalTrans ac |
	ac := ogl camera.
	count := 0.
	"is this inside the viewing pyramid?"
	(ac testSphere: center radius: radius) ifTrue:[
		spheres ifNotNil:[
			spheres do:[:sp | 
				(ac testBounds: sp) ifTrue:[
					ogl pushMatrix.
					ogl glMultMatrixf: sp frame globalTransform transposed.
					sp frame render: ogl. 
					count := count + 1.
					sp frame hasAlpha ifTrue: [ 
						globalTrans := Matrix4x4 new.
						ogl glGetFloatv: GLModelviewMatrix with: globalTrans.
						root addAlphaObject: sp frame 
							transform: globalTrans
							distance: (ac globalPosition - 
								sp globalPosition) squaredLength
							parent: sp frame parent
							transparency: ogl transparency.
						].
					ogl popMatrix.
					].
				].
			].
		qtTL ifNotNil:[ count := count + (qtTL renderTree: ogl root: root).].
		qtTR ifNotNil:[ count := count + (qtTR renderTree: ogl root: root).].
		qtBL ifNotNil:[ count := count + (qtBL renderTree: ogl root: root).].
		qtBR ifNotNil:[ count := count + (qtBR renderTree: ogl root: root).].
		].
	^ count.

! !

!TQuadTree methodsFor: 'accessing' stamp: 'das 10/20/2004 23:36'!
size

	| ttl |
	ttl := 0.
	ttl := spheres ifNil:[0] ifNotNil:[spheres size].
	ttl := ttl + (qtTL ifNil:[0] ifNotNil:[qtTL size]).
	ttl := ttl + (qtTR ifNil:[0] ifNotNil:[qtTR size]).
	ttl := ttl + (qtBL ifNil:[0] ifNotNil:[qtBL size]).
	ttl := ttl + (qtBR ifNil:[0] ifNotNil:[qtBR size]).

	^ ttl.! !

TFrame subclass: #TRay
	instanceVariableNames: 'automatic framePointer framePosition sphereDistSquared currentFrame selection lastSelection doSelect testDistance downRay minDistance minDistanceSquared maxDistance maxDistanceSquared frameScale frameScaleSquared excludeFrame'
	classVariableNames: 'TDownRay'
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TRay commentStamp: '<historical>' prior: 0!
TRay specifies a position and orientation, usually inherited from it's parent TFrame. Though you can modify this. It is used to determine the closest object in a particular direction and the distance to that object. 

This is a factored TPointer change set. The new TRay class is concerned only with picking an object, while the new TPointer class is a subclass of TRay, but includes the events model. This means that the TCamera uses a TRay for the downPointer, and it also makes it easier to use a TRay for arbitrary objects looking for the floor. I also completely rewrote the actual matrix math. Essentially, a TRay is a TFrame with the ability to pick. Now it actually uses the local and global orientation matrices to perform the test. 

TRay tests are fairly expensive, so be careful with them. Also, I don't think you want to use a translation with them. I think more work needs to be done to make this robust.

The way a TRay works is you simply place it as a child frame, and it will automatically get called at the next render loop. You can use it directly by calling the following methods:

#pointerPick: or 
#pointerPickFloor: 

with a TBoundSphere as the argument (which in turn points to a TFrame) as the argument. If the return is true, then the TRay intersected the object.

You can also call one of the #frame:xxxx:xxxx: messages if you know the type of object you are attempting to select. If you wish to check a number of objects and keep only the closest, call the message #testDistance:false. Make sure you call #testDistance: true when you are done.

The following variables are for internal use only:
	automatic - indicates whether a ray is tested automatically during rendering.
	framePointer - is the ray orientation relative to the object we are testing against.
	framePosition - is the ray location relative to the object we are testing against.
	sphereDistSquared - is the distance to the recently tested sphere. 
	currentFrame - this is the frame we are currently testing against.
	selection - the selected frame information. See TSelection for more information.
	doSelect - turns the selection of the TRay on/off. This is used by the TPointer to keep the pointer from selecting other objects between a mouseDown/mouseUp.
	testDistance - a boolean that allows us to ignore the current distance when testing for object intersection. It is used when the user needs to determine intersection and to a specific object.
	downRay - boolean indicates whether this is a floor test ray or not.
The following variables are for external use:
	minDistance - the minimum distance we care about.
	minDistanceSquared - the above squared for performance
	maxDistance - the furthest away we care
	maxDistanceSquared - the above squared for performance.
	frameScale - scaling value for testing picks with micro worlds.

DAS!

TRay subclass: #TPointer
	instanceVariableNames: 'focusSelected pointerDown isOverlay'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TPointer commentStamp: '<historical>' prior: 0!
TPointer is simply a TRay where the  instance methods are redefined to convert to TFarRefs or TObjectIDs. It is used as the main user pointer when selecting an object in a scene. TPointer is not intended to be a part of an Island.

DAS!

!TPointer methodsFor: 'accessing' stamp: 'das 1/9/2006 11:39'!
currentSelection 

	focusSelected ifNil:[^selection].
	^focusSelected = selection frame ifTrue:[selection] ifFalse:[lastSelection].! !

!TPointer methodsFor: 'picking' stamp: 'das 1/8/2006 15:43'!
focusSelected

	^focusSelected.
! !

!TPointer methodsFor: 'picking' stamp: 'das 1/8/2006 14:50'!
focusSelected: frame

	focusSelected := frame.
! !

!TPointer methodsFor: 'initialize' stamp: 'das 5/8/2006 07:09'!
initialize
	super initialize.
	pointerDown := false.
	isOverlay := false.! !

!TPointer methodsFor: 'accessing' stamp: 'das 5/8/2006 07:09'!
isOverlay

	^ isOverlay.! !

!TPointer methodsFor: 'accessing' stamp: 'das 5/8/2006 07:09'!
isOverlay: bool

	isOverlay := bool.! !

!TPointer methodsFor: 'initialize' stamp: 'ar 1/14/2006 14:33'!
newSelection
	"Answer a new selection holder"
	^TXSelection new! !

!TPointer methodsFor: 'accessing' stamp: 'das 3/4/2006 23:33'!
pick: bnds

	focusSelected ifNil:[ ^ super pick: bnds].
	bnds ifNil:[^ false ].
	bnds frame objectOwner = focusSelected ifTrue:[
		^super pick: bnds.].
	^ false.
! !

!TPointer methodsFor: 'accessing' stamp: 'das 3/14/2006 14:50'!
pointerDown

	^pointerDown.! !

!TPointer methodsFor: 'accessing' stamp: 'das 3/13/2006 20:57'!
pointerDown: pd

	pointerDown := pd.! !

!TPointer methodsFor: 'picking' stamp: 'das 3/26/2006 16:20'!
pointerPick: bnds
	| position frame gpAt dp d |

	bnds ifNil:[^ false.].

	focusSelected ifNil:[^ super pointerPick: bnds].

	focusSelected = bnds frame objectOwner asFarRef ifTrue:[
		position := bnds globalPosition - self globalPosition.
		self selectedRayTransform: self localTransform.
		self selectedDistance > ((position length - bnds radius) * frameScale) 		ifTrue:[
				gpAt := self globalTransform column3.
				dp := position - ((gpAt dot: position)*gpAt).
				d := dp dot: dp.
				d < bnds radiusSquared ifFalse:[ ^ false ].
				self sphereDistSquared: d.
				frame := bnds frame.
				self currentFrame: frame.
				(frame pick: self) ifTrue:[
					self selectedPoint ifNil:[self selectedPoint: Vector3 new.].
					self selectedGlobalPoint: 
						(frame globalTransform localPointToGlobal:self selectedPoint).
					self selectedFrame: frame.
					self selectedParent: frame parent.
					self selectedFramePosition:  self framePosition.
					self selectedTarget: frame objectOwner.
					self selectedEventMask: frame objectOwner getAllEventsHandled.
					self selectedScale: frameScale.
	"				self selectedCameraTransform: self parent globalTransform."
					^ true.
				].
			].
		].
	^ false.! !

!TPointer methodsFor: 'picking' stamp: 'das 4/23/2006 16:31'!
resetSelected: event
	| lastSel sel eventHandled lastEventHandled lastTarget target |

	sel := selection frameID.
	target := selection target.
	selection target: nil.
	eventHandled := selection eventMask.
	lastSel := lastSelection frameID.
	lastTarget := lastSelection target.
	lastSelection target: nil.
	lastEventHandled := lastSelection eventMask.
	pointerDown ifFalse:[
"This is a bit trick when we take tasks into account. I need to add a task model to this without interfering with the logic. To be continued..."
	lastSel ~= sel ifTrue:[
" ------ pointer has left the lastSel "
		lastSel ifNotNil:[
			(lastEventHandled anyMask:EventPointerOver) ifTrue:[
				event selection: selection.
				lastTarget future doPointerLeave: event.
			].
		].
" ------ pointer has entered the sel "
		sel ifNotNil:[
			(eventHandled anyMask:EventPointerOver) ifTrue:[
				event selection: selection.
				target future doPointerEnter: event.
			].
		].
	] ifFalse: [
" ------ pointer is still inside the same target "
		sel ifNotNil:[
			(eventHandled anyMask:EventPointerOver) ifTrue: [
				event selection: selection.
				target future doPointerOver: event.
			].
		].
	].
	].
	selection target: target.
	lastSelection target: lastTarget.
pointerDown ifFalse:[
	lastSelection := selection.
	selection := self newSelection.
].
	selection distance: Float infinity.
	self maxDistance: Float infinity.
! !

!TPointer methodsFor: 'picking' stamp: 'das 4/23/2006 16:26'!
resetSelected: event task: task
	| lastSel sel lastTarget target |

	sel := selection frameID.
	target := selection target.
	selection target: nil.
	lastSel := lastSelection frameID.
	lastTarget := lastSelection target.
	lastSelection target: nil.
	pointerDown ifFalse:[
	lastSel ~= sel ifTrue:[
" ------ pointer has left the lastSel "
		lastSel ifNotNil:[
			event selection: selection.
			task doPointerLeave: event.
		].
" ------ pointer has entered the sel "
		sel ifNotNil:[
			event selection: selection.
			task doPointerEnter: event.
		].
	] ifFalse: [
" ------ pointer is still inside the same target "
		sel ifNotNil:[
			event selection: selection.
			task doPointerOver: event.
		].
	].
	].
	selection target: target.
	lastSelection target: lastTarget.
pointerDown ifFalse:[
	lastSelection := selection.
	selection := self newSelection.
].
	selection distance: Float infinity.
	self maxDistance: Float infinity.
! !

!TPointer methodsFor: 'accessing' stamp: 'ar 1/19/2006 10:34'!
selectedFrame: sf

	selection frameID: sf objectID.! !

!TPointer methodsFor: 'accessing' stamp: 'das 3/10/2006 18:30'!
selectedFrameID

	selection frameID! !

!TPointer methodsFor: 'accessing' stamp: 'das 12/13/2005 09:07'!
selectedParent: sp

	selection parent: sp objectID.! !

!TPointer methodsFor: 'accessing' stamp: 'das 12/13/2005 09:08'!
selectedPortal: sp point: pos

	sp ifNil:[selection portal: nil. selection portalPoint:nil. ^ self].
	selection portal ifNil:[selection portal: sp objectID. selection portalPoint: pos.].! !

!TPointer methodsFor: 'accessing' stamp: 'das 1/21/2006 15:22'!
selectedTarget: st

	selection target: st asFarRef.! !

!TRay class methodsFor: 'class initialization' stamp: 'das 1/9/2006 14:07'!
initialize

	TDownRay := Matrix4x4 identity rotationAroundX: -90.
! !

!TRay methodsFor: 'accessing' stamp: 'das 12/30/2002 16:27'!
automatic

	^ automatic.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/16/2006 13:27'!
copiedSelection
	^selection copy.
! !

!TRay methodsFor: 'accessing' stamp: 'das 1/16/2006 13:27'!
currentFrame

	^ currentFrame.! !

!TRay methodsFor: 'accessing' stamp: 'das 12/13/2005 08:52'!
currentFrame: cf

	| inv |
	inv := cf inverseGlobalTransform.
	self framePosition: (inv localPointToGlobal: self globalPosition).
	self framePointerTrans: cf globalOrientation.
! !

!TRay methodsFor: 'accessing' stamp: 'das 12/30/2002 16:27'!
doSelect

	^ doSelect.! !

!TRay methodsFor: 'accessing' stamp: 'das 12/14/2002 12:16'!
doSelect: bool

"Turns the ray test on/off so that the ray test will be focused on a selected item. This is done when we have a mouseDown and want to track the selected object."
	doSelect := bool.! !

!TRay methodsFor: 'accessing' stamp: 'das 11/9/2002 16:23'!
downRay

	^ downRay.! !

!TRay methodsFor: 'accessing' stamp: 'das 11/9/2002 16:23'!
downRay: bool

	downRay := bool.! !

!TRay methodsFor: 'picking' stamp: 'das 8/12/2002 21:40'!
frame: frame pickCylinderFrom: pnt1 radius: rad1 to: pnt2 radius: rad2
	| rval |
	self testDistance: false.
	self currentFrame: frame. 
	rval := self pickCylinderFrom: pnt1 radius: rad1 to: pnt2 radius: rad2.
	self testDistance: true.
	^ rval.
! !

!TRay methodsFor: 'picking' stamp: 'das 9/3/2006 12:16'!
frame: frame pickCylinderFrom: pnt1 to: pnt2 radius: rad
	| rval |
	self testDistance: false.
	self currentFrame: frame. 
	rval := self pickRightCylinderFrom: pnt1 to: pnt2 radius: rad.
	self testDistance: true.
	^ rval.
! !

!TRay methodsFor: 'picking' stamp: 'das 1/26/2005 22:14'!
frame: frame pickFrontPlane: position normal: normal
	| rval |
	self testDistance: false.
	self currentFrame: frame. 
	rval := self pickFrontPlane: position normal: normal.
	self testDistance: true.
	^ rval.
! !

!TRay methodsFor: 'picking' stamp: 'das 8/12/2002 21:41'!
frame: frame pickPlane: position normal: normal
	| rval |
	self testDistance: false.
	self currentFrame: frame. 
	rval := self pickPlane: position normal: normal.
	self testDistance: true.
	^ rval.
! !

!TRay methodsFor: 'picking' stamp: 'das 8/12/2002 21:40'!
frame: frame pickQuad: norm q1: p1 q2: p2 q3: p3 q4: p4
	| rval |
	self testDistance: false.
	self currentFrame: frame. 
	rval := self pickQuad: norm q1: p1 q2: p2 q3: p3 q4: p4.
	self testDistance: true.
	^ rval.
! !

!TRay methodsFor: 'picking' stamp: 'das 8/12/2002 21:42'!
frame: frame pickQuad: p1 q2: p2 q3: p3 q4: p4
	| rval |
	self testDistance: false.
	self currentFrame: frame. 
	rval := self pickQuad: p1 q2: p2 q3: p3 q4: p4.
	self testDistance: true.
	^ rval.
! !

!TRay methodsFor: 'picking' stamp: 'das 9/4/2006 18:50'!
frame: frame pickRightCylinderFrom: pnt1 to: pnt2 radius: rad
	| rval |
	self testDistance: false.
	self currentFrame: frame. 
	rval := self pickRightCylinderFrom: pnt1 to: pnt2 radius: rad.
	self testDistance: true.
	^ rval.
! !

!TRay methodsFor: 'picking' stamp: 'das 8/12/2002 21:42'!
frame: frame pickSphere: loc radiusSquared: rs

	| rval |
	
	self testDistance: false.
	self currentFrame: frame.
	rval := self pickSphere: loc radiusSquared: rs.
	self testDistance: true.
	^ rval

! !

!TRay methodsFor: 'picking' stamp: 'das 8/12/2002 21:42'!
frame: frame pickTriangle: p1 tri: p2 tri: p3
	| rval |
	self testDistance: false.
	self currentFrame: frame. 
	rval := self pickTriangle: p1 tri: p2 tri: p3.
	self testDistance: true.
	^ rval.
! !

!TRay methodsFor: 'picking' stamp: 'das 8/12/2002 21:43'!
frame: frame pickTriangle: norm tri: p1 tri: p2 tri: p3
	| rval |
	self testDistance: false.
	self currentFrame: frame. 
	rval := self pickTriangle: norm tri: p1 tri: p2 tri: p3.
	self testDistance: true.
	^ rval.
! !

!TRay methodsFor: 'accessing' stamp: 'das 12/13/2005 09:01'!
framePointer

	^ framePointer.! !

!TRay methodsFor: 'accessing' stamp: ''!
framePointer: fp
	
	framePointer := fp.
! !

!TRay methodsFor: 'accessing' stamp: 'das 8/12/2002 16:27'!
framePointerTrans: trans

	" convert from global to local target frame coordinates. "
	framePointer := self globalTransform transposed composeWith: trans.! !

!TRay methodsFor: 'accessing' stamp: ''!
framePosition
	self framePointer.
	^ framePosition.
! !

!TRay methodsFor: 'accessing' stamp: ''!
framePosition: fp

	framePosition := fp.
! !

!TRay methodsFor: 'accessing' stamp: 'das 3/23/2006 17:25'!
frameScale: fs

	frameScale := fs.
	frameScaleSquared := fs*fs.! !

!TRay methodsFor: 'picking' stamp: 'sn 2/23/2018 12:44'!
globalTransformUpdate
	
	|trans|

	downRay ifTrue:[
		trans := self globalPosition.
		globalTransform := TDownRay shallowCopy.
		globalTransform translation: trans.
	].! !

!TRay methodsFor: 'initialize' stamp: 'bgf 7/15/2006 15:23'!
initialize
	super initialize.
	self resetTotal.
	self doSelect: true.
	self testDistance: true.
	self visible: false.
	self downRay: false.
	self setAutomatic: true.
	frameScale := 1.
	frameScaleSquared := 1.
	excludeFrame := nil.

	^self
! !

!TRay methodsFor: 'testing' stamp: 'das 11/9/2002 15:03'!
isRay

	^ true.! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:56'!
lastSelection
	^lastSelection! !

!TRay methodsFor: 'accessing' stamp: 'das 6/18/2003 13:35'!
maxDistance

	^ maxDistance.! !

!TRay methodsFor: 'accessing' stamp: 'das 6/18/2003 13:35'!
maxDistance: md

	maxDistance := md.
	maxDistanceSquared := md*md.! !

!TRay methodsFor: 'accessing' stamp: 'das 6/10/2003 11:11'!
minDistance

	^ minDistance.! !

!TRay methodsFor: 'accessing' stamp: 'das 6/10/2003 11:26'!
minDistance: md

	minDistance := md.
	minDistanceSquared := md*md.! !

!TRay methodsFor: 'initialize' stamp: 'ar 1/14/2006 14:32'!
newSelection
	"Answer a new selection holder"
	^TSelection new! !

!TRay methodsFor: 'picking' stamp: 'das 2/1/2006 16:04'!
pick: bnds

	| position |

	bnds ifNil:[^ false ].
	bnds frame ifNotNil:[
		bnds frame isSolid ifTrue:[
			self downRay ifFalse:[ self pointerPick: bnds.] 
			ifTrue:[
				position := (bnds globalPosition - self globalPosition).
				position *= position.
				(position x + position z < bnds radiusSquared) ifTrue:[
					(self pickDownSphere: bnds) ifTrue:[
						(self pointerPickFloor: bnds frame) ifTrue:[ ^ true. ].
					].
				].
			].
		].
	].
	^ false.! !

!TRay methodsFor: 'picking' stamp: 'das 6/18/2003 21:18'!
pickBoundSphere: bnds 
	| position dp d gpAt |

	gpAt := self globalTransform column3.
	position := bnds globalPosition - self globalPosition.
	dp := position - ((gpAt dot: position)*gpAt).
	d := dp dot: dp.
	d < bnds radiusSquared ifTrue:[
		self sphereDistSquared: d.
		^ true.].
	^ false.! !

!TRay methodsFor: 'picking' stamp: 'das 5/27/2006 15:21'!
pickCircle: position normal: normal radius: r

	| div sp d |
	div := normal dot: self framePointer row3. 
	div <= 0 ifTrue:[ ^ false ].
	d := ((self framePosition-position) dot: normal)/div.
	sp:= self framePosition -  (self framePointer row3 * d).
	(self testSelectedDistance: d)=0 ifFalse:[^false].
	(sp - position) length > r ifTrue:[^ false ].
	self selectedDistance: d.
	self selectedPoint: sp.
	self selectedNormal: normal.
	^ true.! !

!TRay methodsFor: 'picking' stamp: 'das 9/3/2006 12:50'!
pickCylinderFrom: pnt1 radius: rad1 to: pnt2 radius: rad2

	| p1 p2 r1 r2 vn lv lx dx pv n nvn pd v ht cosCone cosPlane ip d1 d2 dd |

	"right cylinder"
	rad1 = rad2 ifTrue: [ ^ self pickRightCylinderFrom: pnt1 to: pnt2 radius: rad1.]
	ifFalse: [ "cone"
" Figure out where the vertex is.
	r1,r2 - top and bottom radii. r2 > r1
	p1,p2 - top and bottom center points.
	v - vertex of the cone
	ht - true height of the cone
	vn - vertex normal (up-down normal)
	n - normal plane vector
	nvn - vertex normal dot with plane normal
	ip - actual intersection point"

	rad2 < rad1 ifTrue:[
		r1 := rad2.
		r2 := rad1.
		p1 := pnt2 - self framePosition.	
		p2 := pnt1 - self framePosition.]
	ifFalse:[
		r1 := rad1.
		r2 := rad2.
		p1 := pnt1 - self framePosition.
		p2 := pnt2 - self framePosition.].

	rad1 = 0 ifTrue:[ 
		v := p1.
		ht := (p1- p2) length.
		vn := (p2 - p1) normalized.
		] 
	ifFalse:[
		vn := (p2 - p1) normalized.
		pd := (p1-p2)length.
		ht := (r2*pd)/(r2-r1).
		v := p2 - (vn * ht).
		].


" Calculate the plane that cuts the cone. This is from Graphics Gems V pg 227."
	n := self framePointer row3 cross: v.
	lv := n length.
	lv = 0 ifTrue:[^ false.] ifFalse:[ n := n/lv.].
" We hit the vertex point."
	(n dot: n)=0 ifTrue:[
		ip := v.]
	ifFalse:[
" Test the plane to see if it intersects the cone."
		cosCone := ht/(((ht*ht) + (r2*r2))sqrt).
		nvn := n dot: vn.
		cosPlane := (1- (nvn * nvn)) sqrt.
		cosCone > cosPlane ifTrue:[ ^ false. "no intersection."].

		lx := n dot: (p2-v). " distance of p2 from plane "
		dx := (lx*lx)/ht. " difference in height of closest point to p2"
		lv := (p2-v) - (n * lx). " actual location of closest point of p2 to plane "
		pv := v+ (lv * (ht/(ht-dx))). " location of this point on the plane defined by p2. "

		dx:=((r2*r2)-((pv-p2)squaredLength))sqrt. " distance from pv to r2 on circle at p2"
		pv := pv + (dx * ((n cross: vn) normalized)).

		pv:=(pv-v)normalized. "vector from v ON THE CYLINDER"

		d1 := self framePointer row3.
		d2 := pv.
		dd := d1 cross: d2.
		dx := dd dot: dd.
		dx = 0 ifTrue:[^false].
		dx := ( (v cross: d2) dot: dd )/dx.
		(self testSelectedDistance: dx abs)=0 ifFalse:[^ false].
		ip := self framePosition + (d1*dx).
		].

" Now that we (finally) have ip - the intersection of the pointer with the -infinite- cylinder, we need to test to see if it is between the top and bottom planes,"

	(((ip - self framePosition - p1) dot: vn) < 0.0 or:[((ip - self framePosition - p2) dot: vn) > 0.0]) ifTrue:[ 
		"the point is outside the bounds"
		^ false.].
	self selectedPoint: ip.
	self selectedDistance: dx abs.
	n:= (ip-self framePosition-v)normalized.
	pv := p2-v.
	self selectedNormal: (((pv dot:n)*n)-pv)normalized.
	^ true.
	].
	

	
	! !

!TRay methodsFor: 'picking' stamp: 'das 8/11/2002 17:00'!
pickDownSphere: bnds 
	| position |

	position := bnds globalPosition - self globalPosition.
	^ (((position x *position x) + (position z*position z)) < bnds radiusSquared).
! !

!TRay methodsFor: 'picking' stamp: 'das 12/7/2005 16:27'!
pickFrame: framePtr tri: p1 tri: p2 tri: p3


" This method works by using the orthogonal picking matrix as orthogonal planes going through the origin at 'origin'. The points of the triangle are tested to the two planes orthoganal to the direction of picking. If they are on either side of both planes, then there is a very high probability that our picking ray interesects the triangle, and we do the actual calculation of the intersection point. Note the comments below. I use a vertical bar to indicate which side of the plane a point is on.  ***** p1 p2 | ***** means that p1 and p2 are on the same side, 
***** p1 | p2 ***** indicates that they are on opposite sides. p1 is always on the left side, because it is the first one I test. This really just means - same side as p1 - no meaning other than that. 

This can be optimized for tri-strips, tri-fans, and even full face based meshes. When I have the time... "

	| d1 d2 d3 pp1 pp2 dd1 dd2 xPlane yPlane  |
		
	xPlane := framePtr row1.
	d1 := xPlane dot: p1.
	d2 := xPlane dot: p2.
	d3 := xPlane dot: p3.
	dd1 := d1*d2.
	dd2 := d1*d3.

" test if all points are on the same side of the x plane."
	dd1 < 0 ifFalse:[ 				"***** p1 p2 |  *****"
		dd2 < 0  ifFalse:[  ^ false		"***** p1 p2 p3 | *****" ] 
				ifTrue:[ 			"***** p1 p2 | p3 *****"
						pp1 := p1 + ((p3-p1) * d1/(d1-d3)).
						pp2 := p2 + ((p3-p2) * d2/(d2-d3)).]]
		ifTrue:[ 				"***** p1 | p2 *****"
			dd2 < 0  ifTrue:[ 		"***** p1 | p2 p3 *****" 
						pp1 := p1 + ((p2-p1) * d1/(d1-d2)).
						pp2 := p1 + ((p3-p1) * d1/(d1-d3)).] 
					ifFalse:[ 		"***** p1 p3 | p2 *****"
						pp1 := p1 + ((p2-p1) * d1/(d1-d2)).
						pp2 := p3 + ((p2-p3) * d3/(d3-d2)).].].
" At this point, the triangle is intersected by the x plane and we know where. If both points are on the same side of the y plane, we are done. "
	yPlane := framePtr row2.

	d1 := yPlane dot: pp1.
	d2 := yPlane dot: pp2.
" test if both points are on the same side of the y plane."
	d1 * d2 < 0 ifFalse:[ 
	 ^ false ].
	pp1 := pp1 + ((pp2-pp1) *  d1/ (d1-d2)).
	d1 := pp1 squaredLength.
	(self testSelectedDistanceSquared: d1) ifTrue:
		[
			(self verifySelectedPoint: self framePosition-pp1) ifTrue:[
				self selectedDistanceSquared: d1.
				^ true.
			].
		].

	^ false.

	
	


	


	! !

!TRay methodsFor: 'picking' stamp: 'das 5/27/2006 15:21'!
pickFrontPlane: position normal: normal

	| div sp d |

	div := normal dot: self framePointer row3. 
	div <= 0 ifTrue:[ ^ false ].
	d := ((position-self framePosition) dot: normal)/div.
	sp:= self framePosition +  (self framePointer row3 * d).
	
	(self testSelectedDistance: d)=0 ifFalse:[^false].
	self selectedDistance: d.
	self selectedPoint: sp.
	self selectedNormal: normal.
	^ true.! !

!TRay methodsFor: 'picking' stamp: 'das 6/24/2002 18:26'!
pickLocalBoundSphere: bnds 
	| lpAt position dp dd |

	lpAt := self framePointer row3.
	position := bnds localPosition - self framePosition.
" project center of sphere onto the ray"
	dp := position - (lpAt * (lpAt dot: position)).
	dd := dp dot: dp.

" distance of ray to center - if greater than radius, return false, otherwise, the ray intersects."
	^ (dd < bnds radiusSquared)

! !

!TRay methodsFor: 'picking' stamp: 'das 11/9/2002 16:23'!
pickLocalNorm: norm 

	^ (self framePointer row3 dot: norm) < 0.! !

!TRay methodsFor: 'picking' stamp: 'das 5/27/2006 15:21'!
pickPlane: position normal: normal

	| div sp d |

	div := normal dot: self framePointer row3. 
	div = 0 ifTrue:[ ^ false ].
	d := ((position-self framePosition) dot: normal)/div.
	sp:= self framePosition +  (self framePointer row3 * d).
	
	(self testSelectedDistance: d)=0 ifFalse:[^false].
	self selectedDistance: d.
	self selectedPoint: sp.
	self selectedNormal: normal.
	^ true.! !

!TRay methodsFor: 'picking' stamp: 'das 5/27/2006 14:58'!
pickQuad: norm q1: p1 q2: p2 q3: p3 q4: p4

	| pointAt po1 po2 po3 po4 d1 d2 d3 d4 tval |
	pointAt := self framePointer row3.
	norm ifNotNil:[
		((pointAt dot: norm) < 0) ifTrue:[^ false ].].
	po1 := self framePosition - p1.
	po2 := self framePosition - p2.
	po3 := self framePosition - p3.
	po4 := self framePosition - p4.

	d1 := pointAt dot: po1.
	d2 := pointAt dot: po2.
	d3 := pointAt dot: po3.
	d4 := pointAt dot: po4.

" Are any of the points actually in front of the pointer?"
	d1 < 0 ifTrue:[
		d2 < 0 ifTrue:[
			d3 < 0 ifTrue:[
				d4 < 0 ifTrue:[^ false.]]]].

"	(norm dot: po1) > 0 ifTrue:[ ^ false.]."
" If all of the points are further than the current selected distance, then we are done."
	tval := (self testSelectedDistance: d1) +
		(self testSelectedDistance: d2) +
		(self testSelectedDistance: d3) +
		(self testSelectedDistance: d4).
"Are they all behind or all in front?"
	(tval = 4 or:[tval = -4])ifTrue:[^false].

" At some point, I need to write a special pickFrame for quads, because I calculate one of the vertices twice this way. In Squeak, every cycle counts!!."
  	(self pickFrame: self framePointer tri: po1 tri: po2 tri: po3) ifFalse:[ 
		(self pickFrame: self framePointer tri: po3 tri: po4 tri: po1) ifFalse:[^ false]].
"	self selectedPoint: (self framePosition - self selectedPoint)."
	self selectedNormal: norm.
	^ true.
	
! !

!TRay methodsFor: 'picking' stamp: ''!
pickQuad: p1 q2: p2 q3: p3 q4: p4

	^ self pickQuad: nil q1: p1 q2: p2 q3: p3 q4: p4.
	
! !

!TRay methodsFor: 'picking' stamp: 'das 9/4/2006 21:02'!
pickRightCylinderFrom: pnt1 to: pnt2 radius: rad

	| p1 p2 pd d ps s pn pa dd ip |

" 	Pick a right cylinder.

	p1, p2 axis points of the cylinder relative to the framePosition.
	pd - the cutting plane. The ray is on this plane, which is parallel to the cylinder axis.
	d - distance of plane from cylinder axis.
	s - distance from d point (p1 + d*p) to cylinder surface.
	ps - vector on plane, perpendicular to axis.
	pn - normal vector.
	ip - selected point "

	p1 := pnt1 - self framePosition.
	p2 := pnt2 - self framePosition.

" Calculate the cutting plane between the ray and parallel to the axis of the cylinder. "
	pd := (self framePointer row3 cross: (p1-p2)).
" Test if the ray is parallel to axis."
	(pd dot: pd)=0 ifTrue:[ ^ false.].
	pd normalize.

" Test to see if the plane (and hence the ray) distance is less than the rad distance to the center of the cylinder."
	d := pd dot: p1.
	d abs >rad ifTrue:[ ^ false.]. 

" Calculate the perpendicular normal to the axis on the plane."
	s := ((rad*rad)-(d*d))sqrt.
	ps := ((p1-p2) cross: pd) normalized.

" Calculate the normal on the cylinder at the intersection point."
	pn := ((s*ps)-(d*pd)).
	
" Now we just calculate the intersections between two lines."
	p1 := p1 + pn.
	p2 := p2 + pn.
	pn normalize.

	pa := (p2-p1) normalized.

" We know that the two lines can't be parallel because we tested earlier."
	dd := (self framePointer row3) cross: pa.
	
	s:= (((p1 cross: pa) dot: dd)/(dd dot: dd)).
	ip := framePointer row3 * s.

" Test if we are between the two ends of the cylinder."
	((ip-p1) dot: pa)*((ip-p2) dot: pa)>0 ifTrue:[
		"The point is outside the bounds"
		^ false.].

	s := s abs.
	(self testSelectedDistance: s)=0 ifFalse:[^false].
	self selectedDistance: s.
	self selectedPoint: (ip + self framePosition).
	self selectedNormal: pn.
	^ true.
	
	





! !

!TRay methodsFor: 'picking' stamp: 'das 12/7/2005 16:27'!
pickSphere: loc radiusSquared: rs

	| p dp d dd c |

" This method is used to find the surface of the actual sphere, not just determine if the pointer intersects it."
	rs = 0.0 ifTrue:[^ false.].
	p := loc - self framePosition.
	
" project center of sphere onto the ray"
	dp := self framePointer row3 * (self framePointer row3 dot: p).
	dd := dp - p.
" distance of ray to center - if greater than radius, exit"
	d := dd dot: dd.
	d > rs ifTrue:[^false.].
" calculate the point on the sphere"
	c := (rs-d)sqrt.
	dp := dp + (c * self framePointer row3).
	d := dp dot: dp.
" if it is closer than the current selected object, save and calc norm."
	(self testSelectedDistanceSquared: d) ifTrue: [
		self selectedDistanceSquared: d.
		self selectedPoint: (self framePosition + dp).
		self selectedNormal: (self selectedPoint - loc) normalized.
		^ true.		
		].
	^ false.
! !

!TRay methodsFor: 'picking' stamp: 'ar 6/13/2003 19:42'!
pickTriangle: p1 tri: p2 tri: p3

	^ self pickTriangle: nil tri: p1 tri: p2 tri: p3.
! !

!TRay methodsFor: 'picking' stamp: 'das 5/27/2006 15:01'!
pickTriangle: norm tri: p1 tri: p2 tri: p3

	| pointAt po1 po2 po3 d1 d2 d3 |

	pointAt := self framePointer row3.
	norm ifNotNil:[
		((pointAt dot: norm) < 0) ifTrue:[^ false ].].
	po1 := self framePosition - p1.
	po2 := self framePosition - p2.
	po3 := self framePosition - p3.

" Are any of the points actually in front of the pointer?"
	d1 := pointAt dot: po1.
	d2 := pointAt dot: po2.
	d3 := pointAt dot: po3.
	d1 < 0 ifTrue:[
		d2< 0 ifTrue:[
			d3 < 0 ifTrue:[^ false.]]].

" If all of the points are further than the current selected distance, then we are done."
	((self testSelectedDistance: d1) +
		(self testSelectedDistance: d2) +
		(self testSelectedDistance: d3))abs = 3 ifTrue:[^ false].

	(self pickFrame: self framePointer tri: po1 tri: po2 tri: po3) ifTrue: [
	"	self selectedPoint: (self framePosition - self selectedPoint)."
		self selectedNormal: norm.
		self selectedTriangle: (Array with: p1 with: p2 with: p3).
		^ true.].
	 ^ false.
! !

!TRay methodsFor: 'picking' stamp: 'das 5/27/2006 15:02'!
pickTriangles: vertices list: triList
	"Pick an indexed triangle array"
	| pointAt po1 po2 po3 d1 d2 d3 fp index v0 v1 v2 normal |

	fp := self framePosition.
	pointAt := self framePointer row3.
	index := 0.
	[index < triList size] whileTrue:[
		v0 := vertices at: (triList at: (index := index+1))+1.
		v1 := vertices at: (triList at: (index := index+1))+1.
		v2 := vertices at: (triList at: (index := index+1))+1.
		po1 := fp - v0.
		po2 := fp - v1.
		po3 := fp - v2.

		"Are any of the points actually in front of the pointer?"
		d1 := pointAt dot: po1.
		d2 := pointAt dot: po2.
		d3 := pointAt dot: po3.
		(d1 >= 0.0 or:[d2 >= 0.0 or:[d3 >= 0.0]]) ifTrue:[
			" If all of the points are further than the current selected distance, then we are done."
			((self testSelectedDistance: d1)+
				(self testSelectedDistance: d2)+
				(self testSelectedDistance: d3))abs = 3 ifTrue:[^false].

			(self pickFrame: self framePointer tri: po1 tri: po2 tri: po3) ifTrue: [
				normal := ((po2 - po1) cross: (po3 - po1)) normalized.
			"	self selectedPoint: (self framePosition - self selectedPoint)."
				self selectedNormal: normal.
				self selectedTriangle: (Array with: v0 with: v1 with: v2).
				^ true.
			].
		].
	].
	^false! !

!TRay methodsFor: 'picking' stamp: 'das 5/28/2006 17:44'!
pointerPick: bnds
	| position frame gpAt dp d |

	doSelect ifTrue:[
		bnds ifNil:[^ false.].
		position := bnds globalPosition - self globalPosition.
		self selectedDistance > ((position length - bnds radius) * frameScale) 		ifTrue:[
				gpAt := self globalTransform column3.
				dp := position - ((gpAt dot: position)*gpAt).
				d := dp dot: dp.
				d < bnds radiusSquared ifFalse:[ ^ false ].
				self sphereDistSquared: d.
				frame := bnds frame.
				self currentFrame: frame.
				(frame pick: self) ifTrue:[
					self selectedPoint ifNil:[self selectedPoint: Vector3 new.].
					self selectedFrame: frame.
					self selectedGlobalPoint: 
						(frame globalTransform localPointToGlobal:self selectedPoint).
					self selectedParent: frame parent.
					self selectedFramePosition:  self framePosition.
					self selectedTarget: frame objectOwner.
					self selectedEventMask: frame objectOwner getAllEventsHandled.
					self selectedRayTransform: self localTransform.
					self selectedScale: frameScale.
					^ true.
				].
			].
		].
	^ false.! !

!TRay methodsFor: 'picking' stamp: 'das 2/3/2006 22:48'!
pointerPickFloor: frame
	
	self currentFrame: frame.
	(frame pickFloor: self) ifTrue:[
		self selectedFrame: frame.
		self selectedParent: frame parent.
		self selectedFramePosition:  self framePosition.
		self selectedTarget: frame objectOwner.
		^ true.
		].
	^ false.! !

!TRay methodsFor: 'picking' stamp: 'bgf 7/15/2006 15:52'!
pointerPickTree1: frame log: stream

	"This is a the recursive worker method for pointerPickTree:log:,
	 call that instead.   This is a workalike of pointerPickTree:frame,
	 with exhaustive logging of the pick search."

	| rval trans bs |

	frame = excludeFrame ifTrue: [
		stream cr; nextPutAll: ' Skipped excluded ', frame class, ' ', frame objectID.
		^ false.
	].

	rval := false.
	trans := self selectedCameraTransform.
	bs := frame boundSphere.

	frame visible ifTrue:[(self pointerPick: bs) ifTrue:[ rval := true].].
	(frame visible) ifTrue: [
		(bs isNil) ifTrue: [
			stream cr; 
				nextPutAll: 'NIL boundSphere in ', frame class, ' ', frame objectID, ' in ', frame parent.
		] ifFalse: [
			stream cr; 
				nextPutAll: '', frame class, ' ', frame objectID, ' bS=', (bs radius), '@' ,(bs globalPosition), ' hit ? ', rval.
		]
	] ifFalse: [
		stream cr; nextPutAll: ' Skipped invisible ', frame class, ' ', frame objectID.
	].

	frame frameChildren ifNotNil:[
		frame frameChildren do:[ :fc | (self pointerPickTree1: fc log: stream) ifTrue:[rval := true.].].].
	self selectedCameraTransform: trans.
	^ rval.! !

!TRay methodsFor: 'picking' stamp: 'das 7/19/2006 13:43'!
pointerPickTree: frame

	| rval trans |

	frame = excludeFrame ifTrue: [ ^ false ].

	rval := false.
	trans := self selectedCameraTransform.
	(frame visible & frame isSolid) ifTrue:[(self pointerPick: frame boundSphere) ifTrue:[ rval := true].].
	frame frameChildren ifNotNil:[
		frame frameChildren do:[ :fc | (self pointerPickTree: fc) ifTrue:[rval := true.].].].
	self selectedCameraTransform: trans.
	^ rval.! !

!TRay methodsFor: 'picking' stamp: 'bgf 7/15/2006 15:24'!
pointerPickTree: frame excluding: subframe

	| |

	"Do a pointerPickTree search, treating subframe and its subtree as invisible."

	[ 	excludeFrame := subframe.
		^ self pointerPickTree: frame
	] ensure: [
		excludeFrame := nil.
	].
! !

!TRay methodsFor: 'picking' stamp: 'bgf 5/17/2006 02:42'!
pointerPickTree: frame log: stream

	"This is a debugging workalike of #pointerPickTree:frame 
	 that logs its search process to the given stream.  The TRay state,
	 and the boundSphere and pick result
	 for each descendant of the frame, are recorded."

	| |

	"Dump me, and then start the pick..."

	stream cr; nextPutAll: 'TRay:'; cr; nextPutAll: '----'.
	self class allInstVarNames do: [ :name|
		stream cr; nextPutAll: (name asString, ' = ', (self instVarNamed: name)).
	].

	stream cr; cr; nextPutAll: 'Pick trace:'; cr; nextPutAll:'----------'.

	^ self pointerPickTree1: frame log: stream.
! !

!TRay methodsFor: 'picking' stamp: 'das 1/9/2006 14:07'!
portalTest: frame at: at
	| rval saveFP up side |

	saveFP := framePointer.
	rval := false.

	up := Vector3 x: 0.0 y:1.0 z:0.0.
	side := (up cross: at) normalized.
	up := (at cross: side) normalized.
	framePointer := Matrix4x4 identity.
	framePointer at: 1 at: 1 put: side x.
	framePointer at: 1 at: 2 put: side y.
	framePointer at: 1 at: 3 put: side z.
	framePointer at: 2 at: 1 put: up x.
	framePointer at: 2 at: 2 put: up y.
	framePointer at: 2 at: 3 put: up z.
	framePointer at: 3 at: 1 put: at x.
	framePointer at: 3 at: 2 put: at y.
	framePointer at: 3 at: 3 put: at z.

	self testDistance: false.
	(frame pick: self) ifTrue:[ rval := true.].
	self testDistance: true.
	framePointer := saveFP.
	^ rval.
! !

!TRay methodsFor: 'accessing' stamp: 'ar 1/14/2006 14:33'!
resetSelected

	doSelect ifTrue:[
		lastSelection := selection.
		selection := self newSelection.
		selection distance: Float infinity.
	] ifFalse: [
		currentFrame := self selectedFrame.
		framePosition := nil.
		framePointer := nil.
	].
	self maxDistance: Float infinity.
! !

!TRay methodsFor: 'accessing' stamp: 'das 1/21/2006 11:10'!
resetTotal
	lastSelection := self newSelection.
	selection := self newSelection.
	selection distance: Float infinity.
	self maxDistance: Float infinity.
	self minDistance: 0.0.
! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:43'!
selected
	^selection copy! !

!TRay methodsFor: 'accessing' stamp: 'DPR 7/13/2004 10:01'!
selected: sel
	selection := sel.
! !

!TRay methodsFor: 'accessing' stamp: 'das 4/18/2006 21:11'!
selectedCameraTransform
	^ selection cameraTransform.! !

!TRay methodsFor: 'accessing' stamp: 'das 4/18/2006 21:11'!
selectedCameraTransform: spt
	selection cameraTransform: spt.! !

!TRay methodsFor: 'accessing' stamp: 'das 6/17/2003 19:38'!
selectedDistance
	
	^ selection distance.! !

!TRay methodsFor: 'accessing' stamp: 'das 12/11/2005 09:39'!
selectedDistance: dist

	selection distance: dist * frameScale.
! !

!TRay methodsFor: 'accessing' stamp: 'das 6/17/2003 19:38'!
selectedDistanceSquared
	
	^ selection distanceSquared.! !

!TRay methodsFor: 'accessing' stamp: 'das 12/11/2005 16:54'!
selectedDistanceSquared: dist

	selection distanceSquared: dist* frameScaleSquared.
! !

!TRay methodsFor: 'accessing' stamp: 'das 1/29/2006 14:28'!
selectedEventMask

	^ selection eventMask! !

!TRay methodsFor: 'accessing' stamp: 'ar 1/27/2006 14:24'!
selectedEventMask: se

	selection eventMask: se! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:51'!
selectedFrame
	^ selection frame! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:51'!
selectedFrame: sf
	selection frame: sf! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:52'!
selectedFramePosition

	^selection framePosition.! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:52'!
selectedFramePosition: sfp

	selection framePosition: sfp.! !

!TRay methodsFor: 'accessing' stamp: 'das 2/3/2006 22:43'!
selectedGlobalPoint

	^ selection globalPoint! !

!TRay methodsFor: 'accessing' stamp: 'das 2/3/2006 22:43'!
selectedGlobalPoint: sp

	selection globalPoint: sp.! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:53'!
selectedIndex
	^ selection index! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:53'!
selectedIndex: si
	selection index: si! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:54'!
selectedNormal
	^ selection normal! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:53'!
selectedNormal: sn
	selection normal: sn.! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:52'!
selectedParent
	^ selection parent! !

!TRay methodsFor: 'accessing' stamp: 'ar 6/13/2003 19:52'!
selectedParent: sp
	selection parent: sp.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/21/2006 15:23'!
selectedPoint

	^ selection point! !

!TRay methodsFor: 'accessing' stamp: 'das 8/11/2005 18:38'!
selectedPoint: sp

	selection point: sp.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/21/2006 15:23'!
selectedPointAt

	^ (self selectedPoint - self selectedFramePosition)normalized.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/5/2005 14:36'!
selectedPortal

	^ selection portal.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/21/2006 15:23'!
selectedPortal: sp point: pos

	sp ifNil:[selection portal: nil. selection portalPoint:nil. ^ self].
	selection portal ifNil:[selection portal: sp. selection portalPoint: pos.].! !

!TRay methodsFor: 'accessing' stamp: 'das 1/6/2005 11:09'!
selectedPortalPoint

	^ selection portalPoint.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/6/2005 11:10'!
selectedPortalPoint: pPt

	selection portalPoint: pPt.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/21/2006 15:23'!
selectedRayTransform

	^ selection rayTransform.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/15/2006 12:54'!
selectedRayTransform: rt
	
	selection rayTransform: rt.! !

!TRay methodsFor: 'accessing' stamp: 'das 3/26/2006 15:48'!
selectedScale

	^selection scale.! !

!TRay methodsFor: 'accessing' stamp: 'das 3/26/2006 15:48'!
selectedScale: scl

	selection scale: scl.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/21/2006 15:23'!
selectedTarget

	^ selection target! !

!TRay methodsFor: 'accessing' stamp: 'das 1/21/2006 15:23'!
selectedTarget: trgt

	^ selection target: trgt.! !

!TRay methodsFor: 'accessing' stamp: 'das 1/21/2006 15:23'!
selectedTriangle

	^selection triangle! !

!TRay methodsFor: 'accessing' stamp: 'das 1/21/2006 15:23'!
selectedTriangle: aTriangle

	selection triangle: aTriangle! !

!TRay methodsFor: 'accessing' stamp: 'DPR 7/13/2004 10:01'!
selection
	^selection! !

!TRay methodsFor: 'accessing' stamp: 'das 9/25/2004 13:57'!
selection: newSelection
	lastSelection := selection.
	selection:= newSelection.! !

!TRay methodsFor: 'accessing' stamp: 'das 12/30/2002 16:28'!
setAutomatic: bool

	automatic := bool.! !

!TRay methodsFor: 'accessing' stamp: ''!
sphereDistSquared

	^ sphereDistSquared.! !

!TRay methodsFor: 'accessing' stamp: ''!
sphereDistSquared: sds
	sphereDistSquared := sds.! !

!TRay methodsFor: 'accessing' stamp: 'das 8/12/2002 21:38'!
testDistance

	^ testDistance.! !

!TRay methodsFor: 'accessing' stamp: 'das 8/12/2002 21:39'!
testDistance: bool

	testDistance := bool.! !

!TRay methodsFor: 'accessing' stamp: 'das 5/27/2006 15:10'!
testSelectedDistance: dist
 
	| scaledDistance |
	testDistance ifFalse:[ ^ 0. "don't test, just say yes."].
	dist < 0  ifTrue:[^ -1.].
	scaledDistance := dist * frameScale.
	selection distance < scaledDistance ifTrue:[^ 1 ]. "we already have a closer point."
	minDistance > scaledDistance ifTrue:[^ -1.]. "this is too close"
	maxDistance < scaledDistance ifTrue:[^ 1.]. "this is too far"
	^ 0.! !

!TRay methodsFor: 'accessing' stamp: 'das 3/27/2006 16:19'!
testSelectedDistanceSquared: dist

	| scaledDistance |
	testDistance ifFalse:[ ^ true. "don't test, just say yes."].
	scaledDistance := frameScaleSquared * dist.
	selection distanceSquared < scaledDistance ifTrue:[^ false ]. "we already have a closer point."
	minDistanceSquared > scaledDistance ifTrue:[^ false]. "this choice is too close "
	maxDistanceSquared < scaledDistance ifTrue:[^ false]. "this choice is too far "
	^ true.! !

!TRay methodsFor: 'picking' stamp: 'das 12/7/2005 16:28'!
verifyInFront: sp

	^((framePosition - sp) dot: framePointer row3)> 0.0.
! !

!TRay methodsFor: 'accessing' stamp: 'das 8/11/2005 18:04'!
verifySelectedPoint: sp

	((framePosition - sp) dot: framePointer row3)< 0.0 ifTrue:[^ false].
	selection point: sp.
	^ true.! !

TFrame subclass: #TSpace
	instanceVariableNames: 'color lightFrames portalFrames rayFrames tframes updateFrames alphaObjects cullBackFaces fogOn fogStart fogEnd fogDensity ambientSound testRays savedAlphaObjects defaultAvatar scriptContext defaultMaterial defaultLaserColors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TSpace commentStamp: 'das 6/8/2006 17:00' prior: 0!
A TSpace object acts as the root render frame. It is the ultimate container, and is the entry point to a render tree. TSpace objects are contained in CroquetPlace objects, and can be linked via portals. 

croquetPlace - the CroquetPlace object that this TSpace belongs to.
color - the default color of the space. This is the color you see when there are no objects to render.
lightFrames - an OrderedCollection of all of the TLights in the hierarchy inside of this space.
portalFrames - an OrderedCollection of all ofthe TPortals in the hierarchy inside ofthis space.
rayFrames - an OrderedCollection of all of the TRays in the hierarchy inside of this space.
alphaObjects - a temporay OrderedCollection of the alpha objects to be drawn in each rendered frame.
currentParent - the current parent object of the frame being rendered. This is used for instanced objects.
currentTransform - the current transform of the frame.
cullBackFaces - this is a flag to turn this on and off for everything in the space.
fogOn fogStart fogEnd fogDensity - fog state variables.
ambientSound - current local sound - to be deprecated.
locator - url object.
dropInFrame - drop the TAvatar at this location when we enter the TSpace.
testRays - turn ray testing on and off (e.g. for portal rendering)
savedAlphaObjects - push the alpha objects list onto this stack if we ever recurse into the space WHILE we are rendering it. This actually happens with the TPortal3D.
viewingParticipants - participants currently viewing (rendering) the space, directly or
	through portals they view the space from.
	viewingParticipants should be members of the tea party of every object in the space
	and all of the spaces visible through portals.  When a new viewingParticipant is added
	we walk through all the frames in the space, telling them about the new participant.
	When a new child is added, we first tell it about all of the participants that are viewing
	the space containing it, so it can ensure that participant is in its tea party. 
defaultMaterial - if a TFrame object does not specify a material when it is rendering, it will use this material defined by the TSpace.

DAS!

!TSpace methodsFor: 'accessing' stamp: ''!
addAlpha: ao
	alphaObjects class == Array ifTrue:[^self].
alphaObjects add: ao.
! !

!TSpace methodsFor: 'accessing' stamp: 'das 6/7/2006 21:53'!
addAlphaObject: anObject transform: aTransform distance: aDistance parent: aParent transparency: trans
	alphaObjects class == Array ifTrue:[^self].
	alphaObjects add: (TRenderAlpha object: anObject transform: aTransform distance: aDistance parent: aParent transparency: trans).! !

!TSpace methodsFor: 'accessing' stamp: 'das 12/9/2005 11:39'!
addFrame: frame

	tframes at: frame objectID put: frame.! !

!TSpace methodsFor: 'frameManagement' stamp: ''!
addLightFrame: ltfrm

	lightFrames add: ltfrm.! !

!TSpace methodsFor: 'frameManagement' stamp: ''!
addPortalFrame: prtlFrm

	portalFrames add: prtlFrm.! !

!TSpace methodsFor: 'frameManagement' stamp: 'das 11/9/2002 15:01'!
addRayFrame: rayfrm

	rayFrames add: rayfrm.! !

!TSpace methodsFor: 'rootService' stamp: 'das 1/21/2006 16:40'!
addUpdate: frame

	updateFrames ifNotNil:[
		updateFrames do:[ :uf | uf update. ].
		updateFrames := nil.
	].! !

!TSpace methodsFor: 'accessing' stamp: 'das 6/19/2003 08:13'!
alphaObjects

	^ alphaObjects.! !

!TSpace methodsFor: 'accessing' stamp: 'das 6/19/2003 08:13'!
alphaObjects: ao

	alphaObjects := ao.! !

!TSpace methodsFor: 'ambient sound' stamp: ''!
ambientSound
	^ambientSound! !

!TSpace methodsFor: 'ambient sound' stamp: ''!
ambientSound: aSound
	ambientSound := aSound.
! !

!TSpace methodsFor: 'ambient sound' stamp: 'DPR 5/27/2004 10:19'!
ambientSoundFadeFrom: startVolume to: endVolume duration: msec
	| steps stepSize |
	steps := (msec / 20.0) asInteger.
	steps <= 0 ifTrue: [^self].
	stepSize := (endVolume - startVolume) / steps.
	stepSize = 0.0 ifTrue: [^self].

	self ambientSoundFadeVolume: startVolume stepSize: stepSize forSteps: steps finalVolume: endVolume.
! !

!TSpace methodsFor: 'ambient sound' stamp: 'DPR 5/27/2004 09:21'!
ambientSoundFadeVolume: startVolume stepSize: stepSize forSteps: steps finalVolume: endVolume
	steps = 0 ifTrue: [ ambientSound volume: endVolume. ^self ].

	ambientSound volume: startVolume.

	self future: 20.0 deferRelative: 20.0
		perform: #ambientSoundFadeVolume:stepSize:forSteps:finalVolume:
		withArguments: { startVolume + stepSize . stepSize . steps - 1 . endVolume }.
! !

!TSpace methodsFor: 'ambient sound' stamp: 'DPR 5/27/2004 09:03'!
ambientSoundPauseComplete
	ambientSound pause.
	ambientSound volume: 0.0.
! !

!TSpace methodsFor: 'ambient sound' stamp: 'das 6/21/2002 12:42'!
ambientSoundPlaying
	ambientSound ifNotNil:[^ambientSound isPlaying.].
	^ false.
! !

!TSpace methodsFor: 'accessing' stamp: 'ar 5/27/2006 16:33'!
avatars
	"Answer an array of avatars in this space. Only searches top-level right now."
	^Array streamContents:[:s|
		self frameChildren do:[:fc| fc isAvatar ifTrue:[s nextPut: fc]]
	].! !

!TSpace methodsFor: 'accessing' stamp: 'bgf 8/2/2006 19:57'!
checkMotion: aVector forAvatar: anAvatar

	"A filtering opportunity for subclasses to enforce motion constraints on avatars
	 (such as confining the avatars to a particular region, or particular paths.)

	 The input is a vector motion to apply to the avatar;
       the return value is the resulting translation for the avatar.
	 All avatar replica driving operations should be filtered through this method.
	 The base implementation simply allows the motion, unaltered.

	 Note that 'goTo' operations (such as are used for landmarks) are not filtered;
	 this allows the creation of viewpoints outside the drivable bounds of a space.
	 As soon as an avatar at such an outside location is moved, 
	 it will be warped back into a space if the space confines the avatar."

	^ anAvatar localTransform translation + aVector! !

!TSpace methodsFor: 'accessing' stamp: ''!
color

	^ color.! !

!TSpace methodsFor: 'accessing' stamp: 'ar 2/7/2006 17:35'!
color: clr

	color := clr asVectorColor.! !

!TSpace methodsFor: 'accessing' stamp: ''!
cullBackFaces

	^ cullBackFaces.! !

!TSpace methodsFor: 'accessing' stamp: ''!
cullBackFaces: bool

	cullBackFaces := bool.! !

!TSpace methodsFor: 'accessing' stamp: 'ar 2/1/2006 18:07'!
defaultAvatar
	^defaultAvatar! !

!TSpace methodsFor: 'accessing' stamp: 'ar 2/1/2006 18:07'!
defaultAvatar: atarData
	defaultAvatar := atarData! !

!TSpace methodsFor: 'accessing' stamp: 'das 6/8/2006 17:00'!
defaultMaterial

	^ defaultMaterial.! !

!TSpace methodsFor: 'accessing' stamp: 'das 6/8/2006 17:01'!
defaultMaterial: dm

	defaultMaterial := dm.! !

!TSpace methodsFor: 'frameManagement' stamp: 'das 8/3/2006 18:59'!
doStart: frame

	frame do:[:f | (f future: f stepTime)start.].
! !

!TSpace methodsFor: 'frameManagement' stamp: 'das 8/11/2006 14:48'!
doStop: frame

	frame do:[:f | f stop.].! !

!TSpace methodsFor: 'properties' stamp: 'ar 10/28/2002 00:10'!
eventsGenerated
	^super eventsGenerated, #(
		objectEnter objectLeave
	)! !

!TSpace methodsFor: 'accessing' stamp: 'das 12/9/2005 11:41'!
findFrame: key

	^ tframes at: key.! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 10/26/2002 22:33'!
fogColor
	^color! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 10/26/2002 22:33'!
fogColor: newColor
	self color: newColor.! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 9/9/2002 14:22'!
fogDensity

	^ fogDensity.! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 9/9/2002 14:22'!
fogDensity: fd

	fogDensity := fd.! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 9/9/2002 14:22'!
fogEnd

	^ fogEnd.! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 2/7/2006 17:06'!
fogEnd: fe

	fogEnd := fe.! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 9/9/2002 14:22'!
fogOn

	^ fogOn.! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 9/9/2002 14:22'!
fogOn: bool

	fogOn := bool.! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 9/9/2002 14:22'!
fogStart

	^ fogStart.! !

!TSpace methodsFor: 'fog properties' stamp: 'ar 9/9/2002 14:22'!
fogStart: fs

	fogStart:= fs.! !

!TSpace methodsFor: 'render' stamp: 'das 10/18/2002 21:43'!
forceAlpha

	^ false.! !

!TSpace methodsFor: 'accessing' stamp: 'das 5/8/2005 19:29'!
frameChanged

" do nothing - frameChanged must always be false "! !

!TSpace methodsFor: 'render' stamp: 'DPR 7/7/2004 16:27'!
hasAlphaObjectsToRender
	^(alphaObjects size ~= 0 or: [self forceAlpha])! !

!TSpace methodsFor: 'accessing' stamp: 'bgf 3/27/2006 12:36'!
homeTransform: avatar
	"Return the transform used to return an avatar to a safe place."
	^ Matrix4x4 identity.! !

!TSpace methodsFor: 'accessing' stamp: 'das 1/19/2006 13:38'!
inPortal

" returns a portal s.t. the space itself is the viewPoint. This is useful for fixed camera position portals."
	| portal |
	portal := TPortal new.
	portal postcardLink: self postcard.
	^ portal.
! !

!TSpace methodsFor: 'accessing' stamp: 'bgf 3/27/2006 12:35'!
initialAvatarTranslation: avatar
	"Return the initial position for the given avatar.
	 Avatars added to the space with addToNewSpace: will be placed here."
	^ nil! !

!TSpace methodsFor: 'accessing' stamp: 'bgf 3/27/2006 12:35'!
initialAvatarYaw: avatar
	"Return the initial yaw for the given avatar.
	 Avatars added to the space with addToNewSpace: will be rotated to this yaw."
	^ nil! !

!TSpace methodsFor: 'initialize' stamp: 'das 7/10/2006 10:06'!
initialize
	super initialize.
	self color: 0.7@0.9@1.0@1.0.	
	lightFrames := OrderedCollection new: 0.
	portalFrames := OrderedCollection new: 0.
	rayFrames := OrderedCollection new: 0.
	savedAlphaObjects := OrderedCollection new: 0. "alpha save stack"
	alphaObjects := OrderedCollection new: 0.
	cullBackFaces := true.
	fogOn := false.
	fogStart := 1.0.
	fogEnd := 100.0.
	fogDensity := 0.05.
	testRays := true.
	frameChanged := false.
	alphaObjects := OrderedCollection new: 0. 
	tframes := Dictionary new.
	self register.
	defaultMaterial := TMaterial new.
	(self future:100)primitiveStep.
	defaultLaserColors := ReadStream on: (#(
		(0.95 0.1 0.1)
		(0.85 0.85 0.1)
		(0.1 0.9 0.2)
		(0.1 0.9 0.9)
		(0.2 0.1 1.0)
		(0.9 0.04 0.9)
		(0.6 0.25 0.05)
		(0.95 0.5 0.05)
	) collect:[:each| each asVectorColor]) shuffled.
! !

!TSpace methodsFor: 'testing' stamp: 'das 6/3/2003 21:50'!
isSpace

	^ true.! !

!TSpace methodsFor: 'accessing' stamp: 'ar 7/8/2006 17:56'!
nextLaserColor
	"Answer the next default laser color"
	defaultLaserColors atEnd ifTrue:[defaultLaserColors position: 0].
	^defaultLaserColors next! !

!TSpace methodsFor: 'ambient sound' stamp: 'das 11/3/2005 12:23'!
pauseAmbientSound
	"this is really a fade and pause"
	ambientSound ifNil:[^self].
	self ambientSoundFadeFrom: ambientSound volume to: 0.0 duration: 1000.0.
	(self future: 1000.0) ambientSoundPauseComplete.! !

!TSpace methodsFor: 'ambient sound' stamp: 'ar 6/6/2002 15:26'!
playAmbientSound: aBool
	aBool 
		ifTrue:[self startAmbientSound] 
		ifFalse:[self stopAmbientSound].
! !

!TSpace methodsFor: 'render' stamp: 'DPR 7/7/2004 16:21'!
popAlphaObjects
	"restore alpha objects saved while recursively rendering (in TPortal3D)"
	alphaObjects := savedAlphaObjects removeLast.
	! !

!TSpace methodsFor: 'step' stamp: 'das 6/7/2006 14:58'!
primitiveStep

	| bs |
	testRays ifTrue:[ 
		self doForPick:[:frame | 
			bs := frame boundSphere.
			frame isFrameChanged ifTrue:[
			bs ifNotNil:[
				bs transform: frame globalTransform.
				].
			].
			rayFrames do:[ :rf | rf automatic ifTrue:[rf pick: bs.].].
		].
	].	
	(self future:100)primitiveStep.! !

!TSpace methodsFor: 'render' stamp: 'DPR 7/7/2004 16:19'!
pushAlphaObjects
	"temporarily save alpha objects while recursively rendering (in TPortal3D)"
	savedAlphaObjects add: alphaObjects.
	alphaObjects := OrderedCollection new: 0.
	! !

!TSpace methodsFor: 'accessing' stamp: 'das 12/1/2002 22:18'!
rayFrames

	^ rayFrames.! !

!TSpace methodsFor: 'render' stamp: 'das 1/19/2006 14:20'!
rearrange: ogl

" this allows a space to rearrange itself. Called when a portal bounds is resized."! !

!TSpace methodsFor: 'accessing' stamp: 'das 12/9/2005 11:40'!
removeFrame: frame

	tframes removeKey: frame objectID.! !

!TSpace methodsFor: 'frameManagement' stamp: 'ar 10/8/2002 11:53'!
removeLightFrame: ltfrm

	lightFrames remove: ltfrm ifAbsent:[].
"	lightFrames doWithIndex:[ :lf :i| lf = ltfrm ifTrue: [lightFrames remove: i]]."! !

!TSpace methodsFor: 'frameManagement' stamp: 'das 7/11/2002 01:42'!
removePortalFrame: prtlFrm

	portalFrames remove: prtlFrm.
	"portalFrames doWithIndex:[ :pf :i| pf = prtlFrm ifTrue: [portalFrames remove: i]]."! !

!TSpace methodsFor: 'frameManagement' stamp: 'das 11/9/2002 15:01'!
removeRayFrame: rayfrm

	rayFrames remove: rayfrm ifAbsent:[].
! !

!TSpace methodsFor: 'render' stamp: 'das 1/8/2006 13:52'!
render: ogl


	" don't render me "! !

!TSpace methodsFor: 'render' stamp: 'das 5/14/2005 16:26'!
renderAlpha: ogl

	" don't render me "! !

!TSpace methodsFor: 'render' stamp: 'das 5/14/2005 16:46'!
renderClear: ogl

	ogl glClearColor: (color at: 1) with: (color at: 2) with: (color at: 3) with: (color at: 4).
	ogl glClear: (GLColorBufferBit bitOr: (GLDepthBufferBit bitOr: GLStencilBufferBit)).
! !

!TSpace methodsFor: 'render' stamp: 'das 7/10/2005 15:09'!
renderClearAll: ogl

	ogl glClearColor: (color at: 1) with: (color at: 2) with: (color at: 3) with: (color at: 4).
	ogl glClear: (GLColorBufferBit bitOr: (GLDepthBufferBit bitOr: GLStencilBufferBit)).
! !

!TSpace methodsFor: 'render' stamp: 'das 2/26/2006 15:12'!
renderClearBounds: ogl

"This method is used when we have a smaller overlay portal/space in on top of a full portal/space rendering. This is used for picture-in-picture displays - that is, showing a full environment on top of another one"

	| ac h v |
	ac := ogl camera.
" add the background poly, as we can't clear "
	h := ac bounds width/2.0.
	v := ac bounds height/2.0.
	ogl glDepthMask: GLFalse; " set depth buffer to not write. "
		glDisable: GLLighting;
		glColor3f: (color at: 1) with: (color at: 2) with: (color at: 3);
		glBegin: GLPolygon;
		glVertex3f:  h negated with: v with: ac zScreen negated;
		glVertex3f: h with: v with: ac zScreen negated;
		glVertex3f: h with:v negated  with: ac zScreen negated;
		glVertex3f: h negated with: v negated  with: ac zScreen negated;
		glEnd;
		glDepthMask: GLTrue; 
		glEnable: GLLighting.
! !

!TSpace methodsFor: 'render' stamp: 'das 5/28/2006 16:43'!
renderSpace: ogl


"----- test if enter the portal here -----"
	ogl avatar ifNotNil:[
	ogl avatar prepareRayTest ifTrue:[
	portalFrames do: [ :pf | 
		 "------ is the portal visible? ------"
		pf visible ifTrue:[
		" ------ portal near avatar?-----"
			(pf testEnter:ogl avatar:ogl avatar) ifTrue:[ ogl noSwap. ^ 0.].
	].].].].
	^self renderSpace: ogl port: nil depth: 1 ghostFrame: nil.
! !

!TSpace methodsFor: 'render' stamp: 'das 8/28/2006 02:29'!
renderSpace: ogl port: portal depth: depth ghostFrame: ghost

	| currentCamTrans currentPointerTrans currentCameraInv cpt renderPortals renderedObjects ac cp saveEvent |
	ac := ogl camera.

	"ogl glEnable: GLLineSmooth.
	ogl glHint: GLLineSmoothHint with: GLNicest.
	ogl glEnable: GLPolygonSmooth.
	ogl glHint: GLPolygonSmoothHint with: GLNicest.
	ogl glEnable: GLBlend.
	ogl glBlendFunc: GLSrcAlpha with: GLOneMinusSrcAlpha."

"------ The TSpace render method makes two passes. The first pass is used to render the non-alpha objects and to find the alpha objects. The alpha objects and their transforms are put into the SortedCollection alphaObjects. This array is sorted by distance using the bounding spheres of the objects and the second render pass using the renderAlpha method is made in using this sorted order from furthest to closest. Ghost frames are interleaved TFrames that render along side the space. Ghost frames can not contain portals or lights - as these are just ignored. ------"
"------ reset the rayframes if this is the first time through...."
	renderedObjects := 0.
"test ifNotNil:[
ogl eventPointer ifNotNil:[test translation: ogl eventPointer globalPosition + (ogl eventPointer outVector * -50).].]."
	ac transformClipPlanes.
	cpt := ac clipPlanesTransform.
	currentCamTrans := ac globalTransform copy.
	ogl eventPointer ifNotNil:[currentPointerTrans := ogl eventPointer globalTransform copy.].
	currentCameraInv := currentCamTrans orthoNormInverse.
	ogl glLoadTransposeMatrixf: currentCameraInv.
"------ The portals will have to be sorted here.... ------"
"------ Render the portals from back to front. ------"
	(ogl maxPortalDepth > depth and: [ogl suppressPortals not]) ifTrue:[
		renderPortals := OrderedCollection new.
	portalFrames do: [ :pf | 
		(pf isNotPortal3D and:[pf ~= ac currentPortal or:[pf reentrant]])ifTrue:[
		 " ------ is the portal even visible? ------"
		pf visible ifTrue:[
 		" ------ portal facing camera?-----"
		(pf lookAtCamera: ac) ifTrue:[
		"------- is the portal inside the view frustum? ------"
		pf boundSphere transform: pf globalTransform.
		(ac testBounds: pf boundSphere) ifTrue:[
			renderPortals add: pf.
			]. "(ac testBounds: pf boundSphere) ifTrue:"
		]. "(delta dot: pf lookAt)>0 ifTrue:"
		]. "pf visible"
		]. "pf~=ac currentPortal or:[pf reentrant]"
		]. "portalFrames do:"
ogl forcePick ifTrue:[saveEvent:=ogl eventPointer. ogl eventPointer:nil. ].
		(renderPortals asArray sort:[:a1 :a2| a1 cameraDistance < a2 cameraDistance]) do: [:pf | 
			cp := ac currentPortal.
			ac currentPortal: pf.
			renderedObjects := renderedObjects + (pf render: ogl depth: depth). 
			ac currentPortal: cp.
			" ------ restore the camera for the next go round ------ "
			ac globalTransform: currentCamTrans.
			ac clipPlanesTransform: cpt.].
		]. "ogl maxPortalDepth > depth ifTrue:"
 "------ restore camera position and set up clip planes in this pose.------"
	ogl forcePick ifTrue:[ogl eventPointer: saveEvent.].
	ogl eventPointer ifNotNil:[ogl eventPointer globalTransform: currentPointerTrans.].
	ogl glLoadTransposeMatrixf: currentCameraInv.

"------ Once we have rendered all of the subportals, if this is a portal, set its clip plane so that we don't render objects in front of the portal.------"
	portal ifNotNil: [ portal enableClipPlane: ogl. ].
	" create a new array "	

	" turn on the lights and setup the fog "
	self setupLights: ogl.
	fogOn ifTrue:[ self setupFog: ogl. ].
	" cull back faces if necessary."
	self setCull: ogl.
"------ Render non-alpha geometry. While we are at it, find the alpha geometry for the next pass. All TFrames know how to do this, including TSpace. TSpace is a root frame (there is no THE root frame in TEA) so we need only call the TSpace >> renderFrame message to render everything inside of this. ------"
	renderedObjects := renderedObjects + (self renderFrame: ogl ).
	ogl glPushMatrix.
	ghost ifNotNil:[
		renderedObjects := renderedObjects + (ghost renderFrame: ogl ).
	].
	ogl glPopMatrix.
"------ If there are any alpha objects, set render mode to alpha blending and make a second pass. Use alphaLength because alphaObjects size may be bigger than the actual objects rendered. ------"
	(alphaObjects size ~= 0 or:[self forceAlpha])ifTrue:[
		self renderSpaceAlpha: ogl transform: currentCamTrans.
		].

	0 to: 7 do:[ :index | 	ogl glDisable: GLLight0 + index.].
	ogl glDisable: GLCullFace.
	portal ifNotNil: [portal disableClipPlane:ogl.].
	ogl fogOn:false.
"clear this at the end because it contains bad things that can't be replicated"
	alphaObjects := OrderedCollection new: 0. 
	^ renderedObjects. "Return the number of objects we have rendered."! !

!TSpace methodsFor: 'render' stamp: 'das 8/11/2006 10:11'!
renderSpaceAlpha: ogl transform: currentCamTrans
	
	| saveTrans |

	ogl glPushMatrix.
	ogl glMultTransposeMatrixf: currentCamTrans.
	self setupAlpha: ogl.
"------ Sort the alphaObjects by distance - furthest first ------"
	saveTrans := ogl transparency.
	(alphaObjects asArray sort:[:a1 :a2|  a1 distance > a2 distance]) do: [:aObject | 
		ogl glPushMatrix.
		aObject transform ifNotNil:[
			ogl glMultMatrixf: aObject transform.].
		ogl transparency: aObject transparency.
		aObject tObject doRenderAlpha: ogl.
		ogl glPopMatrix.
		].
	ogl transparency: saveTrans.
"------ Set render mode back to original state. ------"
	ogl glPopMatrix.
	self restoreFromAlpha: ogl.
! !

!TSpace methodsFor: 'render' stamp: 'das 8/11/2006 10:11'!
renderSubspace: ogl
"Completely render a space as if it is a single TFrame. This is used by the TPortal3D."
	| rval globalTrans |

	self pushAlphaObjects. "if we are in the middle of rendering this same world"
	rval := 0.
	rval := self renderFrame: ogl.

	self hasAlphaObjectsToRender ifTrue:[
		globalTrans := Matrix4x4 new.
		ogl glGetFloatv: GLModelviewMatrix with: globalTrans.
		globalTrans := globalTrans transposed.
		ogl glLoadMatrixf: globalTrans  transposed.
		self renderSpaceAlpha: ogl transform: (globalTrans inverseTransformation). 
		].
	self popAlphaObjects.
	^ rval.! !

!TSpace methodsFor: 'render' stamp: 'das 3/12/2004 14:02'!
restoreFromAlpha: ogl

	 ogl glDepthMask: GLTrue.
	ogl glDisable: GLBlend.
! !

!TSpace methodsFor: 'ambient sound' stamp: 'DPR 5/27/2004 08:56'!
resumeAmbientSound
	| volume |
	ambientSound ifNil:[^self].
	volume := 0.8.
	ambientSound reset.
	ambientSound volume: 0.
	ambientSound resumePlaying.
	self ambientSoundFadeFrom: 0.0 to: volume duration: 1000.0.
! !

!TSpace methodsFor: 'accessing' stamp: 'das 5/8/2005 15:19'!
root

	"I am my own root"

	^ self.! !

!TSpace methodsFor: 'rootService' stamp: 'das 1/21/2006 15:53'!
rootService

	updateFrames ifNotNil:[
		updateFrames do:[ :uf | uf update. ].
		updateFrames := nil.
	].! !

!TSpace methodsFor: 'accessing' stamp: 'ar 2/22/2006 17:24'!
scriptContext
	"Answer the receiver's script context"
	^scriptContext! !

!TSpace methodsFor: 'accessing' stamp: 'ar 2/22/2006 17:24'!
scriptContext: aScriptContext
	"Answer the receiver's script context"
	scriptContext := aScriptContext! !

!TSpace methodsFor: 'render' stamp: 'das 10/10/2004 14:22'!
setCull: ogl

	cullBackFaces ifTrue: [ogl glEnable: GLCullFace; glFrontFace: ogl frontFace.]
	ifFalse: [ogl glDisable: GLCullFace.].
! !

!TSpace methodsFor: 'render' stamp: 'das 3/12/2004 14:02'!
setupAlpha: ogl

	ogl glEnable: GLBlend.
	ogl glDepthMask: GLFalse. " set depth buffer to read only. "
	ogl glBlendFunc: GLSrcAlpha with: GLOneMinusSrcAlpha.
! !

!TSpace methodsFor: 'render' stamp: 'das 5/25/2006 12:22'!
setupFog: ogl
	
	ogl
		fogOn:true;
		glFogi: GLFogMode with: GLExp;
		glFogfv: GLFogColor with:  (self fogColor asFloatArray);
		glFogf: GLFogDensity with: fogDensity;
		glHint: GLFogHint with: GLDontCare;
		glFogf: GLFogStart with: fogStart;
		glFogf: GLFogEnd with: fogEnd.
! !

!TSpace methodsFor: 'render' stamp: 'bgf 7/13/2006 01:17'!
setupLights: ogl

" Enable the closest six lights at full strength, set the next two as a ratio if distance squared, and ignore the rest. The ratios would work as follows:
		distanceRatio := 0.5 * furthest distance/closest distance.
This guarantees the following:
		distanceRatio is always less than 1.0.
		if the two values are very close to each other, then the ratio is about 0.5 : 0.5.
		if one is much closer than the other, then the ratio approaches 0.0 : 1.0."

	| sortedFrames distanceRatio |
	lightFrames size <= 8 ifTrue:[
		lightFrames doWithIndex: [ :lf :i | lf enable: ogl index: i ].
		lightFrames size + 1 to: 8 do: [ :i |
			ogl glDisable: (GLLight0 + i - 1).
		].

 	] ifFalse:[
		lightFrames do:[ :lf | 
			lf distance: (ogl camera globalPosition - lf globalPosition ) squaredLength.].
		sortedFrames := lightFrames sort:[ :a1 :a2 | a1 distance > a2 distance].
		1 to: 6 do:[ :index | (sortedFrames at: index) enable: ogl index: index.].
		distanceRatio := 0.5* ((sortedFrames at: 8) distance)/((sortedFrames at: 7) distance).
		(sortedFrames at: 7) enable: ogl index: 7 scaled: 1-distanceRatio.
		(sortedFrames at: 8) enable: ogl index: 8 scaled: distanceRatio.
	].
! !

!TSpace methodsFor: 'ambient sound' stamp: 'das 5/11/2004 14:36'!
startAmbientSound
	ambientSound ifNotNil:[
		ambientSound isLoaded ifFalse:[^self].
		ambientSound resumePlaying].! !

!TSpace methodsFor: 'ambient sound' stamp: 'ar 6/6/2002 15:27'!
stopAmbientSound
	ambientSound ifNotNil:[
		ambientSound isLoaded ifFalse:[^self].
		ambientSound pause].! !

!TSpace methodsFor: 'ambient sound' stamp: 'das 11/3/2005 12:23'!
switchToAmbientSoundFrom: prevSpace
	(self future: 1000.0) resumeAmbientSound.
	prevSpace pauseAmbientSound.! !

!TSpace methodsFor: 'render' stamp: 'das 1/26/2006 14:30'!
testRayFrames: bs

	rayFrames do:[ :rf | rf automatic ifTrue:[rf pick: bs.].].! !

!TSpace methodsFor: 'render' stamp: 'DPR 7/6/2004 16:00'!
testRayFramesQuadTree: qt
	testRays ifTrue:[ rayFrames do:[ :rf | (rf automatic and:[rf downRay]) ifTrue: [
		rf currentFrame: qt.
		qt quadPickFloor: rf location: rf globalPosition ].].].! !

!TSpace methodsFor: 'accessing' stamp: 'das 12/2/2002 00:55'!
testRays

	 ^ testRays.! !

!TSpace methodsFor: 'accessing' stamp: 'das 12/3/2002 09:33'!
testRays: bool
 " This is to turn off testing if we re-enter the same space. I should only test once. "
	testRays := bool.! !

TFrame subclass: #TSpinTest
	instanceVariableNames: 'angle'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Objects'!

!TSpinTest methodsFor: 'as yet unclassified' stamp: 'das 11/4/2005 08:15'!
initialize

	super initialize.
	angle := 0.
	(self future: (250 atRandom+250)) step.! !

!TSpinTest methodsFor: 'as yet unclassified' stamp: 'das 11/8/2005 20:43'!
step

	self addRotationAroundY: angle.
	angle := angle + 0.1.
	angle >= 360.0 ifTrue:[angle := 0.0].

	(self future: 500) step.! !
TBox initialize!
TFrame initialize!
TMaterial initialize!
TRay initialize!
